import{r as mw,a as te,R as $,g as Cm}from"./icons-MaIKpfYO.js";import{i as wn,a as gw,b as Ci,g as Mt,c as be,d as $i,e as he,S as bw,u as vu,s as ww,f as xw,h as Ow,j as Sw,k as Pw,l as Aw,m as $m,n as Qv,o as Kf,t as Im,p as Ew,q as jw,r as _w,v as kw,w as Tw,x as Cw,y as $w,z as yu,A as qr,B as li,C as Zv,D as Jv,E as Tl,F as ey,G as Iw,H as Mw,I as Mm,J as wl,K as Nw,L as Dw,M as Lw,N as Rw,O as Bw,P as zw,Q as Fw,R as Nm,T as Ww,U as Uw,V as Vw,W as Kw,X as Hw,Y as Gw,Z as Xw,_ as Yw,$ as qw,a0 as Qw,a1 as Zw,a2 as Cl,a3 as Dm,a4 as Jw,a5 as ex,a6 as tx,a7 as rx}from"./d3-vendor-BB8wXuZZ.js";var wc={exports:{}},gt={},xc={exports:{}},Oc={};/**
 * @license React
 * scheduler.production.min.js
 *
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var ty;function nx(){return ty||(ty=1,(function(t){function e(z,X){var K=z.length;z.push(X);e:for(;0<K;){var ae=K-1>>>1,re=z[ae];if(0<l(re,X))z[ae]=X,z[K]=re,K=ae;else break e}}function r(z){return z.length===0?null:z[0]}function i(z){if(z.length===0)return null;var X=z[0],K=z.pop();if(K!==X){z[0]=K;e:for(var ae=0,re=z.length,de=re>>>1;ae<de;){var we=2*(ae+1)-1,Ce=z[we],Ie=we+1,At=z[Ie];if(0>l(Ce,K))Ie<re&&0>l(At,Ce)?(z[ae]=At,z[Ie]=K,ae=Ie):(z[ae]=Ce,z[we]=K,ae=we);else if(Ie<re&&0>l(At,K))z[ae]=At,z[Ie]=K,ae=Ie;else break e}}return X}function l(z,X){var K=z.sortIndex-X.sortIndex;return K!==0?K:z.id-X.id}if(typeof performance=="object"&&typeof performance.now=="function"){var s=performance;t.unstable_now=function(){return s.now()}}else{var c=Date,d=c.now();t.unstable_now=function(){return c.now()-d}}var p=[],v=[],h=1,m=null,b=3,O=!1,S=!1,P=!1,g=typeof setTimeout=="function"?setTimeout:null,A=typeof clearTimeout=="function"?clearTimeout:null,j=typeof setImmediate<"u"?setImmediate:null;typeof navigator<"u"&&navigator.scheduling!==void 0&&navigator.scheduling.isInputPending!==void 0&&navigator.scheduling.isInputPending.bind(navigator.scheduling);function _(z){for(var X=r(v);X!==null;){if(X.callback===null)i(v);else if(X.startTime<=z)i(v),X.sortIndex=X.expirationTime,e(p,X);else break;X=r(v)}}function k(z){if(P=!1,_(z),!S)if(r(p)!==null)S=!0,G(w);else{var X=r(v);X!==null&&ee(k,X.startTime-z)}}function w(z,X){S=!1,P&&(P=!1,A(M),M=-1),O=!0;var K=b;try{for(_(X),m=r(p);m!==null&&(!(m.expirationTime>X)||z&&!F());){var ae=m.callback;if(typeof ae=="function"){m.callback=null,b=m.priorityLevel;var re=ae(m.expirationTime<=X);X=t.unstable_now(),typeof re=="function"?m.callback=re:m===r(p)&&i(p),_(X)}else i(p);m=r(p)}if(m!==null)var de=!0;else{var we=r(v);we!==null&&ee(k,we.startTime-X),de=!1}return de}finally{m=null,b=K,O=!1}}var E=!1,C=null,M=-1,D=5,U=-1;function F(){return!(t.unstable_now()-U<D)}function H(){if(C!==null){var z=t.unstable_now();U=z;var X=!0;try{X=C(!0,z)}finally{X?q():(E=!1,C=null)}}else E=!1}var q;if(typeof j=="function")q=function(){j(H)};else if(typeof MessageChannel<"u"){var Y=new MessageChannel,V=Y.port2;Y.port1.onmessage=H,q=function(){V.postMessage(null)}}else q=function(){g(H,0)};function G(z){C=z,E||(E=!0,q())}function ee(z,X){M=g(function(){z(t.unstable_now())},X)}t.unstable_IdlePriority=5,t.unstable_ImmediatePriority=1,t.unstable_LowPriority=4,t.unstable_NormalPriority=3,t.unstable_Profiling=null,t.unstable_UserBlockingPriority=2,t.unstable_cancelCallback=function(z){z.callback=null},t.unstable_continueExecution=function(){S||O||(S=!0,G(w))},t.unstable_forceFrameRate=function(z){0>z||125<z?console.error("forceFrameRate takes a positive int between 0 and 125, forcing frame rates higher than 125 fps is not supported"):D=0<z?Math.floor(1e3/z):5},t.unstable_getCurrentPriorityLevel=function(){return b},t.unstable_getFirstCallbackNode=function(){return r(p)},t.unstable_next=function(z){switch(b){case 1:case 2:case 3:var X=3;break;default:X=b}var K=b;b=X;try{return z()}finally{b=K}},t.unstable_pauseExecution=function(){},t.unstable_requestPaint=function(){},t.unstable_runWithPriority=function(z,X){switch(z){case 1:case 2:case 3:case 4:case 5:break;default:z=3}var K=b;b=z;try{return X()}finally{b=K}},t.unstable_scheduleCallback=function(z,X,K){var ae=t.unstable_now();switch(typeof K=="object"&&K!==null?(K=K.delay,K=typeof K=="number"&&0<K?ae+K:ae):K=ae,z){case 1:var re=-1;break;case 2:re=250;break;case 5:re=1073741823;break;case 4:re=1e4;break;default:re=5e3}return re=K+re,z={id:h++,callback:X,priorityLevel:z,startTime:K,expirationTime:re,sortIndex:-1},K>ae?(z.sortIndex=K,e(v,z),r(p)===null&&z===r(v)&&(P?(A(M),M=-1):P=!0,ee(k,K-ae))):(z.sortIndex=re,e(p,z),S||O||(S=!0,G(w))),z},t.unstable_shouldYield=F,t.unstable_wrapCallback=function(z){var X=b;return function(){var K=b;b=X;try{return z.apply(this,arguments)}finally{b=K}}}})(Oc)),Oc}var ry;function ix(){return ry||(ry=1,xc.exports=nx()),xc.exports}/**
 * @license React
 * react-dom.production.min.js
 *
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var ny;function ax(){if(ny)return gt;ny=1;var t=mw(),e=ix();function r(n){for(var a="https://reactjs.org/docs/error-decoder.html?invariant="+n,o=1;o<arguments.length;o++)a+="&args[]="+encodeURIComponent(arguments[o]);return"Minified React error #"+n+"; visit "+a+" for the full message or use the non-minified dev environment for full errors and additional helpful warnings."}var i=new Set,l={};function s(n,a){c(n,a),c(n+"Capture",a)}function c(n,a){for(l[n]=a,n=0;n<a.length;n++)i.add(a[n])}var d=!(typeof window>"u"||typeof window.document>"u"||typeof window.document.createElement>"u"),p=Object.prototype.hasOwnProperty,v=/^[:A-Z_a-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD][:A-Z_a-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\-.0-9\u00B7\u0300-\u036F\u203F-\u2040]*$/,h={},m={};function b(n){return p.call(m,n)?!0:p.call(h,n)?!1:v.test(n)?m[n]=!0:(h[n]=!0,!1)}function O(n,a,o,u){if(o!==null&&o.type===0)return!1;switch(typeof a){case"function":case"symbol":return!0;case"boolean":return u?!1:o!==null?!o.acceptsBooleans:(n=n.toLowerCase().slice(0,5),n!=="data-"&&n!=="aria-");default:return!1}}function S(n,a,o,u){if(a===null||typeof a>"u"||O(n,a,o,u))return!0;if(u)return!1;if(o!==null)switch(o.type){case 3:return!a;case 4:return a===!1;case 5:return isNaN(a);case 6:return isNaN(a)||1>a}return!1}function P(n,a,o,u,f,y,x){this.acceptsBooleans=a===2||a===3||a===4,this.attributeName=u,this.attributeNamespace=f,this.mustUseProperty=o,this.propertyName=n,this.type=a,this.sanitizeURL=y,this.removeEmptyString=x}var g={};"children dangerouslySetInnerHTML defaultValue defaultChecked innerHTML suppressContentEditableWarning suppressHydrationWarning style".split(" ").forEach(function(n){g[n]=new P(n,0,!1,n,null,!1,!1)}),[["acceptCharset","accept-charset"],["className","class"],["htmlFor","for"],["httpEquiv","http-equiv"]].forEach(function(n){var a=n[0];g[a]=new P(a,1,!1,n[1],null,!1,!1)}),["contentEditable","draggable","spellCheck","value"].forEach(function(n){g[n]=new P(n,2,!1,n.toLowerCase(),null,!1,!1)}),["autoReverse","externalResourcesRequired","focusable","preserveAlpha"].forEach(function(n){g[n]=new P(n,2,!1,n,null,!1,!1)}),"allowFullScreen async autoFocus autoPlay controls default defer disabled disablePictureInPicture disableRemotePlayback formNoValidate hidden loop noModule noValidate open playsInline readOnly required reversed scoped seamless itemScope".split(" ").forEach(function(n){g[n]=new P(n,3,!1,n.toLowerCase(),null,!1,!1)}),["checked","multiple","muted","selected"].forEach(function(n){g[n]=new P(n,3,!0,n,null,!1,!1)}),["capture","download"].forEach(function(n){g[n]=new P(n,4,!1,n,null,!1,!1)}),["cols","rows","size","span"].forEach(function(n){g[n]=new P(n,6,!1,n,null,!1,!1)}),["rowSpan","start"].forEach(function(n){g[n]=new P(n,5,!1,n.toLowerCase(),null,!1,!1)});var A=/[\-:]([a-z])/g;function j(n){return n[1].toUpperCase()}"accent-height alignment-baseline arabic-form baseline-shift cap-height clip-path clip-rule color-interpolation color-interpolation-filters color-profile color-rendering dominant-baseline enable-background fill-opacity fill-rule flood-color flood-opacity font-family font-size font-size-adjust font-stretch font-style font-variant font-weight glyph-name glyph-orientation-horizontal glyph-orientation-vertical horiz-adv-x horiz-origin-x image-rendering letter-spacing lighting-color marker-end marker-mid marker-start overline-position overline-thickness paint-order panose-1 pointer-events rendering-intent shape-rendering stop-color stop-opacity strikethrough-position strikethrough-thickness stroke-dasharray stroke-dashoffset stroke-linecap stroke-linejoin stroke-miterlimit stroke-opacity stroke-width text-anchor text-decoration text-rendering underline-position underline-thickness unicode-bidi unicode-range units-per-em v-alphabetic v-hanging v-ideographic v-mathematical vector-effect vert-adv-y vert-origin-x vert-origin-y word-spacing writing-mode xmlns:xlink x-height".split(" ").forEach(function(n){var a=n.replace(A,j);g[a]=new P(a,1,!1,n,null,!1,!1)}),"xlink:actuate xlink:arcrole xlink:role xlink:show xlink:title xlink:type".split(" ").forEach(function(n){var a=n.replace(A,j);g[a]=new P(a,1,!1,n,"http://www.w3.org/1999/xlink",!1,!1)}),["xml:base","xml:lang","xml:space"].forEach(function(n){var a=n.replace(A,j);g[a]=new P(a,1,!1,n,"http://www.w3.org/XML/1998/namespace",!1,!1)}),["tabIndex","crossOrigin"].forEach(function(n){g[n]=new P(n,1,!1,n.toLowerCase(),null,!1,!1)}),g.xlinkHref=new P("xlinkHref",1,!1,"xlink:href","http://www.w3.org/1999/xlink",!0,!1),["src","href","action","formAction"].forEach(function(n){g[n]=new P(n,1,!1,n.toLowerCase(),null,!0,!0)});function _(n,a,o,u){var f=g.hasOwnProperty(a)?g[a]:null;(f!==null?f.type!==0:u||!(2<a.length)||a[0]!=="o"&&a[0]!=="O"||a[1]!=="n"&&a[1]!=="N")&&(S(a,o,f,u)&&(o=null),u||f===null?b(a)&&(o===null?n.removeAttribute(a):n.setAttribute(a,""+o)):f.mustUseProperty?n[f.propertyName]=o===null?f.type===3?!1:"":o:(a=f.attributeName,u=f.attributeNamespace,o===null?n.removeAttribute(a):(f=f.type,o=f===3||f===4&&o===!0?"":""+o,u?n.setAttributeNS(u,a,o):n.setAttribute(a,o))))}var k=t.__SECRET_INTERNALS_DO_NOT_USE_OR_YOU_WILL_BE_FIRED,w=Symbol.for("react.element"),E=Symbol.for("react.portal"),C=Symbol.for("react.fragment"),M=Symbol.for("react.strict_mode"),D=Symbol.for("react.profiler"),U=Symbol.for("react.provider"),F=Symbol.for("react.context"),H=Symbol.for("react.forward_ref"),q=Symbol.for("react.suspense"),Y=Symbol.for("react.suspense_list"),V=Symbol.for("react.memo"),G=Symbol.for("react.lazy"),ee=Symbol.for("react.offscreen"),z=Symbol.iterator;function X(n){return n===null||typeof n!="object"?null:(n=z&&n[z]||n["@@iterator"],typeof n=="function"?n:null)}var K=Object.assign,ae;function re(n){if(ae===void 0)try{throw Error()}catch(o){var a=o.stack.trim().match(/\n( *(at )?)/);ae=a&&a[1]||""}return`
`+ae+n}var de=!1;function we(n,a){if(!n||de)return"";de=!0;var o=Error.prepareStackTrace;Error.prepareStackTrace=void 0;try{if(a)if(a=function(){throw Error()},Object.defineProperty(a.prototype,"props",{set:function(){throw Error()}}),typeof Reflect=="object"&&Reflect.construct){try{Reflect.construct(a,[])}catch(B){var u=B}Reflect.construct(n,[],a)}else{try{a.call()}catch(B){u=B}n.call(a.prototype)}else{try{throw Error()}catch(B){u=B}n()}}catch(B){if(B&&u&&typeof B.stack=="string"){for(var f=B.stack.split(`
`),y=u.stack.split(`
`),x=f.length-1,T=y.length-1;1<=x&&0<=T&&f[x]!==y[T];)T--;for(;1<=x&&0<=T;x--,T--)if(f[x]!==y[T]){if(x!==1||T!==1)do if(x--,T--,0>T||f[x]!==y[T]){var I=`
`+f[x].replace(" at new "," at ");return n.displayName&&I.includes("<anonymous>")&&(I=I.replace("<anonymous>",n.displayName)),I}while(1<=x&&0<=T);break}}}finally{de=!1,Error.prepareStackTrace=o}return(n=n?n.displayName||n.name:"")?re(n):""}function Ce(n){switch(n.tag){case 5:return re(n.type);case 16:return re("Lazy");case 13:return re("Suspense");case 19:return re("SuspenseList");case 0:case 2:case 15:return n=we(n.type,!1),n;case 11:return n=we(n.type.render,!1),n;case 1:return n=we(n.type,!0),n;default:return""}}function Ie(n){if(n==null)return null;if(typeof n=="function")return n.displayName||n.name||null;if(typeof n=="string")return n;switch(n){case C:return"Fragment";case E:return"Portal";case D:return"Profiler";case M:return"StrictMode";case q:return"Suspense";case Y:return"SuspenseList"}if(typeof n=="object")switch(n.$$typeof){case F:return(n.displayName||"Context")+".Consumer";case U:return(n._context.displayName||"Context")+".Provider";case H:var a=n.render;return n=n.displayName,n||(n=a.displayName||a.name||"",n=n!==""?"ForwardRef("+n+")":"ForwardRef"),n;case V:return a=n.displayName||null,a!==null?a:Ie(n.type)||"Memo";case G:a=n._payload,n=n._init;try{return Ie(n(a))}catch{}}return null}function At(n){var a=n.type;switch(n.tag){case 24:return"Cache";case 9:return(a.displayName||"Context")+".Consumer";case 10:return(a._context.displayName||"Context")+".Provider";case 18:return"DehydratedFragment";case 11:return n=a.render,n=n.displayName||n.name||"",a.displayName||(n!==""?"ForwardRef("+n+")":"ForwardRef");case 7:return"Fragment";case 5:return a;case 4:return"Portal";case 3:return"Root";case 6:return"Text";case 16:return Ie(a);case 8:return a===M?"StrictMode":"Mode";case 22:return"Offscreen";case 12:return"Profiler";case 21:return"Scope";case 13:return"Suspense";case 19:return"SuspenseList";case 25:return"TracingMarker";case 1:case 0:case 17:case 2:case 14:case 15:if(typeof a=="function")return a.displayName||a.name||null;if(typeof a=="string")return a}return null}function Re(n){switch(typeof n){case"boolean":case"number":case"string":case"undefined":return n;case"object":return n;default:return""}}function Yt(n){var a=n.type;return(n=n.nodeName)&&n.toLowerCase()==="input"&&(a==="checkbox"||a==="radio")}function En(n){var a=Yt(n)?"checked":"value",o=Object.getOwnPropertyDescriptor(n.constructor.prototype,a),u=""+n[a];if(!n.hasOwnProperty(a)&&typeof o<"u"&&typeof o.get=="function"&&typeof o.set=="function"){var f=o.get,y=o.set;return Object.defineProperty(n,a,{configurable:!0,get:function(){return f.call(this)},set:function(x){u=""+x,y.call(this,x)}}),Object.defineProperty(n,a,{enumerable:o.enumerable}),{getValue:function(){return u},setValue:function(x){u=""+x},stopTracking:function(){n._valueTracker=null,delete n[a]}}}}function Nt(n){n._valueTracker||(n._valueTracker=En(n))}function Ar(n){if(!n)return!1;var a=n._valueTracker;if(!a)return!0;var o=a.getValue(),u="";return n&&(u=Yt(n)?n.checked?"true":"false":n.value),n=u,n!==o?(a.setValue(n),!0):!1}function Et(n){if(n=n||(typeof document<"u"?document:void 0),typeof n>"u")return null;try{return n.activeElement||n.body}catch{return n.body}}function qt(n,a){var o=a.checked;return K({},a,{defaultChecked:void 0,defaultValue:void 0,value:void 0,checked:o??n._wrapperState.initialChecked})}function Er(n,a){var o=a.defaultValue==null?"":a.defaultValue,u=a.checked!=null?a.checked:a.defaultChecked;o=Re(a.value!=null?a.value:o),n._wrapperState={initialChecked:u,initialValue:o,controlled:a.type==="checkbox"||a.type==="radio"?a.checked!=null:a.value!=null}}function jr(n,a){a=a.checked,a!=null&&_(n,"checked",a,!1)}function Qt(n,a){jr(n,a);var o=Re(a.value),u=a.type;if(o!=null)u==="number"?(o===0&&n.value===""||n.value!=o)&&(n.value=""+o):n.value!==""+o&&(n.value=""+o);else if(u==="submit"||u==="reset"){n.removeAttribute("value");return}a.hasOwnProperty("value")?Zt(n,a.type,o):a.hasOwnProperty("defaultValue")&&Zt(n,a.type,Re(a.defaultValue)),a.checked==null&&a.defaultChecked!=null&&(n.defaultChecked=!!a.defaultChecked)}function _r(n,a,o){if(a.hasOwnProperty("value")||a.hasOwnProperty("defaultValue")){var u=a.type;if(!(u!=="submit"&&u!=="reset"||a.value!==void 0&&a.value!==null))return;a=""+n._wrapperState.initialValue,o||a===n.value||(n.value=a),n.defaultValue=a}o=n.name,o!==""&&(n.name=""),n.defaultChecked=!!n._wrapperState.initialChecked,o!==""&&(n.name=o)}function Zt(n,a,o){(a!=="number"||Et(n.ownerDocument)!==n)&&(o==null?n.defaultValue=""+n._wrapperState.initialValue:n.defaultValue!==""+o&&(n.defaultValue=""+o))}var Jt=Array.isArray;function dt(n,a,o,u){if(n=n.options,a){a={};for(var f=0;f<o.length;f++)a["$"+o[f]]=!0;for(o=0;o<n.length;o++)f=a.hasOwnProperty("$"+n[o].value),n[o].selected!==f&&(n[o].selected=f),f&&u&&(n[o].defaultSelected=!0)}else{for(o=""+Re(o),a=null,f=0;f<n.length;f++){if(n[f].value===o){n[f].selected=!0,u&&(n[f].defaultSelected=!0);return}a!==null||n[f].disabled||(a=n[f])}a!==null&&(a.selected=!0)}}function Dt(n,a){if(a.dangerouslySetInnerHTML!=null)throw Error(r(91));return K({},a,{value:void 0,defaultValue:void 0,children:""+n._wrapperState.initialValue})}function jn(n,a){var o=a.value;if(o==null){if(o=a.children,a=a.defaultValue,o!=null){if(a!=null)throw Error(r(92));if(Jt(o)){if(1<o.length)throw Error(r(93));o=o[0]}a=o}a==null&&(a=""),o=a}n._wrapperState={initialValue:Re(o)}}function Mi(n,a){var o=Re(a.value),u=Re(a.defaultValue);o!=null&&(o=""+o,o!==n.value&&(n.value=o),a.defaultValue==null&&n.defaultValue!==o&&(n.defaultValue=o)),u!=null&&(n.defaultValue=""+u)}function Ni(n){var a=n.textContent;a===n._wrapperState.initialValue&&a!==""&&a!==null&&(n.value=a)}function cd(n){switch(n){case"svg":return"http://www.w3.org/2000/svg";case"math":return"http://www.w3.org/1998/Math/MathML";default:return"http://www.w3.org/1999/xhtml"}}function Cu(n,a){return n==null||n==="http://www.w3.org/1999/xhtml"?cd(a):n==="http://www.w3.org/2000/svg"&&a==="foreignObject"?"http://www.w3.org/1999/xhtml":n}var fo,fd=(function(n){return typeof MSApp<"u"&&MSApp.execUnsafeLocalFunction?function(a,o,u,f){MSApp.execUnsafeLocalFunction(function(){return n(a,o,u,f)})}:n})(function(n,a){if(n.namespaceURI!=="http://www.w3.org/2000/svg"||"innerHTML"in n)n.innerHTML=a;else{for(fo=fo||document.createElement("div"),fo.innerHTML="<svg>"+a.valueOf().toString()+"</svg>",a=fo.firstChild;n.firstChild;)n.removeChild(n.firstChild);for(;a.firstChild;)n.appendChild(a.firstChild)}});function Di(n,a){if(a){var o=n.firstChild;if(o&&o===n.lastChild&&o.nodeType===3){o.nodeValue=a;return}}n.textContent=a}var Li={animationIterationCount:!0,aspectRatio:!0,borderImageOutset:!0,borderImageSlice:!0,borderImageWidth:!0,boxFlex:!0,boxFlexGroup:!0,boxOrdinalGroup:!0,columnCount:!0,columns:!0,flex:!0,flexGrow:!0,flexPositive:!0,flexShrink:!0,flexNegative:!0,flexOrder:!0,gridArea:!0,gridRow:!0,gridRowEnd:!0,gridRowSpan:!0,gridRowStart:!0,gridColumn:!0,gridColumnEnd:!0,gridColumnSpan:!0,gridColumnStart:!0,fontWeight:!0,lineClamp:!0,lineHeight:!0,opacity:!0,order:!0,orphans:!0,tabSize:!0,widows:!0,zIndex:!0,zoom:!0,fillOpacity:!0,floodOpacity:!0,stopOpacity:!0,strokeDasharray:!0,strokeDashoffset:!0,strokeMiterlimit:!0,strokeOpacity:!0,strokeWidth:!0},xb=["Webkit","ms","Moz","O"];Object.keys(Li).forEach(function(n){xb.forEach(function(a){a=a+n.charAt(0).toUpperCase()+n.substring(1),Li[a]=Li[n]})});function dd(n,a,o){return a==null||typeof a=="boolean"||a===""?"":o||typeof a!="number"||a===0||Li.hasOwnProperty(n)&&Li[n]?(""+a).trim():a+"px"}function pd(n,a){n=n.style;for(var o in a)if(a.hasOwnProperty(o)){var u=o.indexOf("--")===0,f=dd(o,a[o],u);o==="float"&&(o="cssFloat"),u?n.setProperty(o,f):n[o]=f}}var Ob=K({menuitem:!0},{area:!0,base:!0,br:!0,col:!0,embed:!0,hr:!0,img:!0,input:!0,keygen:!0,link:!0,meta:!0,param:!0,source:!0,track:!0,wbr:!0});function $u(n,a){if(a){if(Ob[n]&&(a.children!=null||a.dangerouslySetInnerHTML!=null))throw Error(r(137,n));if(a.dangerouslySetInnerHTML!=null){if(a.children!=null)throw Error(r(60));if(typeof a.dangerouslySetInnerHTML!="object"||!("__html"in a.dangerouslySetInnerHTML))throw Error(r(61))}if(a.style!=null&&typeof a.style!="object")throw Error(r(62))}}function Iu(n,a){if(n.indexOf("-")===-1)return typeof a.is=="string";switch(n){case"annotation-xml":case"color-profile":case"font-face":case"font-face-src":case"font-face-uri":case"font-face-format":case"font-face-name":case"missing-glyph":return!1;default:return!0}}var Mu=null;function Nu(n){return n=n.target||n.srcElement||window,n.correspondingUseElement&&(n=n.correspondingUseElement),n.nodeType===3?n.parentNode:n}var Du=null,_n=null,kn=null;function vd(n){if(n=aa(n)){if(typeof Du!="function")throw Error(r(280));var a=n.stateNode;a&&(a=No(a),Du(n.stateNode,n.type,a))}}function yd(n){_n?kn?kn.push(n):kn=[n]:_n=n}function hd(){if(_n){var n=_n,a=kn;if(kn=_n=null,vd(n),a)for(n=0;n<a.length;n++)vd(a[n])}}function md(n,a){return n(a)}function gd(){}var Lu=!1;function bd(n,a,o){if(Lu)return n(a,o);Lu=!0;try{return md(n,a,o)}finally{Lu=!1,(_n!==null||kn!==null)&&(gd(),hd())}}function Ri(n,a){var o=n.stateNode;if(o===null)return null;var u=No(o);if(u===null)return null;o=u[a];e:switch(a){case"onClick":case"onClickCapture":case"onDoubleClick":case"onDoubleClickCapture":case"onMouseDown":case"onMouseDownCapture":case"onMouseMove":case"onMouseMoveCapture":case"onMouseUp":case"onMouseUpCapture":case"onMouseEnter":(u=!u.disabled)||(n=n.type,u=!(n==="button"||n==="input"||n==="select"||n==="textarea")),n=!u;break e;default:n=!1}if(n)return null;if(o&&typeof o!="function")throw Error(r(231,a,typeof o));return o}var Ru=!1;if(d)try{var Bi={};Object.defineProperty(Bi,"passive",{get:function(){Ru=!0}}),window.addEventListener("test",Bi,Bi),window.removeEventListener("test",Bi,Bi)}catch{Ru=!1}function Sb(n,a,o,u,f,y,x,T,I){var B=Array.prototype.slice.call(arguments,3);try{a.apply(o,B)}catch(Z){this.onError(Z)}}var zi=!1,po=null,vo=!1,Bu=null,Pb={onError:function(n){zi=!0,po=n}};function Ab(n,a,o,u,f,y,x,T,I){zi=!1,po=null,Sb.apply(Pb,arguments)}function Eb(n,a,o,u,f,y,x,T,I){if(Ab.apply(this,arguments),zi){if(zi){var B=po;zi=!1,po=null}else throw Error(r(198));vo||(vo=!0,Bu=B)}}function Jr(n){var a=n,o=n;if(n.alternate)for(;a.return;)a=a.return;else{n=a;do a=n,(a.flags&4098)!==0&&(o=a.return),n=a.return;while(n)}return a.tag===3?o:null}function wd(n){if(n.tag===13){var a=n.memoizedState;if(a===null&&(n=n.alternate,n!==null&&(a=n.memoizedState)),a!==null)return a.dehydrated}return null}function xd(n){if(Jr(n)!==n)throw Error(r(188))}function jb(n){var a=n.alternate;if(!a){if(a=Jr(n),a===null)throw Error(r(188));return a!==n?null:n}for(var o=n,u=a;;){var f=o.return;if(f===null)break;var y=f.alternate;if(y===null){if(u=f.return,u!==null){o=u;continue}break}if(f.child===y.child){for(y=f.child;y;){if(y===o)return xd(f),n;if(y===u)return xd(f),a;y=y.sibling}throw Error(r(188))}if(o.return!==u.return)o=f,u=y;else{for(var x=!1,T=f.child;T;){if(T===o){x=!0,o=f,u=y;break}if(T===u){x=!0,u=f,o=y;break}T=T.sibling}if(!x){for(T=y.child;T;){if(T===o){x=!0,o=y,u=f;break}if(T===u){x=!0,u=y,o=f;break}T=T.sibling}if(!x)throw Error(r(189))}}if(o.alternate!==u)throw Error(r(190))}if(o.tag!==3)throw Error(r(188));return o.stateNode.current===o?n:a}function Od(n){return n=jb(n),n!==null?Sd(n):null}function Sd(n){if(n.tag===5||n.tag===6)return n;for(n=n.child;n!==null;){var a=Sd(n);if(a!==null)return a;n=n.sibling}return null}var Pd=e.unstable_scheduleCallback,Ad=e.unstable_cancelCallback,_b=e.unstable_shouldYield,kb=e.unstable_requestPaint,Be=e.unstable_now,Tb=e.unstable_getCurrentPriorityLevel,zu=e.unstable_ImmediatePriority,Ed=e.unstable_UserBlockingPriority,yo=e.unstable_NormalPriority,Cb=e.unstable_LowPriority,jd=e.unstable_IdlePriority,ho=null,er=null;function $b(n){if(er&&typeof er.onCommitFiberRoot=="function")try{er.onCommitFiberRoot(ho,n,void 0,(n.current.flags&128)===128)}catch{}}var Lt=Math.clz32?Math.clz32:Nb,Ib=Math.log,Mb=Math.LN2;function Nb(n){return n>>>=0,n===0?32:31-(Ib(n)/Mb|0)|0}var mo=64,go=4194304;function Fi(n){switch(n&-n){case 1:return 1;case 2:return 2;case 4:return 4;case 8:return 8;case 16:return 16;case 32:return 32;case 64:case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:return n&4194240;case 4194304:case 8388608:case 16777216:case 33554432:case 67108864:return n&130023424;case 134217728:return 134217728;case 268435456:return 268435456;case 536870912:return 536870912;case 1073741824:return 1073741824;default:return n}}function bo(n,a){var o=n.pendingLanes;if(o===0)return 0;var u=0,f=n.suspendedLanes,y=n.pingedLanes,x=o&268435455;if(x!==0){var T=x&~f;T!==0?u=Fi(T):(y&=x,y!==0&&(u=Fi(y)))}else x=o&~f,x!==0?u=Fi(x):y!==0&&(u=Fi(y));if(u===0)return 0;if(a!==0&&a!==u&&(a&f)===0&&(f=u&-u,y=a&-a,f>=y||f===16&&(y&4194240)!==0))return a;if((u&4)!==0&&(u|=o&16),a=n.entangledLanes,a!==0)for(n=n.entanglements,a&=u;0<a;)o=31-Lt(a),f=1<<o,u|=n[o],a&=~f;return u}function Db(n,a){switch(n){case 1:case 2:case 4:return a+250;case 8:case 16:case 32:case 64:case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:return a+5e3;case 4194304:case 8388608:case 16777216:case 33554432:case 67108864:return-1;case 134217728:case 268435456:case 536870912:case 1073741824:return-1;default:return-1}}function Lb(n,a){for(var o=n.suspendedLanes,u=n.pingedLanes,f=n.expirationTimes,y=n.pendingLanes;0<y;){var x=31-Lt(y),T=1<<x,I=f[x];I===-1?((T&o)===0||(T&u)!==0)&&(f[x]=Db(T,a)):I<=a&&(n.expiredLanes|=T),y&=~T}}function Fu(n){return n=n.pendingLanes&-1073741825,n!==0?n:n&1073741824?1073741824:0}function _d(){var n=mo;return mo<<=1,(mo&4194240)===0&&(mo=64),n}function Wu(n){for(var a=[],o=0;31>o;o++)a.push(n);return a}function Wi(n,a,o){n.pendingLanes|=a,a!==536870912&&(n.suspendedLanes=0,n.pingedLanes=0),n=n.eventTimes,a=31-Lt(a),n[a]=o}function Rb(n,a){var o=n.pendingLanes&~a;n.pendingLanes=a,n.suspendedLanes=0,n.pingedLanes=0,n.expiredLanes&=a,n.mutableReadLanes&=a,n.entangledLanes&=a,a=n.entanglements;var u=n.eventTimes;for(n=n.expirationTimes;0<o;){var f=31-Lt(o),y=1<<f;a[f]=0,u[f]=-1,n[f]=-1,o&=~y}}function Uu(n,a){var o=n.entangledLanes|=a;for(n=n.entanglements;o;){var u=31-Lt(o),f=1<<u;f&a|n[u]&a&&(n[u]|=a),o&=~f}}var Ee=0;function kd(n){return n&=-n,1<n?4<n?(n&268435455)!==0?16:536870912:4:1}var Td,Vu,Cd,$d,Id,Ku=!1,wo=[],kr=null,Tr=null,Cr=null,Ui=new Map,Vi=new Map,$r=[],Bb="mousedown mouseup touchcancel touchend touchstart auxclick dblclick pointercancel pointerdown pointerup dragend dragstart drop compositionend compositionstart keydown keypress keyup input textInput copy cut paste click change contextmenu reset submit".split(" ");function Md(n,a){switch(n){case"focusin":case"focusout":kr=null;break;case"dragenter":case"dragleave":Tr=null;break;case"mouseover":case"mouseout":Cr=null;break;case"pointerover":case"pointerout":Ui.delete(a.pointerId);break;case"gotpointercapture":case"lostpointercapture":Vi.delete(a.pointerId)}}function Ki(n,a,o,u,f,y){return n===null||n.nativeEvent!==y?(n={blockedOn:a,domEventName:o,eventSystemFlags:u,nativeEvent:y,targetContainers:[f]},a!==null&&(a=aa(a),a!==null&&Vu(a)),n):(n.eventSystemFlags|=u,a=n.targetContainers,f!==null&&a.indexOf(f)===-1&&a.push(f),n)}function zb(n,a,o,u,f){switch(a){case"focusin":return kr=Ki(kr,n,a,o,u,f),!0;case"dragenter":return Tr=Ki(Tr,n,a,o,u,f),!0;case"mouseover":return Cr=Ki(Cr,n,a,o,u,f),!0;case"pointerover":var y=f.pointerId;return Ui.set(y,Ki(Ui.get(y)||null,n,a,o,u,f)),!0;case"gotpointercapture":return y=f.pointerId,Vi.set(y,Ki(Vi.get(y)||null,n,a,o,u,f)),!0}return!1}function Nd(n){var a=en(n.target);if(a!==null){var o=Jr(a);if(o!==null){if(a=o.tag,a===13){if(a=wd(o),a!==null){n.blockedOn=a,Id(n.priority,function(){Cd(o)});return}}else if(a===3&&o.stateNode.current.memoizedState.isDehydrated){n.blockedOn=o.tag===3?o.stateNode.containerInfo:null;return}}}n.blockedOn=null}function xo(n){if(n.blockedOn!==null)return!1;for(var a=n.targetContainers;0<a.length;){var o=Gu(n.domEventName,n.eventSystemFlags,a[0],n.nativeEvent);if(o===null){o=n.nativeEvent;var u=new o.constructor(o.type,o);Mu=u,o.target.dispatchEvent(u),Mu=null}else return a=aa(o),a!==null&&Vu(a),n.blockedOn=o,!1;a.shift()}return!0}function Dd(n,a,o){xo(n)&&o.delete(a)}function Fb(){Ku=!1,kr!==null&&xo(kr)&&(kr=null),Tr!==null&&xo(Tr)&&(Tr=null),Cr!==null&&xo(Cr)&&(Cr=null),Ui.forEach(Dd),Vi.forEach(Dd)}function Hi(n,a){n.blockedOn===a&&(n.blockedOn=null,Ku||(Ku=!0,e.unstable_scheduleCallback(e.unstable_NormalPriority,Fb)))}function Gi(n){function a(f){return Hi(f,n)}if(0<wo.length){Hi(wo[0],n);for(var o=1;o<wo.length;o++){var u=wo[o];u.blockedOn===n&&(u.blockedOn=null)}}for(kr!==null&&Hi(kr,n),Tr!==null&&Hi(Tr,n),Cr!==null&&Hi(Cr,n),Ui.forEach(a),Vi.forEach(a),o=0;o<$r.length;o++)u=$r[o],u.blockedOn===n&&(u.blockedOn=null);for(;0<$r.length&&(o=$r[0],o.blockedOn===null);)Nd(o),o.blockedOn===null&&$r.shift()}var Tn=k.ReactCurrentBatchConfig,Oo=!0;function Wb(n,a,o,u){var f=Ee,y=Tn.transition;Tn.transition=null;try{Ee=1,Hu(n,a,o,u)}finally{Ee=f,Tn.transition=y}}function Ub(n,a,o,u){var f=Ee,y=Tn.transition;Tn.transition=null;try{Ee=4,Hu(n,a,o,u)}finally{Ee=f,Tn.transition=y}}function Hu(n,a,o,u){if(Oo){var f=Gu(n,a,o,u);if(f===null)cs(n,a,u,So,o),Md(n,u);else if(zb(f,n,a,o,u))u.stopPropagation();else if(Md(n,u),a&4&&-1<Bb.indexOf(n)){for(;f!==null;){var y=aa(f);if(y!==null&&Td(y),y=Gu(n,a,o,u),y===null&&cs(n,a,u,So,o),y===f)break;f=y}f!==null&&u.stopPropagation()}else cs(n,a,u,null,o)}}var So=null;function Gu(n,a,o,u){if(So=null,n=Nu(u),n=en(n),n!==null)if(a=Jr(n),a===null)n=null;else if(o=a.tag,o===13){if(n=wd(a),n!==null)return n;n=null}else if(o===3){if(a.stateNode.current.memoizedState.isDehydrated)return a.tag===3?a.stateNode.containerInfo:null;n=null}else a!==n&&(n=null);return So=n,null}function Ld(n){switch(n){case"cancel":case"click":case"close":case"contextmenu":case"copy":case"cut":case"auxclick":case"dblclick":case"dragend":case"dragstart":case"drop":case"focusin":case"focusout":case"input":case"invalid":case"keydown":case"keypress":case"keyup":case"mousedown":case"mouseup":case"paste":case"pause":case"play":case"pointercancel":case"pointerdown":case"pointerup":case"ratechange":case"reset":case"resize":case"seeked":case"submit":case"touchcancel":case"touchend":case"touchstart":case"volumechange":case"change":case"selectionchange":case"textInput":case"compositionstart":case"compositionend":case"compositionupdate":case"beforeblur":case"afterblur":case"beforeinput":case"blur":case"fullscreenchange":case"focus":case"hashchange":case"popstate":case"select":case"selectstart":return 1;case"drag":case"dragenter":case"dragexit":case"dragleave":case"dragover":case"mousemove":case"mouseout":case"mouseover":case"pointermove":case"pointerout":case"pointerover":case"scroll":case"toggle":case"touchmove":case"wheel":case"mouseenter":case"mouseleave":case"pointerenter":case"pointerleave":return 4;case"message":switch(Tb()){case zu:return 1;case Ed:return 4;case yo:case Cb:return 16;case jd:return 536870912;default:return 16}default:return 16}}var Ir=null,Xu=null,Po=null;function Rd(){if(Po)return Po;var n,a=Xu,o=a.length,u,f="value"in Ir?Ir.value:Ir.textContent,y=f.length;for(n=0;n<o&&a[n]===f[n];n++);var x=o-n;for(u=1;u<=x&&a[o-u]===f[y-u];u++);return Po=f.slice(n,1<u?1-u:void 0)}function Ao(n){var a=n.keyCode;return"charCode"in n?(n=n.charCode,n===0&&a===13&&(n=13)):n=a,n===10&&(n=13),32<=n||n===13?n:0}function Eo(){return!0}function Bd(){return!1}function bt(n){function a(o,u,f,y,x){this._reactName=o,this._targetInst=f,this.type=u,this.nativeEvent=y,this.target=x,this.currentTarget=null;for(var T in n)n.hasOwnProperty(T)&&(o=n[T],this[T]=o?o(y):y[T]);return this.isDefaultPrevented=(y.defaultPrevented!=null?y.defaultPrevented:y.returnValue===!1)?Eo:Bd,this.isPropagationStopped=Bd,this}return K(a.prototype,{preventDefault:function(){this.defaultPrevented=!0;var o=this.nativeEvent;o&&(o.preventDefault?o.preventDefault():typeof o.returnValue!="unknown"&&(o.returnValue=!1),this.isDefaultPrevented=Eo)},stopPropagation:function(){var o=this.nativeEvent;o&&(o.stopPropagation?o.stopPropagation():typeof o.cancelBubble!="unknown"&&(o.cancelBubble=!0),this.isPropagationStopped=Eo)},persist:function(){},isPersistent:Eo}),a}var Cn={eventPhase:0,bubbles:0,cancelable:0,timeStamp:function(n){return n.timeStamp||Date.now()},defaultPrevented:0,isTrusted:0},Yu=bt(Cn),Xi=K({},Cn,{view:0,detail:0}),Vb=bt(Xi),qu,Qu,Yi,jo=K({},Xi,{screenX:0,screenY:0,clientX:0,clientY:0,pageX:0,pageY:0,ctrlKey:0,shiftKey:0,altKey:0,metaKey:0,getModifierState:Ju,button:0,buttons:0,relatedTarget:function(n){return n.relatedTarget===void 0?n.fromElement===n.srcElement?n.toElement:n.fromElement:n.relatedTarget},movementX:function(n){return"movementX"in n?n.movementX:(n!==Yi&&(Yi&&n.type==="mousemove"?(qu=n.screenX-Yi.screenX,Qu=n.screenY-Yi.screenY):Qu=qu=0,Yi=n),qu)},movementY:function(n){return"movementY"in n?n.movementY:Qu}}),zd=bt(jo),Kb=K({},jo,{dataTransfer:0}),Hb=bt(Kb),Gb=K({},Xi,{relatedTarget:0}),Zu=bt(Gb),Xb=K({},Cn,{animationName:0,elapsedTime:0,pseudoElement:0}),Yb=bt(Xb),qb=K({},Cn,{clipboardData:function(n){return"clipboardData"in n?n.clipboardData:window.clipboardData}}),Qb=bt(qb),Zb=K({},Cn,{data:0}),Fd=bt(Zb),Jb={Esc:"Escape",Spacebar:" ",Left:"ArrowLeft",Up:"ArrowUp",Right:"ArrowRight",Down:"ArrowDown",Del:"Delete",Win:"OS",Menu:"ContextMenu",Apps:"ContextMenu",Scroll:"ScrollLock",MozPrintableKey:"Unidentified"},e0={8:"Backspace",9:"Tab",12:"Clear",13:"Enter",16:"Shift",17:"Control",18:"Alt",19:"Pause",20:"CapsLock",27:"Escape",32:" ",33:"PageUp",34:"PageDown",35:"End",36:"Home",37:"ArrowLeft",38:"ArrowUp",39:"ArrowRight",40:"ArrowDown",45:"Insert",46:"Delete",112:"F1",113:"F2",114:"F3",115:"F4",116:"F5",117:"F6",118:"F7",119:"F8",120:"F9",121:"F10",122:"F11",123:"F12",144:"NumLock",145:"ScrollLock",224:"Meta"},t0={Alt:"altKey",Control:"ctrlKey",Meta:"metaKey",Shift:"shiftKey"};function r0(n){var a=this.nativeEvent;return a.getModifierState?a.getModifierState(n):(n=t0[n])?!!a[n]:!1}function Ju(){return r0}var n0=K({},Xi,{key:function(n){if(n.key){var a=Jb[n.key]||n.key;if(a!=="Unidentified")return a}return n.type==="keypress"?(n=Ao(n),n===13?"Enter":String.fromCharCode(n)):n.type==="keydown"||n.type==="keyup"?e0[n.keyCode]||"Unidentified":""},code:0,location:0,ctrlKey:0,shiftKey:0,altKey:0,metaKey:0,repeat:0,locale:0,getModifierState:Ju,charCode:function(n){return n.type==="keypress"?Ao(n):0},keyCode:function(n){return n.type==="keydown"||n.type==="keyup"?n.keyCode:0},which:function(n){return n.type==="keypress"?Ao(n):n.type==="keydown"||n.type==="keyup"?n.keyCode:0}}),i0=bt(n0),a0=K({},jo,{pointerId:0,width:0,height:0,pressure:0,tangentialPressure:0,tiltX:0,tiltY:0,twist:0,pointerType:0,isPrimary:0}),Wd=bt(a0),o0=K({},Xi,{touches:0,targetTouches:0,changedTouches:0,altKey:0,metaKey:0,ctrlKey:0,shiftKey:0,getModifierState:Ju}),l0=bt(o0),u0=K({},Cn,{propertyName:0,elapsedTime:0,pseudoElement:0}),s0=bt(u0),c0=K({},jo,{deltaX:function(n){return"deltaX"in n?n.deltaX:"wheelDeltaX"in n?-n.wheelDeltaX:0},deltaY:function(n){return"deltaY"in n?n.deltaY:"wheelDeltaY"in n?-n.wheelDeltaY:"wheelDelta"in n?-n.wheelDelta:0},deltaZ:0,deltaMode:0}),f0=bt(c0),d0=[9,13,27,32],es=d&&"CompositionEvent"in window,qi=null;d&&"documentMode"in document&&(qi=document.documentMode);var p0=d&&"TextEvent"in window&&!qi,Ud=d&&(!es||qi&&8<qi&&11>=qi),Vd=" ",Kd=!1;function Hd(n,a){switch(n){case"keyup":return d0.indexOf(a.keyCode)!==-1;case"keydown":return a.keyCode!==229;case"keypress":case"mousedown":case"focusout":return!0;default:return!1}}function Gd(n){return n=n.detail,typeof n=="object"&&"data"in n?n.data:null}var $n=!1;function v0(n,a){switch(n){case"compositionend":return Gd(a);case"keypress":return a.which!==32?null:(Kd=!0,Vd);case"textInput":return n=a.data,n===Vd&&Kd?null:n;default:return null}}function y0(n,a){if($n)return n==="compositionend"||!es&&Hd(n,a)?(n=Rd(),Po=Xu=Ir=null,$n=!1,n):null;switch(n){case"paste":return null;case"keypress":if(!(a.ctrlKey||a.altKey||a.metaKey)||a.ctrlKey&&a.altKey){if(a.char&&1<a.char.length)return a.char;if(a.which)return String.fromCharCode(a.which)}return null;case"compositionend":return Ud&&a.locale!=="ko"?null:a.data;default:return null}}var h0={color:!0,date:!0,datetime:!0,"datetime-local":!0,email:!0,month:!0,number:!0,password:!0,range:!0,search:!0,tel:!0,text:!0,time:!0,url:!0,week:!0};function Xd(n){var a=n&&n.nodeName&&n.nodeName.toLowerCase();return a==="input"?!!h0[n.type]:a==="textarea"}function Yd(n,a,o,u){yd(u),a=$o(a,"onChange"),0<a.length&&(o=new Yu("onChange","change",null,o,u),n.push({event:o,listeners:a}))}var Qi=null,Zi=null;function m0(n){pp(n,0)}function _o(n){var a=Ln(n);if(Ar(a))return n}function g0(n,a){if(n==="change")return a}var qd=!1;if(d){var ts;if(d){var rs="oninput"in document;if(!rs){var Qd=document.createElement("div");Qd.setAttribute("oninput","return;"),rs=typeof Qd.oninput=="function"}ts=rs}else ts=!1;qd=ts&&(!document.documentMode||9<document.documentMode)}function Zd(){Qi&&(Qi.detachEvent("onpropertychange",Jd),Zi=Qi=null)}function Jd(n){if(n.propertyName==="value"&&_o(Zi)){var a=[];Yd(a,Zi,n,Nu(n)),bd(m0,a)}}function b0(n,a,o){n==="focusin"?(Zd(),Qi=a,Zi=o,Qi.attachEvent("onpropertychange",Jd)):n==="focusout"&&Zd()}function w0(n){if(n==="selectionchange"||n==="keyup"||n==="keydown")return _o(Zi)}function x0(n,a){if(n==="click")return _o(a)}function O0(n,a){if(n==="input"||n==="change")return _o(a)}function S0(n,a){return n===a&&(n!==0||1/n===1/a)||n!==n&&a!==a}var Rt=typeof Object.is=="function"?Object.is:S0;function Ji(n,a){if(Rt(n,a))return!0;if(typeof n!="object"||n===null||typeof a!="object"||a===null)return!1;var o=Object.keys(n),u=Object.keys(a);if(o.length!==u.length)return!1;for(u=0;u<o.length;u++){var f=o[u];if(!p.call(a,f)||!Rt(n[f],a[f]))return!1}return!0}function ep(n){for(;n&&n.firstChild;)n=n.firstChild;return n}function tp(n,a){var o=ep(n);n=0;for(var u;o;){if(o.nodeType===3){if(u=n+o.textContent.length,n<=a&&u>=a)return{node:o,offset:a-n};n=u}e:{for(;o;){if(o.nextSibling){o=o.nextSibling;break e}o=o.parentNode}o=void 0}o=ep(o)}}function rp(n,a){return n&&a?n===a?!0:n&&n.nodeType===3?!1:a&&a.nodeType===3?rp(n,a.parentNode):"contains"in n?n.contains(a):n.compareDocumentPosition?!!(n.compareDocumentPosition(a)&16):!1:!1}function np(){for(var n=window,a=Et();a instanceof n.HTMLIFrameElement;){try{var o=typeof a.contentWindow.location.href=="string"}catch{o=!1}if(o)n=a.contentWindow;else break;a=Et(n.document)}return a}function ns(n){var a=n&&n.nodeName&&n.nodeName.toLowerCase();return a&&(a==="input"&&(n.type==="text"||n.type==="search"||n.type==="tel"||n.type==="url"||n.type==="password")||a==="textarea"||n.contentEditable==="true")}function P0(n){var a=np(),o=n.focusedElem,u=n.selectionRange;if(a!==o&&o&&o.ownerDocument&&rp(o.ownerDocument.documentElement,o)){if(u!==null&&ns(o)){if(a=u.start,n=u.end,n===void 0&&(n=a),"selectionStart"in o)o.selectionStart=a,o.selectionEnd=Math.min(n,o.value.length);else if(n=(a=o.ownerDocument||document)&&a.defaultView||window,n.getSelection){n=n.getSelection();var f=o.textContent.length,y=Math.min(u.start,f);u=u.end===void 0?y:Math.min(u.end,f),!n.extend&&y>u&&(f=u,u=y,y=f),f=tp(o,y);var x=tp(o,u);f&&x&&(n.rangeCount!==1||n.anchorNode!==f.node||n.anchorOffset!==f.offset||n.focusNode!==x.node||n.focusOffset!==x.offset)&&(a=a.createRange(),a.setStart(f.node,f.offset),n.removeAllRanges(),y>u?(n.addRange(a),n.extend(x.node,x.offset)):(a.setEnd(x.node,x.offset),n.addRange(a)))}}for(a=[],n=o;n=n.parentNode;)n.nodeType===1&&a.push({element:n,left:n.scrollLeft,top:n.scrollTop});for(typeof o.focus=="function"&&o.focus(),o=0;o<a.length;o++)n=a[o],n.element.scrollLeft=n.left,n.element.scrollTop=n.top}}var A0=d&&"documentMode"in document&&11>=document.documentMode,In=null,is=null,ea=null,as=!1;function ip(n,a,o){var u=o.window===o?o.document:o.nodeType===9?o:o.ownerDocument;as||In==null||In!==Et(u)||(u=In,"selectionStart"in u&&ns(u)?u={start:u.selectionStart,end:u.selectionEnd}:(u=(u.ownerDocument&&u.ownerDocument.defaultView||window).getSelection(),u={anchorNode:u.anchorNode,anchorOffset:u.anchorOffset,focusNode:u.focusNode,focusOffset:u.focusOffset}),ea&&Ji(ea,u)||(ea=u,u=$o(is,"onSelect"),0<u.length&&(a=new Yu("onSelect","select",null,a,o),n.push({event:a,listeners:u}),a.target=In)))}function ko(n,a){var o={};return o[n.toLowerCase()]=a.toLowerCase(),o["Webkit"+n]="webkit"+a,o["Moz"+n]="moz"+a,o}var Mn={animationend:ko("Animation","AnimationEnd"),animationiteration:ko("Animation","AnimationIteration"),animationstart:ko("Animation","AnimationStart"),transitionend:ko("Transition","TransitionEnd")},os={},ap={};d&&(ap=document.createElement("div").style,"AnimationEvent"in window||(delete Mn.animationend.animation,delete Mn.animationiteration.animation,delete Mn.animationstart.animation),"TransitionEvent"in window||delete Mn.transitionend.transition);function To(n){if(os[n])return os[n];if(!Mn[n])return n;var a=Mn[n],o;for(o in a)if(a.hasOwnProperty(o)&&o in ap)return os[n]=a[o];return n}var op=To("animationend"),lp=To("animationiteration"),up=To("animationstart"),sp=To("transitionend"),cp=new Map,fp="abort auxClick cancel canPlay canPlayThrough click close contextMenu copy cut drag dragEnd dragEnter dragExit dragLeave dragOver dragStart drop durationChange emptied encrypted ended error gotPointerCapture input invalid keyDown keyPress keyUp load loadedData loadedMetadata loadStart lostPointerCapture mouseDown mouseMove mouseOut mouseOver mouseUp paste pause play playing pointerCancel pointerDown pointerMove pointerOut pointerOver pointerUp progress rateChange reset resize seeked seeking stalled submit suspend timeUpdate touchCancel touchEnd touchStart volumeChange scroll toggle touchMove waiting wheel".split(" ");function Mr(n,a){cp.set(n,a),s(a,[n])}for(var ls=0;ls<fp.length;ls++){var us=fp[ls],E0=us.toLowerCase(),j0=us[0].toUpperCase()+us.slice(1);Mr(E0,"on"+j0)}Mr(op,"onAnimationEnd"),Mr(lp,"onAnimationIteration"),Mr(up,"onAnimationStart"),Mr("dblclick","onDoubleClick"),Mr("focusin","onFocus"),Mr("focusout","onBlur"),Mr(sp,"onTransitionEnd"),c("onMouseEnter",["mouseout","mouseover"]),c("onMouseLeave",["mouseout","mouseover"]),c("onPointerEnter",["pointerout","pointerover"]),c("onPointerLeave",["pointerout","pointerover"]),s("onChange","change click focusin focusout input keydown keyup selectionchange".split(" ")),s("onSelect","focusout contextmenu dragend focusin keydown keyup mousedown mouseup selectionchange".split(" ")),s("onBeforeInput",["compositionend","keypress","textInput","paste"]),s("onCompositionEnd","compositionend focusout keydown keypress keyup mousedown".split(" ")),s("onCompositionStart","compositionstart focusout keydown keypress keyup mousedown".split(" ")),s("onCompositionUpdate","compositionupdate focusout keydown keypress keyup mousedown".split(" "));var ta="abort canplay canplaythrough durationchange emptied encrypted ended error loadeddata loadedmetadata loadstart pause play playing progress ratechange resize seeked seeking stalled suspend timeupdate volumechange waiting".split(" "),_0=new Set("cancel close invalid load scroll toggle".split(" ").concat(ta));function dp(n,a,o){var u=n.type||"unknown-event";n.currentTarget=o,Eb(u,a,void 0,n),n.currentTarget=null}function pp(n,a){a=(a&4)!==0;for(var o=0;o<n.length;o++){var u=n[o],f=u.event;u=u.listeners;e:{var y=void 0;if(a)for(var x=u.length-1;0<=x;x--){var T=u[x],I=T.instance,B=T.currentTarget;if(T=T.listener,I!==y&&f.isPropagationStopped())break e;dp(f,T,B),y=I}else for(x=0;x<u.length;x++){if(T=u[x],I=T.instance,B=T.currentTarget,T=T.listener,I!==y&&f.isPropagationStopped())break e;dp(f,T,B),y=I}}}if(vo)throw n=Bu,vo=!1,Bu=null,n}function ke(n,a){var o=a[hs];o===void 0&&(o=a[hs]=new Set);var u=n+"__bubble";o.has(u)||(vp(a,n,2,!1),o.add(u))}function ss(n,a,o){var u=0;a&&(u|=4),vp(o,n,u,a)}var Co="_reactListening"+Math.random().toString(36).slice(2);function ra(n){if(!n[Co]){n[Co]=!0,i.forEach(function(o){o!=="selectionchange"&&(_0.has(o)||ss(o,!1,n),ss(o,!0,n))});var a=n.nodeType===9?n:n.ownerDocument;a===null||a[Co]||(a[Co]=!0,ss("selectionchange",!1,a))}}function vp(n,a,o,u){switch(Ld(a)){case 1:var f=Wb;break;case 4:f=Ub;break;default:f=Hu}o=f.bind(null,a,o,n),f=void 0,!Ru||a!=="touchstart"&&a!=="touchmove"&&a!=="wheel"||(f=!0),u?f!==void 0?n.addEventListener(a,o,{capture:!0,passive:f}):n.addEventListener(a,o,!0):f!==void 0?n.addEventListener(a,o,{passive:f}):n.addEventListener(a,o,!1)}function cs(n,a,o,u,f){var y=u;if((a&1)===0&&(a&2)===0&&u!==null)e:for(;;){if(u===null)return;var x=u.tag;if(x===3||x===4){var T=u.stateNode.containerInfo;if(T===f||T.nodeType===8&&T.parentNode===f)break;if(x===4)for(x=u.return;x!==null;){var I=x.tag;if((I===3||I===4)&&(I=x.stateNode.containerInfo,I===f||I.nodeType===8&&I.parentNode===f))return;x=x.return}for(;T!==null;){if(x=en(T),x===null)return;if(I=x.tag,I===5||I===6){u=y=x;continue e}T=T.parentNode}}u=u.return}bd(function(){var B=y,Z=Nu(o),J=[];e:{var Q=cp.get(n);if(Q!==void 0){var oe=Yu,ue=n;switch(n){case"keypress":if(Ao(o)===0)break e;case"keydown":case"keyup":oe=i0;break;case"focusin":ue="focus",oe=Zu;break;case"focusout":ue="blur",oe=Zu;break;case"beforeblur":case"afterblur":oe=Zu;break;case"click":if(o.button===2)break e;case"auxclick":case"dblclick":case"mousedown":case"mousemove":case"mouseup":case"mouseout":case"mouseover":case"contextmenu":oe=zd;break;case"drag":case"dragend":case"dragenter":case"dragexit":case"dragleave":case"dragover":case"dragstart":case"drop":oe=Hb;break;case"touchcancel":case"touchend":case"touchmove":case"touchstart":oe=l0;break;case op:case lp:case up:oe=Yb;break;case sp:oe=s0;break;case"scroll":oe=Vb;break;case"wheel":oe=f0;break;case"copy":case"cut":case"paste":oe=Qb;break;case"gotpointercapture":case"lostpointercapture":case"pointercancel":case"pointerdown":case"pointermove":case"pointerout":case"pointerover":case"pointerup":oe=Wd}var se=(a&4)!==0,ze=!se&&n==="scroll",L=se?Q!==null?Q+"Capture":null:Q;se=[];for(var N=B,R;N!==null;){R=N;var ne=R.stateNode;if(R.tag===5&&ne!==null&&(R=ne,L!==null&&(ne=Ri(N,L),ne!=null&&se.push(na(N,ne,R)))),ze)break;N=N.return}0<se.length&&(Q=new oe(Q,ue,null,o,Z),J.push({event:Q,listeners:se}))}}if((a&7)===0){e:{if(Q=n==="mouseover"||n==="pointerover",oe=n==="mouseout"||n==="pointerout",Q&&o!==Mu&&(ue=o.relatedTarget||o.fromElement)&&(en(ue)||ue[ur]))break e;if((oe||Q)&&(Q=Z.window===Z?Z:(Q=Z.ownerDocument)?Q.defaultView||Q.parentWindow:window,oe?(ue=o.relatedTarget||o.toElement,oe=B,ue=ue?en(ue):null,ue!==null&&(ze=Jr(ue),ue!==ze||ue.tag!==5&&ue.tag!==6)&&(ue=null)):(oe=null,ue=B),oe!==ue)){if(se=zd,ne="onMouseLeave",L="onMouseEnter",N="mouse",(n==="pointerout"||n==="pointerover")&&(se=Wd,ne="onPointerLeave",L="onPointerEnter",N="pointer"),ze=oe==null?Q:Ln(oe),R=ue==null?Q:Ln(ue),Q=new se(ne,N+"leave",oe,o,Z),Q.target=ze,Q.relatedTarget=R,ne=null,en(Z)===B&&(se=new se(L,N+"enter",ue,o,Z),se.target=R,se.relatedTarget=ze,ne=se),ze=ne,oe&&ue)t:{for(se=oe,L=ue,N=0,R=se;R;R=Nn(R))N++;for(R=0,ne=L;ne;ne=Nn(ne))R++;for(;0<N-R;)se=Nn(se),N--;for(;0<R-N;)L=Nn(L),R--;for(;N--;){if(se===L||L!==null&&se===L.alternate)break t;se=Nn(se),L=Nn(L)}se=null}else se=null;oe!==null&&yp(J,Q,oe,se,!1),ue!==null&&ze!==null&&yp(J,ze,ue,se,!0)}}e:{if(Q=B?Ln(B):window,oe=Q.nodeName&&Q.nodeName.toLowerCase(),oe==="select"||oe==="input"&&Q.type==="file")var ce=g0;else if(Xd(Q))if(qd)ce=O0;else{ce=w0;var pe=b0}else(oe=Q.nodeName)&&oe.toLowerCase()==="input"&&(Q.type==="checkbox"||Q.type==="radio")&&(ce=x0);if(ce&&(ce=ce(n,B))){Yd(J,ce,o,Z);break e}pe&&pe(n,Q,B),n==="focusout"&&(pe=Q._wrapperState)&&pe.controlled&&Q.type==="number"&&Zt(Q,"number",Q.value)}switch(pe=B?Ln(B):window,n){case"focusin":(Xd(pe)||pe.contentEditable==="true")&&(In=pe,is=B,ea=null);break;case"focusout":ea=is=In=null;break;case"mousedown":as=!0;break;case"contextmenu":case"mouseup":case"dragend":as=!1,ip(J,o,Z);break;case"selectionchange":if(A0)break;case"keydown":case"keyup":ip(J,o,Z)}var ve;if(es)e:{switch(n){case"compositionstart":var me="onCompositionStart";break e;case"compositionend":me="onCompositionEnd";break e;case"compositionupdate":me="onCompositionUpdate";break e}me=void 0}else $n?Hd(n,o)&&(me="onCompositionEnd"):n==="keydown"&&o.keyCode===229&&(me="onCompositionStart");me&&(Ud&&o.locale!=="ko"&&($n||me!=="onCompositionStart"?me==="onCompositionEnd"&&$n&&(ve=Rd()):(Ir=Z,Xu="value"in Ir?Ir.value:Ir.textContent,$n=!0)),pe=$o(B,me),0<pe.length&&(me=new Fd(me,n,null,o,Z),J.push({event:me,listeners:pe}),ve?me.data=ve:(ve=Gd(o),ve!==null&&(me.data=ve)))),(ve=p0?v0(n,o):y0(n,o))&&(B=$o(B,"onBeforeInput"),0<B.length&&(Z=new Fd("onBeforeInput","beforeinput",null,o,Z),J.push({event:Z,listeners:B}),Z.data=ve))}pp(J,a)})}function na(n,a,o){return{instance:n,listener:a,currentTarget:o}}function $o(n,a){for(var o=a+"Capture",u=[];n!==null;){var f=n,y=f.stateNode;f.tag===5&&y!==null&&(f=y,y=Ri(n,o),y!=null&&u.unshift(na(n,y,f)),y=Ri(n,a),y!=null&&u.push(na(n,y,f))),n=n.return}return u}function Nn(n){if(n===null)return null;do n=n.return;while(n&&n.tag!==5);return n||null}function yp(n,a,o,u,f){for(var y=a._reactName,x=[];o!==null&&o!==u;){var T=o,I=T.alternate,B=T.stateNode;if(I!==null&&I===u)break;T.tag===5&&B!==null&&(T=B,f?(I=Ri(o,y),I!=null&&x.unshift(na(o,I,T))):f||(I=Ri(o,y),I!=null&&x.push(na(o,I,T)))),o=o.return}x.length!==0&&n.push({event:a,listeners:x})}var k0=/\r\n?/g,T0=/\u0000|\uFFFD/g;function hp(n){return(typeof n=="string"?n:""+n).replace(k0,`
`).replace(T0,"")}function Io(n,a,o){if(a=hp(a),hp(n)!==a&&o)throw Error(r(425))}function Mo(){}var fs=null,ds=null;function ps(n,a){return n==="textarea"||n==="noscript"||typeof a.children=="string"||typeof a.children=="number"||typeof a.dangerouslySetInnerHTML=="object"&&a.dangerouslySetInnerHTML!==null&&a.dangerouslySetInnerHTML.__html!=null}var vs=typeof setTimeout=="function"?setTimeout:void 0,C0=typeof clearTimeout=="function"?clearTimeout:void 0,mp=typeof Promise=="function"?Promise:void 0,$0=typeof queueMicrotask=="function"?queueMicrotask:typeof mp<"u"?function(n){return mp.resolve(null).then(n).catch(I0)}:vs;function I0(n){setTimeout(function(){throw n})}function ys(n,a){var o=a,u=0;do{var f=o.nextSibling;if(n.removeChild(o),f&&f.nodeType===8)if(o=f.data,o==="/$"){if(u===0){n.removeChild(f),Gi(a);return}u--}else o!=="$"&&o!=="$?"&&o!=="$!"||u++;o=f}while(o);Gi(a)}function Nr(n){for(;n!=null;n=n.nextSibling){var a=n.nodeType;if(a===1||a===3)break;if(a===8){if(a=n.data,a==="$"||a==="$!"||a==="$?")break;if(a==="/$")return null}}return n}function gp(n){n=n.previousSibling;for(var a=0;n;){if(n.nodeType===8){var o=n.data;if(o==="$"||o==="$!"||o==="$?"){if(a===0)return n;a--}else o==="/$"&&a++}n=n.previousSibling}return null}var Dn=Math.random().toString(36).slice(2),tr="__reactFiber$"+Dn,ia="__reactProps$"+Dn,ur="__reactContainer$"+Dn,hs="__reactEvents$"+Dn,M0="__reactListeners$"+Dn,N0="__reactHandles$"+Dn;function en(n){var a=n[tr];if(a)return a;for(var o=n.parentNode;o;){if(a=o[ur]||o[tr]){if(o=a.alternate,a.child!==null||o!==null&&o.child!==null)for(n=gp(n);n!==null;){if(o=n[tr])return o;n=gp(n)}return a}n=o,o=n.parentNode}return null}function aa(n){return n=n[tr]||n[ur],!n||n.tag!==5&&n.tag!==6&&n.tag!==13&&n.tag!==3?null:n}function Ln(n){if(n.tag===5||n.tag===6)return n.stateNode;throw Error(r(33))}function No(n){return n[ia]||null}var ms=[],Rn=-1;function Dr(n){return{current:n}}function Te(n){0>Rn||(n.current=ms[Rn],ms[Rn]=null,Rn--)}function _e(n,a){Rn++,ms[Rn]=n.current,n.current=a}var Lr={},tt=Dr(Lr),pt=Dr(!1),tn=Lr;function Bn(n,a){var o=n.type.contextTypes;if(!o)return Lr;var u=n.stateNode;if(u&&u.__reactInternalMemoizedUnmaskedChildContext===a)return u.__reactInternalMemoizedMaskedChildContext;var f={},y;for(y in o)f[y]=a[y];return u&&(n=n.stateNode,n.__reactInternalMemoizedUnmaskedChildContext=a,n.__reactInternalMemoizedMaskedChildContext=f),f}function vt(n){return n=n.childContextTypes,n!=null}function Do(){Te(pt),Te(tt)}function bp(n,a,o){if(tt.current!==Lr)throw Error(r(168));_e(tt,a),_e(pt,o)}function wp(n,a,o){var u=n.stateNode;if(a=a.childContextTypes,typeof u.getChildContext!="function")return o;u=u.getChildContext();for(var f in u)if(!(f in a))throw Error(r(108,At(n)||"Unknown",f));return K({},o,u)}function Lo(n){return n=(n=n.stateNode)&&n.__reactInternalMemoizedMergedChildContext||Lr,tn=tt.current,_e(tt,n),_e(pt,pt.current),!0}function xp(n,a,o){var u=n.stateNode;if(!u)throw Error(r(169));o?(n=wp(n,a,tn),u.__reactInternalMemoizedMergedChildContext=n,Te(pt),Te(tt),_e(tt,n)):Te(pt),_e(pt,o)}var sr=null,Ro=!1,gs=!1;function Op(n){sr===null?sr=[n]:sr.push(n)}function D0(n){Ro=!0,Op(n)}function Rr(){if(!gs&&sr!==null){gs=!0;var n=0,a=Ee;try{var o=sr;for(Ee=1;n<o.length;n++){var u=o[n];do u=u(!0);while(u!==null)}sr=null,Ro=!1}catch(f){throw sr!==null&&(sr=sr.slice(n+1)),Pd(zu,Rr),f}finally{Ee=a,gs=!1}}return null}var zn=[],Fn=0,Bo=null,zo=0,jt=[],_t=0,rn=null,cr=1,fr="";function nn(n,a){zn[Fn++]=zo,zn[Fn++]=Bo,Bo=n,zo=a}function Sp(n,a,o){jt[_t++]=cr,jt[_t++]=fr,jt[_t++]=rn,rn=n;var u=cr;n=fr;var f=32-Lt(u)-1;u&=~(1<<f),o+=1;var y=32-Lt(a)+f;if(30<y){var x=f-f%5;y=(u&(1<<x)-1).toString(32),u>>=x,f-=x,cr=1<<32-Lt(a)+f|o<<f|u,fr=y+n}else cr=1<<y|o<<f|u,fr=n}function bs(n){n.return!==null&&(nn(n,1),Sp(n,1,0))}function ws(n){for(;n===Bo;)Bo=zn[--Fn],zn[Fn]=null,zo=zn[--Fn],zn[Fn]=null;for(;n===rn;)rn=jt[--_t],jt[_t]=null,fr=jt[--_t],jt[_t]=null,cr=jt[--_t],jt[_t]=null}var wt=null,xt=null,$e=!1,Bt=null;function Pp(n,a){var o=$t(5,null,null,0);o.elementType="DELETED",o.stateNode=a,o.return=n,a=n.deletions,a===null?(n.deletions=[o],n.flags|=16):a.push(o)}function Ap(n,a){switch(n.tag){case 5:var o=n.type;return a=a.nodeType!==1||o.toLowerCase()!==a.nodeName.toLowerCase()?null:a,a!==null?(n.stateNode=a,wt=n,xt=Nr(a.firstChild),!0):!1;case 6:return a=n.pendingProps===""||a.nodeType!==3?null:a,a!==null?(n.stateNode=a,wt=n,xt=null,!0):!1;case 13:return a=a.nodeType!==8?null:a,a!==null?(o=rn!==null?{id:cr,overflow:fr}:null,n.memoizedState={dehydrated:a,treeContext:o,retryLane:1073741824},o=$t(18,null,null,0),o.stateNode=a,o.return=n,n.child=o,wt=n,xt=null,!0):!1;default:return!1}}function xs(n){return(n.mode&1)!==0&&(n.flags&128)===0}function Os(n){if($e){var a=xt;if(a){var o=a;if(!Ap(n,a)){if(xs(n))throw Error(r(418));a=Nr(o.nextSibling);var u=wt;a&&Ap(n,a)?Pp(u,o):(n.flags=n.flags&-4097|2,$e=!1,wt=n)}}else{if(xs(n))throw Error(r(418));n.flags=n.flags&-4097|2,$e=!1,wt=n}}}function Ep(n){for(n=n.return;n!==null&&n.tag!==5&&n.tag!==3&&n.tag!==13;)n=n.return;wt=n}function Fo(n){if(n!==wt)return!1;if(!$e)return Ep(n),$e=!0,!1;var a;if((a=n.tag!==3)&&!(a=n.tag!==5)&&(a=n.type,a=a!=="head"&&a!=="body"&&!ps(n.type,n.memoizedProps)),a&&(a=xt)){if(xs(n))throw jp(),Error(r(418));for(;a;)Pp(n,a),a=Nr(a.nextSibling)}if(Ep(n),n.tag===13){if(n=n.memoizedState,n=n!==null?n.dehydrated:null,!n)throw Error(r(317));e:{for(n=n.nextSibling,a=0;n;){if(n.nodeType===8){var o=n.data;if(o==="/$"){if(a===0){xt=Nr(n.nextSibling);break e}a--}else o!=="$"&&o!=="$!"&&o!=="$?"||a++}n=n.nextSibling}xt=null}}else xt=wt?Nr(n.stateNode.nextSibling):null;return!0}function jp(){for(var n=xt;n;)n=Nr(n.nextSibling)}function Wn(){xt=wt=null,$e=!1}function Ss(n){Bt===null?Bt=[n]:Bt.push(n)}var L0=k.ReactCurrentBatchConfig;function oa(n,a,o){if(n=o.ref,n!==null&&typeof n!="function"&&typeof n!="object"){if(o._owner){if(o=o._owner,o){if(o.tag!==1)throw Error(r(309));var u=o.stateNode}if(!u)throw Error(r(147,n));var f=u,y=""+n;return a!==null&&a.ref!==null&&typeof a.ref=="function"&&a.ref._stringRef===y?a.ref:(a=function(x){var T=f.refs;x===null?delete T[y]:T[y]=x},a._stringRef=y,a)}if(typeof n!="string")throw Error(r(284));if(!o._owner)throw Error(r(290,n))}return n}function Wo(n,a){throw n=Object.prototype.toString.call(a),Error(r(31,n==="[object Object]"?"object with keys {"+Object.keys(a).join(", ")+"}":n))}function _p(n){var a=n._init;return a(n._payload)}function kp(n){function a(L,N){if(n){var R=L.deletions;R===null?(L.deletions=[N],L.flags|=16):R.push(N)}}function o(L,N){if(!n)return null;for(;N!==null;)a(L,N),N=N.sibling;return null}function u(L,N){for(L=new Map;N!==null;)N.key!==null?L.set(N.key,N):L.set(N.index,N),N=N.sibling;return L}function f(L,N){return L=Hr(L,N),L.index=0,L.sibling=null,L}function y(L,N,R){return L.index=R,n?(R=L.alternate,R!==null?(R=R.index,R<N?(L.flags|=2,N):R):(L.flags|=2,N)):(L.flags|=1048576,N)}function x(L){return n&&L.alternate===null&&(L.flags|=2),L}function T(L,N,R,ne){return N===null||N.tag!==6?(N=vc(R,L.mode,ne),N.return=L,N):(N=f(N,R),N.return=L,N)}function I(L,N,R,ne){var ce=R.type;return ce===C?Z(L,N,R.props.children,ne,R.key):N!==null&&(N.elementType===ce||typeof ce=="object"&&ce!==null&&ce.$$typeof===G&&_p(ce)===N.type)?(ne=f(N,R.props),ne.ref=oa(L,N,R),ne.return=L,ne):(ne=dl(R.type,R.key,R.props,null,L.mode,ne),ne.ref=oa(L,N,R),ne.return=L,ne)}function B(L,N,R,ne){return N===null||N.tag!==4||N.stateNode.containerInfo!==R.containerInfo||N.stateNode.implementation!==R.implementation?(N=yc(R,L.mode,ne),N.return=L,N):(N=f(N,R.children||[]),N.return=L,N)}function Z(L,N,R,ne,ce){return N===null||N.tag!==7?(N=dn(R,L.mode,ne,ce),N.return=L,N):(N=f(N,R),N.return=L,N)}function J(L,N,R){if(typeof N=="string"&&N!==""||typeof N=="number")return N=vc(""+N,L.mode,R),N.return=L,N;if(typeof N=="object"&&N!==null){switch(N.$$typeof){case w:return R=dl(N.type,N.key,N.props,null,L.mode,R),R.ref=oa(L,null,N),R.return=L,R;case E:return N=yc(N,L.mode,R),N.return=L,N;case G:var ne=N._init;return J(L,ne(N._payload),R)}if(Jt(N)||X(N))return N=dn(N,L.mode,R,null),N.return=L,N;Wo(L,N)}return null}function Q(L,N,R,ne){var ce=N!==null?N.key:null;if(typeof R=="string"&&R!==""||typeof R=="number")return ce!==null?null:T(L,N,""+R,ne);if(typeof R=="object"&&R!==null){switch(R.$$typeof){case w:return R.key===ce?I(L,N,R,ne):null;case E:return R.key===ce?B(L,N,R,ne):null;case G:return ce=R._init,Q(L,N,ce(R._payload),ne)}if(Jt(R)||X(R))return ce!==null?null:Z(L,N,R,ne,null);Wo(L,R)}return null}function oe(L,N,R,ne,ce){if(typeof ne=="string"&&ne!==""||typeof ne=="number")return L=L.get(R)||null,T(N,L,""+ne,ce);if(typeof ne=="object"&&ne!==null){switch(ne.$$typeof){case w:return L=L.get(ne.key===null?R:ne.key)||null,I(N,L,ne,ce);case E:return L=L.get(ne.key===null?R:ne.key)||null,B(N,L,ne,ce);case G:var pe=ne._init;return oe(L,N,R,pe(ne._payload),ce)}if(Jt(ne)||X(ne))return L=L.get(R)||null,Z(N,L,ne,ce,null);Wo(N,ne)}return null}function ue(L,N,R,ne){for(var ce=null,pe=null,ve=N,me=N=0,qe=null;ve!==null&&me<R.length;me++){ve.index>me?(qe=ve,ve=null):qe=ve.sibling;var Oe=Q(L,ve,R[me],ne);if(Oe===null){ve===null&&(ve=qe);break}n&&ve&&Oe.alternate===null&&a(L,ve),N=y(Oe,N,me),pe===null?ce=Oe:pe.sibling=Oe,pe=Oe,ve=qe}if(me===R.length)return o(L,ve),$e&&nn(L,me),ce;if(ve===null){for(;me<R.length;me++)ve=J(L,R[me],ne),ve!==null&&(N=y(ve,N,me),pe===null?ce=ve:pe.sibling=ve,pe=ve);return $e&&nn(L,me),ce}for(ve=u(L,ve);me<R.length;me++)qe=oe(ve,L,me,R[me],ne),qe!==null&&(n&&qe.alternate!==null&&ve.delete(qe.key===null?me:qe.key),N=y(qe,N,me),pe===null?ce=qe:pe.sibling=qe,pe=qe);return n&&ve.forEach(function(Gr){return a(L,Gr)}),$e&&nn(L,me),ce}function se(L,N,R,ne){var ce=X(R);if(typeof ce!="function")throw Error(r(150));if(R=ce.call(R),R==null)throw Error(r(151));for(var pe=ce=null,ve=N,me=N=0,qe=null,Oe=R.next();ve!==null&&!Oe.done;me++,Oe=R.next()){ve.index>me?(qe=ve,ve=null):qe=ve.sibling;var Gr=Q(L,ve,Oe.value,ne);if(Gr===null){ve===null&&(ve=qe);break}n&&ve&&Gr.alternate===null&&a(L,ve),N=y(Gr,N,me),pe===null?ce=Gr:pe.sibling=Gr,pe=Gr,ve=qe}if(Oe.done)return o(L,ve),$e&&nn(L,me),ce;if(ve===null){for(;!Oe.done;me++,Oe=R.next())Oe=J(L,Oe.value,ne),Oe!==null&&(N=y(Oe,N,me),pe===null?ce=Oe:pe.sibling=Oe,pe=Oe);return $e&&nn(L,me),ce}for(ve=u(L,ve);!Oe.done;me++,Oe=R.next())Oe=oe(ve,L,me,Oe.value,ne),Oe!==null&&(n&&Oe.alternate!==null&&ve.delete(Oe.key===null?me:Oe.key),N=y(Oe,N,me),pe===null?ce=Oe:pe.sibling=Oe,pe=Oe);return n&&ve.forEach(function(hw){return a(L,hw)}),$e&&nn(L,me),ce}function ze(L,N,R,ne){if(typeof R=="object"&&R!==null&&R.type===C&&R.key===null&&(R=R.props.children),typeof R=="object"&&R!==null){switch(R.$$typeof){case w:e:{for(var ce=R.key,pe=N;pe!==null;){if(pe.key===ce){if(ce=R.type,ce===C){if(pe.tag===7){o(L,pe.sibling),N=f(pe,R.props.children),N.return=L,L=N;break e}}else if(pe.elementType===ce||typeof ce=="object"&&ce!==null&&ce.$$typeof===G&&_p(ce)===pe.type){o(L,pe.sibling),N=f(pe,R.props),N.ref=oa(L,pe,R),N.return=L,L=N;break e}o(L,pe);break}else a(L,pe);pe=pe.sibling}R.type===C?(N=dn(R.props.children,L.mode,ne,R.key),N.return=L,L=N):(ne=dl(R.type,R.key,R.props,null,L.mode,ne),ne.ref=oa(L,N,R),ne.return=L,L=ne)}return x(L);case E:e:{for(pe=R.key;N!==null;){if(N.key===pe)if(N.tag===4&&N.stateNode.containerInfo===R.containerInfo&&N.stateNode.implementation===R.implementation){o(L,N.sibling),N=f(N,R.children||[]),N.return=L,L=N;break e}else{o(L,N);break}else a(L,N);N=N.sibling}N=yc(R,L.mode,ne),N.return=L,L=N}return x(L);case G:return pe=R._init,ze(L,N,pe(R._payload),ne)}if(Jt(R))return ue(L,N,R,ne);if(X(R))return se(L,N,R,ne);Wo(L,R)}return typeof R=="string"&&R!==""||typeof R=="number"?(R=""+R,N!==null&&N.tag===6?(o(L,N.sibling),N=f(N,R),N.return=L,L=N):(o(L,N),N=vc(R,L.mode,ne),N.return=L,L=N),x(L)):o(L,N)}return ze}var Un=kp(!0),Tp=kp(!1),Uo=Dr(null),Vo=null,Vn=null,Ps=null;function As(){Ps=Vn=Vo=null}function Es(n){var a=Uo.current;Te(Uo),n._currentValue=a}function js(n,a,o){for(;n!==null;){var u=n.alternate;if((n.childLanes&a)!==a?(n.childLanes|=a,u!==null&&(u.childLanes|=a)):u!==null&&(u.childLanes&a)!==a&&(u.childLanes|=a),n===o)break;n=n.return}}function Kn(n,a){Vo=n,Ps=Vn=null,n=n.dependencies,n!==null&&n.firstContext!==null&&((n.lanes&a)!==0&&(yt=!0),n.firstContext=null)}function kt(n){var a=n._currentValue;if(Ps!==n)if(n={context:n,memoizedValue:a,next:null},Vn===null){if(Vo===null)throw Error(r(308));Vn=n,Vo.dependencies={lanes:0,firstContext:n}}else Vn=Vn.next=n;return a}var an=null;function _s(n){an===null?an=[n]:an.push(n)}function Cp(n,a,o,u){var f=a.interleaved;return f===null?(o.next=o,_s(a)):(o.next=f.next,f.next=o),a.interleaved=o,dr(n,u)}function dr(n,a){n.lanes|=a;var o=n.alternate;for(o!==null&&(o.lanes|=a),o=n,n=n.return;n!==null;)n.childLanes|=a,o=n.alternate,o!==null&&(o.childLanes|=a),o=n,n=n.return;return o.tag===3?o.stateNode:null}var Br=!1;function ks(n){n.updateQueue={baseState:n.memoizedState,firstBaseUpdate:null,lastBaseUpdate:null,shared:{pending:null,interleaved:null,lanes:0},effects:null}}function $p(n,a){n=n.updateQueue,a.updateQueue===n&&(a.updateQueue={baseState:n.baseState,firstBaseUpdate:n.firstBaseUpdate,lastBaseUpdate:n.lastBaseUpdate,shared:n.shared,effects:n.effects})}function pr(n,a){return{eventTime:n,lane:a,tag:0,payload:null,callback:null,next:null}}function zr(n,a,o){var u=n.updateQueue;if(u===null)return null;if(u=u.shared,(xe&2)!==0){var f=u.pending;return f===null?a.next=a:(a.next=f.next,f.next=a),u.pending=a,dr(n,o)}return f=u.interleaved,f===null?(a.next=a,_s(u)):(a.next=f.next,f.next=a),u.interleaved=a,dr(n,o)}function Ko(n,a,o){if(a=a.updateQueue,a!==null&&(a=a.shared,(o&4194240)!==0)){var u=a.lanes;u&=n.pendingLanes,o|=u,a.lanes=o,Uu(n,o)}}function Ip(n,a){var o=n.updateQueue,u=n.alternate;if(u!==null&&(u=u.updateQueue,o===u)){var f=null,y=null;if(o=o.firstBaseUpdate,o!==null){do{var x={eventTime:o.eventTime,lane:o.lane,tag:o.tag,payload:o.payload,callback:o.callback,next:null};y===null?f=y=x:y=y.next=x,o=o.next}while(o!==null);y===null?f=y=a:y=y.next=a}else f=y=a;o={baseState:u.baseState,firstBaseUpdate:f,lastBaseUpdate:y,shared:u.shared,effects:u.effects},n.updateQueue=o;return}n=o.lastBaseUpdate,n===null?o.firstBaseUpdate=a:n.next=a,o.lastBaseUpdate=a}function Ho(n,a,o,u){var f=n.updateQueue;Br=!1;var y=f.firstBaseUpdate,x=f.lastBaseUpdate,T=f.shared.pending;if(T!==null){f.shared.pending=null;var I=T,B=I.next;I.next=null,x===null?y=B:x.next=B,x=I;var Z=n.alternate;Z!==null&&(Z=Z.updateQueue,T=Z.lastBaseUpdate,T!==x&&(T===null?Z.firstBaseUpdate=B:T.next=B,Z.lastBaseUpdate=I))}if(y!==null){var J=f.baseState;x=0,Z=B=I=null,T=y;do{var Q=T.lane,oe=T.eventTime;if((u&Q)===Q){Z!==null&&(Z=Z.next={eventTime:oe,lane:0,tag:T.tag,payload:T.payload,callback:T.callback,next:null});e:{var ue=n,se=T;switch(Q=a,oe=o,se.tag){case 1:if(ue=se.payload,typeof ue=="function"){J=ue.call(oe,J,Q);break e}J=ue;break e;case 3:ue.flags=ue.flags&-65537|128;case 0:if(ue=se.payload,Q=typeof ue=="function"?ue.call(oe,J,Q):ue,Q==null)break e;J=K({},J,Q);break e;case 2:Br=!0}}T.callback!==null&&T.lane!==0&&(n.flags|=64,Q=f.effects,Q===null?f.effects=[T]:Q.push(T))}else oe={eventTime:oe,lane:Q,tag:T.tag,payload:T.payload,callback:T.callback,next:null},Z===null?(B=Z=oe,I=J):Z=Z.next=oe,x|=Q;if(T=T.next,T===null){if(T=f.shared.pending,T===null)break;Q=T,T=Q.next,Q.next=null,f.lastBaseUpdate=Q,f.shared.pending=null}}while(!0);if(Z===null&&(I=J),f.baseState=I,f.firstBaseUpdate=B,f.lastBaseUpdate=Z,a=f.shared.interleaved,a!==null){f=a;do x|=f.lane,f=f.next;while(f!==a)}else y===null&&(f.shared.lanes=0);un|=x,n.lanes=x,n.memoizedState=J}}function Mp(n,a,o){if(n=a.effects,a.effects=null,n!==null)for(a=0;a<n.length;a++){var u=n[a],f=u.callback;if(f!==null){if(u.callback=null,u=o,typeof f!="function")throw Error(r(191,f));f.call(u)}}}var la={},rr=Dr(la),ua=Dr(la),sa=Dr(la);function on(n){if(n===la)throw Error(r(174));return n}function Ts(n,a){switch(_e(sa,a),_e(ua,n),_e(rr,la),n=a.nodeType,n){case 9:case 11:a=(a=a.documentElement)?a.namespaceURI:Cu(null,"");break;default:n=n===8?a.parentNode:a,a=n.namespaceURI||null,n=n.tagName,a=Cu(a,n)}Te(rr),_e(rr,a)}function Hn(){Te(rr),Te(ua),Te(sa)}function Np(n){on(sa.current);var a=on(rr.current),o=Cu(a,n.type);a!==o&&(_e(ua,n),_e(rr,o))}function Cs(n){ua.current===n&&(Te(rr),Te(ua))}var Me=Dr(0);function Go(n){for(var a=n;a!==null;){if(a.tag===13){var o=a.memoizedState;if(o!==null&&(o=o.dehydrated,o===null||o.data==="$?"||o.data==="$!"))return a}else if(a.tag===19&&a.memoizedProps.revealOrder!==void 0){if((a.flags&128)!==0)return a}else if(a.child!==null){a.child.return=a,a=a.child;continue}if(a===n)break;for(;a.sibling===null;){if(a.return===null||a.return===n)return null;a=a.return}a.sibling.return=a.return,a=a.sibling}return null}var $s=[];function Is(){for(var n=0;n<$s.length;n++)$s[n]._workInProgressVersionPrimary=null;$s.length=0}var Xo=k.ReactCurrentDispatcher,Ms=k.ReactCurrentBatchConfig,ln=0,Ne=null,Ue=null,Xe=null,Yo=!1,ca=!1,fa=0,R0=0;function rt(){throw Error(r(321))}function Ns(n,a){if(a===null)return!1;for(var o=0;o<a.length&&o<n.length;o++)if(!Rt(n[o],a[o]))return!1;return!0}function Ds(n,a,o,u,f,y){if(ln=y,Ne=a,a.memoizedState=null,a.updateQueue=null,a.lanes=0,Xo.current=n===null||n.memoizedState===null?W0:U0,n=o(u,f),ca){y=0;do{if(ca=!1,fa=0,25<=y)throw Error(r(301));y+=1,Xe=Ue=null,a.updateQueue=null,Xo.current=V0,n=o(u,f)}while(ca)}if(Xo.current=Zo,a=Ue!==null&&Ue.next!==null,ln=0,Xe=Ue=Ne=null,Yo=!1,a)throw Error(r(300));return n}function Ls(){var n=fa!==0;return fa=0,n}function nr(){var n={memoizedState:null,baseState:null,baseQueue:null,queue:null,next:null};return Xe===null?Ne.memoizedState=Xe=n:Xe=Xe.next=n,Xe}function Tt(){if(Ue===null){var n=Ne.alternate;n=n!==null?n.memoizedState:null}else n=Ue.next;var a=Xe===null?Ne.memoizedState:Xe.next;if(a!==null)Xe=a,Ue=n;else{if(n===null)throw Error(r(310));Ue=n,n={memoizedState:Ue.memoizedState,baseState:Ue.baseState,baseQueue:Ue.baseQueue,queue:Ue.queue,next:null},Xe===null?Ne.memoizedState=Xe=n:Xe=Xe.next=n}return Xe}function da(n,a){return typeof a=="function"?a(n):a}function Rs(n){var a=Tt(),o=a.queue;if(o===null)throw Error(r(311));o.lastRenderedReducer=n;var u=Ue,f=u.baseQueue,y=o.pending;if(y!==null){if(f!==null){var x=f.next;f.next=y.next,y.next=x}u.baseQueue=f=y,o.pending=null}if(f!==null){y=f.next,u=u.baseState;var T=x=null,I=null,B=y;do{var Z=B.lane;if((ln&Z)===Z)I!==null&&(I=I.next={lane:0,action:B.action,hasEagerState:B.hasEagerState,eagerState:B.eagerState,next:null}),u=B.hasEagerState?B.eagerState:n(u,B.action);else{var J={lane:Z,action:B.action,hasEagerState:B.hasEagerState,eagerState:B.eagerState,next:null};I===null?(T=I=J,x=u):I=I.next=J,Ne.lanes|=Z,un|=Z}B=B.next}while(B!==null&&B!==y);I===null?x=u:I.next=T,Rt(u,a.memoizedState)||(yt=!0),a.memoizedState=u,a.baseState=x,a.baseQueue=I,o.lastRenderedState=u}if(n=o.interleaved,n!==null){f=n;do y=f.lane,Ne.lanes|=y,un|=y,f=f.next;while(f!==n)}else f===null&&(o.lanes=0);return[a.memoizedState,o.dispatch]}function Bs(n){var a=Tt(),o=a.queue;if(o===null)throw Error(r(311));o.lastRenderedReducer=n;var u=o.dispatch,f=o.pending,y=a.memoizedState;if(f!==null){o.pending=null;var x=f=f.next;do y=n(y,x.action),x=x.next;while(x!==f);Rt(y,a.memoizedState)||(yt=!0),a.memoizedState=y,a.baseQueue===null&&(a.baseState=y),o.lastRenderedState=y}return[y,u]}function Dp(){}function Lp(n,a){var o=Ne,u=Tt(),f=a(),y=!Rt(u.memoizedState,f);if(y&&(u.memoizedState=f,yt=!0),u=u.queue,zs(zp.bind(null,o,u,n),[n]),u.getSnapshot!==a||y||Xe!==null&&Xe.memoizedState.tag&1){if(o.flags|=2048,pa(9,Bp.bind(null,o,u,f,a),void 0,null),Ye===null)throw Error(r(349));(ln&30)!==0||Rp(o,a,f)}return f}function Rp(n,a,o){n.flags|=16384,n={getSnapshot:a,value:o},a=Ne.updateQueue,a===null?(a={lastEffect:null,stores:null},Ne.updateQueue=a,a.stores=[n]):(o=a.stores,o===null?a.stores=[n]:o.push(n))}function Bp(n,a,o,u){a.value=o,a.getSnapshot=u,Fp(a)&&Wp(n)}function zp(n,a,o){return o(function(){Fp(a)&&Wp(n)})}function Fp(n){var a=n.getSnapshot;n=n.value;try{var o=a();return!Rt(n,o)}catch{return!0}}function Wp(n){var a=dr(n,1);a!==null&&Ut(a,n,1,-1)}function Up(n){var a=nr();return typeof n=="function"&&(n=n()),a.memoizedState=a.baseState=n,n={pending:null,interleaved:null,lanes:0,dispatch:null,lastRenderedReducer:da,lastRenderedState:n},a.queue=n,n=n.dispatch=F0.bind(null,Ne,n),[a.memoizedState,n]}function pa(n,a,o,u){return n={tag:n,create:a,destroy:o,deps:u,next:null},a=Ne.updateQueue,a===null?(a={lastEffect:null,stores:null},Ne.updateQueue=a,a.lastEffect=n.next=n):(o=a.lastEffect,o===null?a.lastEffect=n.next=n:(u=o.next,o.next=n,n.next=u,a.lastEffect=n)),n}function Vp(){return Tt().memoizedState}function qo(n,a,o,u){var f=nr();Ne.flags|=n,f.memoizedState=pa(1|a,o,void 0,u===void 0?null:u)}function Qo(n,a,o,u){var f=Tt();u=u===void 0?null:u;var y=void 0;if(Ue!==null){var x=Ue.memoizedState;if(y=x.destroy,u!==null&&Ns(u,x.deps)){f.memoizedState=pa(a,o,y,u);return}}Ne.flags|=n,f.memoizedState=pa(1|a,o,y,u)}function Kp(n,a){return qo(8390656,8,n,a)}function zs(n,a){return Qo(2048,8,n,a)}function Hp(n,a){return Qo(4,2,n,a)}function Gp(n,a){return Qo(4,4,n,a)}function Xp(n,a){if(typeof a=="function")return n=n(),a(n),function(){a(null)};if(a!=null)return n=n(),a.current=n,function(){a.current=null}}function Yp(n,a,o){return o=o!=null?o.concat([n]):null,Qo(4,4,Xp.bind(null,a,n),o)}function Fs(){}function qp(n,a){var o=Tt();a=a===void 0?null:a;var u=o.memoizedState;return u!==null&&a!==null&&Ns(a,u[1])?u[0]:(o.memoizedState=[n,a],n)}function Qp(n,a){var o=Tt();a=a===void 0?null:a;var u=o.memoizedState;return u!==null&&a!==null&&Ns(a,u[1])?u[0]:(n=n(),o.memoizedState=[n,a],n)}function Zp(n,a,o){return(ln&21)===0?(n.baseState&&(n.baseState=!1,yt=!0),n.memoizedState=o):(Rt(o,a)||(o=_d(),Ne.lanes|=o,un|=o,n.baseState=!0),a)}function B0(n,a){var o=Ee;Ee=o!==0&&4>o?o:4,n(!0);var u=Ms.transition;Ms.transition={};try{n(!1),a()}finally{Ee=o,Ms.transition=u}}function Jp(){return Tt().memoizedState}function z0(n,a,o){var u=Vr(n);if(o={lane:u,action:o,hasEagerState:!1,eagerState:null,next:null},ev(n))tv(a,o);else if(o=Cp(n,a,o,u),o!==null){var f=st();Ut(o,n,u,f),rv(o,a,u)}}function F0(n,a,o){var u=Vr(n),f={lane:u,action:o,hasEagerState:!1,eagerState:null,next:null};if(ev(n))tv(a,f);else{var y=n.alternate;if(n.lanes===0&&(y===null||y.lanes===0)&&(y=a.lastRenderedReducer,y!==null))try{var x=a.lastRenderedState,T=y(x,o);if(f.hasEagerState=!0,f.eagerState=T,Rt(T,x)){var I=a.interleaved;I===null?(f.next=f,_s(a)):(f.next=I.next,I.next=f),a.interleaved=f;return}}catch{}finally{}o=Cp(n,a,f,u),o!==null&&(f=st(),Ut(o,n,u,f),rv(o,a,u))}}function ev(n){var a=n.alternate;return n===Ne||a!==null&&a===Ne}function tv(n,a){ca=Yo=!0;var o=n.pending;o===null?a.next=a:(a.next=o.next,o.next=a),n.pending=a}function rv(n,a,o){if((o&4194240)!==0){var u=a.lanes;u&=n.pendingLanes,o|=u,a.lanes=o,Uu(n,o)}}var Zo={readContext:kt,useCallback:rt,useContext:rt,useEffect:rt,useImperativeHandle:rt,useInsertionEffect:rt,useLayoutEffect:rt,useMemo:rt,useReducer:rt,useRef:rt,useState:rt,useDebugValue:rt,useDeferredValue:rt,useTransition:rt,useMutableSource:rt,useSyncExternalStore:rt,useId:rt,unstable_isNewReconciler:!1},W0={readContext:kt,useCallback:function(n,a){return nr().memoizedState=[n,a===void 0?null:a],n},useContext:kt,useEffect:Kp,useImperativeHandle:function(n,a,o){return o=o!=null?o.concat([n]):null,qo(4194308,4,Xp.bind(null,a,n),o)},useLayoutEffect:function(n,a){return qo(4194308,4,n,a)},useInsertionEffect:function(n,a){return qo(4,2,n,a)},useMemo:function(n,a){var o=nr();return a=a===void 0?null:a,n=n(),o.memoizedState=[n,a],n},useReducer:function(n,a,o){var u=nr();return a=o!==void 0?o(a):a,u.memoizedState=u.baseState=a,n={pending:null,interleaved:null,lanes:0,dispatch:null,lastRenderedReducer:n,lastRenderedState:a},u.queue=n,n=n.dispatch=z0.bind(null,Ne,n),[u.memoizedState,n]},useRef:function(n){var a=nr();return n={current:n},a.memoizedState=n},useState:Up,useDebugValue:Fs,useDeferredValue:function(n){return nr().memoizedState=n},useTransition:function(){var n=Up(!1),a=n[0];return n=B0.bind(null,n[1]),nr().memoizedState=n,[a,n]},useMutableSource:function(){},useSyncExternalStore:function(n,a,o){var u=Ne,f=nr();if($e){if(o===void 0)throw Error(r(407));o=o()}else{if(o=a(),Ye===null)throw Error(r(349));(ln&30)!==0||Rp(u,a,o)}f.memoizedState=o;var y={value:o,getSnapshot:a};return f.queue=y,Kp(zp.bind(null,u,y,n),[n]),u.flags|=2048,pa(9,Bp.bind(null,u,y,o,a),void 0,null),o},useId:function(){var n=nr(),a=Ye.identifierPrefix;if($e){var o=fr,u=cr;o=(u&~(1<<32-Lt(u)-1)).toString(32)+o,a=":"+a+"R"+o,o=fa++,0<o&&(a+="H"+o.toString(32)),a+=":"}else o=R0++,a=":"+a+"r"+o.toString(32)+":";return n.memoizedState=a},unstable_isNewReconciler:!1},U0={readContext:kt,useCallback:qp,useContext:kt,useEffect:zs,useImperativeHandle:Yp,useInsertionEffect:Hp,useLayoutEffect:Gp,useMemo:Qp,useReducer:Rs,useRef:Vp,useState:function(){return Rs(da)},useDebugValue:Fs,useDeferredValue:function(n){var a=Tt();return Zp(a,Ue.memoizedState,n)},useTransition:function(){var n=Rs(da)[0],a=Tt().memoizedState;return[n,a]},useMutableSource:Dp,useSyncExternalStore:Lp,useId:Jp,unstable_isNewReconciler:!1},V0={readContext:kt,useCallback:qp,useContext:kt,useEffect:zs,useImperativeHandle:Yp,useInsertionEffect:Hp,useLayoutEffect:Gp,useMemo:Qp,useReducer:Bs,useRef:Vp,useState:function(){return Bs(da)},useDebugValue:Fs,useDeferredValue:function(n){var a=Tt();return Ue===null?a.memoizedState=n:Zp(a,Ue.memoizedState,n)},useTransition:function(){var n=Bs(da)[0],a=Tt().memoizedState;return[n,a]},useMutableSource:Dp,useSyncExternalStore:Lp,useId:Jp,unstable_isNewReconciler:!1};function zt(n,a){if(n&&n.defaultProps){a=K({},a),n=n.defaultProps;for(var o in n)a[o]===void 0&&(a[o]=n[o]);return a}return a}function Ws(n,a,o,u){a=n.memoizedState,o=o(u,a),o=o==null?a:K({},a,o),n.memoizedState=o,n.lanes===0&&(n.updateQueue.baseState=o)}var Jo={isMounted:function(n){return(n=n._reactInternals)?Jr(n)===n:!1},enqueueSetState:function(n,a,o){n=n._reactInternals;var u=st(),f=Vr(n),y=pr(u,f);y.payload=a,o!=null&&(y.callback=o),a=zr(n,y,f),a!==null&&(Ut(a,n,f,u),Ko(a,n,f))},enqueueReplaceState:function(n,a,o){n=n._reactInternals;var u=st(),f=Vr(n),y=pr(u,f);y.tag=1,y.payload=a,o!=null&&(y.callback=o),a=zr(n,y,f),a!==null&&(Ut(a,n,f,u),Ko(a,n,f))},enqueueForceUpdate:function(n,a){n=n._reactInternals;var o=st(),u=Vr(n),f=pr(o,u);f.tag=2,a!=null&&(f.callback=a),a=zr(n,f,u),a!==null&&(Ut(a,n,u,o),Ko(a,n,u))}};function nv(n,a,o,u,f,y,x){return n=n.stateNode,typeof n.shouldComponentUpdate=="function"?n.shouldComponentUpdate(u,y,x):a.prototype&&a.prototype.isPureReactComponent?!Ji(o,u)||!Ji(f,y):!0}function iv(n,a,o){var u=!1,f=Lr,y=a.contextType;return typeof y=="object"&&y!==null?y=kt(y):(f=vt(a)?tn:tt.current,u=a.contextTypes,y=(u=u!=null)?Bn(n,f):Lr),a=new a(o,y),n.memoizedState=a.state!==null&&a.state!==void 0?a.state:null,a.updater=Jo,n.stateNode=a,a._reactInternals=n,u&&(n=n.stateNode,n.__reactInternalMemoizedUnmaskedChildContext=f,n.__reactInternalMemoizedMaskedChildContext=y),a}function av(n,a,o,u){n=a.state,typeof a.componentWillReceiveProps=="function"&&a.componentWillReceiveProps(o,u),typeof a.UNSAFE_componentWillReceiveProps=="function"&&a.UNSAFE_componentWillReceiveProps(o,u),a.state!==n&&Jo.enqueueReplaceState(a,a.state,null)}function Us(n,a,o,u){var f=n.stateNode;f.props=o,f.state=n.memoizedState,f.refs={},ks(n);var y=a.contextType;typeof y=="object"&&y!==null?f.context=kt(y):(y=vt(a)?tn:tt.current,f.context=Bn(n,y)),f.state=n.memoizedState,y=a.getDerivedStateFromProps,typeof y=="function"&&(Ws(n,a,y,o),f.state=n.memoizedState),typeof a.getDerivedStateFromProps=="function"||typeof f.getSnapshotBeforeUpdate=="function"||typeof f.UNSAFE_componentWillMount!="function"&&typeof f.componentWillMount!="function"||(a=f.state,typeof f.componentWillMount=="function"&&f.componentWillMount(),typeof f.UNSAFE_componentWillMount=="function"&&f.UNSAFE_componentWillMount(),a!==f.state&&Jo.enqueueReplaceState(f,f.state,null),Ho(n,o,f,u),f.state=n.memoizedState),typeof f.componentDidMount=="function"&&(n.flags|=4194308)}function Gn(n,a){try{var o="",u=a;do o+=Ce(u),u=u.return;while(u);var f=o}catch(y){f=`
Error generating stack: `+y.message+`
`+y.stack}return{value:n,source:a,stack:f,digest:null}}function Vs(n,a,o){return{value:n,source:null,stack:o??null,digest:a??null}}function Ks(n,a){try{console.error(a.value)}catch(o){setTimeout(function(){throw o})}}var K0=typeof WeakMap=="function"?WeakMap:Map;function ov(n,a,o){o=pr(-1,o),o.tag=3,o.payload={element:null};var u=a.value;return o.callback=function(){ol||(ol=!0,oc=u),Ks(n,a)},o}function lv(n,a,o){o=pr(-1,o),o.tag=3;var u=n.type.getDerivedStateFromError;if(typeof u=="function"){var f=a.value;o.payload=function(){return u(f)},o.callback=function(){Ks(n,a)}}var y=n.stateNode;return y!==null&&typeof y.componentDidCatch=="function"&&(o.callback=function(){Ks(n,a),typeof u!="function"&&(Wr===null?Wr=new Set([this]):Wr.add(this));var x=a.stack;this.componentDidCatch(a.value,{componentStack:x!==null?x:""})}),o}function uv(n,a,o){var u=n.pingCache;if(u===null){u=n.pingCache=new K0;var f=new Set;u.set(a,f)}else f=u.get(a),f===void 0&&(f=new Set,u.set(a,f));f.has(o)||(f.add(o),n=aw.bind(null,n,a,o),a.then(n,n))}function sv(n){do{var a;if((a=n.tag===13)&&(a=n.memoizedState,a=a!==null?a.dehydrated!==null:!0),a)return n;n=n.return}while(n!==null);return null}function cv(n,a,o,u,f){return(n.mode&1)===0?(n===a?n.flags|=65536:(n.flags|=128,o.flags|=131072,o.flags&=-52805,o.tag===1&&(o.alternate===null?o.tag=17:(a=pr(-1,1),a.tag=2,zr(o,a,1))),o.lanes|=1),n):(n.flags|=65536,n.lanes=f,n)}var H0=k.ReactCurrentOwner,yt=!1;function ut(n,a,o,u){a.child=n===null?Tp(a,null,o,u):Un(a,n.child,o,u)}function fv(n,a,o,u,f){o=o.render;var y=a.ref;return Kn(a,f),u=Ds(n,a,o,u,y,f),o=Ls(),n!==null&&!yt?(a.updateQueue=n.updateQueue,a.flags&=-2053,n.lanes&=~f,vr(n,a,f)):($e&&o&&bs(a),a.flags|=1,ut(n,a,u,f),a.child)}function dv(n,a,o,u,f){if(n===null){var y=o.type;return typeof y=="function"&&!pc(y)&&y.defaultProps===void 0&&o.compare===null&&o.defaultProps===void 0?(a.tag=15,a.type=y,pv(n,a,y,u,f)):(n=dl(o.type,null,u,a,a.mode,f),n.ref=a.ref,n.return=a,a.child=n)}if(y=n.child,(n.lanes&f)===0){var x=y.memoizedProps;if(o=o.compare,o=o!==null?o:Ji,o(x,u)&&n.ref===a.ref)return vr(n,a,f)}return a.flags|=1,n=Hr(y,u),n.ref=a.ref,n.return=a,a.child=n}function pv(n,a,o,u,f){if(n!==null){var y=n.memoizedProps;if(Ji(y,u)&&n.ref===a.ref)if(yt=!1,a.pendingProps=u=y,(n.lanes&f)!==0)(n.flags&131072)!==0&&(yt=!0);else return a.lanes=n.lanes,vr(n,a,f)}return Hs(n,a,o,u,f)}function vv(n,a,o){var u=a.pendingProps,f=u.children,y=n!==null?n.memoizedState:null;if(u.mode==="hidden")if((a.mode&1)===0)a.memoizedState={baseLanes:0,cachePool:null,transitions:null},_e(Yn,Ot),Ot|=o;else{if((o&1073741824)===0)return n=y!==null?y.baseLanes|o:o,a.lanes=a.childLanes=1073741824,a.memoizedState={baseLanes:n,cachePool:null,transitions:null},a.updateQueue=null,_e(Yn,Ot),Ot|=n,null;a.memoizedState={baseLanes:0,cachePool:null,transitions:null},u=y!==null?y.baseLanes:o,_e(Yn,Ot),Ot|=u}else y!==null?(u=y.baseLanes|o,a.memoizedState=null):u=o,_e(Yn,Ot),Ot|=u;return ut(n,a,f,o),a.child}function yv(n,a){var o=a.ref;(n===null&&o!==null||n!==null&&n.ref!==o)&&(a.flags|=512,a.flags|=2097152)}function Hs(n,a,o,u,f){var y=vt(o)?tn:tt.current;return y=Bn(a,y),Kn(a,f),o=Ds(n,a,o,u,y,f),u=Ls(),n!==null&&!yt?(a.updateQueue=n.updateQueue,a.flags&=-2053,n.lanes&=~f,vr(n,a,f)):($e&&u&&bs(a),a.flags|=1,ut(n,a,o,f),a.child)}function hv(n,a,o,u,f){if(vt(o)){var y=!0;Lo(a)}else y=!1;if(Kn(a,f),a.stateNode===null)tl(n,a),iv(a,o,u),Us(a,o,u,f),u=!0;else if(n===null){var x=a.stateNode,T=a.memoizedProps;x.props=T;var I=x.context,B=o.contextType;typeof B=="object"&&B!==null?B=kt(B):(B=vt(o)?tn:tt.current,B=Bn(a,B));var Z=o.getDerivedStateFromProps,J=typeof Z=="function"||typeof x.getSnapshotBeforeUpdate=="function";J||typeof x.UNSAFE_componentWillReceiveProps!="function"&&typeof x.componentWillReceiveProps!="function"||(T!==u||I!==B)&&av(a,x,u,B),Br=!1;var Q=a.memoizedState;x.state=Q,Ho(a,u,x,f),I=a.memoizedState,T!==u||Q!==I||pt.current||Br?(typeof Z=="function"&&(Ws(a,o,Z,u),I=a.memoizedState),(T=Br||nv(a,o,T,u,Q,I,B))?(J||typeof x.UNSAFE_componentWillMount!="function"&&typeof x.componentWillMount!="function"||(typeof x.componentWillMount=="function"&&x.componentWillMount(),typeof x.UNSAFE_componentWillMount=="function"&&x.UNSAFE_componentWillMount()),typeof x.componentDidMount=="function"&&(a.flags|=4194308)):(typeof x.componentDidMount=="function"&&(a.flags|=4194308),a.memoizedProps=u,a.memoizedState=I),x.props=u,x.state=I,x.context=B,u=T):(typeof x.componentDidMount=="function"&&(a.flags|=4194308),u=!1)}else{x=a.stateNode,$p(n,a),T=a.memoizedProps,B=a.type===a.elementType?T:zt(a.type,T),x.props=B,J=a.pendingProps,Q=x.context,I=o.contextType,typeof I=="object"&&I!==null?I=kt(I):(I=vt(o)?tn:tt.current,I=Bn(a,I));var oe=o.getDerivedStateFromProps;(Z=typeof oe=="function"||typeof x.getSnapshotBeforeUpdate=="function")||typeof x.UNSAFE_componentWillReceiveProps!="function"&&typeof x.componentWillReceiveProps!="function"||(T!==J||Q!==I)&&av(a,x,u,I),Br=!1,Q=a.memoizedState,x.state=Q,Ho(a,u,x,f);var ue=a.memoizedState;T!==J||Q!==ue||pt.current||Br?(typeof oe=="function"&&(Ws(a,o,oe,u),ue=a.memoizedState),(B=Br||nv(a,o,B,u,Q,ue,I)||!1)?(Z||typeof x.UNSAFE_componentWillUpdate!="function"&&typeof x.componentWillUpdate!="function"||(typeof x.componentWillUpdate=="function"&&x.componentWillUpdate(u,ue,I),typeof x.UNSAFE_componentWillUpdate=="function"&&x.UNSAFE_componentWillUpdate(u,ue,I)),typeof x.componentDidUpdate=="function"&&(a.flags|=4),typeof x.getSnapshotBeforeUpdate=="function"&&(a.flags|=1024)):(typeof x.componentDidUpdate!="function"||T===n.memoizedProps&&Q===n.memoizedState||(a.flags|=4),typeof x.getSnapshotBeforeUpdate!="function"||T===n.memoizedProps&&Q===n.memoizedState||(a.flags|=1024),a.memoizedProps=u,a.memoizedState=ue),x.props=u,x.state=ue,x.context=I,u=B):(typeof x.componentDidUpdate!="function"||T===n.memoizedProps&&Q===n.memoizedState||(a.flags|=4),typeof x.getSnapshotBeforeUpdate!="function"||T===n.memoizedProps&&Q===n.memoizedState||(a.flags|=1024),u=!1)}return Gs(n,a,o,u,y,f)}function Gs(n,a,o,u,f,y){yv(n,a);var x=(a.flags&128)!==0;if(!u&&!x)return f&&xp(a,o,!1),vr(n,a,y);u=a.stateNode,H0.current=a;var T=x&&typeof o.getDerivedStateFromError!="function"?null:u.render();return a.flags|=1,n!==null&&x?(a.child=Un(a,n.child,null,y),a.child=Un(a,null,T,y)):ut(n,a,T,y),a.memoizedState=u.state,f&&xp(a,o,!0),a.child}function mv(n){var a=n.stateNode;a.pendingContext?bp(n,a.pendingContext,a.pendingContext!==a.context):a.context&&bp(n,a.context,!1),Ts(n,a.containerInfo)}function gv(n,a,o,u,f){return Wn(),Ss(f),a.flags|=256,ut(n,a,o,u),a.child}var Xs={dehydrated:null,treeContext:null,retryLane:0};function Ys(n){return{baseLanes:n,cachePool:null,transitions:null}}function bv(n,a,o){var u=a.pendingProps,f=Me.current,y=!1,x=(a.flags&128)!==0,T;if((T=x)||(T=n!==null&&n.memoizedState===null?!1:(f&2)!==0),T?(y=!0,a.flags&=-129):(n===null||n.memoizedState!==null)&&(f|=1),_e(Me,f&1),n===null)return Os(a),n=a.memoizedState,n!==null&&(n=n.dehydrated,n!==null)?((a.mode&1)===0?a.lanes=1:n.data==="$!"?a.lanes=8:a.lanes=1073741824,null):(x=u.children,n=u.fallback,y?(u=a.mode,y=a.child,x={mode:"hidden",children:x},(u&1)===0&&y!==null?(y.childLanes=0,y.pendingProps=x):y=pl(x,u,0,null),n=dn(n,u,o,null),y.return=a,n.return=a,y.sibling=n,a.child=y,a.child.memoizedState=Ys(o),a.memoizedState=Xs,n):qs(a,x));if(f=n.memoizedState,f!==null&&(T=f.dehydrated,T!==null))return G0(n,a,x,u,T,f,o);if(y){y=u.fallback,x=a.mode,f=n.child,T=f.sibling;var I={mode:"hidden",children:u.children};return(x&1)===0&&a.child!==f?(u=a.child,u.childLanes=0,u.pendingProps=I,a.deletions=null):(u=Hr(f,I),u.subtreeFlags=f.subtreeFlags&14680064),T!==null?y=Hr(T,y):(y=dn(y,x,o,null),y.flags|=2),y.return=a,u.return=a,u.sibling=y,a.child=u,u=y,y=a.child,x=n.child.memoizedState,x=x===null?Ys(o):{baseLanes:x.baseLanes|o,cachePool:null,transitions:x.transitions},y.memoizedState=x,y.childLanes=n.childLanes&~o,a.memoizedState=Xs,u}return y=n.child,n=y.sibling,u=Hr(y,{mode:"visible",children:u.children}),(a.mode&1)===0&&(u.lanes=o),u.return=a,u.sibling=null,n!==null&&(o=a.deletions,o===null?(a.deletions=[n],a.flags|=16):o.push(n)),a.child=u,a.memoizedState=null,u}function qs(n,a){return a=pl({mode:"visible",children:a},n.mode,0,null),a.return=n,n.child=a}function el(n,a,o,u){return u!==null&&Ss(u),Un(a,n.child,null,o),n=qs(a,a.pendingProps.children),n.flags|=2,a.memoizedState=null,n}function G0(n,a,o,u,f,y,x){if(o)return a.flags&256?(a.flags&=-257,u=Vs(Error(r(422))),el(n,a,x,u)):a.memoizedState!==null?(a.child=n.child,a.flags|=128,null):(y=u.fallback,f=a.mode,u=pl({mode:"visible",children:u.children},f,0,null),y=dn(y,f,x,null),y.flags|=2,u.return=a,y.return=a,u.sibling=y,a.child=u,(a.mode&1)!==0&&Un(a,n.child,null,x),a.child.memoizedState=Ys(x),a.memoizedState=Xs,y);if((a.mode&1)===0)return el(n,a,x,null);if(f.data==="$!"){if(u=f.nextSibling&&f.nextSibling.dataset,u)var T=u.dgst;return u=T,y=Error(r(419)),u=Vs(y,u,void 0),el(n,a,x,u)}if(T=(x&n.childLanes)!==0,yt||T){if(u=Ye,u!==null){switch(x&-x){case 4:f=2;break;case 16:f=8;break;case 64:case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:case 4194304:case 8388608:case 16777216:case 33554432:case 67108864:f=32;break;case 536870912:f=268435456;break;default:f=0}f=(f&(u.suspendedLanes|x))!==0?0:f,f!==0&&f!==y.retryLane&&(y.retryLane=f,dr(n,f),Ut(u,n,f,-1))}return dc(),u=Vs(Error(r(421))),el(n,a,x,u)}return f.data==="$?"?(a.flags|=128,a.child=n.child,a=ow.bind(null,n),f._reactRetry=a,null):(n=y.treeContext,xt=Nr(f.nextSibling),wt=a,$e=!0,Bt=null,n!==null&&(jt[_t++]=cr,jt[_t++]=fr,jt[_t++]=rn,cr=n.id,fr=n.overflow,rn=a),a=qs(a,u.children),a.flags|=4096,a)}function wv(n,a,o){n.lanes|=a;var u=n.alternate;u!==null&&(u.lanes|=a),js(n.return,a,o)}function Qs(n,a,o,u,f){var y=n.memoizedState;y===null?n.memoizedState={isBackwards:a,rendering:null,renderingStartTime:0,last:u,tail:o,tailMode:f}:(y.isBackwards=a,y.rendering=null,y.renderingStartTime=0,y.last=u,y.tail=o,y.tailMode=f)}function xv(n,a,o){var u=a.pendingProps,f=u.revealOrder,y=u.tail;if(ut(n,a,u.children,o),u=Me.current,(u&2)!==0)u=u&1|2,a.flags|=128;else{if(n!==null&&(n.flags&128)!==0)e:for(n=a.child;n!==null;){if(n.tag===13)n.memoizedState!==null&&wv(n,o,a);else if(n.tag===19)wv(n,o,a);else if(n.child!==null){n.child.return=n,n=n.child;continue}if(n===a)break e;for(;n.sibling===null;){if(n.return===null||n.return===a)break e;n=n.return}n.sibling.return=n.return,n=n.sibling}u&=1}if(_e(Me,u),(a.mode&1)===0)a.memoizedState=null;else switch(f){case"forwards":for(o=a.child,f=null;o!==null;)n=o.alternate,n!==null&&Go(n)===null&&(f=o),o=o.sibling;o=f,o===null?(f=a.child,a.child=null):(f=o.sibling,o.sibling=null),Qs(a,!1,f,o,y);break;case"backwards":for(o=null,f=a.child,a.child=null;f!==null;){if(n=f.alternate,n!==null&&Go(n)===null){a.child=f;break}n=f.sibling,f.sibling=o,o=f,f=n}Qs(a,!0,o,null,y);break;case"together":Qs(a,!1,null,null,void 0);break;default:a.memoizedState=null}return a.child}function tl(n,a){(a.mode&1)===0&&n!==null&&(n.alternate=null,a.alternate=null,a.flags|=2)}function vr(n,a,o){if(n!==null&&(a.dependencies=n.dependencies),un|=a.lanes,(o&a.childLanes)===0)return null;if(n!==null&&a.child!==n.child)throw Error(r(153));if(a.child!==null){for(n=a.child,o=Hr(n,n.pendingProps),a.child=o,o.return=a;n.sibling!==null;)n=n.sibling,o=o.sibling=Hr(n,n.pendingProps),o.return=a;o.sibling=null}return a.child}function X0(n,a,o){switch(a.tag){case 3:mv(a),Wn();break;case 5:Np(a);break;case 1:vt(a.type)&&Lo(a);break;case 4:Ts(a,a.stateNode.containerInfo);break;case 10:var u=a.type._context,f=a.memoizedProps.value;_e(Uo,u._currentValue),u._currentValue=f;break;case 13:if(u=a.memoizedState,u!==null)return u.dehydrated!==null?(_e(Me,Me.current&1),a.flags|=128,null):(o&a.child.childLanes)!==0?bv(n,a,o):(_e(Me,Me.current&1),n=vr(n,a,o),n!==null?n.sibling:null);_e(Me,Me.current&1);break;case 19:if(u=(o&a.childLanes)!==0,(n.flags&128)!==0){if(u)return xv(n,a,o);a.flags|=128}if(f=a.memoizedState,f!==null&&(f.rendering=null,f.tail=null,f.lastEffect=null),_e(Me,Me.current),u)break;return null;case 22:case 23:return a.lanes=0,vv(n,a,o)}return vr(n,a,o)}var Ov,Zs,Sv,Pv;Ov=function(n,a){for(var o=a.child;o!==null;){if(o.tag===5||o.tag===6)n.appendChild(o.stateNode);else if(o.tag!==4&&o.child!==null){o.child.return=o,o=o.child;continue}if(o===a)break;for(;o.sibling===null;){if(o.return===null||o.return===a)return;o=o.return}o.sibling.return=o.return,o=o.sibling}},Zs=function(){},Sv=function(n,a,o,u){var f=n.memoizedProps;if(f!==u){n=a.stateNode,on(rr.current);var y=null;switch(o){case"input":f=qt(n,f),u=qt(n,u),y=[];break;case"select":f=K({},f,{value:void 0}),u=K({},u,{value:void 0}),y=[];break;case"textarea":f=Dt(n,f),u=Dt(n,u),y=[];break;default:typeof f.onClick!="function"&&typeof u.onClick=="function"&&(n.onclick=Mo)}$u(o,u);var x;o=null;for(B in f)if(!u.hasOwnProperty(B)&&f.hasOwnProperty(B)&&f[B]!=null)if(B==="style"){var T=f[B];for(x in T)T.hasOwnProperty(x)&&(o||(o={}),o[x]="")}else B!=="dangerouslySetInnerHTML"&&B!=="children"&&B!=="suppressContentEditableWarning"&&B!=="suppressHydrationWarning"&&B!=="autoFocus"&&(l.hasOwnProperty(B)?y||(y=[]):(y=y||[]).push(B,null));for(B in u){var I=u[B];if(T=f!=null?f[B]:void 0,u.hasOwnProperty(B)&&I!==T&&(I!=null||T!=null))if(B==="style")if(T){for(x in T)!T.hasOwnProperty(x)||I&&I.hasOwnProperty(x)||(o||(o={}),o[x]="");for(x in I)I.hasOwnProperty(x)&&T[x]!==I[x]&&(o||(o={}),o[x]=I[x])}else o||(y||(y=[]),y.push(B,o)),o=I;else B==="dangerouslySetInnerHTML"?(I=I?I.__html:void 0,T=T?T.__html:void 0,I!=null&&T!==I&&(y=y||[]).push(B,I)):B==="children"?typeof I!="string"&&typeof I!="number"||(y=y||[]).push(B,""+I):B!=="suppressContentEditableWarning"&&B!=="suppressHydrationWarning"&&(l.hasOwnProperty(B)?(I!=null&&B==="onScroll"&&ke("scroll",n),y||T===I||(y=[])):(y=y||[]).push(B,I))}o&&(y=y||[]).push("style",o);var B=y;(a.updateQueue=B)&&(a.flags|=4)}},Pv=function(n,a,o,u){o!==u&&(a.flags|=4)};function va(n,a){if(!$e)switch(n.tailMode){case"hidden":a=n.tail;for(var o=null;a!==null;)a.alternate!==null&&(o=a),a=a.sibling;o===null?n.tail=null:o.sibling=null;break;case"collapsed":o=n.tail;for(var u=null;o!==null;)o.alternate!==null&&(u=o),o=o.sibling;u===null?a||n.tail===null?n.tail=null:n.tail.sibling=null:u.sibling=null}}function nt(n){var a=n.alternate!==null&&n.alternate.child===n.child,o=0,u=0;if(a)for(var f=n.child;f!==null;)o|=f.lanes|f.childLanes,u|=f.subtreeFlags&14680064,u|=f.flags&14680064,f.return=n,f=f.sibling;else for(f=n.child;f!==null;)o|=f.lanes|f.childLanes,u|=f.subtreeFlags,u|=f.flags,f.return=n,f=f.sibling;return n.subtreeFlags|=u,n.childLanes=o,a}function Y0(n,a,o){var u=a.pendingProps;switch(ws(a),a.tag){case 2:case 16:case 15:case 0:case 11:case 7:case 8:case 12:case 9:case 14:return nt(a),null;case 1:return vt(a.type)&&Do(),nt(a),null;case 3:return u=a.stateNode,Hn(),Te(pt),Te(tt),Is(),u.pendingContext&&(u.context=u.pendingContext,u.pendingContext=null),(n===null||n.child===null)&&(Fo(a)?a.flags|=4:n===null||n.memoizedState.isDehydrated&&(a.flags&256)===0||(a.flags|=1024,Bt!==null&&(sc(Bt),Bt=null))),Zs(n,a),nt(a),null;case 5:Cs(a);var f=on(sa.current);if(o=a.type,n!==null&&a.stateNode!=null)Sv(n,a,o,u,f),n.ref!==a.ref&&(a.flags|=512,a.flags|=2097152);else{if(!u){if(a.stateNode===null)throw Error(r(166));return nt(a),null}if(n=on(rr.current),Fo(a)){u=a.stateNode,o=a.type;var y=a.memoizedProps;switch(u[tr]=a,u[ia]=y,n=(a.mode&1)!==0,o){case"dialog":ke("cancel",u),ke("close",u);break;case"iframe":case"object":case"embed":ke("load",u);break;case"video":case"audio":for(f=0;f<ta.length;f++)ke(ta[f],u);break;case"source":ke("error",u);break;case"img":case"image":case"link":ke("error",u),ke("load",u);break;case"details":ke("toggle",u);break;case"input":Er(u,y),ke("invalid",u);break;case"select":u._wrapperState={wasMultiple:!!y.multiple},ke("invalid",u);break;case"textarea":jn(u,y),ke("invalid",u)}$u(o,y),f=null;for(var x in y)if(y.hasOwnProperty(x)){var T=y[x];x==="children"?typeof T=="string"?u.textContent!==T&&(y.suppressHydrationWarning!==!0&&Io(u.textContent,T,n),f=["children",T]):typeof T=="number"&&u.textContent!==""+T&&(y.suppressHydrationWarning!==!0&&Io(u.textContent,T,n),f=["children",""+T]):l.hasOwnProperty(x)&&T!=null&&x==="onScroll"&&ke("scroll",u)}switch(o){case"input":Nt(u),_r(u,y,!0);break;case"textarea":Nt(u),Ni(u);break;case"select":case"option":break;default:typeof y.onClick=="function"&&(u.onclick=Mo)}u=f,a.updateQueue=u,u!==null&&(a.flags|=4)}else{x=f.nodeType===9?f:f.ownerDocument,n==="http://www.w3.org/1999/xhtml"&&(n=cd(o)),n==="http://www.w3.org/1999/xhtml"?o==="script"?(n=x.createElement("div"),n.innerHTML="<script><\/script>",n=n.removeChild(n.firstChild)):typeof u.is=="string"?n=x.createElement(o,{is:u.is}):(n=x.createElement(o),o==="select"&&(x=n,u.multiple?x.multiple=!0:u.size&&(x.size=u.size))):n=x.createElementNS(n,o),n[tr]=a,n[ia]=u,Ov(n,a,!1,!1),a.stateNode=n;e:{switch(x=Iu(o,u),o){case"dialog":ke("cancel",n),ke("close",n),f=u;break;case"iframe":case"object":case"embed":ke("load",n),f=u;break;case"video":case"audio":for(f=0;f<ta.length;f++)ke(ta[f],n);f=u;break;case"source":ke("error",n),f=u;break;case"img":case"image":case"link":ke("error",n),ke("load",n),f=u;break;case"details":ke("toggle",n),f=u;break;case"input":Er(n,u),f=qt(n,u),ke("invalid",n);break;case"option":f=u;break;case"select":n._wrapperState={wasMultiple:!!u.multiple},f=K({},u,{value:void 0}),ke("invalid",n);break;case"textarea":jn(n,u),f=Dt(n,u),ke("invalid",n);break;default:f=u}$u(o,f),T=f;for(y in T)if(T.hasOwnProperty(y)){var I=T[y];y==="style"?pd(n,I):y==="dangerouslySetInnerHTML"?(I=I?I.__html:void 0,I!=null&&fd(n,I)):y==="children"?typeof I=="string"?(o!=="textarea"||I!=="")&&Di(n,I):typeof I=="number"&&Di(n,""+I):y!=="suppressContentEditableWarning"&&y!=="suppressHydrationWarning"&&y!=="autoFocus"&&(l.hasOwnProperty(y)?I!=null&&y==="onScroll"&&ke("scroll",n):I!=null&&_(n,y,I,x))}switch(o){case"input":Nt(n),_r(n,u,!1);break;case"textarea":Nt(n),Ni(n);break;case"option":u.value!=null&&n.setAttribute("value",""+Re(u.value));break;case"select":n.multiple=!!u.multiple,y=u.value,y!=null?dt(n,!!u.multiple,y,!1):u.defaultValue!=null&&dt(n,!!u.multiple,u.defaultValue,!0);break;default:typeof f.onClick=="function"&&(n.onclick=Mo)}switch(o){case"button":case"input":case"select":case"textarea":u=!!u.autoFocus;break e;case"img":u=!0;break e;default:u=!1}}u&&(a.flags|=4)}a.ref!==null&&(a.flags|=512,a.flags|=2097152)}return nt(a),null;case 6:if(n&&a.stateNode!=null)Pv(n,a,n.memoizedProps,u);else{if(typeof u!="string"&&a.stateNode===null)throw Error(r(166));if(o=on(sa.current),on(rr.current),Fo(a)){if(u=a.stateNode,o=a.memoizedProps,u[tr]=a,(y=u.nodeValue!==o)&&(n=wt,n!==null))switch(n.tag){case 3:Io(u.nodeValue,o,(n.mode&1)!==0);break;case 5:n.memoizedProps.suppressHydrationWarning!==!0&&Io(u.nodeValue,o,(n.mode&1)!==0)}y&&(a.flags|=4)}else u=(o.nodeType===9?o:o.ownerDocument).createTextNode(u),u[tr]=a,a.stateNode=u}return nt(a),null;case 13:if(Te(Me),u=a.memoizedState,n===null||n.memoizedState!==null&&n.memoizedState.dehydrated!==null){if($e&&xt!==null&&(a.mode&1)!==0&&(a.flags&128)===0)jp(),Wn(),a.flags|=98560,y=!1;else if(y=Fo(a),u!==null&&u.dehydrated!==null){if(n===null){if(!y)throw Error(r(318));if(y=a.memoizedState,y=y!==null?y.dehydrated:null,!y)throw Error(r(317));y[tr]=a}else Wn(),(a.flags&128)===0&&(a.memoizedState=null),a.flags|=4;nt(a),y=!1}else Bt!==null&&(sc(Bt),Bt=null),y=!0;if(!y)return a.flags&65536?a:null}return(a.flags&128)!==0?(a.lanes=o,a):(u=u!==null,u!==(n!==null&&n.memoizedState!==null)&&u&&(a.child.flags|=8192,(a.mode&1)!==0&&(n===null||(Me.current&1)!==0?Ve===0&&(Ve=3):dc())),a.updateQueue!==null&&(a.flags|=4),nt(a),null);case 4:return Hn(),Zs(n,a),n===null&&ra(a.stateNode.containerInfo),nt(a),null;case 10:return Es(a.type._context),nt(a),null;case 17:return vt(a.type)&&Do(),nt(a),null;case 19:if(Te(Me),y=a.memoizedState,y===null)return nt(a),null;if(u=(a.flags&128)!==0,x=y.rendering,x===null)if(u)va(y,!1);else{if(Ve!==0||n!==null&&(n.flags&128)!==0)for(n=a.child;n!==null;){if(x=Go(n),x!==null){for(a.flags|=128,va(y,!1),u=x.updateQueue,u!==null&&(a.updateQueue=u,a.flags|=4),a.subtreeFlags=0,u=o,o=a.child;o!==null;)y=o,n=u,y.flags&=14680066,x=y.alternate,x===null?(y.childLanes=0,y.lanes=n,y.child=null,y.subtreeFlags=0,y.memoizedProps=null,y.memoizedState=null,y.updateQueue=null,y.dependencies=null,y.stateNode=null):(y.childLanes=x.childLanes,y.lanes=x.lanes,y.child=x.child,y.subtreeFlags=0,y.deletions=null,y.memoizedProps=x.memoizedProps,y.memoizedState=x.memoizedState,y.updateQueue=x.updateQueue,y.type=x.type,n=x.dependencies,y.dependencies=n===null?null:{lanes:n.lanes,firstContext:n.firstContext}),o=o.sibling;return _e(Me,Me.current&1|2),a.child}n=n.sibling}y.tail!==null&&Be()>qn&&(a.flags|=128,u=!0,va(y,!1),a.lanes=4194304)}else{if(!u)if(n=Go(x),n!==null){if(a.flags|=128,u=!0,o=n.updateQueue,o!==null&&(a.updateQueue=o,a.flags|=4),va(y,!0),y.tail===null&&y.tailMode==="hidden"&&!x.alternate&&!$e)return nt(a),null}else 2*Be()-y.renderingStartTime>qn&&o!==1073741824&&(a.flags|=128,u=!0,va(y,!1),a.lanes=4194304);y.isBackwards?(x.sibling=a.child,a.child=x):(o=y.last,o!==null?o.sibling=x:a.child=x,y.last=x)}return y.tail!==null?(a=y.tail,y.rendering=a,y.tail=a.sibling,y.renderingStartTime=Be(),a.sibling=null,o=Me.current,_e(Me,u?o&1|2:o&1),a):(nt(a),null);case 22:case 23:return fc(),u=a.memoizedState!==null,n!==null&&n.memoizedState!==null!==u&&(a.flags|=8192),u&&(a.mode&1)!==0?(Ot&1073741824)!==0&&(nt(a),a.subtreeFlags&6&&(a.flags|=8192)):nt(a),null;case 24:return null;case 25:return null}throw Error(r(156,a.tag))}function q0(n,a){switch(ws(a),a.tag){case 1:return vt(a.type)&&Do(),n=a.flags,n&65536?(a.flags=n&-65537|128,a):null;case 3:return Hn(),Te(pt),Te(tt),Is(),n=a.flags,(n&65536)!==0&&(n&128)===0?(a.flags=n&-65537|128,a):null;case 5:return Cs(a),null;case 13:if(Te(Me),n=a.memoizedState,n!==null&&n.dehydrated!==null){if(a.alternate===null)throw Error(r(340));Wn()}return n=a.flags,n&65536?(a.flags=n&-65537|128,a):null;case 19:return Te(Me),null;case 4:return Hn(),null;case 10:return Es(a.type._context),null;case 22:case 23:return fc(),null;case 24:return null;default:return null}}var rl=!1,it=!1,Q0=typeof WeakSet=="function"?WeakSet:Set,le=null;function Xn(n,a){var o=n.ref;if(o!==null)if(typeof o=="function")try{o(null)}catch(u){De(n,a,u)}else o.current=null}function Js(n,a,o){try{o()}catch(u){De(n,a,u)}}var Av=!1;function Z0(n,a){if(fs=Oo,n=np(),ns(n)){if("selectionStart"in n)var o={start:n.selectionStart,end:n.selectionEnd};else e:{o=(o=n.ownerDocument)&&o.defaultView||window;var u=o.getSelection&&o.getSelection();if(u&&u.rangeCount!==0){o=u.anchorNode;var f=u.anchorOffset,y=u.focusNode;u=u.focusOffset;try{o.nodeType,y.nodeType}catch{o=null;break e}var x=0,T=-1,I=-1,B=0,Z=0,J=n,Q=null;t:for(;;){for(var oe;J!==o||f!==0&&J.nodeType!==3||(T=x+f),J!==y||u!==0&&J.nodeType!==3||(I=x+u),J.nodeType===3&&(x+=J.nodeValue.length),(oe=J.firstChild)!==null;)Q=J,J=oe;for(;;){if(J===n)break t;if(Q===o&&++B===f&&(T=x),Q===y&&++Z===u&&(I=x),(oe=J.nextSibling)!==null)break;J=Q,Q=J.parentNode}J=oe}o=T===-1||I===-1?null:{start:T,end:I}}else o=null}o=o||{start:0,end:0}}else o=null;for(ds={focusedElem:n,selectionRange:o},Oo=!1,le=a;le!==null;)if(a=le,n=a.child,(a.subtreeFlags&1028)!==0&&n!==null)n.return=a,le=n;else for(;le!==null;){a=le;try{var ue=a.alternate;if((a.flags&1024)!==0)switch(a.tag){case 0:case 11:case 15:break;case 1:if(ue!==null){var se=ue.memoizedProps,ze=ue.memoizedState,L=a.stateNode,N=L.getSnapshotBeforeUpdate(a.elementType===a.type?se:zt(a.type,se),ze);L.__reactInternalSnapshotBeforeUpdate=N}break;case 3:var R=a.stateNode.containerInfo;R.nodeType===1?R.textContent="":R.nodeType===9&&R.documentElement&&R.removeChild(R.documentElement);break;case 5:case 6:case 4:case 17:break;default:throw Error(r(163))}}catch(ne){De(a,a.return,ne)}if(n=a.sibling,n!==null){n.return=a.return,le=n;break}le=a.return}return ue=Av,Av=!1,ue}function ya(n,a,o){var u=a.updateQueue;if(u=u!==null?u.lastEffect:null,u!==null){var f=u=u.next;do{if((f.tag&n)===n){var y=f.destroy;f.destroy=void 0,y!==void 0&&Js(a,o,y)}f=f.next}while(f!==u)}}function nl(n,a){if(a=a.updateQueue,a=a!==null?a.lastEffect:null,a!==null){var o=a=a.next;do{if((o.tag&n)===n){var u=o.create;o.destroy=u()}o=o.next}while(o!==a)}}function ec(n){var a=n.ref;if(a!==null){var o=n.stateNode;switch(n.tag){case 5:n=o;break;default:n=o}typeof a=="function"?a(n):a.current=n}}function Ev(n){var a=n.alternate;a!==null&&(n.alternate=null,Ev(a)),n.child=null,n.deletions=null,n.sibling=null,n.tag===5&&(a=n.stateNode,a!==null&&(delete a[tr],delete a[ia],delete a[hs],delete a[M0],delete a[N0])),n.stateNode=null,n.return=null,n.dependencies=null,n.memoizedProps=null,n.memoizedState=null,n.pendingProps=null,n.stateNode=null,n.updateQueue=null}function jv(n){return n.tag===5||n.tag===3||n.tag===4}function _v(n){e:for(;;){for(;n.sibling===null;){if(n.return===null||jv(n.return))return null;n=n.return}for(n.sibling.return=n.return,n=n.sibling;n.tag!==5&&n.tag!==6&&n.tag!==18;){if(n.flags&2||n.child===null||n.tag===4)continue e;n.child.return=n,n=n.child}if(!(n.flags&2))return n.stateNode}}function tc(n,a,o){var u=n.tag;if(u===5||u===6)n=n.stateNode,a?o.nodeType===8?o.parentNode.insertBefore(n,a):o.insertBefore(n,a):(o.nodeType===8?(a=o.parentNode,a.insertBefore(n,o)):(a=o,a.appendChild(n)),o=o._reactRootContainer,o!=null||a.onclick!==null||(a.onclick=Mo));else if(u!==4&&(n=n.child,n!==null))for(tc(n,a,o),n=n.sibling;n!==null;)tc(n,a,o),n=n.sibling}function rc(n,a,o){var u=n.tag;if(u===5||u===6)n=n.stateNode,a?o.insertBefore(n,a):o.appendChild(n);else if(u!==4&&(n=n.child,n!==null))for(rc(n,a,o),n=n.sibling;n!==null;)rc(n,a,o),n=n.sibling}var Qe=null,Ft=!1;function Fr(n,a,o){for(o=o.child;o!==null;)kv(n,a,o),o=o.sibling}function kv(n,a,o){if(er&&typeof er.onCommitFiberUnmount=="function")try{er.onCommitFiberUnmount(ho,o)}catch{}switch(o.tag){case 5:it||Xn(o,a);case 6:var u=Qe,f=Ft;Qe=null,Fr(n,a,o),Qe=u,Ft=f,Qe!==null&&(Ft?(n=Qe,o=o.stateNode,n.nodeType===8?n.parentNode.removeChild(o):n.removeChild(o)):Qe.removeChild(o.stateNode));break;case 18:Qe!==null&&(Ft?(n=Qe,o=o.stateNode,n.nodeType===8?ys(n.parentNode,o):n.nodeType===1&&ys(n,o),Gi(n)):ys(Qe,o.stateNode));break;case 4:u=Qe,f=Ft,Qe=o.stateNode.containerInfo,Ft=!0,Fr(n,a,o),Qe=u,Ft=f;break;case 0:case 11:case 14:case 15:if(!it&&(u=o.updateQueue,u!==null&&(u=u.lastEffect,u!==null))){f=u=u.next;do{var y=f,x=y.destroy;y=y.tag,x!==void 0&&((y&2)!==0||(y&4)!==0)&&Js(o,a,x),f=f.next}while(f!==u)}Fr(n,a,o);break;case 1:if(!it&&(Xn(o,a),u=o.stateNode,typeof u.componentWillUnmount=="function"))try{u.props=o.memoizedProps,u.state=o.memoizedState,u.componentWillUnmount()}catch(T){De(o,a,T)}Fr(n,a,o);break;case 21:Fr(n,a,o);break;case 22:o.mode&1?(it=(u=it)||o.memoizedState!==null,Fr(n,a,o),it=u):Fr(n,a,o);break;default:Fr(n,a,o)}}function Tv(n){var a=n.updateQueue;if(a!==null){n.updateQueue=null;var o=n.stateNode;o===null&&(o=n.stateNode=new Q0),a.forEach(function(u){var f=lw.bind(null,n,u);o.has(u)||(o.add(u),u.then(f,f))})}}function Wt(n,a){var o=a.deletions;if(o!==null)for(var u=0;u<o.length;u++){var f=o[u];try{var y=n,x=a,T=x;e:for(;T!==null;){switch(T.tag){case 5:Qe=T.stateNode,Ft=!1;break e;case 3:Qe=T.stateNode.containerInfo,Ft=!0;break e;case 4:Qe=T.stateNode.containerInfo,Ft=!0;break e}T=T.return}if(Qe===null)throw Error(r(160));kv(y,x,f),Qe=null,Ft=!1;var I=f.alternate;I!==null&&(I.return=null),f.return=null}catch(B){De(f,a,B)}}if(a.subtreeFlags&12854)for(a=a.child;a!==null;)Cv(a,n),a=a.sibling}function Cv(n,a){var o=n.alternate,u=n.flags;switch(n.tag){case 0:case 11:case 14:case 15:if(Wt(a,n),ir(n),u&4){try{ya(3,n,n.return),nl(3,n)}catch(se){De(n,n.return,se)}try{ya(5,n,n.return)}catch(se){De(n,n.return,se)}}break;case 1:Wt(a,n),ir(n),u&512&&o!==null&&Xn(o,o.return);break;case 5:if(Wt(a,n),ir(n),u&512&&o!==null&&Xn(o,o.return),n.flags&32){var f=n.stateNode;try{Di(f,"")}catch(se){De(n,n.return,se)}}if(u&4&&(f=n.stateNode,f!=null)){var y=n.memoizedProps,x=o!==null?o.memoizedProps:y,T=n.type,I=n.updateQueue;if(n.updateQueue=null,I!==null)try{T==="input"&&y.type==="radio"&&y.name!=null&&jr(f,y),Iu(T,x);var B=Iu(T,y);for(x=0;x<I.length;x+=2){var Z=I[x],J=I[x+1];Z==="style"?pd(f,J):Z==="dangerouslySetInnerHTML"?fd(f,J):Z==="children"?Di(f,J):_(f,Z,J,B)}switch(T){case"input":Qt(f,y);break;case"textarea":Mi(f,y);break;case"select":var Q=f._wrapperState.wasMultiple;f._wrapperState.wasMultiple=!!y.multiple;var oe=y.value;oe!=null?dt(f,!!y.multiple,oe,!1):Q!==!!y.multiple&&(y.defaultValue!=null?dt(f,!!y.multiple,y.defaultValue,!0):dt(f,!!y.multiple,y.multiple?[]:"",!1))}f[ia]=y}catch(se){De(n,n.return,se)}}break;case 6:if(Wt(a,n),ir(n),u&4){if(n.stateNode===null)throw Error(r(162));f=n.stateNode,y=n.memoizedProps;try{f.nodeValue=y}catch(se){De(n,n.return,se)}}break;case 3:if(Wt(a,n),ir(n),u&4&&o!==null&&o.memoizedState.isDehydrated)try{Gi(a.containerInfo)}catch(se){De(n,n.return,se)}break;case 4:Wt(a,n),ir(n);break;case 13:Wt(a,n),ir(n),f=n.child,f.flags&8192&&(y=f.memoizedState!==null,f.stateNode.isHidden=y,!y||f.alternate!==null&&f.alternate.memoizedState!==null||(ac=Be())),u&4&&Tv(n);break;case 22:if(Z=o!==null&&o.memoizedState!==null,n.mode&1?(it=(B=it)||Z,Wt(a,n),it=B):Wt(a,n),ir(n),u&8192){if(B=n.memoizedState!==null,(n.stateNode.isHidden=B)&&!Z&&(n.mode&1)!==0)for(le=n,Z=n.child;Z!==null;){for(J=le=Z;le!==null;){switch(Q=le,oe=Q.child,Q.tag){case 0:case 11:case 14:case 15:ya(4,Q,Q.return);break;case 1:Xn(Q,Q.return);var ue=Q.stateNode;if(typeof ue.componentWillUnmount=="function"){u=Q,o=Q.return;try{a=u,ue.props=a.memoizedProps,ue.state=a.memoizedState,ue.componentWillUnmount()}catch(se){De(u,o,se)}}break;case 5:Xn(Q,Q.return);break;case 22:if(Q.memoizedState!==null){Mv(J);continue}}oe!==null?(oe.return=Q,le=oe):Mv(J)}Z=Z.sibling}e:for(Z=null,J=n;;){if(J.tag===5){if(Z===null){Z=J;try{f=J.stateNode,B?(y=f.style,typeof y.setProperty=="function"?y.setProperty("display","none","important"):y.display="none"):(T=J.stateNode,I=J.memoizedProps.style,x=I!=null&&I.hasOwnProperty("display")?I.display:null,T.style.display=dd("display",x))}catch(se){De(n,n.return,se)}}}else if(J.tag===6){if(Z===null)try{J.stateNode.nodeValue=B?"":J.memoizedProps}catch(se){De(n,n.return,se)}}else if((J.tag!==22&&J.tag!==23||J.memoizedState===null||J===n)&&J.child!==null){J.child.return=J,J=J.child;continue}if(J===n)break e;for(;J.sibling===null;){if(J.return===null||J.return===n)break e;Z===J&&(Z=null),J=J.return}Z===J&&(Z=null),J.sibling.return=J.return,J=J.sibling}}break;case 19:Wt(a,n),ir(n),u&4&&Tv(n);break;case 21:break;default:Wt(a,n),ir(n)}}function ir(n){var a=n.flags;if(a&2){try{e:{for(var o=n.return;o!==null;){if(jv(o)){var u=o;break e}o=o.return}throw Error(r(160))}switch(u.tag){case 5:var f=u.stateNode;u.flags&32&&(Di(f,""),u.flags&=-33);var y=_v(n);rc(n,y,f);break;case 3:case 4:var x=u.stateNode.containerInfo,T=_v(n);tc(n,T,x);break;default:throw Error(r(161))}}catch(I){De(n,n.return,I)}n.flags&=-3}a&4096&&(n.flags&=-4097)}function J0(n,a,o){le=n,$v(n)}function $v(n,a,o){for(var u=(n.mode&1)!==0;le!==null;){var f=le,y=f.child;if(f.tag===22&&u){var x=f.memoizedState!==null||rl;if(!x){var T=f.alternate,I=T!==null&&T.memoizedState!==null||it;T=rl;var B=it;if(rl=x,(it=I)&&!B)for(le=f;le!==null;)x=le,I=x.child,x.tag===22&&x.memoizedState!==null?Nv(f):I!==null?(I.return=x,le=I):Nv(f);for(;y!==null;)le=y,$v(y),y=y.sibling;le=f,rl=T,it=B}Iv(n)}else(f.subtreeFlags&8772)!==0&&y!==null?(y.return=f,le=y):Iv(n)}}function Iv(n){for(;le!==null;){var a=le;if((a.flags&8772)!==0){var o=a.alternate;try{if((a.flags&8772)!==0)switch(a.tag){case 0:case 11:case 15:it||nl(5,a);break;case 1:var u=a.stateNode;if(a.flags&4&&!it)if(o===null)u.componentDidMount();else{var f=a.elementType===a.type?o.memoizedProps:zt(a.type,o.memoizedProps);u.componentDidUpdate(f,o.memoizedState,u.__reactInternalSnapshotBeforeUpdate)}var y=a.updateQueue;y!==null&&Mp(a,y,u);break;case 3:var x=a.updateQueue;if(x!==null){if(o=null,a.child!==null)switch(a.child.tag){case 5:o=a.child.stateNode;break;case 1:o=a.child.stateNode}Mp(a,x,o)}break;case 5:var T=a.stateNode;if(o===null&&a.flags&4){o=T;var I=a.memoizedProps;switch(a.type){case"button":case"input":case"select":case"textarea":I.autoFocus&&o.focus();break;case"img":I.src&&(o.src=I.src)}}break;case 6:break;case 4:break;case 12:break;case 13:if(a.memoizedState===null){var B=a.alternate;if(B!==null){var Z=B.memoizedState;if(Z!==null){var J=Z.dehydrated;J!==null&&Gi(J)}}}break;case 19:case 17:case 21:case 22:case 23:case 25:break;default:throw Error(r(163))}it||a.flags&512&&ec(a)}catch(Q){De(a,a.return,Q)}}if(a===n){le=null;break}if(o=a.sibling,o!==null){o.return=a.return,le=o;break}le=a.return}}function Mv(n){for(;le!==null;){var a=le;if(a===n){le=null;break}var o=a.sibling;if(o!==null){o.return=a.return,le=o;break}le=a.return}}function Nv(n){for(;le!==null;){var a=le;try{switch(a.tag){case 0:case 11:case 15:var o=a.return;try{nl(4,a)}catch(I){De(a,o,I)}break;case 1:var u=a.stateNode;if(typeof u.componentDidMount=="function"){var f=a.return;try{u.componentDidMount()}catch(I){De(a,f,I)}}var y=a.return;try{ec(a)}catch(I){De(a,y,I)}break;case 5:var x=a.return;try{ec(a)}catch(I){De(a,x,I)}}}catch(I){De(a,a.return,I)}if(a===n){le=null;break}var T=a.sibling;if(T!==null){T.return=a.return,le=T;break}le=a.return}}var ew=Math.ceil,il=k.ReactCurrentDispatcher,nc=k.ReactCurrentOwner,Ct=k.ReactCurrentBatchConfig,xe=0,Ye=null,Fe=null,Ze=0,Ot=0,Yn=Dr(0),Ve=0,ha=null,un=0,al=0,ic=0,ma=null,ht=null,ac=0,qn=1/0,yr=null,ol=!1,oc=null,Wr=null,ll=!1,Ur=null,ul=0,ga=0,lc=null,sl=-1,cl=0;function st(){return(xe&6)!==0?Be():sl!==-1?sl:sl=Be()}function Vr(n){return(n.mode&1)===0?1:(xe&2)!==0&&Ze!==0?Ze&-Ze:L0.transition!==null?(cl===0&&(cl=_d()),cl):(n=Ee,n!==0||(n=window.event,n=n===void 0?16:Ld(n.type)),n)}function Ut(n,a,o,u){if(50<ga)throw ga=0,lc=null,Error(r(185));Wi(n,o,u),((xe&2)===0||n!==Ye)&&(n===Ye&&((xe&2)===0&&(al|=o),Ve===4&&Kr(n,Ze)),mt(n,u),o===1&&xe===0&&(a.mode&1)===0&&(qn=Be()+500,Ro&&Rr()))}function mt(n,a){var o=n.callbackNode;Lb(n,a);var u=bo(n,n===Ye?Ze:0);if(u===0)o!==null&&Ad(o),n.callbackNode=null,n.callbackPriority=0;else if(a=u&-u,n.callbackPriority!==a){if(o!=null&&Ad(o),a===1)n.tag===0?D0(Lv.bind(null,n)):Op(Lv.bind(null,n)),$0(function(){(xe&6)===0&&Rr()}),o=null;else{switch(kd(u)){case 1:o=zu;break;case 4:o=Ed;break;case 16:o=yo;break;case 536870912:o=jd;break;default:o=yo}o=Kv(o,Dv.bind(null,n))}n.callbackPriority=a,n.callbackNode=o}}function Dv(n,a){if(sl=-1,cl=0,(xe&6)!==0)throw Error(r(327));var o=n.callbackNode;if(Qn()&&n.callbackNode!==o)return null;var u=bo(n,n===Ye?Ze:0);if(u===0)return null;if((u&30)!==0||(u&n.expiredLanes)!==0||a)a=fl(n,u);else{a=u;var f=xe;xe|=2;var y=Bv();(Ye!==n||Ze!==a)&&(yr=null,qn=Be()+500,cn(n,a));do try{nw();break}catch(T){Rv(n,T)}while(!0);As(),il.current=y,xe=f,Fe!==null?a=0:(Ye=null,Ze=0,a=Ve)}if(a!==0){if(a===2&&(f=Fu(n),f!==0&&(u=f,a=uc(n,f))),a===1)throw o=ha,cn(n,0),Kr(n,u),mt(n,Be()),o;if(a===6)Kr(n,u);else{if(f=n.current.alternate,(u&30)===0&&!tw(f)&&(a=fl(n,u),a===2&&(y=Fu(n),y!==0&&(u=y,a=uc(n,y))),a===1))throw o=ha,cn(n,0),Kr(n,u),mt(n,Be()),o;switch(n.finishedWork=f,n.finishedLanes=u,a){case 0:case 1:throw Error(r(345));case 2:fn(n,ht,yr);break;case 3:if(Kr(n,u),(u&130023424)===u&&(a=ac+500-Be(),10<a)){if(bo(n,0)!==0)break;if(f=n.suspendedLanes,(f&u)!==u){st(),n.pingedLanes|=n.suspendedLanes&f;break}n.timeoutHandle=vs(fn.bind(null,n,ht,yr),a);break}fn(n,ht,yr);break;case 4:if(Kr(n,u),(u&4194240)===u)break;for(a=n.eventTimes,f=-1;0<u;){var x=31-Lt(u);y=1<<x,x=a[x],x>f&&(f=x),u&=~y}if(u=f,u=Be()-u,u=(120>u?120:480>u?480:1080>u?1080:1920>u?1920:3e3>u?3e3:4320>u?4320:1960*ew(u/1960))-u,10<u){n.timeoutHandle=vs(fn.bind(null,n,ht,yr),u);break}fn(n,ht,yr);break;case 5:fn(n,ht,yr);break;default:throw Error(r(329))}}}return mt(n,Be()),n.callbackNode===o?Dv.bind(null,n):null}function uc(n,a){var o=ma;return n.current.memoizedState.isDehydrated&&(cn(n,a).flags|=256),n=fl(n,a),n!==2&&(a=ht,ht=o,a!==null&&sc(a)),n}function sc(n){ht===null?ht=n:ht.push.apply(ht,n)}function tw(n){for(var a=n;;){if(a.flags&16384){var o=a.updateQueue;if(o!==null&&(o=o.stores,o!==null))for(var u=0;u<o.length;u++){var f=o[u],y=f.getSnapshot;f=f.value;try{if(!Rt(y(),f))return!1}catch{return!1}}}if(o=a.child,a.subtreeFlags&16384&&o!==null)o.return=a,a=o;else{if(a===n)break;for(;a.sibling===null;){if(a.return===null||a.return===n)return!0;a=a.return}a.sibling.return=a.return,a=a.sibling}}return!0}function Kr(n,a){for(a&=~ic,a&=~al,n.suspendedLanes|=a,n.pingedLanes&=~a,n=n.expirationTimes;0<a;){var o=31-Lt(a),u=1<<o;n[o]=-1,a&=~u}}function Lv(n){if((xe&6)!==0)throw Error(r(327));Qn();var a=bo(n,0);if((a&1)===0)return mt(n,Be()),null;var o=fl(n,a);if(n.tag!==0&&o===2){var u=Fu(n);u!==0&&(a=u,o=uc(n,u))}if(o===1)throw o=ha,cn(n,0),Kr(n,a),mt(n,Be()),o;if(o===6)throw Error(r(345));return n.finishedWork=n.current.alternate,n.finishedLanes=a,fn(n,ht,yr),mt(n,Be()),null}function cc(n,a){var o=xe;xe|=1;try{return n(a)}finally{xe=o,xe===0&&(qn=Be()+500,Ro&&Rr())}}function sn(n){Ur!==null&&Ur.tag===0&&(xe&6)===0&&Qn();var a=xe;xe|=1;var o=Ct.transition,u=Ee;try{if(Ct.transition=null,Ee=1,n)return n()}finally{Ee=u,Ct.transition=o,xe=a,(xe&6)===0&&Rr()}}function fc(){Ot=Yn.current,Te(Yn)}function cn(n,a){n.finishedWork=null,n.finishedLanes=0;var o=n.timeoutHandle;if(o!==-1&&(n.timeoutHandle=-1,C0(o)),Fe!==null)for(o=Fe.return;o!==null;){var u=o;switch(ws(u),u.tag){case 1:u=u.type.childContextTypes,u!=null&&Do();break;case 3:Hn(),Te(pt),Te(tt),Is();break;case 5:Cs(u);break;case 4:Hn();break;case 13:Te(Me);break;case 19:Te(Me);break;case 10:Es(u.type._context);break;case 22:case 23:fc()}o=o.return}if(Ye=n,Fe=n=Hr(n.current,null),Ze=Ot=a,Ve=0,ha=null,ic=al=un=0,ht=ma=null,an!==null){for(a=0;a<an.length;a++)if(o=an[a],u=o.interleaved,u!==null){o.interleaved=null;var f=u.next,y=o.pending;if(y!==null){var x=y.next;y.next=f,u.next=x}o.pending=u}an=null}return n}function Rv(n,a){do{var o=Fe;try{if(As(),Xo.current=Zo,Yo){for(var u=Ne.memoizedState;u!==null;){var f=u.queue;f!==null&&(f.pending=null),u=u.next}Yo=!1}if(ln=0,Xe=Ue=Ne=null,ca=!1,fa=0,nc.current=null,o===null||o.return===null){Ve=1,ha=a,Fe=null;break}e:{var y=n,x=o.return,T=o,I=a;if(a=Ze,T.flags|=32768,I!==null&&typeof I=="object"&&typeof I.then=="function"){var B=I,Z=T,J=Z.tag;if((Z.mode&1)===0&&(J===0||J===11||J===15)){var Q=Z.alternate;Q?(Z.updateQueue=Q.updateQueue,Z.memoizedState=Q.memoizedState,Z.lanes=Q.lanes):(Z.updateQueue=null,Z.memoizedState=null)}var oe=sv(x);if(oe!==null){oe.flags&=-257,cv(oe,x,T,y,a),oe.mode&1&&uv(y,B,a),a=oe,I=B;var ue=a.updateQueue;if(ue===null){var se=new Set;se.add(I),a.updateQueue=se}else ue.add(I);break e}else{if((a&1)===0){uv(y,B,a),dc();break e}I=Error(r(426))}}else if($e&&T.mode&1){var ze=sv(x);if(ze!==null){(ze.flags&65536)===0&&(ze.flags|=256),cv(ze,x,T,y,a),Ss(Gn(I,T));break e}}y=I=Gn(I,T),Ve!==4&&(Ve=2),ma===null?ma=[y]:ma.push(y),y=x;do{switch(y.tag){case 3:y.flags|=65536,a&=-a,y.lanes|=a;var L=ov(y,I,a);Ip(y,L);break e;case 1:T=I;var N=y.type,R=y.stateNode;if((y.flags&128)===0&&(typeof N.getDerivedStateFromError=="function"||R!==null&&typeof R.componentDidCatch=="function"&&(Wr===null||!Wr.has(R)))){y.flags|=65536,a&=-a,y.lanes|=a;var ne=lv(y,T,a);Ip(y,ne);break e}}y=y.return}while(y!==null)}Fv(o)}catch(ce){a=ce,Fe===o&&o!==null&&(Fe=o=o.return);continue}break}while(!0)}function Bv(){var n=il.current;return il.current=Zo,n===null?Zo:n}function dc(){(Ve===0||Ve===3||Ve===2)&&(Ve=4),Ye===null||(un&268435455)===0&&(al&268435455)===0||Kr(Ye,Ze)}function fl(n,a){var o=xe;xe|=2;var u=Bv();(Ye!==n||Ze!==a)&&(yr=null,cn(n,a));do try{rw();break}catch(f){Rv(n,f)}while(!0);if(As(),xe=o,il.current=u,Fe!==null)throw Error(r(261));return Ye=null,Ze=0,Ve}function rw(){for(;Fe!==null;)zv(Fe)}function nw(){for(;Fe!==null&&!_b();)zv(Fe)}function zv(n){var a=Vv(n.alternate,n,Ot);n.memoizedProps=n.pendingProps,a===null?Fv(n):Fe=a,nc.current=null}function Fv(n){var a=n;do{var o=a.alternate;if(n=a.return,(a.flags&32768)===0){if(o=Y0(o,a,Ot),o!==null){Fe=o;return}}else{if(o=q0(o,a),o!==null){o.flags&=32767,Fe=o;return}if(n!==null)n.flags|=32768,n.subtreeFlags=0,n.deletions=null;else{Ve=6,Fe=null;return}}if(a=a.sibling,a!==null){Fe=a;return}Fe=a=n}while(a!==null);Ve===0&&(Ve=5)}function fn(n,a,o){var u=Ee,f=Ct.transition;try{Ct.transition=null,Ee=1,iw(n,a,o,u)}finally{Ct.transition=f,Ee=u}return null}function iw(n,a,o,u){do Qn();while(Ur!==null);if((xe&6)!==0)throw Error(r(327));o=n.finishedWork;var f=n.finishedLanes;if(o===null)return null;if(n.finishedWork=null,n.finishedLanes=0,o===n.current)throw Error(r(177));n.callbackNode=null,n.callbackPriority=0;var y=o.lanes|o.childLanes;if(Rb(n,y),n===Ye&&(Fe=Ye=null,Ze=0),(o.subtreeFlags&2064)===0&&(o.flags&2064)===0||ll||(ll=!0,Kv(yo,function(){return Qn(),null})),y=(o.flags&15990)!==0,(o.subtreeFlags&15990)!==0||y){y=Ct.transition,Ct.transition=null;var x=Ee;Ee=1;var T=xe;xe|=4,nc.current=null,Z0(n,o),Cv(o,n),P0(ds),Oo=!!fs,ds=fs=null,n.current=o,J0(o),kb(),xe=T,Ee=x,Ct.transition=y}else n.current=o;if(ll&&(ll=!1,Ur=n,ul=f),y=n.pendingLanes,y===0&&(Wr=null),$b(o.stateNode),mt(n,Be()),a!==null)for(u=n.onRecoverableError,o=0;o<a.length;o++)f=a[o],u(f.value,{componentStack:f.stack,digest:f.digest});if(ol)throw ol=!1,n=oc,oc=null,n;return(ul&1)!==0&&n.tag!==0&&Qn(),y=n.pendingLanes,(y&1)!==0?n===lc?ga++:(ga=0,lc=n):ga=0,Rr(),null}function Qn(){if(Ur!==null){var n=kd(ul),a=Ct.transition,o=Ee;try{if(Ct.transition=null,Ee=16>n?16:n,Ur===null)var u=!1;else{if(n=Ur,Ur=null,ul=0,(xe&6)!==0)throw Error(r(331));var f=xe;for(xe|=4,le=n.current;le!==null;){var y=le,x=y.child;if((le.flags&16)!==0){var T=y.deletions;if(T!==null){for(var I=0;I<T.length;I++){var B=T[I];for(le=B;le!==null;){var Z=le;switch(Z.tag){case 0:case 11:case 15:ya(8,Z,y)}var J=Z.child;if(J!==null)J.return=Z,le=J;else for(;le!==null;){Z=le;var Q=Z.sibling,oe=Z.return;if(Ev(Z),Z===B){le=null;break}if(Q!==null){Q.return=oe,le=Q;break}le=oe}}}var ue=y.alternate;if(ue!==null){var se=ue.child;if(se!==null){ue.child=null;do{var ze=se.sibling;se.sibling=null,se=ze}while(se!==null)}}le=y}}if((y.subtreeFlags&2064)!==0&&x!==null)x.return=y,le=x;else e:for(;le!==null;){if(y=le,(y.flags&2048)!==0)switch(y.tag){case 0:case 11:case 15:ya(9,y,y.return)}var L=y.sibling;if(L!==null){L.return=y.return,le=L;break e}le=y.return}}var N=n.current;for(le=N;le!==null;){x=le;var R=x.child;if((x.subtreeFlags&2064)!==0&&R!==null)R.return=x,le=R;else e:for(x=N;le!==null;){if(T=le,(T.flags&2048)!==0)try{switch(T.tag){case 0:case 11:case 15:nl(9,T)}}catch(ce){De(T,T.return,ce)}if(T===x){le=null;break e}var ne=T.sibling;if(ne!==null){ne.return=T.return,le=ne;break e}le=T.return}}if(xe=f,Rr(),er&&typeof er.onPostCommitFiberRoot=="function")try{er.onPostCommitFiberRoot(ho,n)}catch{}u=!0}return u}finally{Ee=o,Ct.transition=a}}return!1}function Wv(n,a,o){a=Gn(o,a),a=ov(n,a,1),n=zr(n,a,1),a=st(),n!==null&&(Wi(n,1,a),mt(n,a))}function De(n,a,o){if(n.tag===3)Wv(n,n,o);else for(;a!==null;){if(a.tag===3){Wv(a,n,o);break}else if(a.tag===1){var u=a.stateNode;if(typeof a.type.getDerivedStateFromError=="function"||typeof u.componentDidCatch=="function"&&(Wr===null||!Wr.has(u))){n=Gn(o,n),n=lv(a,n,1),a=zr(a,n,1),n=st(),a!==null&&(Wi(a,1,n),mt(a,n));break}}a=a.return}}function aw(n,a,o){var u=n.pingCache;u!==null&&u.delete(a),a=st(),n.pingedLanes|=n.suspendedLanes&o,Ye===n&&(Ze&o)===o&&(Ve===4||Ve===3&&(Ze&130023424)===Ze&&500>Be()-ac?cn(n,0):ic|=o),mt(n,a)}function Uv(n,a){a===0&&((n.mode&1)===0?a=1:(a=go,go<<=1,(go&130023424)===0&&(go=4194304)));var o=st();n=dr(n,a),n!==null&&(Wi(n,a,o),mt(n,o))}function ow(n){var a=n.memoizedState,o=0;a!==null&&(o=a.retryLane),Uv(n,o)}function lw(n,a){var o=0;switch(n.tag){case 13:var u=n.stateNode,f=n.memoizedState;f!==null&&(o=f.retryLane);break;case 19:u=n.stateNode;break;default:throw Error(r(314))}u!==null&&u.delete(a),Uv(n,o)}var Vv;Vv=function(n,a,o){if(n!==null)if(n.memoizedProps!==a.pendingProps||pt.current)yt=!0;else{if((n.lanes&o)===0&&(a.flags&128)===0)return yt=!1,X0(n,a,o);yt=(n.flags&131072)!==0}else yt=!1,$e&&(a.flags&1048576)!==0&&Sp(a,zo,a.index);switch(a.lanes=0,a.tag){case 2:var u=a.type;tl(n,a),n=a.pendingProps;var f=Bn(a,tt.current);Kn(a,o),f=Ds(null,a,u,n,f,o);var y=Ls();return a.flags|=1,typeof f=="object"&&f!==null&&typeof f.render=="function"&&f.$$typeof===void 0?(a.tag=1,a.memoizedState=null,a.updateQueue=null,vt(u)?(y=!0,Lo(a)):y=!1,a.memoizedState=f.state!==null&&f.state!==void 0?f.state:null,ks(a),f.updater=Jo,a.stateNode=f,f._reactInternals=a,Us(a,u,n,o),a=Gs(null,a,u,!0,y,o)):(a.tag=0,$e&&y&&bs(a),ut(null,a,f,o),a=a.child),a;case 16:u=a.elementType;e:{switch(tl(n,a),n=a.pendingProps,f=u._init,u=f(u._payload),a.type=u,f=a.tag=sw(u),n=zt(u,n),f){case 0:a=Hs(null,a,u,n,o);break e;case 1:a=hv(null,a,u,n,o);break e;case 11:a=fv(null,a,u,n,o);break e;case 14:a=dv(null,a,u,zt(u.type,n),o);break e}throw Error(r(306,u,""))}return a;case 0:return u=a.type,f=a.pendingProps,f=a.elementType===u?f:zt(u,f),Hs(n,a,u,f,o);case 1:return u=a.type,f=a.pendingProps,f=a.elementType===u?f:zt(u,f),hv(n,a,u,f,o);case 3:e:{if(mv(a),n===null)throw Error(r(387));u=a.pendingProps,y=a.memoizedState,f=y.element,$p(n,a),Ho(a,u,null,o);var x=a.memoizedState;if(u=x.element,y.isDehydrated)if(y={element:u,isDehydrated:!1,cache:x.cache,pendingSuspenseBoundaries:x.pendingSuspenseBoundaries,transitions:x.transitions},a.updateQueue.baseState=y,a.memoizedState=y,a.flags&256){f=Gn(Error(r(423)),a),a=gv(n,a,u,o,f);break e}else if(u!==f){f=Gn(Error(r(424)),a),a=gv(n,a,u,o,f);break e}else for(xt=Nr(a.stateNode.containerInfo.firstChild),wt=a,$e=!0,Bt=null,o=Tp(a,null,u,o),a.child=o;o;)o.flags=o.flags&-3|4096,o=o.sibling;else{if(Wn(),u===f){a=vr(n,a,o);break e}ut(n,a,u,o)}a=a.child}return a;case 5:return Np(a),n===null&&Os(a),u=a.type,f=a.pendingProps,y=n!==null?n.memoizedProps:null,x=f.children,ps(u,f)?x=null:y!==null&&ps(u,y)&&(a.flags|=32),yv(n,a),ut(n,a,x,o),a.child;case 6:return n===null&&Os(a),null;case 13:return bv(n,a,o);case 4:return Ts(a,a.stateNode.containerInfo),u=a.pendingProps,n===null?a.child=Un(a,null,u,o):ut(n,a,u,o),a.child;case 11:return u=a.type,f=a.pendingProps,f=a.elementType===u?f:zt(u,f),fv(n,a,u,f,o);case 7:return ut(n,a,a.pendingProps,o),a.child;case 8:return ut(n,a,a.pendingProps.children,o),a.child;case 12:return ut(n,a,a.pendingProps.children,o),a.child;case 10:e:{if(u=a.type._context,f=a.pendingProps,y=a.memoizedProps,x=f.value,_e(Uo,u._currentValue),u._currentValue=x,y!==null)if(Rt(y.value,x)){if(y.children===f.children&&!pt.current){a=vr(n,a,o);break e}}else for(y=a.child,y!==null&&(y.return=a);y!==null;){var T=y.dependencies;if(T!==null){x=y.child;for(var I=T.firstContext;I!==null;){if(I.context===u){if(y.tag===1){I=pr(-1,o&-o),I.tag=2;var B=y.updateQueue;if(B!==null){B=B.shared;var Z=B.pending;Z===null?I.next=I:(I.next=Z.next,Z.next=I),B.pending=I}}y.lanes|=o,I=y.alternate,I!==null&&(I.lanes|=o),js(y.return,o,a),T.lanes|=o;break}I=I.next}}else if(y.tag===10)x=y.type===a.type?null:y.child;else if(y.tag===18){if(x=y.return,x===null)throw Error(r(341));x.lanes|=o,T=x.alternate,T!==null&&(T.lanes|=o),js(x,o,a),x=y.sibling}else x=y.child;if(x!==null)x.return=y;else for(x=y;x!==null;){if(x===a){x=null;break}if(y=x.sibling,y!==null){y.return=x.return,x=y;break}x=x.return}y=x}ut(n,a,f.children,o),a=a.child}return a;case 9:return f=a.type,u=a.pendingProps.children,Kn(a,o),f=kt(f),u=u(f),a.flags|=1,ut(n,a,u,o),a.child;case 14:return u=a.type,f=zt(u,a.pendingProps),f=zt(u.type,f),dv(n,a,u,f,o);case 15:return pv(n,a,a.type,a.pendingProps,o);case 17:return u=a.type,f=a.pendingProps,f=a.elementType===u?f:zt(u,f),tl(n,a),a.tag=1,vt(u)?(n=!0,Lo(a)):n=!1,Kn(a,o),iv(a,u,f),Us(a,u,f,o),Gs(null,a,u,!0,n,o);case 19:return xv(n,a,o);case 22:return vv(n,a,o)}throw Error(r(156,a.tag))};function Kv(n,a){return Pd(n,a)}function uw(n,a,o,u){this.tag=n,this.key=o,this.sibling=this.child=this.return=this.stateNode=this.type=this.elementType=null,this.index=0,this.ref=null,this.pendingProps=a,this.dependencies=this.memoizedState=this.updateQueue=this.memoizedProps=null,this.mode=u,this.subtreeFlags=this.flags=0,this.deletions=null,this.childLanes=this.lanes=0,this.alternate=null}function $t(n,a,o,u){return new uw(n,a,o,u)}function pc(n){return n=n.prototype,!(!n||!n.isReactComponent)}function sw(n){if(typeof n=="function")return pc(n)?1:0;if(n!=null){if(n=n.$$typeof,n===H)return 11;if(n===V)return 14}return 2}function Hr(n,a){var o=n.alternate;return o===null?(o=$t(n.tag,a,n.key,n.mode),o.elementType=n.elementType,o.type=n.type,o.stateNode=n.stateNode,o.alternate=n,n.alternate=o):(o.pendingProps=a,o.type=n.type,o.flags=0,o.subtreeFlags=0,o.deletions=null),o.flags=n.flags&14680064,o.childLanes=n.childLanes,o.lanes=n.lanes,o.child=n.child,o.memoizedProps=n.memoizedProps,o.memoizedState=n.memoizedState,o.updateQueue=n.updateQueue,a=n.dependencies,o.dependencies=a===null?null:{lanes:a.lanes,firstContext:a.firstContext},o.sibling=n.sibling,o.index=n.index,o.ref=n.ref,o}function dl(n,a,o,u,f,y){var x=2;if(u=n,typeof n=="function")pc(n)&&(x=1);else if(typeof n=="string")x=5;else e:switch(n){case C:return dn(o.children,f,y,a);case M:x=8,f|=8;break;case D:return n=$t(12,o,a,f|2),n.elementType=D,n.lanes=y,n;case q:return n=$t(13,o,a,f),n.elementType=q,n.lanes=y,n;case Y:return n=$t(19,o,a,f),n.elementType=Y,n.lanes=y,n;case ee:return pl(o,f,y,a);default:if(typeof n=="object"&&n!==null)switch(n.$$typeof){case U:x=10;break e;case F:x=9;break e;case H:x=11;break e;case V:x=14;break e;case G:x=16,u=null;break e}throw Error(r(130,n==null?n:typeof n,""))}return a=$t(x,o,a,f),a.elementType=n,a.type=u,a.lanes=y,a}function dn(n,a,o,u){return n=$t(7,n,u,a),n.lanes=o,n}function pl(n,a,o,u){return n=$t(22,n,u,a),n.elementType=ee,n.lanes=o,n.stateNode={isHidden:!1},n}function vc(n,a,o){return n=$t(6,n,null,a),n.lanes=o,n}function yc(n,a,o){return a=$t(4,n.children!==null?n.children:[],n.key,a),a.lanes=o,a.stateNode={containerInfo:n.containerInfo,pendingChildren:null,implementation:n.implementation},a}function cw(n,a,o,u,f){this.tag=a,this.containerInfo=n,this.finishedWork=this.pingCache=this.current=this.pendingChildren=null,this.timeoutHandle=-1,this.callbackNode=this.pendingContext=this.context=null,this.callbackPriority=0,this.eventTimes=Wu(0),this.expirationTimes=Wu(-1),this.entangledLanes=this.finishedLanes=this.mutableReadLanes=this.expiredLanes=this.pingedLanes=this.suspendedLanes=this.pendingLanes=0,this.entanglements=Wu(0),this.identifierPrefix=u,this.onRecoverableError=f,this.mutableSourceEagerHydrationData=null}function hc(n,a,o,u,f,y,x,T,I){return n=new cw(n,a,o,T,I),a===1?(a=1,y===!0&&(a|=8)):a=0,y=$t(3,null,null,a),n.current=y,y.stateNode=n,y.memoizedState={element:u,isDehydrated:o,cache:null,transitions:null,pendingSuspenseBoundaries:null},ks(y),n}function fw(n,a,o){var u=3<arguments.length&&arguments[3]!==void 0?arguments[3]:null;return{$$typeof:E,key:u==null?null:""+u,children:n,containerInfo:a,implementation:o}}function Hv(n){if(!n)return Lr;n=n._reactInternals;e:{if(Jr(n)!==n||n.tag!==1)throw Error(r(170));var a=n;do{switch(a.tag){case 3:a=a.stateNode.context;break e;case 1:if(vt(a.type)){a=a.stateNode.__reactInternalMemoizedMergedChildContext;break e}}a=a.return}while(a!==null);throw Error(r(171))}if(n.tag===1){var o=n.type;if(vt(o))return wp(n,o,a)}return a}function Gv(n,a,o,u,f,y,x,T,I){return n=hc(o,u,!0,n,f,y,x,T,I),n.context=Hv(null),o=n.current,u=st(),f=Vr(o),y=pr(u,f),y.callback=a??null,zr(o,y,f),n.current.lanes=f,Wi(n,f,u),mt(n,u),n}function vl(n,a,o,u){var f=a.current,y=st(),x=Vr(f);return o=Hv(o),a.context===null?a.context=o:a.pendingContext=o,a=pr(y,x),a.payload={element:n},u=u===void 0?null:u,u!==null&&(a.callback=u),n=zr(f,a,x),n!==null&&(Ut(n,f,x,y),Ko(n,f,x)),x}function yl(n){if(n=n.current,!n.child)return null;switch(n.child.tag){case 5:return n.child.stateNode;default:return n.child.stateNode}}function Xv(n,a){if(n=n.memoizedState,n!==null&&n.dehydrated!==null){var o=n.retryLane;n.retryLane=o!==0&&o<a?o:a}}function mc(n,a){Xv(n,a),(n=n.alternate)&&Xv(n,a)}function dw(){return null}var Yv=typeof reportError=="function"?reportError:function(n){console.error(n)};function gc(n){this._internalRoot=n}hl.prototype.render=gc.prototype.render=function(n){var a=this._internalRoot;if(a===null)throw Error(r(409));vl(n,a,null,null)},hl.prototype.unmount=gc.prototype.unmount=function(){var n=this._internalRoot;if(n!==null){this._internalRoot=null;var a=n.containerInfo;sn(function(){vl(null,n,null,null)}),a[ur]=null}};function hl(n){this._internalRoot=n}hl.prototype.unstable_scheduleHydration=function(n){if(n){var a=$d();n={blockedOn:null,target:n,priority:a};for(var o=0;o<$r.length&&a!==0&&a<$r[o].priority;o++);$r.splice(o,0,n),o===0&&Nd(n)}};function bc(n){return!(!n||n.nodeType!==1&&n.nodeType!==9&&n.nodeType!==11)}function ml(n){return!(!n||n.nodeType!==1&&n.nodeType!==9&&n.nodeType!==11&&(n.nodeType!==8||n.nodeValue!==" react-mount-point-unstable "))}function qv(){}function pw(n,a,o,u,f){if(f){if(typeof u=="function"){var y=u;u=function(){var B=yl(x);y.call(B)}}var x=Gv(a,u,n,0,null,!1,!1,"",qv);return n._reactRootContainer=x,n[ur]=x.current,ra(n.nodeType===8?n.parentNode:n),sn(),x}for(;f=n.lastChild;)n.removeChild(f);if(typeof u=="function"){var T=u;u=function(){var B=yl(I);T.call(B)}}var I=hc(n,0,!1,null,null,!1,!1,"",qv);return n._reactRootContainer=I,n[ur]=I.current,ra(n.nodeType===8?n.parentNode:n),sn(function(){vl(a,I,o,u)}),I}function gl(n,a,o,u,f){var y=o._reactRootContainer;if(y){var x=y;if(typeof f=="function"){var T=f;f=function(){var I=yl(x);T.call(I)}}vl(a,x,n,f)}else x=pw(o,a,n,f,u);return yl(x)}Td=function(n){switch(n.tag){case 3:var a=n.stateNode;if(a.current.memoizedState.isDehydrated){var o=Fi(a.pendingLanes);o!==0&&(Uu(a,o|1),mt(a,Be()),(xe&6)===0&&(qn=Be()+500,Rr()))}break;case 13:sn(function(){var u=dr(n,1);if(u!==null){var f=st();Ut(u,n,1,f)}}),mc(n,1)}},Vu=function(n){if(n.tag===13){var a=dr(n,134217728);if(a!==null){var o=st();Ut(a,n,134217728,o)}mc(n,134217728)}},Cd=function(n){if(n.tag===13){var a=Vr(n),o=dr(n,a);if(o!==null){var u=st();Ut(o,n,a,u)}mc(n,a)}},$d=function(){return Ee},Id=function(n,a){var o=Ee;try{return Ee=n,a()}finally{Ee=o}},Du=function(n,a,o){switch(a){case"input":if(Qt(n,o),a=o.name,o.type==="radio"&&a!=null){for(o=n;o.parentNode;)o=o.parentNode;for(o=o.querySelectorAll("input[name="+JSON.stringify(""+a)+'][type="radio"]'),a=0;a<o.length;a++){var u=o[a];if(u!==n&&u.form===n.form){var f=No(u);if(!f)throw Error(r(90));Ar(u),Qt(u,f)}}}break;case"textarea":Mi(n,o);break;case"select":a=o.value,a!=null&&dt(n,!!o.multiple,a,!1)}},md=cc,gd=sn;var vw={usingClientEntryPoint:!1,Events:[aa,Ln,No,yd,hd,cc]},ba={findFiberByHostInstance:en,bundleType:0,version:"18.3.1",rendererPackageName:"react-dom"},yw={bundleType:ba.bundleType,version:ba.version,rendererPackageName:ba.rendererPackageName,rendererConfig:ba.rendererConfig,overrideHookState:null,overrideHookStateDeletePath:null,overrideHookStateRenamePath:null,overrideProps:null,overridePropsDeletePath:null,overridePropsRenamePath:null,setErrorHandler:null,setSuspenseHandler:null,scheduleUpdate:null,currentDispatcherRef:k.ReactCurrentDispatcher,findHostInstanceByFiber:function(n){return n=Od(n),n===null?null:n.stateNode},findFiberByHostInstance:ba.findFiberByHostInstance||dw,findHostInstancesForRefresh:null,scheduleRefresh:null,scheduleRoot:null,setRefreshHandler:null,getCurrentFiber:null,reconcilerVersion:"18.3.1-next-f1338f8080-20240426"};if(typeof __REACT_DEVTOOLS_GLOBAL_HOOK__<"u"){var bl=__REACT_DEVTOOLS_GLOBAL_HOOK__;if(!bl.isDisabled&&bl.supportsFiber)try{ho=bl.inject(yw),er=bl}catch{}}return gt.__SECRET_INTERNALS_DO_NOT_USE_OR_YOU_WILL_BE_FIRED=vw,gt.createPortal=function(n,a){var o=2<arguments.length&&arguments[2]!==void 0?arguments[2]:null;if(!bc(a))throw Error(r(200));return fw(n,a,null,o)},gt.createRoot=function(n,a){if(!bc(n))throw Error(r(299));var o=!1,u="",f=Yv;return a!=null&&(a.unstable_strictMode===!0&&(o=!0),a.identifierPrefix!==void 0&&(u=a.identifierPrefix),a.onRecoverableError!==void 0&&(f=a.onRecoverableError)),a=hc(n,1,!1,null,null,o,!1,u,f),n[ur]=a.current,ra(n.nodeType===8?n.parentNode:n),new gc(a)},gt.findDOMNode=function(n){if(n==null)return null;if(n.nodeType===1)return n;var a=n._reactInternals;if(a===void 0)throw typeof n.render=="function"?Error(r(188)):(n=Object.keys(n).join(","),Error(r(268,n)));return n=Od(a),n=n===null?null:n.stateNode,n},gt.flushSync=function(n){return sn(n)},gt.hydrate=function(n,a,o){if(!ml(a))throw Error(r(200));return gl(null,n,a,!0,o)},gt.hydrateRoot=function(n,a,o){if(!bc(n))throw Error(r(405));var u=o!=null&&o.hydratedSources||null,f=!1,y="",x=Yv;if(o!=null&&(o.unstable_strictMode===!0&&(f=!0),o.identifierPrefix!==void 0&&(y=o.identifierPrefix),o.onRecoverableError!==void 0&&(x=o.onRecoverableError)),a=Gv(a,null,n,1,o??null,f,!1,y,x),n[ur]=a.current,ra(n),u)for(n=0;n<u.length;n++)o=u[n],f=o._getVersion,f=f(o._source),a.mutableSourceEagerHydrationData==null?a.mutableSourceEagerHydrationData=[o,f]:a.mutableSourceEagerHydrationData.push(o,f);return new hl(a)},gt.render=function(n,a,o){if(!ml(a))throw Error(r(200));return gl(null,n,a,!1,o)},gt.unmountComponentAtNode=function(n){if(!ml(n))throw Error(r(40));return n._reactRootContainer?(sn(function(){gl(null,null,n,!1,function(){n._reactRootContainer=null,n[ur]=null})}),!0):!1},gt.unstable_batchedUpdates=cc,gt.unstable_renderSubtreeIntoContainer=function(n,a,o,u){if(!ml(o))throw Error(r(200));if(n==null||n._reactInternals===void 0)throw Error(r(38));return gl(n,a,o,!1,u)},gt.version="18.3.1-next-f1338f8080-20240426",gt}var iy;function ox(){if(iy)return wc.exports;iy=1;function t(){if(!(typeof __REACT_DEVTOOLS_GLOBAL_HOOK__>"u"||typeof __REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE!="function"))try{__REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE(t)}catch(e){console.error(e)}}return t(),wc.exports=ax(),wc.exports}ox();function Lm(t){var e,r,i="";if(typeof t=="string"||typeof t=="number")i+=t;else if(typeof t=="object")if(Array.isArray(t)){var l=t.length;for(e=0;e<l;e++)t[e]&&(r=Lm(t[e]))&&(i&&(i+=" "),i+=r)}else for(r in t)t[r]&&(i&&(i+=" "),i+=r);return i}function ge(){for(var t,e,r=0,i="",l=arguments.length;r<l;r++)(t=arguments[r])&&(e=Lm(t))&&(i&&(i+=" "),i+=e);return i}var Sc={exports:{}},Pe={};/**
 * @license React
 * react-is.production.min.js
 *
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var ay;function lx(){if(ay)return Pe;ay=1;var t=Symbol.for("react.element"),e=Symbol.for("react.portal"),r=Symbol.for("react.fragment"),i=Symbol.for("react.strict_mode"),l=Symbol.for("react.profiler"),s=Symbol.for("react.provider"),c=Symbol.for("react.context"),d=Symbol.for("react.server_context"),p=Symbol.for("react.forward_ref"),v=Symbol.for("react.suspense"),h=Symbol.for("react.suspense_list"),m=Symbol.for("react.memo"),b=Symbol.for("react.lazy"),O=Symbol.for("react.offscreen"),S;S=Symbol.for("react.module.reference");function P(g){if(typeof g=="object"&&g!==null){var A=g.$$typeof;switch(A){case t:switch(g=g.type,g){case r:case l:case i:case v:case h:return g;default:switch(g=g&&g.$$typeof,g){case d:case c:case p:case b:case m:case s:return g;default:return A}}case e:return A}}}return Pe.ContextConsumer=c,Pe.ContextProvider=s,Pe.Element=t,Pe.ForwardRef=p,Pe.Fragment=r,Pe.Lazy=b,Pe.Memo=m,Pe.Portal=e,Pe.Profiler=l,Pe.StrictMode=i,Pe.Suspense=v,Pe.SuspenseList=h,Pe.isAsyncMode=function(){return!1},Pe.isConcurrentMode=function(){return!1},Pe.isContextConsumer=function(g){return P(g)===c},Pe.isContextProvider=function(g){return P(g)===s},Pe.isElement=function(g){return typeof g=="object"&&g!==null&&g.$$typeof===t},Pe.isForwardRef=function(g){return P(g)===p},Pe.isFragment=function(g){return P(g)===r},Pe.isLazy=function(g){return P(g)===b},Pe.isMemo=function(g){return P(g)===m},Pe.isPortal=function(g){return P(g)===e},Pe.isProfiler=function(g){return P(g)===l},Pe.isStrictMode=function(g){return P(g)===i},Pe.isSuspense=function(g){return P(g)===v},Pe.isSuspenseList=function(g){return P(g)===h},Pe.isValidElementType=function(g){return typeof g=="string"||typeof g=="function"||g===r||g===l||g===i||g===v||g===h||g===O||typeof g=="object"&&g!==null&&(g.$$typeof===b||g.$$typeof===m||g.$$typeof===s||g.$$typeof===c||g.$$typeof===p||g.$$typeof===S||g.getModuleId!==void 0)},Pe.typeOf=P,Pe}var oy;function ux(){return oy||(oy=1,Sc.exports=lx()),Sc.exports}var sx=ux(),Ht=function(e){return e===0?0:e>0?1:-1},mn=function(e){return wn(e)&&e.indexOf("%")===e.length-1},ie=function(e){return gw(e)&&!Ci(e)},cx=function(e){return be(e)},Ge=function(e){return ie(e)||wn(e)},fx=0,uo=function(e){var r=++fx;return"".concat(e||"").concat(r)},Gt=function(e,r){var i=arguments.length>2&&arguments[2]!==void 0?arguments[2]:0,l=arguments.length>3&&arguments[3]!==void 0?arguments[3]:!1;if(!ie(e)&&!wn(e))return i;var s;if(mn(e)){var c=e.indexOf("%");s=r*parseFloat(e.slice(0,c))/100}else s=+e;return Ci(s)&&(s=i),l&&s>r&&(s=r),s},Yr=function(e){if(!e)return null;var r=Object.keys(e);return r&&r.length?e[r[0]]:null},dx=function(e){if(!Array.isArray(e))return!1;for(var r=e.length,i={},l=0;l<r;l++)if(!i[e[l]])i[e[l]]=!0;else return!0;return!1},ot=function(e,r){return ie(e)&&ie(r)?function(i){return e+i*(r-e)}:function(){return r}};function $l(t,e,r){return!t||!t.length?null:t.find(function(i){return i&&(typeof e=="function"?e(i):Mt(i,e))===r})}var px=function(e,r){return ie(e)&&ie(r)?e-r:wn(e)&&wn(r)?e.localeCompare(r):e instanceof Date&&r instanceof Date?e.getTime()-r.getTime():String(e).localeCompare(String(r))};function ii(t,e){for(var r in t)if({}.hasOwnProperty.call(t,r)&&(!{}.hasOwnProperty.call(e,r)||t[r]!==e[r]))return!1;for(var i in e)if({}.hasOwnProperty.call(e,i)&&!{}.hasOwnProperty.call(t,i))return!1;return!0}function zc(t){"@babel/helpers - typeof";return zc=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},zc(t)}var vx=["viewBox","children"],yx=["aria-activedescendant","aria-atomic","aria-autocomplete","aria-busy","aria-checked","aria-colcount","aria-colindex","aria-colspan","aria-controls","aria-current","aria-describedby","aria-details","aria-disabled","aria-errormessage","aria-expanded","aria-flowto","aria-haspopup","aria-hidden","aria-invalid","aria-keyshortcuts","aria-label","aria-labelledby","aria-level","aria-live","aria-modal","aria-multiline","aria-multiselectable","aria-orientation","aria-owns","aria-placeholder","aria-posinset","aria-pressed","aria-readonly","aria-relevant","aria-required","aria-roledescription","aria-rowcount","aria-rowindex","aria-rowspan","aria-selected","aria-setsize","aria-sort","aria-valuemax","aria-valuemin","aria-valuenow","aria-valuetext","className","color","height","id","lang","max","media","method","min","name","style","target","width","role","tabIndex","accentHeight","accumulate","additive","alignmentBaseline","allowReorder","alphabetic","amplitude","arabicForm","ascent","attributeName","attributeType","autoReverse","azimuth","baseFrequency","baselineShift","baseProfile","bbox","begin","bias","by","calcMode","capHeight","clip","clipPath","clipPathUnits","clipRule","colorInterpolation","colorInterpolationFilters","colorProfile","colorRendering","contentScriptType","contentStyleType","cursor","cx","cy","d","decelerate","descent","diffuseConstant","direction","display","divisor","dominantBaseline","dur","dx","dy","edgeMode","elevation","enableBackground","end","exponent","externalResourcesRequired","fill","fillOpacity","fillRule","filter","filterRes","filterUnits","floodColor","floodOpacity","focusable","fontFamily","fontSize","fontSizeAdjust","fontStretch","fontStyle","fontVariant","fontWeight","format","from","fx","fy","g1","g2","glyphName","glyphOrientationHorizontal","glyphOrientationVertical","glyphRef","gradientTransform","gradientUnits","hanging","horizAdvX","horizOriginX","href","ideographic","imageRendering","in2","in","intercept","k1","k2","k3","k4","k","kernelMatrix","kernelUnitLength","kerning","keyPoints","keySplines","keyTimes","lengthAdjust","letterSpacing","lightingColor","limitingConeAngle","local","markerEnd","markerHeight","markerMid","markerStart","markerUnits","markerWidth","mask","maskContentUnits","maskUnits","mathematical","mode","numOctaves","offset","opacity","operator","order","orient","orientation","origin","overflow","overlinePosition","overlineThickness","paintOrder","panose1","pathLength","patternContentUnits","patternTransform","patternUnits","pointerEvents","pointsAtX","pointsAtY","pointsAtZ","preserveAlpha","preserveAspectRatio","primitiveUnits","r","radius","refX","refY","renderingIntent","repeatCount","repeatDur","requiredExtensions","requiredFeatures","restart","result","rotate","rx","ry","seed","shapeRendering","slope","spacing","specularConstant","specularExponent","speed","spreadMethod","startOffset","stdDeviation","stemh","stemv","stitchTiles","stopColor","stopOpacity","strikethroughPosition","strikethroughThickness","string","stroke","strokeDasharray","strokeDashoffset","strokeLinecap","strokeLinejoin","strokeMiterlimit","strokeOpacity","strokeWidth","surfaceScale","systemLanguage","tableValues","targetX","targetY","textAnchor","textDecoration","textLength","textRendering","to","transform","u1","u2","underlinePosition","underlineThickness","unicode","unicodeBidi","unicodeRange","unitsPerEm","vAlphabetic","values","vectorEffect","version","vertAdvY","vertOriginX","vertOriginY","vHanging","vIdeographic","viewTarget","visibility","vMathematical","widths","wordSpacing","writingMode","x1","x2","x","xChannelSelector","xHeight","xlinkActuate","xlinkArcrole","xlinkHref","xlinkRole","xlinkShow","xlinkTitle","xlinkType","xmlBase","xmlLang","xmlns","xmlnsXlink","xmlSpace","y1","y2","y","yChannelSelector","z","zoomAndPan","ref","key","angle"],ly=["points","pathLength"],Pc={svg:vx,polygon:ly,polyline:ly},Hf=["dangerouslySetInnerHTML","onCopy","onCopyCapture","onCut","onCutCapture","onPaste","onPasteCapture","onCompositionEnd","onCompositionEndCapture","onCompositionStart","onCompositionStartCapture","onCompositionUpdate","onCompositionUpdateCapture","onFocus","onFocusCapture","onBlur","onBlurCapture","onChange","onChangeCapture","onBeforeInput","onBeforeInputCapture","onInput","onInputCapture","onReset","onResetCapture","onSubmit","onSubmitCapture","onInvalid","onInvalidCapture","onLoad","onLoadCapture","onError","onErrorCapture","onKeyDown","onKeyDownCapture","onKeyPress","onKeyPressCapture","onKeyUp","onKeyUpCapture","onAbort","onAbortCapture","onCanPlay","onCanPlayCapture","onCanPlayThrough","onCanPlayThroughCapture","onDurationChange","onDurationChangeCapture","onEmptied","onEmptiedCapture","onEncrypted","onEncryptedCapture","onEnded","onEndedCapture","onLoadedData","onLoadedDataCapture","onLoadedMetadata","onLoadedMetadataCapture","onLoadStart","onLoadStartCapture","onPause","onPauseCapture","onPlay","onPlayCapture","onPlaying","onPlayingCapture","onProgress","onProgressCapture","onRateChange","onRateChangeCapture","onSeeked","onSeekedCapture","onSeeking","onSeekingCapture","onStalled","onStalledCapture","onSuspend","onSuspendCapture","onTimeUpdate","onTimeUpdateCapture","onVolumeChange","onVolumeChangeCapture","onWaiting","onWaitingCapture","onAuxClick","onAuxClickCapture","onClick","onClickCapture","onContextMenu","onContextMenuCapture","onDoubleClick","onDoubleClickCapture","onDrag","onDragCapture","onDragEnd","onDragEndCapture","onDragEnter","onDragEnterCapture","onDragExit","onDragExitCapture","onDragLeave","onDragLeaveCapture","onDragOver","onDragOverCapture","onDragStart","onDragStartCapture","onDrop","onDropCapture","onMouseDown","onMouseDownCapture","onMouseEnter","onMouseLeave","onMouseMove","onMouseMoveCapture","onMouseOut","onMouseOutCapture","onMouseOver","onMouseOverCapture","onMouseUp","onMouseUpCapture","onSelect","onSelectCapture","onTouchCancel","onTouchCancelCapture","onTouchEnd","onTouchEndCapture","onTouchMove","onTouchMoveCapture","onTouchStart","onTouchStartCapture","onPointerDown","onPointerDownCapture","onPointerMove","onPointerMoveCapture","onPointerUp","onPointerUpCapture","onPointerCancel","onPointerCancelCapture","onPointerEnter","onPointerEnterCapture","onPointerLeave","onPointerLeaveCapture","onPointerOver","onPointerOverCapture","onPointerOut","onPointerOutCapture","onGotPointerCapture","onGotPointerCaptureCapture","onLostPointerCapture","onLostPointerCaptureCapture","onScroll","onScrollCapture","onWheel","onWheelCapture","onAnimationStart","onAnimationStartCapture","onAnimationEnd","onAnimationEndCapture","onAnimationIteration","onAnimationIterationCapture","onTransitionEnd","onTransitionEndCapture"],Il=function(e,r){if(!e||typeof e=="function"||typeof e=="boolean")return null;var i=e;if(te.isValidElement(e)&&(i=e.props),!$i(i))return null;var l={};return Object.keys(i).forEach(function(s){Hf.includes(s)&&(l[s]=r||function(c){return i[s](i,c)})}),l},hx=function(e,r,i){return function(l){return e(r,i,l),null}},ui=function(e,r,i){if(!$i(e)||zc(e)!=="object")return null;var l=null;return Object.keys(e).forEach(function(s){var c=e[s];Hf.includes(s)&&typeof c=="function"&&(l||(l={}),l[s]=hx(c,r,i))}),l},mx=["children"],gx=["children"];function uy(t,e){if(t==null)return{};var r=bx(t,e),i,l;if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(t);for(l=0;l<s.length;l++)i=s[l],!(e.indexOf(i)>=0)&&Object.prototype.propertyIsEnumerable.call(t,i)&&(r[i]=t[i])}return r}function bx(t,e){if(t==null)return{};var r={};for(var i in t)if(Object.prototype.hasOwnProperty.call(t,i)){if(e.indexOf(i)>=0)continue;r[i]=t[i]}return r}function Fc(t){"@babel/helpers - typeof";return Fc=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},Fc(t)}var sy={click:"onClick",mousedown:"onMouseDown",mouseup:"onMouseUp",mouseover:"onMouseOver",mousemove:"onMouseMove",mouseout:"onMouseOut",mouseenter:"onMouseEnter",mouseleave:"onMouseLeave",touchcancel:"onTouchCancel",touchend:"onTouchEnd",touchmove:"onTouchMove",touchstart:"onTouchStart",contextmenu:"onContextMenu",dblclick:"onDoubleClick"},xr=function(e){return typeof e=="string"?e:e?e.displayName||e.name||"Component":""},cy=null,Ac=null,Gf=function t(e){if(e===cy&&Array.isArray(Ac))return Ac;var r=[];return te.Children.forEach(e,function(i){be(i)||(sx.isFragment(i)?r=r.concat(t(i.props.children)):r.push(i))}),Ac=r,cy=e,r};function Xt(t,e){var r=[],i=[];return Array.isArray(e)?i=e.map(function(l){return xr(l)}):i=[xr(e)],Gf(t).forEach(function(l){var s=Mt(l,"type.displayName")||Mt(l,"type.name");i.indexOf(s)!==-1&&r.push(l)}),r}function Pt(t,e){var r=Xt(t,e);return r&&r[0]}var fy=function(e){if(!e||!e.props)return!1;var r=e.props,i=r.width,l=r.height;return!(!ie(i)||i<=0||!ie(l)||l<=0)},wx=["a","altGlyph","altGlyphDef","altGlyphItem","animate","animateColor","animateMotion","animateTransform","circle","clipPath","color-profile","cursor","defs","desc","ellipse","feBlend","feColormatrix","feComponentTransfer","feComposite","feConvolveMatrix","feDiffuseLighting","feDisplacementMap","feDistantLight","feFlood","feFuncA","feFuncB","feFuncG","feFuncR","feGaussianBlur","feImage","feMerge","feMergeNode","feMorphology","feOffset","fePointLight","feSpecularLighting","feSpotLight","feTile","feTurbulence","filter","font","font-face","font-face-format","font-face-name","font-face-url","foreignObject","g","glyph","glyphRef","hkern","image","line","lineGradient","marker","mask","metadata","missing-glyph","mpath","path","pattern","polygon","polyline","radialGradient","rect","script","set","stop","style","svg","switch","symbol","text","textPath","title","tref","tspan","use","view","vkern"],xx=function(e){return e&&e.type&&wn(e.type)&&wx.indexOf(e.type)>=0},Ox=function(e){return e&&Fc(e)==="object"&&"clipDot"in e},Sx=function(e,r,i,l){var s,c=(s=Pc==null?void 0:Pc[l])!==null&&s!==void 0?s:[];return r.startsWith("data-")||!he(e)&&(l&&c.includes(r)||yx.includes(r))||i&&Hf.includes(r)},fe=function(e,r,i){if(!e||typeof e=="function"||typeof e=="boolean")return null;var l=e;if(te.isValidElement(e)&&(l=e.props),!$i(l))return null;var s={};return Object.keys(l).forEach(function(c){var d;Sx((d=l)===null||d===void 0?void 0:d[c],c,r,i)&&(s[c]=l[c])}),s},Wc=function t(e,r){if(e===r)return!0;var i=te.Children.count(e);if(i!==te.Children.count(r))return!1;if(i===0)return!0;if(i===1)return dy(Array.isArray(e)?e[0]:e,Array.isArray(r)?r[0]:r);for(var l=0;l<i;l++){var s=e[l],c=r[l];if(Array.isArray(s)||Array.isArray(c)){if(!t(s,c))return!1}else if(!dy(s,c))return!1}return!0},dy=function(e,r){if(be(e)&&be(r))return!0;if(!be(e)&&!be(r)){var i=e.props||{},l=i.children,s=uy(i,mx),c=r.props||{},d=c.children,p=uy(c,gx);return l&&d?ii(s,p)&&Wc(l,d):!l&&!d?ii(s,p):!1}return!1},py=function(e,r){var i=[],l={};return Gf(e).forEach(function(s,c){if(xx(s))i.push(s);else if(s){var d=xr(s.type),p=r[d]||{},v=p.handler,h=p.once;if(v&&(!h||!l[d])){var m=v(s,d,c);i.push(m),l[d]=!0}}}),i},Px=function(e){var r=e&&e.type;return r&&sy[r]?sy[r]:null},Ax=function(e,r){return Gf(r).indexOf(e)},Ex=["children","width","height","viewBox","className","style","title","desc"];function Uc(){return Uc=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},Uc.apply(this,arguments)}function jx(t,e){if(t==null)return{};var r=_x(t,e),i,l;if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(t);for(l=0;l<s.length;l++)i=s[l],!(e.indexOf(i)>=0)&&Object.prototype.propertyIsEnumerable.call(t,i)&&(r[i]=t[i])}return r}function _x(t,e){if(t==null)return{};var r={};for(var i in t)if(Object.prototype.hasOwnProperty.call(t,i)){if(e.indexOf(i)>=0)continue;r[i]=t[i]}return r}function Vc(t){var e=t.children,r=t.width,i=t.height,l=t.viewBox,s=t.className,c=t.style,d=t.title,p=t.desc,v=jx(t,Ex),h=l||{width:r,height:i,x:0,y:0},m=ge("recharts-surface",s);return $.createElement("svg",Uc({},fe(v,!0,"svg"),{className:m,width:r,height:i,style:c,viewBox:"".concat(h.x," ").concat(h.y," ").concat(h.width," ").concat(h.height)}),$.createElement("title",null,d),$.createElement("desc",null,p),e)}var kx=["children","className"];function Kc(){return Kc=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},Kc.apply(this,arguments)}function Tx(t,e){if(t==null)return{};var r=Cx(t,e),i,l;if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(t);for(l=0;l<s.length;l++)i=s[l],!(e.indexOf(i)>=0)&&Object.prototype.propertyIsEnumerable.call(t,i)&&(r[i]=t[i])}return r}function Cx(t,e){if(t==null)return{};var r={};for(var i in t)if(Object.prototype.hasOwnProperty.call(t,i)){if(e.indexOf(i)>=0)continue;r[i]=t[i]}return r}var Se=$.forwardRef(function(t,e){var r=t.children,i=t.className,l=Tx(t,kx),s=ge("recharts-layer",i);return $.createElement("g",Kc({className:s},fe(l,!0),{ref:e}),r)}),Or=function(e,r){for(var i=arguments.length,l=new Array(i>2?i-2:0),s=2;s<i;s++)l[s-2]=arguments[s]};function $a(t){"@babel/helpers - typeof";return $a=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},$a(t)}var $x=["type","size","sizeType"];function Hc(){return Hc=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},Hc.apply(this,arguments)}function vy(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function yy(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?vy(Object(r),!0).forEach(function(i){Ix(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):vy(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function Ix(t,e,r){return e=Mx(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function Mx(t){var e=Nx(t,"string");return $a(e)=="symbol"?e:e+""}function Nx(t,e){if($a(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if($a(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}function Dx(t,e){if(t==null)return{};var r=Lx(t,e),i,l;if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(t);for(l=0;l<s.length;l++)i=s[l],!(e.indexOf(i)>=0)&&Object.prototype.propertyIsEnumerable.call(t,i)&&(r[i]=t[i])}return r}function Lx(t,e){if(t==null)return{};var r={};for(var i in t)if(Object.prototype.hasOwnProperty.call(t,i)){if(e.indexOf(i)>=0)continue;r[i]=t[i]}return r}var Rm={symbolCircle:$m,symbolCross:Aw,symbolDiamond:Pw,symbolSquare:Sw,symbolStar:Ow,symbolTriangle:xw,symbolWye:ww},Rx=Math.PI/180,Bx=function(e){var r="symbol".concat(vu(e));return Rm[r]||$m},zx=function(e,r,i){if(r==="area")return e;switch(i){case"cross":return 5*e*e/9;case"diamond":return .5*e*e/Math.sqrt(3);case"square":return e*e;case"star":{var l=18*Rx;return 1.25*e*e*(Math.tan(l)-Math.tan(l*2)*Math.pow(Math.tan(l),2))}case"triangle":return Math.sqrt(3)*e*e/4;case"wye":return(21-10*Math.sqrt(3))*e*e/8;default:return Math.PI*e*e/4}},Fx=function(e,r){Rm["symbol".concat(vu(e))]=r},Xf=function(e){var r=e.type,i=r===void 0?"circle":r,l=e.size,s=l===void 0?64:l,c=e.sizeType,d=c===void 0?"area":c,p=Dx(e,$x),v=yy(yy({},p),{},{type:i,size:s,sizeType:d}),h=function(){var g=Bx(i),A=bw().type(g).size(zx(s,d,i));return A()},m=v.className,b=v.cx,O=v.cy,S=fe(v,!0);return b===+b&&O===+O&&s===+s?$.createElement("path",Hc({},S,{className:ge("recharts-symbols",m),transform:"translate(".concat(b,", ").concat(O,")"),d:h()})):null};Xf.registerSymbol=Fx;function si(t){"@babel/helpers - typeof";return si=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},si(t)}function Gc(){return Gc=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},Gc.apply(this,arguments)}function hy(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function Wx(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?hy(Object(r),!0).forEach(function(i){Ia(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):hy(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function Ux(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")}function Vx(t,e){for(var r=0;r<e.length;r++){var i=e[r];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(t,zm(i.key),i)}}function Kx(t,e,r){return e&&Vx(t.prototype,e),Object.defineProperty(t,"prototype",{writable:!1}),t}function Hx(t,e,r){return e=Ml(e),Gx(t,Bm()?Reflect.construct(e,r||[],Ml(t).constructor):e.apply(t,r))}function Gx(t,e){if(e&&(si(e)==="object"||typeof e=="function"))return e;if(e!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return Xx(t)}function Xx(t){if(t===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return t}function Bm(){try{var t=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(Bm=function(){return!!t})()}function Ml(t){return Ml=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(r){return r.__proto__||Object.getPrototypeOf(r)},Ml(t)}function Yx(t,e){if(typeof e!="function"&&e!==null)throw new TypeError("Super expression must either be null or a function");t.prototype=Object.create(e&&e.prototype,{constructor:{value:t,writable:!0,configurable:!0}}),Object.defineProperty(t,"prototype",{writable:!1}),e&&Xc(t,e)}function Xc(t,e){return Xc=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(i,l){return i.__proto__=l,i},Xc(t,e)}function Ia(t,e,r){return e=zm(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function zm(t){var e=qx(t,"string");return si(e)=="symbol"?e:e+""}function qx(t,e){if(si(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(si(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(t)}var It=32,Yf=(function(t){function e(){return Ux(this,e),Hx(this,e,arguments)}return Yx(e,t),Kx(e,[{key:"renderIcon",value:function(i){var l=this.props.inactiveColor,s=It/2,c=It/6,d=It/3,p=i.inactive?l:i.color;if(i.type==="plainline")return $.createElement("line",{strokeWidth:4,fill:"none",stroke:p,strokeDasharray:i.payload.strokeDasharray,x1:0,y1:s,x2:It,y2:s,className:"recharts-legend-icon"});if(i.type==="line")return $.createElement("path",{strokeWidth:4,fill:"none",stroke:p,d:"M0,".concat(s,"h").concat(d,`
            A`).concat(c,",").concat(c,",0,1,1,").concat(2*d,",").concat(s,`
            H`).concat(It,"M").concat(2*d,",").concat(s,`
            A`).concat(c,",").concat(c,",0,1,1,").concat(d,",").concat(s),className:"recharts-legend-icon"});if(i.type==="rect")return $.createElement("path",{stroke:"none",fill:p,d:"M0,".concat(It/8,"h").concat(It,"v").concat(It*3/4,"h").concat(-It,"z"),className:"recharts-legend-icon"});if($.isValidElement(i.legendIcon)){var v=Wx({},i);return delete v.legendIcon,$.cloneElement(i.legendIcon,v)}return $.createElement(Xf,{fill:p,cx:s,cy:s,size:It,sizeType:"diameter",type:i.type})}},{key:"renderItems",value:function(){var i=this,l=this.props,s=l.payload,c=l.iconSize,d=l.layout,p=l.formatter,v=l.inactiveColor,h={x:0,y:0,width:It,height:It},m={display:d==="horizontal"?"inline-block":"block",marginRight:10},b={display:"inline-block",verticalAlign:"middle",marginRight:4};return s.map(function(O,S){var P=O.formatter||p,g=ge(Ia(Ia({"recharts-legend-item":!0},"legend-item-".concat(S),!0),"inactive",O.inactive));if(O.type==="none")return null;var A=he(O.value)?null:O.value;Or(!he(O.value),`The name property is also required when using a function for the dataKey of a chart's cartesian components. Ex: <Bar name="Name of my Data"/>`);var j=O.inactive?v:O.color;return $.createElement("li",Gc({className:g,style:m,key:"legend-item-".concat(S)},ui(i.props,O,S)),$.createElement(Vc,{width:c,height:c,viewBox:h,style:b},i.renderIcon(O)),$.createElement("span",{className:"recharts-legend-item-text",style:{color:j}},P?P(A,O,S):A))})}},{key:"render",value:function(){var i=this.props,l=i.payload,s=i.layout,c=i.align;if(!l||!l.length)return null;var d={padding:0,margin:0,textAlign:s==="horizontal"?c:"left"};return $.createElement("ul",{className:"recharts-default-legend",style:d},this.renderItems())}}])})(te.PureComponent);Ia(Yf,"displayName","Legend");Ia(Yf,"defaultProps",{iconSize:14,layout:"horizontal",align:"center",verticalAlign:"middle",inactiveColor:"#ccc"});function Fm(t,e,r){return e===!0?Qv(t,r):he(e)?Qv(t,e):t}function ci(t){"@babel/helpers - typeof";return ci=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},ci(t)}var Qx=["ref"];function my(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function hr(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?my(Object(r),!0).forEach(function(i){hu(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):my(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function Zx(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")}function gy(t,e){for(var r=0;r<e.length;r++){var i=e[r];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(t,Um(i.key),i)}}function Jx(t,e,r){return e&&gy(t.prototype,e),r&&gy(t,r),Object.defineProperty(t,"prototype",{writable:!1}),t}function eO(t,e,r){return e=Nl(e),tO(t,Wm()?Reflect.construct(e,r||[],Nl(t).constructor):e.apply(t,r))}function tO(t,e){if(e&&(ci(e)==="object"||typeof e=="function"))return e;if(e!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return rO(t)}function rO(t){if(t===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return t}function Wm(){try{var t=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(Wm=function(){return!!t})()}function Nl(t){return Nl=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(r){return r.__proto__||Object.getPrototypeOf(r)},Nl(t)}function nO(t,e){if(typeof e!="function"&&e!==null)throw new TypeError("Super expression must either be null or a function");t.prototype=Object.create(e&&e.prototype,{constructor:{value:t,writable:!0,configurable:!0}}),Object.defineProperty(t,"prototype",{writable:!1}),e&&Yc(t,e)}function Yc(t,e){return Yc=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(i,l){return i.__proto__=l,i},Yc(t,e)}function hu(t,e,r){return e=Um(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function Um(t){var e=iO(t,"string");return ci(e)=="symbol"?e:e+""}function iO(t,e){if(ci(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(ci(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(t)}function aO(t,e){if(t==null)return{};var r=oO(t,e),i,l;if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(t);for(l=0;l<s.length;l++)i=s[l],!(e.indexOf(i)>=0)&&Object.prototype.propertyIsEnumerable.call(t,i)&&(r[i]=t[i])}return r}function oO(t,e){if(t==null)return{};var r={};for(var i in t)if(Object.prototype.hasOwnProperty.call(t,i)){if(e.indexOf(i)>=0)continue;r[i]=t[i]}return r}function lO(t){return t.value}function uO(t,e){if($.isValidElement(t))return $.cloneElement(t,e);if(typeof t=="function")return $.createElement(t,e);e.ref;var r=aO(e,Qx);return $.createElement(Yf,r)}var by=1,ai=(function(t){function e(){var r;Zx(this,e);for(var i=arguments.length,l=new Array(i),s=0;s<i;s++)l[s]=arguments[s];return r=eO(this,e,[].concat(l)),hu(r,"lastBoundingBox",{width:-1,height:-1}),r}return nO(e,t),Jx(e,[{key:"componentDidMount",value:function(){this.updateBBox()}},{key:"componentDidUpdate",value:function(){this.updateBBox()}},{key:"getBBox",value:function(){if(this.wrapperNode&&this.wrapperNode.getBoundingClientRect){var i=this.wrapperNode.getBoundingClientRect();return i.height=this.wrapperNode.offsetHeight,i.width=this.wrapperNode.offsetWidth,i}return null}},{key:"updateBBox",value:function(){var i=this.props.onBBoxUpdate,l=this.getBBox();l?(Math.abs(l.width-this.lastBoundingBox.width)>by||Math.abs(l.height-this.lastBoundingBox.height)>by)&&(this.lastBoundingBox.width=l.width,this.lastBoundingBox.height=l.height,i&&i(l)):(this.lastBoundingBox.width!==-1||this.lastBoundingBox.height!==-1)&&(this.lastBoundingBox.width=-1,this.lastBoundingBox.height=-1,i&&i(null))}},{key:"getBBoxSnapshot",value:function(){return this.lastBoundingBox.width>=0&&this.lastBoundingBox.height>=0?hr({},this.lastBoundingBox):{width:0,height:0}}},{key:"getDefaultPosition",value:function(i){var l=this.props,s=l.layout,c=l.align,d=l.verticalAlign,p=l.margin,v=l.chartWidth,h=l.chartHeight,m,b;if(!i||(i.left===void 0||i.left===null)&&(i.right===void 0||i.right===null))if(c==="center"&&s==="vertical"){var O=this.getBBoxSnapshot();m={left:((v||0)-O.width)/2}}else m=c==="right"?{right:p&&p.right||0}:{left:p&&p.left||0};if(!i||(i.top===void 0||i.top===null)&&(i.bottom===void 0||i.bottom===null))if(d==="middle"){var S=this.getBBoxSnapshot();b={top:((h||0)-S.height)/2}}else b=d==="bottom"?{bottom:p&&p.bottom||0}:{top:p&&p.top||0};return hr(hr({},m),b)}},{key:"render",value:function(){var i=this,l=this.props,s=l.content,c=l.width,d=l.height,p=l.wrapperStyle,v=l.payloadUniqBy,h=l.payload,m=hr(hr({position:"absolute",width:c||"auto",height:d||"auto"},this.getDefaultPosition(p)),p);return $.createElement("div",{className:"recharts-legend-wrapper",style:m,ref:function(O){i.wrapperNode=O}},uO(s,hr(hr({},this.props),{},{payload:Fm(h,v,lO)})))}}],[{key:"getWithHeight",value:function(i,l){var s=hr(hr({},this.defaultProps),i.props),c=s.layout;return c==="vertical"&&ie(i.props.height)?{height:i.props.height}:c==="horizontal"?{width:i.props.width||l}:null}}])})(te.PureComponent);hu(ai,"displayName","Legend");hu(ai,"defaultProps",{iconSize:14,layout:"horizontal",align:"center",verticalAlign:"bottom"});function Ma(t){"@babel/helpers - typeof";return Ma=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},Ma(t)}function qc(){return qc=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},qc.apply(this,arguments)}function sO(t,e){return pO(t)||dO(t,e)||fO(t,e)||cO()}function cO(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function fO(t,e){if(t){if(typeof t=="string")return wy(t,e);var r=Object.prototype.toString.call(t).slice(8,-1);if(r==="Object"&&t.constructor&&(r=t.constructor.name),r==="Map"||r==="Set")return Array.from(t);if(r==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(r))return wy(t,e)}}function wy(t,e){(e==null||e>t.length)&&(e=t.length);for(var r=0,i=new Array(e);r<e;r++)i[r]=t[r];return i}function dO(t,e){var r=t==null?null:typeof Symbol<"u"&&t[Symbol.iterator]||t["@@iterator"];if(r!=null){var i,l,s,c,d=[],p=!0,v=!1;try{if(s=(r=r.call(t)).next,e!==0)for(;!(p=(i=s.call(r)).done)&&(d.push(i.value),d.length!==e);p=!0);}catch(h){v=!0,l=h}finally{try{if(!p&&r.return!=null&&(c=r.return(),Object(c)!==c))return}finally{if(v)throw l}}return d}}function pO(t){if(Array.isArray(t))return t}function xy(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function Ec(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?xy(Object(r),!0).forEach(function(i){vO(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):xy(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function vO(t,e,r){return e=yO(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function yO(t){var e=hO(t,"string");return Ma(e)=="symbol"?e:e+""}function hO(t,e){if(Ma(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(Ma(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}function mO(t){return Array.isArray(t)&&Ge(t[0])&&Ge(t[1])?t.join(" ~ "):t}var gO=function(e){var r=e.separator,i=r===void 0?" : ":r,l=e.contentStyle,s=l===void 0?{}:l,c=e.itemStyle,d=c===void 0?{}:c,p=e.labelStyle,v=p===void 0?{}:p,h=e.payload,m=e.formatter,b=e.itemSorter,O=e.wrapperClassName,S=e.labelClassName,P=e.label,g=e.labelFormatter,A=e.accessibilityLayer,j=A===void 0?!1:A,_=function(){if(h&&h.length){var H={padding:0,margin:0},q=(b?Kf(h,b):h).map(function(Y,V){if(Y.type==="none")return null;var G=Ec({display:"block",paddingTop:4,paddingBottom:4,color:Y.color||"#000"},d),ee=Y.formatter||m||mO,z=Y.value,X=Y.name,K=z,ae=X;if(ee&&K!=null&&ae!=null){var re=ee(z,X,Y,V,h);if(Array.isArray(re)){var de=sO(re,2);K=de[0],ae=de[1]}else K=re}return $.createElement("li",{className:"recharts-tooltip-item",key:"tooltip-item-".concat(V),style:G},Ge(ae)?$.createElement("span",{className:"recharts-tooltip-item-name"},ae):null,Ge(ae)?$.createElement("span",{className:"recharts-tooltip-item-separator"},i):null,$.createElement("span",{className:"recharts-tooltip-item-value"},K),$.createElement("span",{className:"recharts-tooltip-item-unit"},Y.unit||""))});return $.createElement("ul",{className:"recharts-tooltip-item-list",style:H},q)}return null},k=Ec({margin:0,padding:10,backgroundColor:"#fff",border:"1px solid #ccc",whiteSpace:"nowrap"},s),w=Ec({margin:0},v),E=!be(P),C=E?P:"",M=ge("recharts-default-tooltip",O),D=ge("recharts-tooltip-label",S);E&&g&&h!==void 0&&h!==null&&(C=g(P,h));var U=j?{role:"status","aria-live":"assertive"}:{};return $.createElement("div",qc({className:M,style:k},U),$.createElement("p",{className:D,style:w},$.isValidElement(C)?C:"".concat(C)),_())};function Na(t){"@babel/helpers - typeof";return Na=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},Na(t)}function xl(t,e,r){return e=bO(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function bO(t){var e=wO(t,"string");return Na(e)=="symbol"?e:e+""}function wO(t,e){if(Na(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(Na(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}var wa="recharts-tooltip-wrapper",xO={visibility:"hidden"};function OO(t){var e=t.coordinate,r=t.translateX,i=t.translateY;return ge(wa,xl(xl(xl(xl({},"".concat(wa,"-right"),ie(r)&&e&&ie(e.x)&&r>=e.x),"".concat(wa,"-left"),ie(r)&&e&&ie(e.x)&&r<e.x),"".concat(wa,"-bottom"),ie(i)&&e&&ie(e.y)&&i>=e.y),"".concat(wa,"-top"),ie(i)&&e&&ie(e.y)&&i<e.y))}function Oy(t){var e=t.allowEscapeViewBox,r=t.coordinate,i=t.key,l=t.offsetTopLeft,s=t.position,c=t.reverseDirection,d=t.tooltipDimension,p=t.viewBox,v=t.viewBoxDimension;if(s&&ie(s[i]))return s[i];var h=r[i]-d-l,m=r[i]+l;if(e[i])return c[i]?h:m;if(c[i]){var b=h,O=p[i];return b<O?Math.max(m,p[i]):Math.max(h,p[i])}var S=m+d,P=p[i]+v;return S>P?Math.max(h,p[i]):Math.max(m,p[i])}function SO(t){var e=t.translateX,r=t.translateY,i=t.useTranslate3d;return{transform:i?"translate3d(".concat(e,"px, ").concat(r,"px, 0)"):"translate(".concat(e,"px, ").concat(r,"px)")}}function PO(t){var e=t.allowEscapeViewBox,r=t.coordinate,i=t.offsetTopLeft,l=t.position,s=t.reverseDirection,c=t.tooltipBox,d=t.useTranslate3d,p=t.viewBox,v,h,m;return c.height>0&&c.width>0&&r?(h=Oy({allowEscapeViewBox:e,coordinate:r,key:"x",offsetTopLeft:i,position:l,reverseDirection:s,tooltipDimension:c.width,viewBox:p,viewBoxDimension:p.width}),m=Oy({allowEscapeViewBox:e,coordinate:r,key:"y",offsetTopLeft:i,position:l,reverseDirection:s,tooltipDimension:c.height,viewBox:p,viewBoxDimension:p.height}),v=SO({translateX:h,translateY:m,useTranslate3d:d})):v=xO,{cssProperties:v,cssClasses:OO({translateX:h,translateY:m,coordinate:r})}}function fi(t){"@babel/helpers - typeof";return fi=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},fi(t)}function Sy(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function Py(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?Sy(Object(r),!0).forEach(function(i){Zc(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):Sy(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function AO(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")}function EO(t,e){for(var r=0;r<e.length;r++){var i=e[r];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(t,Km(i.key),i)}}function jO(t,e,r){return e&&EO(t.prototype,e),Object.defineProperty(t,"prototype",{writable:!1}),t}function _O(t,e,r){return e=Dl(e),kO(t,Vm()?Reflect.construct(e,r||[],Dl(t).constructor):e.apply(t,r))}function kO(t,e){if(e&&(fi(e)==="object"||typeof e=="function"))return e;if(e!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return TO(t)}function TO(t){if(t===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return t}function Vm(){try{var t=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(Vm=function(){return!!t})()}function Dl(t){return Dl=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(r){return r.__proto__||Object.getPrototypeOf(r)},Dl(t)}function CO(t,e){if(typeof e!="function"&&e!==null)throw new TypeError("Super expression must either be null or a function");t.prototype=Object.create(e&&e.prototype,{constructor:{value:t,writable:!0,configurable:!0}}),Object.defineProperty(t,"prototype",{writable:!1}),e&&Qc(t,e)}function Qc(t,e){return Qc=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(i,l){return i.__proto__=l,i},Qc(t,e)}function Zc(t,e,r){return e=Km(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function Km(t){var e=$O(t,"string");return fi(e)=="symbol"?e:e+""}function $O(t,e){if(fi(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(fi(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(t)}var Ay=1,IO=(function(t){function e(){var r;AO(this,e);for(var i=arguments.length,l=new Array(i),s=0;s<i;s++)l[s]=arguments[s];return r=_O(this,e,[].concat(l)),Zc(r,"state",{dismissed:!1,dismissedAtCoordinate:{x:0,y:0},lastBoundingBox:{width:-1,height:-1}}),Zc(r,"handleKeyDown",function(c){if(c.key==="Escape"){var d,p,v,h;r.setState({dismissed:!0,dismissedAtCoordinate:{x:(d=(p=r.props.coordinate)===null||p===void 0?void 0:p.x)!==null&&d!==void 0?d:0,y:(v=(h=r.props.coordinate)===null||h===void 0?void 0:h.y)!==null&&v!==void 0?v:0}})}}),r}return CO(e,t),jO(e,[{key:"updateBBox",value:function(){if(this.wrapperNode&&this.wrapperNode.getBoundingClientRect){var i=this.wrapperNode.getBoundingClientRect();(Math.abs(i.width-this.state.lastBoundingBox.width)>Ay||Math.abs(i.height-this.state.lastBoundingBox.height)>Ay)&&this.setState({lastBoundingBox:{width:i.width,height:i.height}})}else(this.state.lastBoundingBox.width!==-1||this.state.lastBoundingBox.height!==-1)&&this.setState({lastBoundingBox:{width:-1,height:-1}})}},{key:"componentDidMount",value:function(){document.addEventListener("keydown",this.handleKeyDown),this.updateBBox()}},{key:"componentWillUnmount",value:function(){document.removeEventListener("keydown",this.handleKeyDown)}},{key:"componentDidUpdate",value:function(){var i,l;this.props.active&&this.updateBBox(),this.state.dismissed&&(((i=this.props.coordinate)===null||i===void 0?void 0:i.x)!==this.state.dismissedAtCoordinate.x||((l=this.props.coordinate)===null||l===void 0?void 0:l.y)!==this.state.dismissedAtCoordinate.y)&&(this.state.dismissed=!1)}},{key:"render",value:function(){var i=this,l=this.props,s=l.active,c=l.allowEscapeViewBox,d=l.animationDuration,p=l.animationEasing,v=l.children,h=l.coordinate,m=l.hasPayload,b=l.isAnimationActive,O=l.offset,S=l.position,P=l.reverseDirection,g=l.useTranslate3d,A=l.viewBox,j=l.wrapperStyle,_=PO({allowEscapeViewBox:c,coordinate:h,offsetTopLeft:O,position:S,reverseDirection:P,tooltipBox:this.state.lastBoundingBox,useTranslate3d:g,viewBox:A}),k=_.cssClasses,w=_.cssProperties,E=Py(Py({transition:b&&s?"transform ".concat(d,"ms ").concat(p):void 0},w),{},{pointerEvents:"none",visibility:!this.state.dismissed&&s&&m?"visible":"hidden",position:"absolute",top:0,left:0},j);return $.createElement("div",{tabIndex:-1,className:k,style:E,ref:function(M){i.wrapperNode=M}},v)}}])})(te.PureComponent),MO=function(){return!(typeof window<"u"&&window.document&&window.document.createElement&&window.setTimeout)},Sn={isSsr:MO()};function di(t){"@babel/helpers - typeof";return di=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},di(t)}function Ey(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function jy(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?Ey(Object(r),!0).forEach(function(i){qf(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):Ey(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function NO(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")}function DO(t,e){for(var r=0;r<e.length;r++){var i=e[r];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(t,Gm(i.key),i)}}function LO(t,e,r){return e&&DO(t.prototype,e),Object.defineProperty(t,"prototype",{writable:!1}),t}function RO(t,e,r){return e=Ll(e),BO(t,Hm()?Reflect.construct(e,r||[],Ll(t).constructor):e.apply(t,r))}function BO(t,e){if(e&&(di(e)==="object"||typeof e=="function"))return e;if(e!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return zO(t)}function zO(t){if(t===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return t}function Hm(){try{var t=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(Hm=function(){return!!t})()}function Ll(t){return Ll=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(r){return r.__proto__||Object.getPrototypeOf(r)},Ll(t)}function FO(t,e){if(typeof e!="function"&&e!==null)throw new TypeError("Super expression must either be null or a function");t.prototype=Object.create(e&&e.prototype,{constructor:{value:t,writable:!0,configurable:!0}}),Object.defineProperty(t,"prototype",{writable:!1}),e&&Jc(t,e)}function Jc(t,e){return Jc=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(i,l){return i.__proto__=l,i},Jc(t,e)}function qf(t,e,r){return e=Gm(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function Gm(t){var e=WO(t,"string");return di(e)=="symbol"?e:e+""}function WO(t,e){if(di(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(di(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(t)}function UO(t){return t.dataKey}function VO(t,e){return $.isValidElement(t)?$.cloneElement(t,e):typeof t=="function"?$.createElement(t,e):$.createElement(gO,e)}var mr=(function(t){function e(){return NO(this,e),RO(this,e,arguments)}return FO(e,t),LO(e,[{key:"render",value:function(){var i=this,l=this.props,s=l.active,c=l.allowEscapeViewBox,d=l.animationDuration,p=l.animationEasing,v=l.content,h=l.coordinate,m=l.filterNull,b=l.isAnimationActive,O=l.offset,S=l.payload,P=l.payloadUniqBy,g=l.position,A=l.reverseDirection,j=l.useTranslate3d,_=l.viewBox,k=l.wrapperStyle,w=S??[];m&&w.length&&(w=Fm(S.filter(function(C){return C.value!=null&&(C.hide!==!0||i.props.includeHidden)}),P,UO));var E=w.length>0;return $.createElement(IO,{allowEscapeViewBox:c,animationDuration:d,animationEasing:p,isAnimationActive:b,active:s,coordinate:h,hasPayload:E,offset:O,position:g,reverseDirection:A,useTranslate3d:j,viewBox:_,wrapperStyle:k},VO(v,jy(jy({},this.props),{},{payload:w})))}}])})(te.PureComponent);qf(mr,"displayName","Tooltip");qf(mr,"defaultProps",{accessibilityLayer:!1,allowEscapeViewBox:{x:!1,y:!1},animationDuration:400,animationEasing:"ease",contentStyle:{},coordinate:{x:0,y:0},cursor:!0,cursorStyle:{},filterNull:!0,isAnimationActive:!Sn.isSsr,itemStyle:{},labelStyle:{},offset:10,reverseDirection:{x:!1,y:!1},separator:" : ",trigger:"hover",useTranslate3d:!1,viewBox:{x:0,y:0,height:0,width:0},wrapperStyle:{}});function Da(t){"@babel/helpers - typeof";return Da=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},Da(t)}function _y(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function Ol(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?_y(Object(r),!0).forEach(function(i){KO(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):_y(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function KO(t,e,r){return e=HO(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function HO(t){var e=GO(t,"string");return Da(e)=="symbol"?e:e+""}function GO(t,e){if(Da(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(Da(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}function XO(t,e){return ZO(t)||QO(t,e)||qO(t,e)||YO()}function YO(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function qO(t,e){if(t){if(typeof t=="string")return ky(t,e);var r=Object.prototype.toString.call(t).slice(8,-1);if(r==="Object"&&t.constructor&&(r=t.constructor.name),r==="Map"||r==="Set")return Array.from(t);if(r==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(r))return ky(t,e)}}function ky(t,e){(e==null||e>t.length)&&(e=t.length);for(var r=0,i=new Array(e);r<e;r++)i[r]=t[r];return i}function QO(t,e){var r=t==null?null:typeof Symbol<"u"&&t[Symbol.iterator]||t["@@iterator"];if(r!=null){var i,l,s,c,d=[],p=!0,v=!1;try{if(s=(r=r.call(t)).next,e!==0)for(;!(p=(i=s.call(r)).done)&&(d.push(i.value),d.length!==e);p=!0);}catch(h){v=!0,l=h}finally{try{if(!p&&r.return!=null&&(c=r.return(),Object(c)!==c))return}finally{if(v)throw l}}return d}}function ZO(t){if(Array.isArray(t))return t}var gC=te.forwardRef(function(t,e){var r=t.aspect,i=t.initialDimension,l=i===void 0?{width:-1,height:-1}:i,s=t.width,c=s===void 0?"100%":s,d=t.height,p=d===void 0?"100%":d,v=t.minWidth,h=v===void 0?0:v,m=t.minHeight,b=t.maxHeight,O=t.children,S=t.debounce,P=S===void 0?0:S,g=t.id,A=t.className,j=t.onResize,_=t.style,k=_===void 0?{}:_,w=te.useRef(null),E=te.useRef();E.current=j,te.useImperativeHandle(e,function(){return Object.defineProperty(w.current,"current",{get:function(){return console.warn("The usage of ref.current.current is deprecated and will no longer be supported."),w.current},configurable:!0})});var C=te.useState({containerWidth:l.width,containerHeight:l.height}),M=XO(C,2),D=M[0],U=M[1],F=te.useCallback(function(q,Y){U(function(V){var G=Math.round(q),ee=Math.round(Y);return V.containerWidth===G&&V.containerHeight===ee?V:{containerWidth:G,containerHeight:ee}})},[]);te.useEffect(function(){var q=function(X){var K,ae=X[0].contentRect,re=ae.width,de=ae.height;F(re,de),(K=E.current)===null||K===void 0||K.call(E,re,de)};P>0&&(q=Im(q,P,{trailing:!0,leading:!1}));var Y=new ResizeObserver(q),V=w.current.getBoundingClientRect(),G=V.width,ee=V.height;return F(G,ee),Y.observe(w.current),function(){Y.disconnect()}},[F,P]);var H=te.useMemo(function(){var q=D.containerWidth,Y=D.containerHeight;if(q<0||Y<0)return null;Or(mn(c)||mn(p),`The width(%s) and height(%s) are both fixed numbers,
       maybe you don't need to use a ResponsiveContainer.`,c,p),Or(!r||r>0,"The aspect(%s) must be greater than zero.",r);var V=mn(c)?q:c,G=mn(p)?Y:p;r&&r>0&&(V?G=V/r:G&&(V=G*r),b&&G>b&&(G=b)),Or(V>0||G>0,`The width(%s) and height(%s) of chart should be greater than 0,
       please check the style of container, or the props width(%s) and height(%s),
       or add a minWidth(%s) or minHeight(%s) or use aspect(%s) to control the
       height and width.`,V,G,c,p,h,m,r);var ee=!Array.isArray(O)&&xr(O.type).endsWith("Chart");return $.Children.map(O,function(z){return $.isValidElement(z)?te.cloneElement(z,Ol({width:V,height:G},ee?{style:Ol({height:"100%",width:"100%",maxHeight:G,maxWidth:V},z.props.style)}:{})):z})},[r,O,p,b,m,h,D,c]);return $.createElement("div",{id:g?"".concat(g):void 0,className:ge("recharts-responsive-container",A),style:Ol(Ol({},k),{},{width:c,height:p,minWidth:h,minHeight:m,maxHeight:b}),ref:w},H)}),Xm=function(e){return null};Xm.displayName="Cell";function La(t){"@babel/helpers - typeof";return La=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},La(t)}function Ty(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function ef(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?Ty(Object(r),!0).forEach(function(i){JO(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):Ty(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function JO(t,e,r){return e=eS(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function eS(t){var e=tS(t,"string");return La(e)=="symbol"?e:e+""}function tS(t,e){if(La(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(La(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}var Zn={widthCache:{},cacheCount:0},rS=2e3,nS={position:"absolute",top:"-20000px",left:0,padding:0,margin:0,border:"none",whiteSpace:"pre"},Cy="recharts_measurement_span";function iS(t){var e=ef({},t);return Object.keys(e).forEach(function(r){e[r]||delete e[r]}),e}var Ea=function(e){var r=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{};if(e==null||Sn.isSsr)return{width:0,height:0};var i=iS(r),l=JSON.stringify({text:e,copyStyle:i});if(Zn.widthCache[l])return Zn.widthCache[l];try{var s=document.getElementById(Cy);s||(s=document.createElement("span"),s.setAttribute("id",Cy),s.setAttribute("aria-hidden","true"),document.body.appendChild(s));var c=ef(ef({},nS),i);Object.assign(s.style,c),s.textContent="".concat(e);var d=s.getBoundingClientRect(),p={width:d.width,height:d.height};return Zn.widthCache[l]=p,++Zn.cacheCount>rS&&(Zn.cacheCount=0,Zn.widthCache={}),p}catch{return{width:0,height:0}}},aS=function(e){return{top:e.top+window.scrollY-document.documentElement.clientTop,left:e.left+window.scrollX-document.documentElement.clientLeft}};function Ra(t){"@babel/helpers - typeof";return Ra=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},Ra(t)}function Rl(t,e){return sS(t)||uS(t,e)||lS(t,e)||oS()}function oS(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function lS(t,e){if(t){if(typeof t=="string")return $y(t,e);var r=Object.prototype.toString.call(t).slice(8,-1);if(r==="Object"&&t.constructor&&(r=t.constructor.name),r==="Map"||r==="Set")return Array.from(t);if(r==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(r))return $y(t,e)}}function $y(t,e){(e==null||e>t.length)&&(e=t.length);for(var r=0,i=new Array(e);r<e;r++)i[r]=t[r];return i}function uS(t,e){var r=t==null?null:typeof Symbol<"u"&&t[Symbol.iterator]||t["@@iterator"];if(r!=null){var i,l,s,c,d=[],p=!0,v=!1;try{if(s=(r=r.call(t)).next,e===0){if(Object(r)!==r)return;p=!1}else for(;!(p=(i=s.call(r)).done)&&(d.push(i.value),d.length!==e);p=!0);}catch(h){v=!0,l=h}finally{try{if(!p&&r.return!=null&&(c=r.return(),Object(c)!==c))return}finally{if(v)throw l}}return d}}function sS(t){if(Array.isArray(t))return t}function cS(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")}function Iy(t,e){for(var r=0;r<e.length;r++){var i=e[r];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(t,dS(i.key),i)}}function fS(t,e,r){return e&&Iy(t.prototype,e),r&&Iy(t,r),Object.defineProperty(t,"prototype",{writable:!1}),t}function dS(t){var e=pS(t,"string");return Ra(e)=="symbol"?e:e+""}function pS(t,e){if(Ra(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(Ra(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(t)}var My=/(-?\d+(?:\.\d+)?[a-zA-Z%]*)([*/])(-?\d+(?:\.\d+)?[a-zA-Z%]*)/,Ny=/(-?\d+(?:\.\d+)?[a-zA-Z%]*)([+-])(-?\d+(?:\.\d+)?[a-zA-Z%]*)/,vS=/^px|cm|vh|vw|em|rem|%|mm|in|pt|pc|ex|ch|vmin|vmax|Q$/,yS=/(-?\d+(?:\.\d+)?)([a-zA-Z%]+)?/,Ym={cm:96/2.54,mm:96/25.4,pt:96/72,pc:96/6,in:96,Q:96/(2.54*40),px:1},hS=Object.keys(Ym),ei="NaN";function mS(t,e){return t*Ym[e]}var Sl=(function(){function t(e,r){cS(this,t),this.num=e,this.unit=r,this.num=e,this.unit=r,Number.isNaN(e)&&(this.unit=""),r!==""&&!vS.test(r)&&(this.num=NaN,this.unit=""),hS.includes(r)&&(this.num=mS(e,r),this.unit="px")}return fS(t,[{key:"add",value:function(r){return this.unit!==r.unit?new t(NaN,""):new t(this.num+r.num,this.unit)}},{key:"subtract",value:function(r){return this.unit!==r.unit?new t(NaN,""):new t(this.num-r.num,this.unit)}},{key:"multiply",value:function(r){return this.unit!==""&&r.unit!==""&&this.unit!==r.unit?new t(NaN,""):new t(this.num*r.num,this.unit||r.unit)}},{key:"divide",value:function(r){return this.unit!==""&&r.unit!==""&&this.unit!==r.unit?new t(NaN,""):new t(this.num/r.num,this.unit||r.unit)}},{key:"toString",value:function(){return"".concat(this.num).concat(this.unit)}},{key:"isNaN",value:function(){return Number.isNaN(this.num)}}],[{key:"parse",value:function(r){var i,l=(i=yS.exec(r))!==null&&i!==void 0?i:[],s=Rl(l,3),c=s[1],d=s[2];return new t(parseFloat(c),d??"")}}])})();function qm(t){if(t.includes(ei))return ei;for(var e=t;e.includes("*")||e.includes("/");){var r,i=(r=My.exec(e))!==null&&r!==void 0?r:[],l=Rl(i,4),s=l[1],c=l[2],d=l[3],p=Sl.parse(s??""),v=Sl.parse(d??""),h=c==="*"?p.multiply(v):p.divide(v);if(h.isNaN())return ei;e=e.replace(My,h.toString())}for(;e.includes("+")||/.-\d+(?:\.\d+)?/.test(e);){var m,b=(m=Ny.exec(e))!==null&&m!==void 0?m:[],O=Rl(b,4),S=O[1],P=O[2],g=O[3],A=Sl.parse(S??""),j=Sl.parse(g??""),_=P==="+"?A.add(j):A.subtract(j);if(_.isNaN())return ei;e=e.replace(Ny,_.toString())}return e}var Dy=/\(([^()]*)\)/;function gS(t){for(var e=t;e.includes("(");){var r=Dy.exec(e),i=Rl(r,2),l=i[1];e=e.replace(Dy,qm(l))}return e}function bS(t){var e=t.replace(/\s+/g,"");return e=gS(e),e=qm(e),e}function wS(t){try{return bS(t)}catch{return ei}}function jc(t){var e=wS(t.slice(5,-1));return e===ei?"":e}var xS=["x","y","lineHeight","capHeight","scaleToFit","textAnchor","verticalAnchor","fill"],OS=["dx","dy","angle","className","breakAll"];function tf(){return tf=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},tf.apply(this,arguments)}function Ly(t,e){if(t==null)return{};var r=SS(t,e),i,l;if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(t);for(l=0;l<s.length;l++)i=s[l],!(e.indexOf(i)>=0)&&Object.prototype.propertyIsEnumerable.call(t,i)&&(r[i]=t[i])}return r}function SS(t,e){if(t==null)return{};var r={};for(var i in t)if(Object.prototype.hasOwnProperty.call(t,i)){if(e.indexOf(i)>=0)continue;r[i]=t[i]}return r}function Ry(t,e){return jS(t)||ES(t,e)||AS(t,e)||PS()}function PS(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function AS(t,e){if(t){if(typeof t=="string")return By(t,e);var r=Object.prototype.toString.call(t).slice(8,-1);if(r==="Object"&&t.constructor&&(r=t.constructor.name),r==="Map"||r==="Set")return Array.from(t);if(r==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(r))return By(t,e)}}function By(t,e){(e==null||e>t.length)&&(e=t.length);for(var r=0,i=new Array(e);r<e;r++)i[r]=t[r];return i}function ES(t,e){var r=t==null?null:typeof Symbol<"u"&&t[Symbol.iterator]||t["@@iterator"];if(r!=null){var i,l,s,c,d=[],p=!0,v=!1;try{if(s=(r=r.call(t)).next,e===0){if(Object(r)!==r)return;p=!1}else for(;!(p=(i=s.call(r)).done)&&(d.push(i.value),d.length!==e);p=!0);}catch(h){v=!0,l=h}finally{try{if(!p&&r.return!=null&&(c=r.return(),Object(c)!==c))return}finally{if(v)throw l}}return d}}function jS(t){if(Array.isArray(t))return t}var Qm=/[ \f\n\r\t\v\u2028\u2029]+/,Zm=function(e){var r=e.children,i=e.breakAll,l=e.style;try{var s=[];be(r)||(i?s=r.toString().split(""):s=r.toString().split(Qm));var c=s.map(function(p){return{word:p,width:Ea(p,l).width}}),d=i?0:Ea(" ",l).width;return{wordsWithComputedWidth:c,spaceWidth:d}}catch{return null}},_S=function(e,r,i,l,s){var c=e.maxLines,d=e.children,p=e.style,v=e.breakAll,h=ie(c),m=d,b=function(){var V=arguments.length>0&&arguments[0]!==void 0?arguments[0]:[];return V.reduce(function(G,ee){var z=ee.word,X=ee.width,K=G[G.length-1];if(K&&(l==null||s||K.width+X+i<Number(l)))K.words.push(z),K.width+=X+i;else{var ae={words:[z],width:X};G.push(ae)}return G},[])},O=b(r),S=function(V){return V.reduce(function(G,ee){return G.width>ee.width?G:ee})};if(!h)return O;for(var P="…",g=function(V){var G=m.slice(0,V),ee=Zm({breakAll:v,style:p,children:G+P}).wordsWithComputedWidth,z=b(ee),X=z.length>c||S(z).width>Number(l);return[X,z]},A=0,j=m.length-1,_=0,k;A<=j&&_<=m.length-1;){var w=Math.floor((A+j)/2),E=w-1,C=g(E),M=Ry(C,2),D=M[0],U=M[1],F=g(w),H=Ry(F,1),q=H[0];if(!D&&!q&&(A=w+1),D&&q&&(j=w-1),!D&&q){k=U;break}_++}return k||O},zy=function(e){var r=be(e)?[]:e.toString().split(Qm);return[{words:r}]},kS=function(e){var r=e.width,i=e.scaleToFit,l=e.children,s=e.style,c=e.breakAll,d=e.maxLines;if((r||i)&&!Sn.isSsr){var p,v,h=Zm({breakAll:c,children:l,style:s});if(h){var m=h.wordsWithComputedWidth,b=h.spaceWidth;p=m,v=b}else return zy(l);return _S({breakAll:c,children:l,maxLines:d,style:s},p,v,r,i)}return zy(l)},Fy="#808080",pi=function(e){var r=e.x,i=r===void 0?0:r,l=e.y,s=l===void 0?0:l,c=e.lineHeight,d=c===void 0?"1em":c,p=e.capHeight,v=p===void 0?"0.71em":p,h=e.scaleToFit,m=h===void 0?!1:h,b=e.textAnchor,O=b===void 0?"start":b,S=e.verticalAnchor,P=S===void 0?"end":S,g=e.fill,A=g===void 0?Fy:g,j=Ly(e,xS),_=te.useMemo(function(){return kS({breakAll:j.breakAll,children:j.children,maxLines:j.maxLines,scaleToFit:m,style:j.style,width:j.width})},[j.breakAll,j.children,j.maxLines,m,j.style,j.width]),k=j.dx,w=j.dy,E=j.angle,C=j.className,M=j.breakAll,D=Ly(j,OS);if(!Ge(i)||!Ge(s))return null;var U=i+(ie(k)?k:0),F=s+(ie(w)?w:0),H;switch(P){case"start":H=jc("calc(".concat(v,")"));break;case"middle":H=jc("calc(".concat((_.length-1)/2," * -").concat(d," + (").concat(v," / 2))"));break;default:H=jc("calc(".concat(_.length-1," * -").concat(d,")"));break}var q=[];if(m){var Y=_[0].width,V=j.width;q.push("scale(".concat((ie(V)?V/Y:1)/Y,")"))}return E&&q.push("rotate(".concat(E,", ").concat(U,", ").concat(F,")")),q.length&&(D.transform=q.join(" ")),$.createElement("text",tf({},fe(D,!0),{x:U,y:F,className:ge("recharts-text",C),textAnchor:O,fill:A.includes("url")?Fy:A}),_.map(function(G,ee){var z=G.words.join(M?"":" ");return $.createElement("tspan",{x:U,dy:ee===0?H:d,key:"".concat(z,"-").concat(ee)},z)}))},TS="Invariant failed";function xn(t,e){throw new Error(TS)}var CS=["offset","layout","width","dataKey","data","dataPointFormatter","xAxis","yAxis"];function vi(t){"@babel/helpers - typeof";return vi=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},vi(t)}function Bl(){return Bl=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},Bl.apply(this,arguments)}function $S(t,e){return DS(t)||NS(t,e)||MS(t,e)||IS()}function IS(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function MS(t,e){if(t){if(typeof t=="string")return Wy(t,e);var r=Object.prototype.toString.call(t).slice(8,-1);if(r==="Object"&&t.constructor&&(r=t.constructor.name),r==="Map"||r==="Set")return Array.from(t);if(r==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(r))return Wy(t,e)}}function Wy(t,e){(e==null||e>t.length)&&(e=t.length);for(var r=0,i=new Array(e);r<e;r++)i[r]=t[r];return i}function NS(t,e){var r=t==null?null:typeof Symbol<"u"&&t[Symbol.iterator]||t["@@iterator"];if(r!=null){var i,l,s,c,d=[],p=!0,v=!1;try{if(s=(r=r.call(t)).next,e!==0)for(;!(p=(i=s.call(r)).done)&&(d.push(i.value),d.length!==e);p=!0);}catch(h){v=!0,l=h}finally{try{if(!p&&r.return!=null&&(c=r.return(),Object(c)!==c))return}finally{if(v)throw l}}return d}}function DS(t){if(Array.isArray(t))return t}function LS(t,e){if(t==null)return{};var r=RS(t,e),i,l;if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(t);for(l=0;l<s.length;l++)i=s[l],!(e.indexOf(i)>=0)&&Object.prototype.propertyIsEnumerable.call(t,i)&&(r[i]=t[i])}return r}function RS(t,e){if(t==null)return{};var r={};for(var i in t)if(Object.prototype.hasOwnProperty.call(t,i)){if(e.indexOf(i)>=0)continue;r[i]=t[i]}return r}function BS(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")}function zS(t,e){for(var r=0;r<e.length;r++){var i=e[r];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(t,tg(i.key),i)}}function FS(t,e,r){return e&&zS(t.prototype,e),Object.defineProperty(t,"prototype",{writable:!1}),t}function WS(t,e,r){return e=zl(e),US(t,Jm()?Reflect.construct(e,r||[],zl(t).constructor):e.apply(t,r))}function US(t,e){if(e&&(vi(e)==="object"||typeof e=="function"))return e;if(e!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return VS(t)}function VS(t){if(t===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return t}function Jm(){try{var t=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(Jm=function(){return!!t})()}function zl(t){return zl=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(r){return r.__proto__||Object.getPrototypeOf(r)},zl(t)}function KS(t,e){if(typeof e!="function"&&e!==null)throw new TypeError("Super expression must either be null or a function");t.prototype=Object.create(e&&e.prototype,{constructor:{value:t,writable:!0,configurable:!0}}),Object.defineProperty(t,"prototype",{writable:!1}),e&&rf(t,e)}function rf(t,e){return rf=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(i,l){return i.__proto__=l,i},rf(t,e)}function eg(t,e,r){return e=tg(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function tg(t){var e=HS(t,"string");return vi(e)=="symbol"?e:e+""}function HS(t,e){if(vi(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(vi(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(t)}var mu=(function(t){function e(){return BS(this,e),WS(this,e,arguments)}return KS(e,t),FS(e,[{key:"render",value:function(){var i=this.props,l=i.offset,s=i.layout,c=i.width,d=i.dataKey,p=i.data,v=i.dataPointFormatter,h=i.xAxis,m=i.yAxis,b=LS(i,CS),O=fe(b,!1);this.props.direction==="x"&&h.type!=="number"&&xn();var S=p.map(function(P){var g=v(P,d),A=g.x,j=g.y,_=g.value,k=g.errorVal;if(!k)return null;var w=[],E,C;if(Array.isArray(k)){var M=$S(k,2);E=M[0],C=M[1]}else E=C=k;if(s==="vertical"){var D=h.scale,U=j+l,F=U+c,H=U-c,q=D(_-E),Y=D(_+C);w.push({x1:Y,y1:F,x2:Y,y2:H}),w.push({x1:q,y1:U,x2:Y,y2:U}),w.push({x1:q,y1:F,x2:q,y2:H})}else if(s==="horizontal"){var V=m.scale,G=A+l,ee=G-c,z=G+c,X=V(_-E),K=V(_+C);w.push({x1:ee,y1:K,x2:z,y2:K}),w.push({x1:G,y1:X,x2:G,y2:K}),w.push({x1:ee,y1:X,x2:z,y2:X})}return $.createElement(Se,Bl({className:"recharts-errorBar",key:"bar-".concat(w.map(function(ae){return"".concat(ae.x1,"-").concat(ae.x2,"-").concat(ae.y1,"-").concat(ae.y2)}))},O),w.map(function(ae){return $.createElement("line",Bl({},ae,{key:"line-".concat(ae.x1,"-").concat(ae.x2,"-").concat(ae.y1,"-").concat(ae.y2)}))}))});return $.createElement(Se,{className:"recharts-errorBars"},S)}}])})($.Component);eg(mu,"defaultProps",{stroke:"black",strokeWidth:1.5,width:5,offset:0,layout:"horizontal"});eg(mu,"displayName","ErrorBar");function Ba(t){"@babel/helpers - typeof";return Ba=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},Ba(t)}function Uy(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function pn(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?Uy(Object(r),!0).forEach(function(i){GS(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):Uy(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function GS(t,e,r){return e=XS(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function XS(t){var e=YS(t,"string");return Ba(e)=="symbol"?e:e+""}function YS(t,e){if(Ba(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(Ba(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}var rg=function(e){var r=e.children,i=e.formattedGraphicalItems,l=e.legendWidth,s=e.legendContent,c=Pt(r,ai);if(!c)return null;var d=ai.defaultProps,p=d!==void 0?pn(pn({},d),c.props):{},v;return c.props&&c.props.payload?v=c.props&&c.props.payload:s==="children"?v=(i||[]).reduce(function(h,m){var b=m.item,O=m.props,S=O.sectors||O.data||[];return h.concat(S.map(function(P){return{type:c.props.iconType||b.props.legendType,value:P.name,color:P.fill,payload:P}}))},[]):v=(i||[]).map(function(h){var m=h.item,b=m.type.defaultProps,O=b!==void 0?pn(pn({},b),m.props):{},S=O.dataKey,P=O.name,g=O.legendType,A=O.hide;return{inactive:A,dataKey:S,type:p.iconType||g||"square",color:Qf(m),value:P||S,payload:O}}),pn(pn(pn({},p),ai.getWithHeight(c,l)),{},{payload:v,item:c})};function za(t){"@babel/helpers - typeof";return za=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},za(t)}function Vy(t){return JS(t)||ZS(t)||QS(t)||qS()}function qS(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function QS(t,e){if(t){if(typeof t=="string")return nf(t,e);var r=Object.prototype.toString.call(t).slice(8,-1);if(r==="Object"&&t.constructor&&(r=t.constructor.name),r==="Map"||r==="Set")return Array.from(t);if(r==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(r))return nf(t,e)}}function ZS(t){if(typeof Symbol<"u"&&t[Symbol.iterator]!=null||t["@@iterator"]!=null)return Array.from(t)}function JS(t){if(Array.isArray(t))return nf(t)}function nf(t,e){(e==null||e>t.length)&&(e=t.length);for(var r=0,i=new Array(e);r<e;r++)i[r]=t[r];return i}function Ky(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function Le(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?Ky(Object(r),!0).forEach(function(i){oi(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):Ky(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function oi(t,e,r){return e=eP(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function eP(t){var e=tP(t,"string");return za(e)=="symbol"?e:e+""}function tP(t,e){if(za(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(za(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}function ft(t,e,r){return be(t)||be(e)?r:Ge(e)?Mt(t,e,r):he(e)?e(t):r}function ja(t,e,r,i){var l=$w(t,function(d){return ft(d,e)});if(r==="number"){var s=l.filter(function(d){return ie(d)||parseFloat(d)});return s.length?[yu(s),qr(s)]:[1/0,-1/0]}var c=i?l.filter(function(d){return!be(d)}):l;return c.map(function(d){return Ge(d)||d instanceof Date?d:""})}var rP=function(e){var r,i=arguments.length>1&&arguments[1]!==void 0?arguments[1]:[],l=arguments.length>2?arguments[2]:void 0,s=arguments.length>3?arguments[3]:void 0,c=-1,d=(r=i==null?void 0:i.length)!==null&&r!==void 0?r:0;if(d<=1)return 0;if(s&&s.axisType==="angleAxis"&&Math.abs(Math.abs(s.range[1]-s.range[0])-360)<=1e-6)for(var p=s.range,v=0;v<d;v++){var h=v>0?l[v-1].coordinate:l[d-1].coordinate,m=l[v].coordinate,b=v>=d-1?l[0].coordinate:l[v+1].coordinate,O=void 0;if(Ht(m-h)!==Ht(b-m)){var S=[];if(Ht(b-m)===Ht(p[1]-p[0])){O=b;var P=m+p[1]-p[0];S[0]=Math.min(P,(P+h)/2),S[1]=Math.max(P,(P+h)/2)}else{O=h;var g=b+p[1]-p[0];S[0]=Math.min(m,(g+m)/2),S[1]=Math.max(m,(g+m)/2)}var A=[Math.min(m,(O+m)/2),Math.max(m,(O+m)/2)];if(e>A[0]&&e<=A[1]||e>=S[0]&&e<=S[1]){c=l[v].index;break}}else{var j=Math.min(h,b),_=Math.max(h,b);if(e>(j+m)/2&&e<=(_+m)/2){c=l[v].index;break}}}else for(var k=0;k<d;k++)if(k===0&&e<=(i[k].coordinate+i[k+1].coordinate)/2||k>0&&k<d-1&&e>(i[k].coordinate+i[k-1].coordinate)/2&&e<=(i[k].coordinate+i[k+1].coordinate)/2||k===d-1&&e>(i[k].coordinate+i[k-1].coordinate)/2){c=i[k].index;break}return c},Qf=function(e){var r,i=e,l=i.type.displayName,s=(r=e.type)!==null&&r!==void 0&&r.defaultProps?Le(Le({},e.type.defaultProps),e.props):e.props,c=s.stroke,d=s.fill,p;switch(l){case"Line":p=c;break;case"Area":case"Radar":p=c&&c!=="none"?c:d;break;default:p=d;break}return p},nP=function(e){var r=e.barSize,i=e.totalSize,l=e.stackGroups,s=l===void 0?{}:l;if(!s)return{};for(var c={},d=Object.keys(s),p=0,v=d.length;p<v;p++)for(var h=s[d[p]].stackGroups,m=Object.keys(h),b=0,O=m.length;b<O;b++){var S=h[m[b]],P=S.items,g=S.cateAxisId,A=P.filter(function(C){return xr(C.type).indexOf("Bar")>=0});if(A&&A.length){var j=A[0].type.defaultProps,_=j!==void 0?Le(Le({},j),A[0].props):A[0].props,k=_.barSize,w=_[g];c[w]||(c[w]=[]);var E=be(k)?r:k;c[w].push({item:A[0],stackList:A.slice(1),barSize:be(E)?void 0:Gt(E,i,0)})}}return c},iP=function(e){var r=e.barGap,i=e.barCategoryGap,l=e.bandSize,s=e.sizeList,c=s===void 0?[]:s,d=e.maxBarSize,p=c.length;if(p<1)return null;var v=Gt(r,l,0,!0),h,m=[];if(c[0].barSize===+c[0].barSize){var b=!1,O=l/p,S=c.reduce(function(k,w){return k+w.barSize||0},0);S+=(p-1)*v,S>=l&&(S-=(p-1)*v,v=0),S>=l&&O>0&&(b=!0,O*=.9,S=p*O);var P=(l-S)/2>>0,g={offset:P-v,size:0};h=c.reduce(function(k,w){var E={item:w.item,position:{offset:g.offset+g.size+v,size:b?O:w.barSize}},C=[].concat(Vy(k),[E]);return g=C[C.length-1].position,w.stackList&&w.stackList.length&&w.stackList.forEach(function(M){C.push({item:M,position:g})}),C},m)}else{var A=Gt(i,l,0,!0);l-2*A-(p-1)*v<=0&&(v=0);var j=(l-2*A-(p-1)*v)/p;j>1&&(j>>=0);var _=d===+d?Math.min(j,d):j;h=c.reduce(function(k,w,E){var C=[].concat(Vy(k),[{item:w.item,position:{offset:A+(j+v)*E+(j-_)/2,size:_}}]);return w.stackList&&w.stackList.length&&w.stackList.forEach(function(M){C.push({item:M,position:C[C.length-1].position})}),C},m)}return h},aP=function(e,r,i,l){var s=i.children,c=i.width,d=i.margin,p=c-(d.left||0)-(d.right||0),v=rg({children:s,legendWidth:p});if(v){var h=l||{},m=h.width,b=h.height,O=v.align,S=v.verticalAlign,P=v.layout;if((P==="vertical"||P==="horizontal"&&S==="middle")&&O!=="center"&&ie(e[O]))return Le(Le({},e),{},oi({},O,e[O]+(m||0)));if((P==="horizontal"||P==="vertical"&&O==="center")&&S!=="middle"&&ie(e[S]))return Le(Le({},e),{},oi({},S,e[S]+(b||0)))}return e},oP=function(e,r,i){return be(r)?!0:e==="horizontal"?r==="yAxis":e==="vertical"||i==="x"?r==="xAxis":i==="y"?r==="yAxis":!0},ng=function(e,r,i,l,s){var c=r.props.children,d=Xt(c,mu).filter(function(v){return oP(l,s,v.props.direction)});if(d&&d.length){var p=d.map(function(v){return v.props.dataKey});return e.reduce(function(v,h){var m=ft(h,i);if(be(m))return v;var b=Array.isArray(m)?[yu(m),qr(m)]:[m,m],O=p.reduce(function(S,P){var g=ft(h,P,0),A=b[0]-Math.abs(Array.isArray(g)?g[0]:g),j=b[1]+Math.abs(Array.isArray(g)?g[1]:g);return[Math.min(A,S[0]),Math.max(j,S[1])]},[1/0,-1/0]);return[Math.min(O[0],v[0]),Math.max(O[1],v[1])]},[1/0,-1/0])}return null},lP=function(e,r,i,l,s){var c=r.map(function(d){return ng(e,d,i,s,l)}).filter(function(d){return!be(d)});return c&&c.length?c.reduce(function(d,p){return[Math.min(d[0],p[0]),Math.max(d[1],p[1])]},[1/0,-1/0]):null},ig=function(e,r,i,l,s){var c=r.map(function(p){var v=p.props.dataKey;return i==="number"&&v&&ng(e,p,v,l)||ja(e,v,i,s)});if(i==="number")return c.reduce(function(p,v){return[Math.min(p[0],v[0]),Math.max(p[1],v[1])]},[1/0,-1/0]);var d={};return c.reduce(function(p,v){for(var h=0,m=v.length;h<m;h++)d[v[h]]||(d[v[h]]=!0,p.push(v[h]));return p},[])},ag=function(e,r){return e==="horizontal"&&r==="xAxis"||e==="vertical"&&r==="yAxis"||e==="centric"&&r==="angleAxis"||e==="radial"&&r==="radiusAxis"},og=function(e,r,i,l){if(l)return e.map(function(p){return p.coordinate});var s,c,d=e.map(function(p){return p.coordinate===r&&(s=!0),p.coordinate===i&&(c=!0),p.coordinate});return s||d.push(r),c||d.push(i),d},wr=function(e,r,i){if(!e)return null;var l=e.scale,s=e.duplicateDomain,c=e.type,d=e.range,p=e.realScaleType==="scaleBand"?l.bandwidth()/2:2,v=(r||i)&&c==="category"&&l.bandwidth?l.bandwidth()/p:0;if(v=e.axisType==="angleAxis"&&(d==null?void 0:d.length)>=2?Ht(d[0]-d[1])*2*v:v,r&&(e.ticks||e.niceTicks)){var h=(e.ticks||e.niceTicks).map(function(m){var b=s?s.indexOf(m):m;return{coordinate:l(b)+v,value:m,offset:v}});return h.filter(function(m){return!Ci(m.coordinate)})}return e.isCategorical&&e.categoricalDomain?e.categoricalDomain.map(function(m,b){return{coordinate:l(m)+v,value:m,index:b,offset:v}}):l.ticks&&!i?l.ticks(e.tickCount).map(function(m){return{coordinate:l(m)+v,value:m,offset:v}}):l.domain().map(function(m,b){return{coordinate:l(m)+v,value:s?s[m]:m,index:b,offset:v}})},_c=new WeakMap,Pl=function(e,r){if(typeof r!="function")return e;_c.has(e)||_c.set(e,new WeakMap);var i=_c.get(e);if(i.has(r))return i.get(r);var l=function(){e.apply(void 0,arguments),r.apply(void 0,arguments)};return i.set(r,l),l},lg=function(e,r,i){var l=e.scale,s=e.type,c=e.layout,d=e.axisType;if(l==="auto")return c==="radial"&&d==="radiusAxis"?{scale:Zv(),realScaleType:"band"}:c==="radial"&&d==="angleAxis"?{scale:Jv(),realScaleType:"linear"}:s==="category"&&r&&(r.indexOf("LineChart")>=0||r.indexOf("AreaChart")>=0||r.indexOf("ComposedChart")>=0&&!i)?{scale:Tl(),realScaleType:"point"}:s==="category"?{scale:Zv(),realScaleType:"band"}:{scale:Jv(),realScaleType:"linear"};if(wn(l)){var p="scale".concat(vu(l));return{scale:(ey[p]||Tl)(),realScaleType:ey[p]?p:"point"}}return he(l)?{scale:l}:{scale:Tl(),realScaleType:"point"}},Hy=1e-4,ug=function(e){var r=e.domain();if(!(!r||r.length<=2)){var i=r.length,l=e.range(),s=Math.min(l[0],l[1])-Hy,c=Math.max(l[0],l[1])+Hy,d=e(r[0]),p=e(r[i-1]);(d<s||d>c||p<s||p>c)&&e.domain([r[0],r[i-1]])}},uP=function(e,r){if(!e)return null;for(var i=0,l=e.length;i<l;i++)if(e[i].item===r)return e[i].position;return null},sP=function(e,r){if(!r||r.length!==2||!ie(r[0])||!ie(r[1]))return e;var i=Math.min(r[0],r[1]),l=Math.max(r[0],r[1]),s=[e[0],e[1]];return(!ie(e[0])||e[0]<i)&&(s[0]=i),(!ie(e[1])||e[1]>l)&&(s[1]=l),s[0]>l&&(s[0]=l),s[1]<i&&(s[1]=i),s},cP=function(e){var r=e.length;if(!(r<=0))for(var i=0,l=e[0].length;i<l;++i)for(var s=0,c=0,d=0;d<r;++d){var p=Ci(e[d][i][1])?e[d][i][0]:e[d][i][1];p>=0?(e[d][i][0]=s,e[d][i][1]=s+p,s=e[d][i][1]):(e[d][i][0]=c,e[d][i][1]=c+p,c=e[d][i][1])}},fP=function(e){var r=e.length;if(!(r<=0))for(var i=0,l=e[0].length;i<l;++i)for(var s=0,c=0;c<r;++c){var d=Ci(e[c][i][1])?e[c][i][0]:e[c][i][1];d>=0?(e[c][i][0]=s,e[c][i][1]=s+d,s=e[c][i][1]):(e[c][i][0]=0,e[c][i][1]=0)}},dP={sign:cP,expand:Cw,none:Tw,silhouette:kw,wiggle:_w,positive:fP},pP=function(e,r,i){var l=r.map(function(d){return d.props.dataKey}),s=dP[i],c=Ew().keys(l).value(function(d,p){return+ft(d,p,0)}).order(jw).offset(s);return c(e)},vP=function(e,r,i,l,s,c){if(!e)return null;var d=c?r.reverse():r,p={},v=d.reduce(function(m,b){var O,S=(O=b.type)!==null&&O!==void 0&&O.defaultProps?Le(Le({},b.type.defaultProps),b.props):b.props,P=S.stackId,g=S.hide;if(g)return m;var A=S[i],j=m[A]||{hasStack:!1,stackGroups:{}};if(Ge(P)){var _=j.stackGroups[P]||{numericAxisId:i,cateAxisId:l,items:[]};_.items.push(b),j.hasStack=!0,j.stackGroups[P]=_}else j.stackGroups[uo("_stackId_")]={numericAxisId:i,cateAxisId:l,items:[b]};return Le(Le({},m),{},oi({},A,j))},p),h={};return Object.keys(v).reduce(function(m,b){var O=v[b];if(O.hasStack){var S={};O.stackGroups=Object.keys(O.stackGroups).reduce(function(P,g){var A=O.stackGroups[g];return Le(Le({},P),{},oi({},g,{numericAxisId:i,cateAxisId:l,items:A.items,stackedData:pP(e,A.items,s)}))},S)}return Le(Le({},m),{},oi({},b,O))},h)},sg=function(e,r){var i=r.realScaleType,l=r.type,s=r.tickCount,c=r.originalDomain,d=r.allowDecimals,p=i||r.scale;if(p!=="auto"&&p!=="linear")return null;if(s&&l==="number"&&c&&(c[0]==="auto"||c[1]==="auto")){var v=e.domain();if(!v.length)return null;var h=Iw(v,s,d);return e.domain([yu(h),qr(h)]),{niceTicks:h}}if(s&&l==="number"){var m=e.domain(),b=Mw(m,s,d);return{niceTicks:b}}return null};function Gy(t){var e=t.axis,r=t.ticks,i=t.bandSize,l=t.entry,s=t.index,c=t.dataKey;if(e.type==="category"){if(!e.allowDuplicatedCategory&&e.dataKey&&!be(l[e.dataKey])){var d=$l(r,"value",l[e.dataKey]);if(d)return d.coordinate+i/2}return r[s]?r[s].coordinate+i/2:null}var p=ft(l,be(c)?e.dataKey:c);return be(p)?null:e.scale(p)}var Xy=function(e){var r=e.axis,i=e.ticks,l=e.offset,s=e.bandSize,c=e.entry,d=e.index;if(r.type==="category")return i[d]?i[d].coordinate+l:null;var p=ft(c,r.dataKey,r.domain[d]);return be(p)?null:r.scale(p)-s/2+l},yP=function(e){var r=e.numericAxis,i=r.scale.domain();if(r.type==="number"){var l=Math.min(i[0],i[1]),s=Math.max(i[0],i[1]);return l<=0&&s>=0?0:s<0?s:l}return i[0]},hP=function(e,r){var i,l=(i=e.type)!==null&&i!==void 0&&i.defaultProps?Le(Le({},e.type.defaultProps),e.props):e.props,s=l.stackId;if(Ge(s)){var c=r[s];if(c){var d=c.items.indexOf(e);return d>=0?c.stackedData[d]:null}}return null},mP=function(e){return e.reduce(function(r,i){return[yu(i.concat([r[0]]).filter(ie)),qr(i.concat([r[1]]).filter(ie))]},[1/0,-1/0])},cg=function(e,r,i){return Object.keys(e).reduce(function(l,s){var c=e[s],d=c.stackedData,p=d.reduce(function(v,h){var m=mP(h.slice(r,i+1));return[Math.min(v[0],m[0]),Math.max(v[1],m[1])]},[1/0,-1/0]);return[Math.min(p[0],l[0]),Math.max(p[1],l[1])]},[1/0,-1/0]).map(function(l){return l===1/0||l===-1/0?0:l})},Yy=/^dataMin[\s]*-[\s]*([0-9]+([.]{1}[0-9]+){0,1})$/,qy=/^dataMax[\s]*\+[\s]*([0-9]+([.]{1}[0-9]+){0,1})$/,af=function(e,r,i){if(he(e))return e(r,i);if(!Array.isArray(e))return r;var l=[];if(ie(e[0]))l[0]=i?e[0]:Math.min(e[0],r[0]);else if(Yy.test(e[0])){var s=+Yy.exec(e[0])[1];l[0]=r[0]-s}else he(e[0])?l[0]=e[0](r[0]):l[0]=r[0];if(ie(e[1]))l[1]=i?e[1]:Math.max(e[1],r[1]);else if(qy.test(e[1])){var c=+qy.exec(e[1])[1];l[1]=r[1]+c}else he(e[1])?l[1]=e[1](r[1]):l[1]=r[1];return l},Fl=function(e,r,i){if(e&&e.scale&&e.scale.bandwidth){var l=e.scale.bandwidth();if(!i||l>0)return l}if(e&&r&&r.length>=2){for(var s=Kf(r,function(m){return m.coordinate}),c=1/0,d=1,p=s.length;d<p;d++){var v=s[d],h=s[d-1];c=Math.min((v.coordinate||0)-(h.coordinate||0),c)}return c===1/0?0:c}return i?void 0:0},Qy=function(e,r,i){return!e||!e.length||li(e,Mt(i,"type.defaultProps.domain"))?r:e},fg=function(e,r){var i=e.type.defaultProps?Le(Le({},e.type.defaultProps),e.props):e.props,l=i.dataKey,s=i.name,c=i.unit,d=i.formatter,p=i.tooltipType,v=i.chartType,h=i.hide;return Le(Le({},fe(e,!1)),{},{dataKey:l,unit:c,formatter:d,name:s||l,color:Qf(e),value:ft(r,l),type:p,payload:r,chartType:v,hide:h})};function Fa(t){"@babel/helpers - typeof";return Fa=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},Fa(t)}function Zy(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function gr(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?Zy(Object(r),!0).forEach(function(i){dg(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):Zy(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function dg(t,e,r){return e=gP(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function gP(t){var e=bP(t,"string");return Fa(e)=="symbol"?e:e+""}function bP(t,e){if(Fa(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(Fa(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}function wP(t,e){return PP(t)||SP(t,e)||OP(t,e)||xP()}function xP(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function OP(t,e){if(t){if(typeof t=="string")return Jy(t,e);var r=Object.prototype.toString.call(t).slice(8,-1);if(r==="Object"&&t.constructor&&(r=t.constructor.name),r==="Map"||r==="Set")return Array.from(t);if(r==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(r))return Jy(t,e)}}function Jy(t,e){(e==null||e>t.length)&&(e=t.length);for(var r=0,i=new Array(e);r<e;r++)i[r]=t[r];return i}function SP(t,e){var r=t==null?null:typeof Symbol<"u"&&t[Symbol.iterator]||t["@@iterator"];if(r!=null){var i,l,s,c,d=[],p=!0,v=!1;try{if(s=(r=r.call(t)).next,e!==0)for(;!(p=(i=s.call(r)).done)&&(d.push(i.value),d.length!==e);p=!0);}catch(h){v=!0,l=h}finally{try{if(!p&&r.return!=null&&(c=r.return(),Object(c)!==c))return}finally{if(v)throw l}}return d}}function PP(t){if(Array.isArray(t))return t}var Wl=Math.PI/180,AP=function(e){return e*180/Math.PI},je=function(e,r,i,l){return{x:e+Math.cos(-Wl*l)*i,y:r+Math.sin(-Wl*l)*i}},EP=function(e,r){var i=arguments.length>2&&arguments[2]!==void 0?arguments[2]:{top:0,right:0,bottom:0,left:0};return Math.min(Math.abs(e-(i.left||0)-(i.right||0)),Math.abs(r-(i.top||0)-(i.bottom||0)))/2},jP=function(e,r,i,l,s){var c=e.width,d=e.height,p=e.startAngle,v=e.endAngle,h=Gt(e.cx,c,c/2),m=Gt(e.cy,d,d/2),b=EP(c,d,i),O=Gt(e.innerRadius,b,0),S=Gt(e.outerRadius,b,b*.8),P=Object.keys(r);return P.reduce(function(g,A){var j=r[A],_=j.domain,k=j.reversed,w;if(be(j.range))l==="angleAxis"?w=[p,v]:l==="radiusAxis"&&(w=[O,S]),k&&(w=[w[1],w[0]]);else{w=j.range;var E=w,C=wP(E,2);p=C[0],v=C[1]}var M=lg(j,s),D=M.realScaleType,U=M.scale;U.domain(_).range(w),ug(U);var F=sg(U,gr(gr({},j),{},{realScaleType:D})),H=gr(gr(gr({},j),F),{},{range:w,radius:S,realScaleType:D,scale:U,cx:h,cy:m,innerRadius:O,outerRadius:S,startAngle:p,endAngle:v});return gr(gr({},g),{},dg({},A,H))},{})},_P=function(e,r){var i=e.x,l=e.y,s=r.x,c=r.y;return Math.sqrt(Math.pow(i-s,2)+Math.pow(l-c,2))},kP=function(e,r){var i=e.x,l=e.y,s=r.cx,c=r.cy,d=_P({x:i,y:l},{x:s,y:c});if(d<=0)return{radius:d};var p=(i-s)/d,v=Math.acos(p);return l>c&&(v=2*Math.PI-v),{radius:d,angle:AP(v),angleInRadian:v}},TP=function(e){var r=e.startAngle,i=e.endAngle,l=Math.floor(r/360),s=Math.floor(i/360),c=Math.min(l,s);return{startAngle:r-c*360,endAngle:i-c*360}},CP=function(e,r){var i=r.startAngle,l=r.endAngle,s=Math.floor(i/360),c=Math.floor(l/360),d=Math.min(s,c);return e+d*360},eh=function(e,r){var i=e.x,l=e.y,s=kP({x:i,y:l},r),c=s.radius,d=s.angle,p=r.innerRadius,v=r.outerRadius;if(c<p||c>v)return!1;if(c===0)return!0;var h=TP(r),m=h.startAngle,b=h.endAngle,O=d,S;if(m<=b){for(;O>b;)O-=360;for(;O<m;)O+=360;S=O>=m&&O<=b}else{for(;O>m;)O-=360;for(;O<b;)O+=360;S=O>=b&&O<=m}return S?gr(gr({},r),{},{radius:c,angle:CP(O,r)}):null},pg=function(e){return!te.isValidElement(e)&&!he(e)&&typeof e!="boolean"?e.className:""};function Wa(t){"@babel/helpers - typeof";return Wa=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},Wa(t)}var $P=["offset"];function IP(t){return LP(t)||DP(t)||NP(t)||MP()}function MP(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function NP(t,e){if(t){if(typeof t=="string")return of(t,e);var r=Object.prototype.toString.call(t).slice(8,-1);if(r==="Object"&&t.constructor&&(r=t.constructor.name),r==="Map"||r==="Set")return Array.from(t);if(r==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(r))return of(t,e)}}function DP(t){if(typeof Symbol<"u"&&t[Symbol.iterator]!=null||t["@@iterator"]!=null)return Array.from(t)}function LP(t){if(Array.isArray(t))return of(t)}function of(t,e){(e==null||e>t.length)&&(e=t.length);for(var r=0,i=new Array(e);r<e;r++)i[r]=t[r];return i}function RP(t,e){if(t==null)return{};var r=BP(t,e),i,l;if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(t);for(l=0;l<s.length;l++)i=s[l],!(e.indexOf(i)>=0)&&Object.prototype.propertyIsEnumerable.call(t,i)&&(r[i]=t[i])}return r}function BP(t,e){if(t==null)return{};var r={};for(var i in t)if(Object.prototype.hasOwnProperty.call(t,i)){if(e.indexOf(i)>=0)continue;r[i]=t[i]}return r}function th(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function He(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?th(Object(r),!0).forEach(function(i){zP(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):th(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function zP(t,e,r){return e=FP(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function FP(t){var e=WP(t,"string");return Wa(e)=="symbol"?e:e+""}function WP(t,e){if(Wa(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(Wa(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}function Ua(){return Ua=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},Ua.apply(this,arguments)}var UP=function(e){var r=e.value,i=e.formatter,l=be(e.children)?r:e.children;return he(i)?i(l):l},VP=function(e,r){var i=Ht(r-e),l=Math.min(Math.abs(r-e),360);return i*l},KP=function(e,r,i){var l=e.position,s=e.viewBox,c=e.offset,d=e.className,p=s,v=p.cx,h=p.cy,m=p.innerRadius,b=p.outerRadius,O=p.startAngle,S=p.endAngle,P=p.clockWise,g=(m+b)/2,A=VP(O,S),j=A>=0?1:-1,_,k;l==="insideStart"?(_=O+j*c,k=P):l==="insideEnd"?(_=S-j*c,k=!P):l==="end"&&(_=S+j*c,k=P),k=A<=0?k:!k;var w=je(v,h,g,_),E=je(v,h,g,_+(k?1:-1)*359),C="M".concat(w.x,",").concat(w.y,`
    A`).concat(g,",").concat(g,",0,1,").concat(k?0:1,`,
    `).concat(E.x,",").concat(E.y),M=be(e.id)?uo("recharts-radial-line-"):e.id;return $.createElement("text",Ua({},i,{dominantBaseline:"central",className:ge("recharts-radial-bar-label",d)}),$.createElement("defs",null,$.createElement("path",{id:M,d:C})),$.createElement("textPath",{xlinkHref:"#".concat(M)},r))},HP=function(e){var r=e.viewBox,i=e.offset,l=e.position,s=r,c=s.cx,d=s.cy,p=s.innerRadius,v=s.outerRadius,h=s.startAngle,m=s.endAngle,b=(h+m)/2;if(l==="outside"){var O=je(c,d,v+i,b),S=O.x,P=O.y;return{x:S,y:P,textAnchor:S>=c?"start":"end",verticalAnchor:"middle"}}if(l==="center")return{x:c,y:d,textAnchor:"middle",verticalAnchor:"middle"};if(l==="centerTop")return{x:c,y:d,textAnchor:"middle",verticalAnchor:"start"};if(l==="centerBottom")return{x:c,y:d,textAnchor:"middle",verticalAnchor:"end"};var g=(p+v)/2,A=je(c,d,g,b),j=A.x,_=A.y;return{x:j,y:_,textAnchor:"middle",verticalAnchor:"middle"}},GP=function(e){var r=e.viewBox,i=e.parentViewBox,l=e.offset,s=e.position,c=r,d=c.x,p=c.y,v=c.width,h=c.height,m=h>=0?1:-1,b=m*l,O=m>0?"end":"start",S=m>0?"start":"end",P=v>=0?1:-1,g=P*l,A=P>0?"end":"start",j=P>0?"start":"end";if(s==="top"){var _={x:d+v/2,y:p-m*l,textAnchor:"middle",verticalAnchor:O};return He(He({},_),i?{height:Math.max(p-i.y,0),width:v}:{})}if(s==="bottom"){var k={x:d+v/2,y:p+h+b,textAnchor:"middle",verticalAnchor:S};return He(He({},k),i?{height:Math.max(i.y+i.height-(p+h),0),width:v}:{})}if(s==="left"){var w={x:d-g,y:p+h/2,textAnchor:A,verticalAnchor:"middle"};return He(He({},w),i?{width:Math.max(w.x-i.x,0),height:h}:{})}if(s==="right"){var E={x:d+v+g,y:p+h/2,textAnchor:j,verticalAnchor:"middle"};return He(He({},E),i?{width:Math.max(i.x+i.width-E.x,0),height:h}:{})}var C=i?{width:v,height:h}:{};return s==="insideLeft"?He({x:d+g,y:p+h/2,textAnchor:j,verticalAnchor:"middle"},C):s==="insideRight"?He({x:d+v-g,y:p+h/2,textAnchor:A,verticalAnchor:"middle"},C):s==="insideTop"?He({x:d+v/2,y:p+b,textAnchor:"middle",verticalAnchor:S},C):s==="insideBottom"?He({x:d+v/2,y:p+h-b,textAnchor:"middle",verticalAnchor:O},C):s==="insideTopLeft"?He({x:d+g,y:p+b,textAnchor:j,verticalAnchor:S},C):s==="insideTopRight"?He({x:d+v-g,y:p+b,textAnchor:A,verticalAnchor:S},C):s==="insideBottomLeft"?He({x:d+g,y:p+h-b,textAnchor:j,verticalAnchor:O},C):s==="insideBottomRight"?He({x:d+v-g,y:p+h-b,textAnchor:A,verticalAnchor:O},C):$i(s)&&(ie(s.x)||mn(s.x))&&(ie(s.y)||mn(s.y))?He({x:d+Gt(s.x,v),y:p+Gt(s.y,h),textAnchor:"end",verticalAnchor:"end"},C):He({x:d+v/2,y:p+h/2,textAnchor:"middle",verticalAnchor:"middle"},C)},XP=function(e){return"cx"in e&&ie(e.cx)};function et(t){var e=t.offset,r=e===void 0?5:e,i=RP(t,$P),l=He({offset:r},i),s=l.viewBox,c=l.position,d=l.value,p=l.children,v=l.content,h=l.className,m=h===void 0?"":h,b=l.textBreakAll;if(!s||be(d)&&be(p)&&!te.isValidElement(v)&&!he(v))return null;if(te.isValidElement(v))return te.cloneElement(v,l);var O;if(he(v)){if(O=te.createElement(v,l),te.isValidElement(O))return O}else O=UP(l);var S=XP(s),P=fe(l,!0);if(S&&(c==="insideStart"||c==="insideEnd"||c==="end"))return KP(l,O,P);var g=S?HP(l):GP(l);return $.createElement(pi,Ua({className:ge("recharts-label",m)},P,g,{breakAll:b}),O)}et.displayName="Label";var vg=function(e){var r=e.cx,i=e.cy,l=e.angle,s=e.startAngle,c=e.endAngle,d=e.r,p=e.radius,v=e.innerRadius,h=e.outerRadius,m=e.x,b=e.y,O=e.top,S=e.left,P=e.width,g=e.height,A=e.clockWise,j=e.labelViewBox;if(j)return j;if(ie(P)&&ie(g)){if(ie(m)&&ie(b))return{x:m,y:b,width:P,height:g};if(ie(O)&&ie(S))return{x:O,y:S,width:P,height:g}}return ie(m)&&ie(b)?{x:m,y:b,width:0,height:0}:ie(r)&&ie(i)?{cx:r,cy:i,startAngle:s||l||0,endAngle:c||l||0,innerRadius:v||0,outerRadius:h||p||d||0,clockWise:A}:e.viewBox?e.viewBox:{}},YP=function(e,r){return e?e===!0?$.createElement(et,{key:"label-implicit",viewBox:r}):Ge(e)?$.createElement(et,{key:"label-implicit",viewBox:r,value:e}):te.isValidElement(e)?e.type===et?te.cloneElement(e,{key:"label-implicit",viewBox:r}):$.createElement(et,{key:"label-implicit",content:e,viewBox:r}):he(e)?$.createElement(et,{key:"label-implicit",content:e,viewBox:r}):$i(e)?$.createElement(et,Ua({viewBox:r},e,{key:"label-implicit"})):null:null},qP=function(e,r){var i=arguments.length>2&&arguments[2]!==void 0?arguments[2]:!0;if(!e||!e.children&&i&&!e.label)return null;var l=e.children,s=vg(e),c=Xt(l,et).map(function(p,v){return te.cloneElement(p,{viewBox:r||s,key:"label-".concat(v)})});if(!i)return c;var d=YP(e.label,r||s);return[d].concat(IP(c))};et.parseViewBox=vg;et.renderCallByParent=qP;function Va(t){"@babel/helpers - typeof";return Va=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},Va(t)}var QP=["valueAccessor"],ZP=["data","dataKey","clockWise","id","textBreakAll"];function JP(t){return nA(t)||rA(t)||tA(t)||eA()}function eA(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function tA(t,e){if(t){if(typeof t=="string")return lf(t,e);var r=Object.prototype.toString.call(t).slice(8,-1);if(r==="Object"&&t.constructor&&(r=t.constructor.name),r==="Map"||r==="Set")return Array.from(t);if(r==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(r))return lf(t,e)}}function rA(t){if(typeof Symbol<"u"&&t[Symbol.iterator]!=null||t["@@iterator"]!=null)return Array.from(t)}function nA(t){if(Array.isArray(t))return lf(t)}function lf(t,e){(e==null||e>t.length)&&(e=t.length);for(var r=0,i=new Array(e);r<e;r++)i[r]=t[r];return i}function Ul(){return Ul=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},Ul.apply(this,arguments)}function rh(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function nh(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?rh(Object(r),!0).forEach(function(i){iA(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):rh(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function iA(t,e,r){return e=aA(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function aA(t){var e=oA(t,"string");return Va(e)=="symbol"?e:e+""}function oA(t,e){if(Va(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(Va(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}function ih(t,e){if(t==null)return{};var r=lA(t,e),i,l;if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(t);for(l=0;l<s.length;l++)i=s[l],!(e.indexOf(i)>=0)&&Object.prototype.propertyIsEnumerable.call(t,i)&&(r[i]=t[i])}return r}function lA(t,e){if(t==null)return{};var r={};for(var i in t)if(Object.prototype.hasOwnProperty.call(t,i)){if(e.indexOf(i)>=0)continue;r[i]=t[i]}return r}var uA=function(e){return Array.isArray(e.value)?Mm(e.value):e.value};function Sr(t){var e=t.valueAccessor,r=e===void 0?uA:e,i=ih(t,QP),l=i.data,s=i.dataKey,c=i.clockWise,d=i.id,p=i.textBreakAll,v=ih(i,ZP);return!l||!l.length?null:$.createElement(Se,{className:"recharts-label-list"},l.map(function(h,m){var b=be(s)?r(h,m):ft(h&&h.payload,s),O=be(d)?{}:{id:"".concat(d,"-").concat(m)};return $.createElement(et,Ul({},fe(h,!0),v,O,{parentViewBox:h.parentViewBox,value:b,textBreakAll:p,viewBox:et.parseViewBox(be(c)?h:nh(nh({},h),{},{clockWise:c})),key:"label-".concat(m),index:m}))}))}Sr.displayName="LabelList";function sA(t,e){return t?t===!0?$.createElement(Sr,{key:"labelList-implicit",data:e}):$.isValidElement(t)||he(t)?$.createElement(Sr,{key:"labelList-implicit",data:e,content:t}):$i(t)?$.createElement(Sr,Ul({data:e},t,{key:"labelList-implicit"})):null:null}function cA(t,e){var r=arguments.length>2&&arguments[2]!==void 0?arguments[2]:!0;if(!t||!t.children&&r&&!t.label)return null;var i=t.children,l=Xt(i,Sr).map(function(c,d){return te.cloneElement(c,{data:e,key:"labelList-".concat(d)})});if(!r)return l;var s=sA(t.label,e);return[s].concat(JP(l))}Sr.renderCallByParent=cA;function Ka(t){"@babel/helpers - typeof";return Ka=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},Ka(t)}function uf(){return uf=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},uf.apply(this,arguments)}function ah(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function oh(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?ah(Object(r),!0).forEach(function(i){fA(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):ah(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function fA(t,e,r){return e=dA(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function dA(t){var e=pA(t,"string");return Ka(e)=="symbol"?e:e+""}function pA(t,e){if(Ka(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(Ka(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}var vA=function(e,r){var i=Ht(r-e),l=Math.min(Math.abs(r-e),359.999);return i*l},Al=function(e){var r=e.cx,i=e.cy,l=e.radius,s=e.angle,c=e.sign,d=e.isExternal,p=e.cornerRadius,v=e.cornerIsExternal,h=p*(d?1:-1)+l,m=Math.asin(p/h)/Wl,b=v?s:s+c*m,O=je(r,i,h,b),S=je(r,i,l,b),P=v?s-c*m:s,g=je(r,i,h*Math.cos(m*Wl),P);return{center:O,circleTangency:S,lineTangency:g,theta:m}},yg=function(e){var r=e.cx,i=e.cy,l=e.innerRadius,s=e.outerRadius,c=e.startAngle,d=e.endAngle,p=vA(c,d),v=c+p,h=je(r,i,s,c),m=je(r,i,s,v),b="M ".concat(h.x,",").concat(h.y,`
    A `).concat(s,",").concat(s,`,0,
    `).concat(+(Math.abs(p)>180),",").concat(+(c>v),`,
    `).concat(m.x,",").concat(m.y,`
  `);if(l>0){var O=je(r,i,l,c),S=je(r,i,l,v);b+="L ".concat(S.x,",").concat(S.y,`
            A `).concat(l,",").concat(l,`,0,
            `).concat(+(Math.abs(p)>180),",").concat(+(c<=v),`,
            `).concat(O.x,",").concat(O.y," Z")}else b+="L ".concat(r,",").concat(i," Z");return b},yA=function(e){var r=e.cx,i=e.cy,l=e.innerRadius,s=e.outerRadius,c=e.cornerRadius,d=e.forceCornerRadius,p=e.cornerIsExternal,v=e.startAngle,h=e.endAngle,m=Ht(h-v),b=Al({cx:r,cy:i,radius:s,angle:v,sign:m,cornerRadius:c,cornerIsExternal:p}),O=b.circleTangency,S=b.lineTangency,P=b.theta,g=Al({cx:r,cy:i,radius:s,angle:h,sign:-m,cornerRadius:c,cornerIsExternal:p}),A=g.circleTangency,j=g.lineTangency,_=g.theta,k=p?Math.abs(v-h):Math.abs(v-h)-P-_;if(k<0)return d?"M ".concat(S.x,",").concat(S.y,`
        a`).concat(c,",").concat(c,",0,0,1,").concat(c*2,`,0
        a`).concat(c,",").concat(c,",0,0,1,").concat(-c*2,`,0
      `):yg({cx:r,cy:i,innerRadius:l,outerRadius:s,startAngle:v,endAngle:h});var w="M ".concat(S.x,",").concat(S.y,`
    A`).concat(c,",").concat(c,",0,0,").concat(+(m<0),",").concat(O.x,",").concat(O.y,`
    A`).concat(s,",").concat(s,",0,").concat(+(k>180),",").concat(+(m<0),",").concat(A.x,",").concat(A.y,`
    A`).concat(c,",").concat(c,",0,0,").concat(+(m<0),",").concat(j.x,",").concat(j.y,`
  `);if(l>0){var E=Al({cx:r,cy:i,radius:l,angle:v,sign:m,isExternal:!0,cornerRadius:c,cornerIsExternal:p}),C=E.circleTangency,M=E.lineTangency,D=E.theta,U=Al({cx:r,cy:i,radius:l,angle:h,sign:-m,isExternal:!0,cornerRadius:c,cornerIsExternal:p}),F=U.circleTangency,H=U.lineTangency,q=U.theta,Y=p?Math.abs(v-h):Math.abs(v-h)-D-q;if(Y<0&&c===0)return"".concat(w,"L").concat(r,",").concat(i,"Z");w+="L".concat(H.x,",").concat(H.y,`
      A`).concat(c,",").concat(c,",0,0,").concat(+(m<0),",").concat(F.x,",").concat(F.y,`
      A`).concat(l,",").concat(l,",0,").concat(+(Y>180),",").concat(+(m>0),",").concat(C.x,",").concat(C.y,`
      A`).concat(c,",").concat(c,",0,0,").concat(+(m<0),",").concat(M.x,",").concat(M.y,"Z")}else w+="L".concat(r,",").concat(i,"Z");return w},hA={cx:0,cy:0,innerRadius:0,outerRadius:0,startAngle:0,endAngle:0,cornerRadius:0,forceCornerRadius:!1,cornerIsExternal:!1},hg=function(e){var r=oh(oh({},hA),e),i=r.cx,l=r.cy,s=r.innerRadius,c=r.outerRadius,d=r.cornerRadius,p=r.forceCornerRadius,v=r.cornerIsExternal,h=r.startAngle,m=r.endAngle,b=r.className;if(c<s||h===m)return null;var O=ge("recharts-sector",b),S=c-s,P=Gt(d,S,0,!0),g;return P>0&&Math.abs(h-m)<360?g=yA({cx:i,cy:l,innerRadius:s,outerRadius:c,cornerRadius:Math.min(P,S/2),forceCornerRadius:p,cornerIsExternal:v,startAngle:h,endAngle:m}):g=yg({cx:i,cy:l,innerRadius:s,outerRadius:c,startAngle:h,endAngle:m}),$.createElement("path",uf({},fe(r,!0),{className:O,d:g,role:"img"}))};function Ha(t){"@babel/helpers - typeof";return Ha=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},Ha(t)}function sf(){return sf=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},sf.apply(this,arguments)}function lh(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function uh(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?lh(Object(r),!0).forEach(function(i){mA(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):lh(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function mA(t,e,r){return e=gA(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function gA(t){var e=bA(t,"string");return Ha(e)=="symbol"?e:e+""}function bA(t,e){if(Ha(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(Ha(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}var sh={curveBasisClosed:Gw,curveBasisOpen:Hw,curveBasis:Kw,curveBumpX:Vw,curveBumpY:Uw,curveLinearClosed:Ww,curveLinear:Nm,curveMonotoneX:Fw,curveMonotoneY:zw,curveNatural:Bw,curveStep:Rw,curveStepAfter:Lw,curveStepBefore:Dw},El=function(e){return e.x===+e.x&&e.y===+e.y},xa=function(e){return e.x},Oa=function(e){return e.y},wA=function(e,r){if(he(e))return e;var i="curve".concat(vu(e));return(i==="curveMonotone"||i==="curveBump")&&r?sh["".concat(i).concat(r==="vertical"?"Y":"X")]:sh[i]||Nm},xA=function(e){var r=e.type,i=r===void 0?"linear":r,l=e.points,s=l===void 0?[]:l,c=e.baseLine,d=e.layout,p=e.connectNulls,v=p===void 0?!1:p,h=wA(i,d),m=v?s.filter(function(P){return El(P)}):s,b;if(Array.isArray(c)){var O=v?c.filter(function(P){return El(P)}):c,S=m.map(function(P,g){return uh(uh({},P),{},{base:O[g]})});return d==="vertical"?b=wl().y(Oa).x1(xa).x0(function(P){return P.base.x}):b=wl().x(xa).y1(Oa).y0(function(P){return P.base.y}),b.defined(El).curve(h),b(S)}return d==="vertical"&&ie(c)?b=wl().y(Oa).x1(xa).x0(c):ie(c)?b=wl().x(xa).y1(Oa).y0(c):b=Nw().x(xa).y(Oa),b.defined(El).curve(h),b(m)},_a=function(e){var r=e.className,i=e.points,l=e.path,s=e.pathRef;if((!i||!i.length)&&!l)return null;var c=i&&i.length?xA(e):l;return te.createElement("path",sf({},fe(e,!1),Il(e),{className:ge("recharts-curve",r),d:c,ref:s}))},kc={exports:{}},Tc,ch;function OA(){if(ch)return Tc;ch=1;var t="SECRET_DO_NOT_PASS_THIS_OR_YOU_WILL_BE_FIRED";return Tc=t,Tc}var Cc,fh;function SA(){if(fh)return Cc;fh=1;var t=OA();function e(){}function r(){}return r.resetWarningCache=e,Cc=function(){function i(c,d,p,v,h,m){if(m!==t){var b=new Error("Calling PropTypes validators directly is not supported by the `prop-types` package. Use PropTypes.checkPropTypes() to call them. Read more at http://fb.me/use-check-prop-types");throw b.name="Invariant Violation",b}}i.isRequired=i;function l(){return i}var s={array:i,bigint:i,bool:i,func:i,number:i,object:i,string:i,symbol:i,any:i,arrayOf:l,element:i,elementType:i,instanceOf:l,node:i,objectOf:l,oneOf:l,oneOfType:l,shape:l,exact:l,checkPropTypes:r,resetWarningCache:e};return s.PropTypes=s,s},Cc}var dh;function PA(){return dh||(dh=1,kc.exports=SA()()),kc.exports}var AA=PA();const Ae=Cm(AA),{getOwnPropertyNames:EA,getOwnPropertySymbols:jA}=Object,{hasOwnProperty:_A}=Object.prototype;function $c(t,e){return function(i,l,s){return t(i,l,s)&&e(i,l,s)}}function jl(t){return function(r,i,l){if(!r||!i||typeof r!="object"||typeof i!="object")return t(r,i,l);const{cache:s}=l,c=s.get(r),d=s.get(i);if(c&&d)return c===i&&d===r;s.set(r,i),s.set(i,r);const p=t(r,i,l);return s.delete(r),s.delete(i),p}}function kA(t){return t!=null?t[Symbol.toStringTag]:void 0}function ph(t){return EA(t).concat(jA(t))}const TA=Object.hasOwn||((t,e)=>_A.call(t,e));function Pn(t,e){return t===e||!t&&!e&&t!==t&&e!==e}const CA="__v",$A="__o",IA="_owner",{getOwnPropertyDescriptor:vh,keys:yh}=Object;function MA(t,e){return t.byteLength===e.byteLength&&Vl(new Uint8Array(t),new Uint8Array(e))}function NA(t,e,r){let i=t.length;if(e.length!==i)return!1;for(;i-- >0;)if(!r.equals(t[i],e[i],i,i,t,e,r))return!1;return!0}function DA(t,e){return t.byteLength===e.byteLength&&Vl(new Uint8Array(t.buffer,t.byteOffset,t.byteLength),new Uint8Array(e.buffer,e.byteOffset,e.byteLength))}function LA(t,e){return Pn(t.getTime(),e.getTime())}function RA(t,e){return t.name===e.name&&t.message===e.message&&t.cause===e.cause&&t.stack===e.stack}function BA(t,e){return t===e}function hh(t,e,r){const i=t.size;if(i!==e.size)return!1;if(!i)return!0;const l=new Array(i),s=t.entries();let c,d,p=0;for(;(c=s.next())&&!c.done;){const v=e.entries();let h=!1,m=0;for(;(d=v.next())&&!d.done;){if(l[m]){m++;continue}const b=c.value,O=d.value;if(r.equals(b[0],O[0],p,m,t,e,r)&&r.equals(b[1],O[1],b[0],O[0],t,e,r)){h=l[m]=!0;break}m++}if(!h)return!1;p++}return!0}const zA=Pn;function FA(t,e,r){const i=yh(t);let l=i.length;if(yh(e).length!==l)return!1;for(;l-- >0;)if(!mg(t,e,r,i[l]))return!1;return!0}function Sa(t,e,r){const i=ph(t);let l=i.length;if(ph(e).length!==l)return!1;let s,c,d;for(;l-- >0;)if(s=i[l],!mg(t,e,r,s)||(c=vh(t,s),d=vh(e,s),(c||d)&&(!c||!d||c.configurable!==d.configurable||c.enumerable!==d.enumerable||c.writable!==d.writable)))return!1;return!0}function WA(t,e){return Pn(t.valueOf(),e.valueOf())}function UA(t,e){return t.source===e.source&&t.flags===e.flags}function mh(t,e,r){const i=t.size;if(i!==e.size)return!1;if(!i)return!0;const l=new Array(i),s=t.values();let c,d;for(;(c=s.next())&&!c.done;){const p=e.values();let v=!1,h=0;for(;(d=p.next())&&!d.done;){if(!l[h]&&r.equals(c.value,d.value,c.value,d.value,t,e,r)){v=l[h]=!0;break}h++}if(!v)return!1}return!0}function Vl(t,e){let r=t.byteLength;if(e.byteLength!==r||t.byteOffset!==e.byteOffset)return!1;for(;r-- >0;)if(t[r]!==e[r])return!1;return!0}function VA(t,e){return t.hostname===e.hostname&&t.pathname===e.pathname&&t.protocol===e.protocol&&t.port===e.port&&t.hash===e.hash&&t.username===e.username&&t.password===e.password}function mg(t,e,r,i){return(i===IA||i===$A||i===CA)&&(t.$$typeof||e.$$typeof)?!0:TA(e,i)&&r.equals(t[i],e[i],i,i,t,e,r)}const KA="[object ArrayBuffer]",HA="[object Arguments]",GA="[object Boolean]",XA="[object DataView]",YA="[object Date]",qA="[object Error]",QA="[object Map]",ZA="[object Number]",JA="[object Object]",e1="[object RegExp]",t1="[object Set]",r1="[object String]",n1={"[object Int8Array]":!0,"[object Uint8Array]":!0,"[object Uint8ClampedArray]":!0,"[object Int16Array]":!0,"[object Uint16Array]":!0,"[object Int32Array]":!0,"[object Uint32Array]":!0,"[object Float16Array]":!0,"[object Float32Array]":!0,"[object Float64Array]":!0,"[object BigInt64Array]":!0,"[object BigUint64Array]":!0},i1="[object URL]",a1=Object.prototype.toString;function o1({areArrayBuffersEqual:t,areArraysEqual:e,areDataViewsEqual:r,areDatesEqual:i,areErrorsEqual:l,areFunctionsEqual:s,areMapsEqual:c,areNumbersEqual:d,areObjectsEqual:p,arePrimitiveWrappersEqual:v,areRegExpsEqual:h,areSetsEqual:m,areTypedArraysEqual:b,areUrlsEqual:O,unknownTagComparators:S}){return function(g,A,j){if(g===A)return!0;if(g==null||A==null)return!1;const _=typeof g;if(_!==typeof A)return!1;if(_!=="object")return _==="number"?d(g,A,j):_==="function"?s(g,A,j):!1;const k=g.constructor;if(k!==A.constructor)return!1;if(k===Object)return p(g,A,j);if(Array.isArray(g))return e(g,A,j);if(k===Date)return i(g,A,j);if(k===RegExp)return h(g,A,j);if(k===Map)return c(g,A,j);if(k===Set)return m(g,A,j);const w=a1.call(g);if(w===YA)return i(g,A,j);if(w===e1)return h(g,A,j);if(w===QA)return c(g,A,j);if(w===t1)return m(g,A,j);if(w===JA)return typeof g.then!="function"&&typeof A.then!="function"&&p(g,A,j);if(w===i1)return O(g,A,j);if(w===qA)return l(g,A,j);if(w===HA)return p(g,A,j);if(n1[w])return b(g,A,j);if(w===KA)return t(g,A,j);if(w===XA)return r(g,A,j);if(w===GA||w===ZA||w===r1)return v(g,A,j);if(S){let E=S[w];if(!E){const C=kA(g);C&&(E=S[C])}if(E)return E(g,A,j)}return!1}}function l1({circular:t,createCustomConfig:e,strict:r}){let i={areArrayBuffersEqual:MA,areArraysEqual:r?Sa:NA,areDataViewsEqual:DA,areDatesEqual:LA,areErrorsEqual:RA,areFunctionsEqual:BA,areMapsEqual:r?$c(hh,Sa):hh,areNumbersEqual:zA,areObjectsEqual:r?Sa:FA,arePrimitiveWrappersEqual:WA,areRegExpsEqual:UA,areSetsEqual:r?$c(mh,Sa):mh,areTypedArraysEqual:r?$c(Vl,Sa):Vl,areUrlsEqual:VA,unknownTagComparators:void 0};if(e&&(i=Object.assign({},i,e(i))),t){const l=jl(i.areArraysEqual),s=jl(i.areMapsEqual),c=jl(i.areObjectsEqual),d=jl(i.areSetsEqual);i=Object.assign({},i,{areArraysEqual:l,areMapsEqual:s,areObjectsEqual:c,areSetsEqual:d})}return i}function u1(t){return function(e,r,i,l,s,c,d){return t(e,r,d)}}function s1({circular:t,comparator:e,createState:r,equals:i,strict:l}){if(r)return function(d,p){const{cache:v=t?new WeakMap:void 0,meta:h}=r();return e(d,p,{cache:v,equals:i,meta:h,strict:l})};if(t)return function(d,p){return e(d,p,{cache:new WeakMap,equals:i,meta:void 0,strict:l})};const s={cache:void 0,equals:i,meta:void 0,strict:l};return function(d,p){return e(d,p,s)}}const c1=Zr();Zr({strict:!0});Zr({circular:!0});Zr({circular:!0,strict:!0});Zr({createInternalComparator:()=>Pn});Zr({strict:!0,createInternalComparator:()=>Pn});Zr({circular:!0,createInternalComparator:()=>Pn});Zr({circular:!0,createInternalComparator:()=>Pn,strict:!0});function Zr(t={}){const{circular:e=!1,createInternalComparator:r,createState:i,strict:l=!1}=t,s=l1(t),c=o1(s),d=r?r(c):u1(c);return s1({circular:e,comparator:c,createState:i,equals:d,strict:l})}function f1(t){typeof requestAnimationFrame<"u"&&requestAnimationFrame(t)}function gh(t){var e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:0,r=-1,i=function l(s){r<0&&(r=s),s-r>e?(t(s),r=-1):f1(l)};requestAnimationFrame(i)}function cf(t){"@babel/helpers - typeof";return cf=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},cf(t)}function d1(t){return h1(t)||y1(t)||v1(t)||p1()}function p1(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function v1(t,e){if(t){if(typeof t=="string")return bh(t,e);var r=Object.prototype.toString.call(t).slice(8,-1);if(r==="Object"&&t.constructor&&(r=t.constructor.name),r==="Map"||r==="Set")return Array.from(t);if(r==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(r))return bh(t,e)}}function bh(t,e){(e==null||e>t.length)&&(e=t.length);for(var r=0,i=new Array(e);r<e;r++)i[r]=t[r];return i}function y1(t){if(typeof Symbol<"u"&&t[Symbol.iterator]!=null||t["@@iterator"]!=null)return Array.from(t)}function h1(t){if(Array.isArray(t))return t}function m1(){var t={},e=function(){return null},r=!1,i=function l(s){if(!r){if(Array.isArray(s)){if(!s.length)return;var c=s,d=d1(c),p=d[0],v=d.slice(1);if(typeof p=="number"){gh(l.bind(null,v),p);return}l(p),gh(l.bind(null,v));return}cf(s)==="object"&&(t=s,e(t)),typeof s=="function"&&s()}};return{stop:function(){r=!0},start:function(s){r=!1,i(s)},subscribe:function(s){return e=s,function(){e=function(){return null}}}}}function Ga(t){"@babel/helpers - typeof";return Ga=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},Ga(t)}function wh(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function xh(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?wh(Object(r),!0).forEach(function(i){gg(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):wh(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function gg(t,e,r){return e=g1(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function g1(t){var e=b1(t,"string");return Ga(e)==="symbol"?e:String(e)}function b1(t,e){if(Ga(t)!=="object"||t===null)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(Ga(i)!=="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}var w1=function(e,r){return[Object.keys(e),Object.keys(r)].reduce(function(i,l){return i.filter(function(s){return l.includes(s)})})},x1=function(e){return e},O1=function(e){return e.replace(/([A-Z])/g,function(r){return"-".concat(r.toLowerCase())})},ka=function(e,r){return Object.keys(r).reduce(function(i,l){return xh(xh({},i),{},gg({},l,e(l,r[l])))},{})},Oh=function(e,r,i){return e.map(function(l){return"".concat(O1(l)," ").concat(r,"ms ").concat(i)}).join(",")};function S1(t,e){return E1(t)||A1(t,e)||bg(t,e)||P1()}function P1(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function A1(t,e){var r=t==null?null:typeof Symbol<"u"&&t[Symbol.iterator]||t["@@iterator"];if(r!=null){var i,l,s,c,d=[],p=!0,v=!1;try{if(s=(r=r.call(t)).next,e!==0)for(;!(p=(i=s.call(r)).done)&&(d.push(i.value),d.length!==e);p=!0);}catch(h){v=!0,l=h}finally{try{if(!p&&r.return!=null&&(c=r.return(),Object(c)!==c))return}finally{if(v)throw l}}return d}}function E1(t){if(Array.isArray(t))return t}function j1(t){return T1(t)||k1(t)||bg(t)||_1()}function _1(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function bg(t,e){if(t){if(typeof t=="string")return ff(t,e);var r=Object.prototype.toString.call(t).slice(8,-1);if(r==="Object"&&t.constructor&&(r=t.constructor.name),r==="Map"||r==="Set")return Array.from(t);if(r==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(r))return ff(t,e)}}function k1(t){if(typeof Symbol<"u"&&t[Symbol.iterator]!=null||t["@@iterator"]!=null)return Array.from(t)}function T1(t){if(Array.isArray(t))return ff(t)}function ff(t,e){(e==null||e>t.length)&&(e=t.length);for(var r=0,i=new Array(e);r<e;r++)i[r]=t[r];return i}var Kl=1e-4,wg=function(e,r){return[0,3*e,3*r-6*e,3*e-3*r+1]},xg=function(e,r){return e.map(function(i,l){return i*Math.pow(r,l)}).reduce(function(i,l){return i+l})},Sh=function(e,r){return function(i){var l=wg(e,r);return xg(l,i)}},C1=function(e,r){return function(i){var l=wg(e,r),s=[].concat(j1(l.map(function(c,d){return c*d}).slice(1)),[0]);return xg(s,i)}},Ph=function(){for(var e=arguments.length,r=new Array(e),i=0;i<e;i++)r[i]=arguments[i];var l=r[0],s=r[1],c=r[2],d=r[3];if(r.length===1)switch(r[0]){case"linear":l=0,s=0,c=1,d=1;break;case"ease":l=.25,s=.1,c=.25,d=1;break;case"ease-in":l=.42,s=0,c=1,d=1;break;case"ease-out":l=.42,s=0,c=.58,d=1;break;case"ease-in-out":l=0,s=0,c=.58,d=1;break;default:{var p=r[0].split("(");if(p[0]==="cubic-bezier"&&p[1].split(")")[0].split(",").length===4){var v=p[1].split(")")[0].split(",").map(function(g){return parseFloat(g)}),h=S1(v,4);l=h[0],s=h[1],c=h[2],d=h[3]}}}var m=Sh(l,c),b=Sh(s,d),O=C1(l,c),S=function(A){return A>1?1:A<0?0:A},P=function(A){for(var j=A>1?1:A,_=j,k=0;k<8;++k){var w=m(_)-j,E=O(_);if(Math.abs(w-j)<Kl||E<Kl)return b(_);_=S(_-w/E)}return b(_)};return P.isStepper=!1,P},$1=function(){var e=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{},r=e.stiff,i=r===void 0?100:r,l=e.damping,s=l===void 0?8:l,c=e.dt,d=c===void 0?17:c,p=function(h,m,b){var O=-(h-m)*i,S=b*s,P=b+(O-S)*d/1e3,g=b*d/1e3+h;return Math.abs(g-m)<Kl&&Math.abs(P)<Kl?[m,0]:[g,P]};return p.isStepper=!0,p.dt=d,p},I1=function(){for(var e=arguments.length,r=new Array(e),i=0;i<e;i++)r[i]=arguments[i];var l=r[0];if(typeof l=="string")switch(l){case"ease":case"ease-in-out":case"ease-out":case"ease-in":case"linear":return Ph(l);case"spring":return $1();default:if(l.split("(")[0]==="cubic-bezier")return Ph(l)}return typeof l=="function"?l:null};function Xa(t){"@babel/helpers - typeof";return Xa=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},Xa(t)}function Ah(t){return D1(t)||N1(t)||Og(t)||M1()}function M1(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function N1(t){if(typeof Symbol<"u"&&t[Symbol.iterator]!=null||t["@@iterator"]!=null)return Array.from(t)}function D1(t){if(Array.isArray(t))return pf(t)}function Eh(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function Je(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?Eh(Object(r),!0).forEach(function(i){df(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):Eh(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function df(t,e,r){return e=L1(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function L1(t){var e=R1(t,"string");return Xa(e)==="symbol"?e:String(e)}function R1(t,e){if(Xa(t)!=="object"||t===null)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(Xa(i)!=="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}function B1(t,e){return W1(t)||F1(t,e)||Og(t,e)||z1()}function z1(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function Og(t,e){if(t){if(typeof t=="string")return pf(t,e);var r=Object.prototype.toString.call(t).slice(8,-1);if(r==="Object"&&t.constructor&&(r=t.constructor.name),r==="Map"||r==="Set")return Array.from(t);if(r==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(r))return pf(t,e)}}function pf(t,e){(e==null||e>t.length)&&(e=t.length);for(var r=0,i=new Array(e);r<e;r++)i[r]=t[r];return i}function F1(t,e){var r=t==null?null:typeof Symbol<"u"&&t[Symbol.iterator]||t["@@iterator"];if(r!=null){var i,l,s,c,d=[],p=!0,v=!1;try{if(s=(r=r.call(t)).next,e!==0)for(;!(p=(i=s.call(r)).done)&&(d.push(i.value),d.length!==e);p=!0);}catch(h){v=!0,l=h}finally{try{if(!p&&r.return!=null&&(c=r.return(),Object(c)!==c))return}finally{if(v)throw l}}return d}}function W1(t){if(Array.isArray(t))return t}var Hl=function(e,r,i){return e+(r-e)*i},vf=function(e){var r=e.from,i=e.to;return r!==i},U1=function t(e,r,i){var l=ka(function(s,c){if(vf(c)){var d=e(c.from,c.to,c.velocity),p=B1(d,2),v=p[0],h=p[1];return Je(Je({},c),{},{from:v,velocity:h})}return c},r);return i<1?ka(function(s,c){return vf(c)?Je(Je({},c),{},{velocity:Hl(c.velocity,l[s].velocity,i),from:Hl(c.from,l[s].from,i)}):c},r):t(e,l,i-1)};const V1=(function(t,e,r,i,l){var s=w1(t,e),c=s.reduce(function(g,A){return Je(Je({},g),{},df({},A,[t[A],e[A]]))},{}),d=s.reduce(function(g,A){return Je(Je({},g),{},df({},A,{from:t[A],velocity:0,to:e[A]}))},{}),p=-1,v,h,m=function(){return null},b=function(){return ka(function(A,j){return j.from},d)},O=function(){return!Object.values(d).filter(vf).length},S=function(A){v||(v=A);var j=A-v,_=j/r.dt;d=U1(r,d,_),l(Je(Je(Je({},t),e),b())),v=A,O()||(p=requestAnimationFrame(m))},P=function(A){h||(h=A);var j=(A-h)/i,_=ka(function(w,E){return Hl.apply(void 0,Ah(E).concat([r(j)]))},c);if(l(Je(Je(Je({},t),e),_)),j<1)p=requestAnimationFrame(m);else{var k=ka(function(w,E){return Hl.apply(void 0,Ah(E).concat([r(1)]))},c);l(Je(Je(Je({},t),e),k))}};return m=r.isStepper?S:P,function(){return requestAnimationFrame(m),function(){cancelAnimationFrame(p)}}});function yi(t){"@babel/helpers - typeof";return yi=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},yi(t)}var K1=["children","begin","duration","attributeName","easing","isActive","steps","from","to","canBegin","onAnimationEnd","shouldReAnimate","onAnimationReStart"];function H1(t,e){if(t==null)return{};var r=G1(t,e),i,l;if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(t);for(l=0;l<s.length;l++)i=s[l],!(e.indexOf(i)>=0)&&Object.prototype.propertyIsEnumerable.call(t,i)&&(r[i]=t[i])}return r}function G1(t,e){if(t==null)return{};var r={},i=Object.keys(t),l,s;for(s=0;s<i.length;s++)l=i[s],!(e.indexOf(l)>=0)&&(r[l]=t[l]);return r}function Ic(t){return Q1(t)||q1(t)||Y1(t)||X1()}function X1(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function Y1(t,e){if(t){if(typeof t=="string")return yf(t,e);var r=Object.prototype.toString.call(t).slice(8,-1);if(r==="Object"&&t.constructor&&(r=t.constructor.name),r==="Map"||r==="Set")return Array.from(t);if(r==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(r))return yf(t,e)}}function q1(t){if(typeof Symbol<"u"&&t[Symbol.iterator]!=null||t["@@iterator"]!=null)return Array.from(t)}function Q1(t){if(Array.isArray(t))return yf(t)}function yf(t,e){(e==null||e>t.length)&&(e=t.length);for(var r=0,i=new Array(e);r<e;r++)i[r]=t[r];return i}function jh(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function Vt(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?jh(Object(r),!0).forEach(function(i){Aa(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):jh(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function Aa(t,e,r){return e=Sg(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function Z1(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")}function J1(t,e){for(var r=0;r<e.length;r++){var i=e[r];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(t,Sg(i.key),i)}}function eE(t,e,r){return e&&J1(t.prototype,e),Object.defineProperty(t,"prototype",{writable:!1}),t}function Sg(t){var e=tE(t,"string");return yi(e)==="symbol"?e:String(e)}function tE(t,e){if(yi(t)!=="object"||t===null)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(yi(i)!=="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}function rE(t,e){if(typeof e!="function"&&e!==null)throw new TypeError("Super expression must either be null or a function");t.prototype=Object.create(e&&e.prototype,{constructor:{value:t,writable:!0,configurable:!0}}),Object.defineProperty(t,"prototype",{writable:!1}),e&&hf(t,e)}function hf(t,e){return hf=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(i,l){return i.__proto__=l,i},hf(t,e)}function nE(t){var e=iE();return function(){var i=Gl(t),l;if(e){var s=Gl(this).constructor;l=Reflect.construct(i,arguments,s)}else l=i.apply(this,arguments);return mf(this,l)}}function mf(t,e){if(e&&(yi(e)==="object"||typeof e=="function"))return e;if(e!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return gf(t)}function gf(t){if(t===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return t}function iE(){if(typeof Reflect>"u"||!Reflect.construct||Reflect.construct.sham)return!1;if(typeof Proxy=="function")return!0;try{return Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){})),!0}catch{return!1}}function Gl(t){return Gl=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(r){return r.__proto__||Object.getPrototypeOf(r)},Gl(t)}var lr=(function(t){rE(r,t);var e=nE(r);function r(i,l){var s;Z1(this,r),s=e.call(this,i,l);var c=s.props,d=c.isActive,p=c.attributeName,v=c.from,h=c.to,m=c.steps,b=c.children,O=c.duration;if(s.handleStyleChange=s.handleStyleChange.bind(gf(s)),s.changeStyle=s.changeStyle.bind(gf(s)),!d||O<=0)return s.state={style:{}},typeof b=="function"&&(s.state={style:h}),mf(s);if(m&&m.length)s.state={style:m[0].style};else if(v){if(typeof b=="function")return s.state={style:v},mf(s);s.state={style:p?Aa({},p,v):v}}else s.state={style:{}};return s}return eE(r,[{key:"componentDidMount",value:function(){var l=this.props,s=l.isActive,c=l.canBegin;this.mounted=!0,!(!s||!c)&&this.runAnimation(this.props)}},{key:"componentDidUpdate",value:function(l){var s=this.props,c=s.isActive,d=s.canBegin,p=s.attributeName,v=s.shouldReAnimate,h=s.to,m=s.from,b=this.state.style;if(d){if(!c){var O={style:p?Aa({},p,h):h};this.state&&b&&(p&&b[p]!==h||!p&&b!==h)&&this.setState(O);return}if(!(c1(l.to,h)&&l.canBegin&&l.isActive)){var S=!l.canBegin||!l.isActive;this.manager&&this.manager.stop(),this.stopJSAnimation&&this.stopJSAnimation();var P=S||v?m:l.to;if(this.state&&b){var g={style:p?Aa({},p,P):P};(p&&b[p]!==P||!p&&b!==P)&&this.setState(g)}this.runAnimation(Vt(Vt({},this.props),{},{from:P,begin:0}))}}}},{key:"componentWillUnmount",value:function(){this.mounted=!1;var l=this.props.onAnimationEnd;this.unSubscribe&&this.unSubscribe(),this.manager&&(this.manager.stop(),this.manager=null),this.stopJSAnimation&&this.stopJSAnimation(),l&&l()}},{key:"handleStyleChange",value:function(l){this.changeStyle(l)}},{key:"changeStyle",value:function(l){this.mounted&&this.setState({style:l})}},{key:"runJSAnimation",value:function(l){var s=this,c=l.from,d=l.to,p=l.duration,v=l.easing,h=l.begin,m=l.onAnimationEnd,b=l.onAnimationStart,O=V1(c,d,I1(v),p,this.changeStyle),S=function(){s.stopJSAnimation=O()};this.manager.start([b,h,S,p,m])}},{key:"runStepAnimation",value:function(l){var s=this,c=l.steps,d=l.begin,p=l.onAnimationStart,v=c[0],h=v.style,m=v.duration,b=m===void 0?0:m,O=function(P,g,A){if(A===0)return P;var j=g.duration,_=g.easing,k=_===void 0?"ease":_,w=g.style,E=g.properties,C=g.onAnimationEnd,M=A>0?c[A-1]:g,D=E||Object.keys(w);if(typeof k=="function"||k==="spring")return[].concat(Ic(P),[s.runJSAnimation.bind(s,{from:M.style,to:w,duration:j,easing:k}),j]);var U=Oh(D,j,k),F=Vt(Vt(Vt({},M.style),w),{},{transition:U});return[].concat(Ic(P),[F,j,C]).filter(x1)};return this.manager.start([p].concat(Ic(c.reduce(O,[h,Math.max(b,d)])),[l.onAnimationEnd]))}},{key:"runAnimation",value:function(l){this.manager||(this.manager=m1());var s=l.begin,c=l.duration,d=l.attributeName,p=l.to,v=l.easing,h=l.onAnimationStart,m=l.onAnimationEnd,b=l.steps,O=l.children,S=this.manager;if(this.unSubscribe=S.subscribe(this.handleStyleChange),typeof v=="function"||typeof O=="function"||v==="spring"){this.runJSAnimation(l);return}if(b.length>1){this.runStepAnimation(l);return}var P=d?Aa({},d,p):p,g=Oh(Object.keys(P),c,v);S.start([h,s,Vt(Vt({},P),{},{transition:g}),c,m])}},{key:"render",value:function(){var l=this.props,s=l.children;l.begin;var c=l.duration;l.attributeName,l.easing;var d=l.isActive;l.steps,l.from,l.to,l.canBegin,l.onAnimationEnd,l.shouldReAnimate,l.onAnimationReStart;var p=H1(l,K1),v=te.Children.count(s),h=this.state.style;if(typeof s=="function")return s(h);if(!d||v===0||c<=0)return s;var m=function(O){var S=O.props,P=S.style,g=P===void 0?{}:P,A=S.className,j=te.cloneElement(O,Vt(Vt({},p),{},{style:Vt(Vt({},g),h),className:A}));return j};return v===1?m(te.Children.only(s)):$.createElement("div",null,te.Children.map(s,function(b){return m(b)}))}}]),r})(te.PureComponent);lr.displayName="Animate";lr.defaultProps={begin:0,duration:1e3,from:"",to:"",attributeName:"",easing:"ease",isActive:!0,canBegin:!0,steps:[],onAnimationEnd:function(){},onAnimationStart:function(){}};lr.propTypes={from:Ae.oneOfType([Ae.object,Ae.string]),to:Ae.oneOfType([Ae.object,Ae.string]),attributeName:Ae.string,duration:Ae.number,begin:Ae.number,easing:Ae.oneOfType([Ae.string,Ae.func]),steps:Ae.arrayOf(Ae.shape({duration:Ae.number.isRequired,style:Ae.object.isRequired,easing:Ae.oneOfType([Ae.oneOf(["ease","ease-in","ease-out","ease-in-out","linear"]),Ae.func]),properties:Ae.arrayOf("string"),onAnimationEnd:Ae.func})),children:Ae.oneOfType([Ae.node,Ae.func]),isActive:Ae.bool,canBegin:Ae.bool,onAnimationEnd:Ae.func,shouldReAnimate:Ae.bool,onAnimationStart:Ae.func,onAnimationReStart:Ae.func};function Ya(t){"@babel/helpers - typeof";return Ya=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},Ya(t)}function Xl(){return Xl=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},Xl.apply(this,arguments)}function aE(t,e){return sE(t)||uE(t,e)||lE(t,e)||oE()}function oE(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function lE(t,e){if(t){if(typeof t=="string")return _h(t,e);var r=Object.prototype.toString.call(t).slice(8,-1);if(r==="Object"&&t.constructor&&(r=t.constructor.name),r==="Map"||r==="Set")return Array.from(t);if(r==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(r))return _h(t,e)}}function _h(t,e){(e==null||e>t.length)&&(e=t.length);for(var r=0,i=new Array(e);r<e;r++)i[r]=t[r];return i}function uE(t,e){var r=t==null?null:typeof Symbol<"u"&&t[Symbol.iterator]||t["@@iterator"];if(r!=null){var i,l,s,c,d=[],p=!0,v=!1;try{if(s=(r=r.call(t)).next,e!==0)for(;!(p=(i=s.call(r)).done)&&(d.push(i.value),d.length!==e);p=!0);}catch(h){v=!0,l=h}finally{try{if(!p&&r.return!=null&&(c=r.return(),Object(c)!==c))return}finally{if(v)throw l}}return d}}function sE(t){if(Array.isArray(t))return t}function kh(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function Th(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?kh(Object(r),!0).forEach(function(i){cE(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):kh(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function cE(t,e,r){return e=fE(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function fE(t){var e=dE(t,"string");return Ya(e)=="symbol"?e:e+""}function dE(t,e){if(Ya(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(Ya(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}var Ch=function(e,r,i,l,s){var c=Math.min(Math.abs(i)/2,Math.abs(l)/2),d=l>=0?1:-1,p=i>=0?1:-1,v=l>=0&&i>=0||l<0&&i<0?1:0,h;if(c>0&&s instanceof Array){for(var m=[0,0,0,0],b=0,O=4;b<O;b++)m[b]=s[b]>c?c:s[b];h="M".concat(e,",").concat(r+d*m[0]),m[0]>0&&(h+="A ".concat(m[0],",").concat(m[0],",0,0,").concat(v,",").concat(e+p*m[0],",").concat(r)),h+="L ".concat(e+i-p*m[1],",").concat(r),m[1]>0&&(h+="A ".concat(m[1],",").concat(m[1],",0,0,").concat(v,`,
        `).concat(e+i,",").concat(r+d*m[1])),h+="L ".concat(e+i,",").concat(r+l-d*m[2]),m[2]>0&&(h+="A ".concat(m[2],",").concat(m[2],",0,0,").concat(v,`,
        `).concat(e+i-p*m[2],",").concat(r+l)),h+="L ".concat(e+p*m[3],",").concat(r+l),m[3]>0&&(h+="A ".concat(m[3],",").concat(m[3],",0,0,").concat(v,`,
        `).concat(e,",").concat(r+l-d*m[3])),h+="Z"}else if(c>0&&s===+s&&s>0){var S=Math.min(c,s);h="M ".concat(e,",").concat(r+d*S,`
            A `).concat(S,",").concat(S,",0,0,").concat(v,",").concat(e+p*S,",").concat(r,`
            L `).concat(e+i-p*S,",").concat(r,`
            A `).concat(S,",").concat(S,",0,0,").concat(v,",").concat(e+i,",").concat(r+d*S,`
            L `).concat(e+i,",").concat(r+l-d*S,`
            A `).concat(S,",").concat(S,",0,0,").concat(v,",").concat(e+i-p*S,",").concat(r+l,`
            L `).concat(e+p*S,",").concat(r+l,`
            A `).concat(S,",").concat(S,",0,0,").concat(v,",").concat(e,",").concat(r+l-d*S," Z")}else h="M ".concat(e,",").concat(r," h ").concat(i," v ").concat(l," h ").concat(-i," Z");return h},pE=function(e,r){if(!e||!r)return!1;var i=e.x,l=e.y,s=r.x,c=r.y,d=r.width,p=r.height;if(Math.abs(d)>0&&Math.abs(p)>0){var v=Math.min(s,s+d),h=Math.max(s,s+d),m=Math.min(c,c+p),b=Math.max(c,c+p);return i>=v&&i<=h&&l>=m&&l<=b}return!1},vE={x:0,y:0,width:0,height:0,radius:0,isAnimationActive:!1,isUpdateAnimationActive:!1,animationBegin:0,animationDuration:1500,animationEasing:"ease"},Zf=function(e){var r=Th(Th({},vE),e),i=te.useRef(),l=te.useState(-1),s=aE(l,2),c=s[0],d=s[1];te.useEffect(function(){if(i.current&&i.current.getTotalLength)try{var k=i.current.getTotalLength();k&&d(k)}catch{}},[]);var p=r.x,v=r.y,h=r.width,m=r.height,b=r.radius,O=r.className,S=r.animationEasing,P=r.animationDuration,g=r.animationBegin,A=r.isAnimationActive,j=r.isUpdateAnimationActive;if(p!==+p||v!==+v||h!==+h||m!==+m||h===0||m===0)return null;var _=ge("recharts-rectangle",O);return j?$.createElement(lr,{canBegin:c>0,from:{width:h,height:m,x:p,y:v},to:{width:h,height:m,x:p,y:v},duration:P,animationEasing:S,isActive:j},function(k){var w=k.width,E=k.height,C=k.x,M=k.y;return $.createElement(lr,{canBegin:c>0,from:"0px ".concat(c===-1?1:c,"px"),to:"".concat(c,"px 0px"),attributeName:"strokeDasharray",begin:g,duration:P,isActive:A,easing:S},$.createElement("path",Xl({},fe(r,!0),{className:_,d:Ch(C,M,w,E,b),ref:i})))}):$.createElement("path",Xl({},fe(r,!0),{className:_,d:Ch(p,v,h,m,b)}))},yE=["points","className","baseLinePoints","connectNulls"];function ti(){return ti=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},ti.apply(this,arguments)}function hE(t,e){if(t==null)return{};var r=mE(t,e),i,l;if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(t);for(l=0;l<s.length;l++)i=s[l],!(e.indexOf(i)>=0)&&Object.prototype.propertyIsEnumerable.call(t,i)&&(r[i]=t[i])}return r}function mE(t,e){if(t==null)return{};var r={};for(var i in t)if(Object.prototype.hasOwnProperty.call(t,i)){if(e.indexOf(i)>=0)continue;r[i]=t[i]}return r}function $h(t){return xE(t)||wE(t)||bE(t)||gE()}function gE(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function bE(t,e){if(t){if(typeof t=="string")return bf(t,e);var r=Object.prototype.toString.call(t).slice(8,-1);if(r==="Object"&&t.constructor&&(r=t.constructor.name),r==="Map"||r==="Set")return Array.from(t);if(r==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(r))return bf(t,e)}}function wE(t){if(typeof Symbol<"u"&&t[Symbol.iterator]!=null||t["@@iterator"]!=null)return Array.from(t)}function xE(t){if(Array.isArray(t))return bf(t)}function bf(t,e){(e==null||e>t.length)&&(e=t.length);for(var r=0,i=new Array(e);r<e;r++)i[r]=t[r];return i}var Ih=function(e){return e&&e.x===+e.x&&e.y===+e.y},OE=function(){var e=arguments.length>0&&arguments[0]!==void 0?arguments[0]:[],r=[[]];return e.forEach(function(i){Ih(i)?r[r.length-1].push(i):r[r.length-1].length>0&&r.push([])}),Ih(e[0])&&r[r.length-1].push(e[0]),r[r.length-1].length<=0&&(r=r.slice(0,-1)),r},Ta=function(e,r){var i=OE(e);r&&(i=[i.reduce(function(s,c){return[].concat($h(s),$h(c))},[])]);var l=i.map(function(s){return s.reduce(function(c,d,p){return"".concat(c).concat(p===0?"M":"L").concat(d.x,",").concat(d.y)},"")}).join("");return i.length===1?"".concat(l,"Z"):l},SE=function(e,r,i){var l=Ta(e,i);return"".concat(l.slice(-1)==="Z"?l.slice(0,-1):l,"L").concat(Ta(r.reverse(),i).slice(1))},Pg=function(e){var r=e.points,i=e.className,l=e.baseLinePoints,s=e.connectNulls,c=hE(e,yE);if(!r||!r.length)return null;var d=ge("recharts-polygon",i);if(l&&l.length){var p=c.stroke&&c.stroke!=="none",v=SE(r,l,s);return $.createElement("g",{className:d},$.createElement("path",ti({},fe(c,!0),{fill:v.slice(-1)==="Z"?c.fill:"none",stroke:"none",d:v})),p?$.createElement("path",ti({},fe(c,!0),{fill:"none",d:Ta(r,s)})):null,p?$.createElement("path",ti({},fe(c,!0),{fill:"none",d:Ta(l,s)})):null)}var h=Ta(r,s);return $.createElement("path",ti({},fe(c,!0),{fill:h.slice(-1)==="Z"?c.fill:"none",className:d,d:h}))};function wf(){return wf=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},wf.apply(this,arguments)}var so=function(e){var r=e.cx,i=e.cy,l=e.r,s=e.className,c=ge("recharts-dot",s);return r===+r&&i===+i&&l===+l?te.createElement("circle",wf({},fe(e,!1),Il(e),{className:c,cx:r,cy:i,r:l})):null};function qa(t){"@babel/helpers - typeof";return qa=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},qa(t)}var PE=["x","y","top","left","width","height","className"];function xf(){return xf=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},xf.apply(this,arguments)}function Mh(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function AE(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?Mh(Object(r),!0).forEach(function(i){EE(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):Mh(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function EE(t,e,r){return e=jE(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function jE(t){var e=_E(t,"string");return qa(e)=="symbol"?e:e+""}function _E(t,e){if(qa(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(qa(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}function kE(t,e){if(t==null)return{};var r=TE(t,e),i,l;if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(t);for(l=0;l<s.length;l++)i=s[l],!(e.indexOf(i)>=0)&&Object.prototype.propertyIsEnumerable.call(t,i)&&(r[i]=t[i])}return r}function TE(t,e){if(t==null)return{};var r={};for(var i in t)if(Object.prototype.hasOwnProperty.call(t,i)){if(e.indexOf(i)>=0)continue;r[i]=t[i]}return r}var CE=function(e,r,i,l,s,c){return"M".concat(e,",").concat(s,"v").concat(l,"M").concat(c,",").concat(r,"h").concat(i)},$E=function(e){var r=e.x,i=r===void 0?0:r,l=e.y,s=l===void 0?0:l,c=e.top,d=c===void 0?0:c,p=e.left,v=p===void 0?0:p,h=e.width,m=h===void 0?0:h,b=e.height,O=b===void 0?0:b,S=e.className,P=kE(e,PE),g=AE({x:i,y:s,top:d,left:v,width:m,height:O},P);return!ie(i)||!ie(s)||!ie(m)||!ie(O)||!ie(d)||!ie(v)?null:$.createElement("path",xf({},fe(g,!0),{className:ge("recharts-cross",S),d:CE(i,s,m,O,d,v)}))},IE=["cx","cy","innerRadius","outerRadius","gridType","radialLines"];function Qa(t){"@babel/helpers - typeof";return Qa=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},Qa(t)}function ME(t,e){if(t==null)return{};var r=NE(t,e),i,l;if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(t);for(l=0;l<s.length;l++)i=s[l],!(e.indexOf(i)>=0)&&Object.prototype.propertyIsEnumerable.call(t,i)&&(r[i]=t[i])}return r}function NE(t,e){if(t==null)return{};var r={};for(var i in t)if(Object.prototype.hasOwnProperty.call(t,i)){if(e.indexOf(i)>=0)continue;r[i]=t[i]}return r}function Pr(){return Pr=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},Pr.apply(this,arguments)}function Nh(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function Za(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?Nh(Object(r),!0).forEach(function(i){DE(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):Nh(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function DE(t,e,r){return e=LE(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function LE(t){var e=RE(t,"string");return Qa(e)=="symbol"?e:e+""}function RE(t,e){if(Qa(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(Qa(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}var BE=function(e,r,i,l){var s="";return l.forEach(function(c,d){var p=je(r,i,e,c);d?s+="L ".concat(p.x,",").concat(p.y):s+="M ".concat(p.x,",").concat(p.y)}),s+="Z",s},zE=function(e){var r=e.cx,i=e.cy,l=e.innerRadius,s=e.outerRadius,c=e.polarAngles,d=e.radialLines;if(!c||!c.length||!d)return null;var p=Za({stroke:"#ccc"},fe(e,!1));return $.createElement("g",{className:"recharts-polar-grid-angle"},c.map(function(v){var h=je(r,i,l,v),m=je(r,i,s,v);return $.createElement("line",Pr({},p,{key:"line-".concat(v),x1:h.x,y1:h.y,x2:m.x,y2:m.y}))}))},FE=function(e){var r=e.cx,i=e.cy,l=e.radius,s=e.index,c=Za(Za({stroke:"#ccc"},fe(e,!1)),{},{fill:"none"});return $.createElement("circle",Pr({},c,{className:ge("recharts-polar-grid-concentric-circle",e.className),key:"circle-".concat(s),cx:r,cy:i,r:l}))},WE=function(e){var r=e.radius,i=e.index,l=Za(Za({stroke:"#ccc"},fe(e,!1)),{},{fill:"none"});return $.createElement("path",Pr({},l,{className:ge("recharts-polar-grid-concentric-polygon",e.className),key:"path-".concat(i),d:BE(r,e.cx,e.cy,e.polarAngles)}))},UE=function(e){var r=e.polarRadius,i=e.gridType;return!r||!r.length?null:$.createElement("g",{className:"recharts-polar-grid-concentric"},r.map(function(l,s){var c=s;return i==="circle"?$.createElement(FE,Pr({key:c},e,{radius:l,index:s})):$.createElement(WE,Pr({key:c},e,{radius:l,index:s}))}))},VE=function(e){var r=e.cx,i=r===void 0?0:r,l=e.cy,s=l===void 0?0:l,c=e.innerRadius,d=c===void 0?0:c,p=e.outerRadius,v=p===void 0?0:p,h=e.gridType,m=h===void 0?"polygon":h,b=e.radialLines,O=b===void 0?!0:b,S=ME(e,IE);return v<=0?null:$.createElement("g",{className:"recharts-polar-grid"},$.createElement(zE,Pr({cx:i,cy:s,innerRadius:d,outerRadius:v,gridType:m,radialLines:O},S)),$.createElement(UE,Pr({cx:i,cy:s,innerRadius:d,outerRadius:v,gridType:m,radialLines:O},S)))};VE.displayName="PolarGrid";var KE=["cx","cy","angle","ticks","axisLine"],HE=["ticks","tick","angle","tickFormatter","stroke"];function hi(t){"@babel/helpers - typeof";return hi=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},hi(t)}function Ca(){return Ca=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},Ca.apply(this,arguments)}function Dh(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function vn(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?Dh(Object(r),!0).forEach(function(i){gu(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):Dh(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function Lh(t,e){if(t==null)return{};var r=GE(t,e),i,l;if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(t);for(l=0;l<s.length;l++)i=s[l],!(e.indexOf(i)>=0)&&Object.prototype.propertyIsEnumerable.call(t,i)&&(r[i]=t[i])}return r}function GE(t,e){if(t==null)return{};var r={};for(var i in t)if(Object.prototype.hasOwnProperty.call(t,i)){if(e.indexOf(i)>=0)continue;r[i]=t[i]}return r}function XE(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")}function Rh(t,e){for(var r=0;r<e.length;r++){var i=e[r];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(t,Eg(i.key),i)}}function YE(t,e,r){return e&&Rh(t.prototype,e),r&&Rh(t,r),Object.defineProperty(t,"prototype",{writable:!1}),t}function qE(t,e,r){return e=Yl(e),QE(t,Ag()?Reflect.construct(e,r||[],Yl(t).constructor):e.apply(t,r))}function QE(t,e){if(e&&(hi(e)==="object"||typeof e=="function"))return e;if(e!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return ZE(t)}function ZE(t){if(t===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return t}function Ag(){try{var t=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(Ag=function(){return!!t})()}function Yl(t){return Yl=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(r){return r.__proto__||Object.getPrototypeOf(r)},Yl(t)}function JE(t,e){if(typeof e!="function"&&e!==null)throw new TypeError("Super expression must either be null or a function");t.prototype=Object.create(e&&e.prototype,{constructor:{value:t,writable:!0,configurable:!0}}),Object.defineProperty(t,"prototype",{writable:!1}),e&&Of(t,e)}function Of(t,e){return Of=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(i,l){return i.__proto__=l,i},Of(t,e)}function gu(t,e,r){return e=Eg(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function Eg(t){var e=ej(t,"string");return hi(e)=="symbol"?e:e+""}function ej(t,e){if(hi(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(hi(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(t)}var bu=(function(t){function e(){return XE(this,e),qE(this,e,arguments)}return JE(e,t),YE(e,[{key:"getTickValueCoord",value:function(i){var l=i.coordinate,s=this.props,c=s.angle,d=s.cx,p=s.cy;return je(d,p,l,c)}},{key:"getTickTextAnchor",value:function(){var i=this.props.orientation,l;switch(i){case"left":l="end";break;case"right":l="start";break;default:l="middle";break}return l}},{key:"getViewBox",value:function(){var i=this.props,l=i.cx,s=i.cy,c=i.angle,d=i.ticks,p=Xw(d,function(h){return h.coordinate||0}),v=Yw(d,function(h){return h.coordinate||0});return{cx:l,cy:s,startAngle:c,endAngle:c,innerRadius:v.coordinate||0,outerRadius:p.coordinate||0}}},{key:"renderAxisLine",value:function(){var i=this.props,l=i.cx,s=i.cy,c=i.angle,d=i.ticks,p=i.axisLine,v=Lh(i,KE),h=d.reduce(function(S,P){return[Math.min(S[0],P.coordinate),Math.max(S[1],P.coordinate)]},[1/0,-1/0]),m=je(l,s,h[0],c),b=je(l,s,h[1],c),O=vn(vn(vn({},fe(v,!1)),{},{fill:"none"},fe(p,!1)),{},{x1:m.x,y1:m.y,x2:b.x,y2:b.y});return $.createElement("line",Ca({className:"recharts-polar-radius-axis-line"},O))}},{key:"renderTicks",value:function(){var i=this,l=this.props,s=l.ticks,c=l.tick,d=l.angle,p=l.tickFormatter,v=l.stroke,h=Lh(l,HE),m=this.getTickTextAnchor(),b=fe(h,!1),O=fe(c,!1),S=s.map(function(P,g){var A=i.getTickValueCoord(P),j=vn(vn(vn(vn({textAnchor:m,transform:"rotate(".concat(90-d,", ").concat(A.x,", ").concat(A.y,")")},b),{},{stroke:"none",fill:v},O),{},{index:g},A),{},{payload:P});return $.createElement(Se,Ca({className:ge("recharts-polar-radius-axis-tick",pg(c)),key:"tick-".concat(P.coordinate)},ui(i.props,P,g)),e.renderTickItem(c,j,p?p(P.value,g):P.value))});return $.createElement(Se,{className:"recharts-polar-radius-axis-ticks"},S)}},{key:"render",value:function(){var i=this.props,l=i.ticks,s=i.axisLine,c=i.tick;return!l||!l.length?null:$.createElement(Se,{className:ge("recharts-polar-radius-axis",this.props.className)},s&&this.renderAxisLine(),c&&this.renderTicks(),et.renderCallByParent(this.props,this.getViewBox()))}}],[{key:"renderTickItem",value:function(i,l,s){var c;return $.isValidElement(i)?c=$.cloneElement(i,l):he(i)?c=i(l):c=$.createElement(pi,Ca({},l,{className:"recharts-polar-radius-axis-tick-value"}),s),c}}])})(te.PureComponent);gu(bu,"displayName","PolarRadiusAxis");gu(bu,"axisType","radiusAxis");gu(bu,"defaultProps",{type:"number",radiusAxisId:0,cx:0,cy:0,angle:0,orientation:"right",stroke:"#ccc",axisLine:!0,tick:!0,tickCount:5,allowDataOverflow:!1,scale:"auto",allowDuplicatedCategory:!0});function mi(t){"@babel/helpers - typeof";return mi=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},mi(t)}function hn(){return hn=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},hn.apply(this,arguments)}function Bh(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function yn(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?Bh(Object(r),!0).forEach(function(i){wu(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):Bh(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function tj(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")}function zh(t,e){for(var r=0;r<e.length;r++){var i=e[r];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(t,_g(i.key),i)}}function rj(t,e,r){return e&&zh(t.prototype,e),r&&zh(t,r),Object.defineProperty(t,"prototype",{writable:!1}),t}function nj(t,e,r){return e=ql(e),ij(t,jg()?Reflect.construct(e,r||[],ql(t).constructor):e.apply(t,r))}function ij(t,e){if(e&&(mi(e)==="object"||typeof e=="function"))return e;if(e!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return aj(t)}function aj(t){if(t===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return t}function jg(){try{var t=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(jg=function(){return!!t})()}function ql(t){return ql=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(r){return r.__proto__||Object.getPrototypeOf(r)},ql(t)}function oj(t,e){if(typeof e!="function"&&e!==null)throw new TypeError("Super expression must either be null or a function");t.prototype=Object.create(e&&e.prototype,{constructor:{value:t,writable:!0,configurable:!0}}),Object.defineProperty(t,"prototype",{writable:!1}),e&&Sf(t,e)}function Sf(t,e){return Sf=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(i,l){return i.__proto__=l,i},Sf(t,e)}function wu(t,e,r){return e=_g(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function _g(t){var e=lj(t,"string");return mi(e)=="symbol"?e:e+""}function lj(t,e){if(mi(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(mi(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(t)}var uj=Math.PI/180,Fh=1e-5,xu=(function(t){function e(){return tj(this,e),nj(this,e,arguments)}return oj(e,t),rj(e,[{key:"getTickLineCoord",value:function(i){var l=this.props,s=l.cx,c=l.cy,d=l.radius,p=l.orientation,v=l.tickSize,h=v||8,m=je(s,c,d,i.coordinate),b=je(s,c,d+(p==="inner"?-1:1)*h,i.coordinate);return{x1:m.x,y1:m.y,x2:b.x,y2:b.y}}},{key:"getTickTextAnchor",value:function(i){var l=this.props.orientation,s=Math.cos(-i.coordinate*uj),c;return s>Fh?c=l==="outer"?"start":"end":s<-Fh?c=l==="outer"?"end":"start":c="middle",c}},{key:"renderAxisLine",value:function(){var i=this.props,l=i.cx,s=i.cy,c=i.radius,d=i.axisLine,p=i.axisLineType,v=yn(yn({},fe(this.props,!1)),{},{fill:"none"},fe(d,!1));if(p==="circle")return $.createElement(so,hn({className:"recharts-polar-angle-axis-line"},v,{cx:l,cy:s,r:c}));var h=this.props.ticks,m=h.map(function(b){return je(l,s,c,b.coordinate)});return $.createElement(Pg,hn({className:"recharts-polar-angle-axis-line"},v,{points:m}))}},{key:"renderTicks",value:function(){var i=this,l=this.props,s=l.ticks,c=l.tick,d=l.tickLine,p=l.tickFormatter,v=l.stroke,h=fe(this.props,!1),m=fe(c,!1),b=yn(yn({},h),{},{fill:"none"},fe(d,!1)),O=s.map(function(S,P){var g=i.getTickLineCoord(S),A=i.getTickTextAnchor(S),j=yn(yn(yn({textAnchor:A},h),{},{stroke:"none",fill:v},m),{},{index:P,payload:S,x:g.x2,y:g.y2});return $.createElement(Se,hn({className:ge("recharts-polar-angle-axis-tick",pg(c)),key:"tick-".concat(S.coordinate)},ui(i.props,S,P)),d&&$.createElement("line",hn({className:"recharts-polar-angle-axis-tick-line"},b,g)),c&&e.renderTickItem(c,j,p?p(S.value,P):S.value))});return $.createElement(Se,{className:"recharts-polar-angle-axis-ticks"},O)}},{key:"render",value:function(){var i=this.props,l=i.ticks,s=i.radius,c=i.axisLine;return s<=0||!l||!l.length?null:$.createElement(Se,{className:ge("recharts-polar-angle-axis",this.props.className)},c&&this.renderAxisLine(),this.renderTicks())}}],[{key:"renderTickItem",value:function(i,l,s){var c;return $.isValidElement(i)?c=$.cloneElement(i,l):he(i)?c=i(l):c=$.createElement(pi,hn({},l,{className:"recharts-polar-angle-axis-tick-value"}),s),c}}])})(te.PureComponent);wu(xu,"displayName","PolarAngleAxis");wu(xu,"axisType","angleAxis");wu(xu,"defaultProps",{type:"category",angleAxisId:0,scale:"auto",cx:0,cy:0,orientation:"outer",axisLine:!0,tickLine:!0,tickSize:8,tick:!0,hide:!1,allowDuplicatedCategory:!0});function Ja(t){"@babel/helpers - typeof";return Ja=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},Ja(t)}function Ql(){return Ql=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},Ql.apply(this,arguments)}function sj(t,e){return pj(t)||dj(t,e)||fj(t,e)||cj()}function cj(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function fj(t,e){if(t){if(typeof t=="string")return Wh(t,e);var r=Object.prototype.toString.call(t).slice(8,-1);if(r==="Object"&&t.constructor&&(r=t.constructor.name),r==="Map"||r==="Set")return Array.from(t);if(r==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(r))return Wh(t,e)}}function Wh(t,e){(e==null||e>t.length)&&(e=t.length);for(var r=0,i=new Array(e);r<e;r++)i[r]=t[r];return i}function dj(t,e){var r=t==null?null:typeof Symbol<"u"&&t[Symbol.iterator]||t["@@iterator"];if(r!=null){var i,l,s,c,d=[],p=!0,v=!1;try{if(s=(r=r.call(t)).next,e!==0)for(;!(p=(i=s.call(r)).done)&&(d.push(i.value),d.length!==e);p=!0);}catch(h){v=!0,l=h}finally{try{if(!p&&r.return!=null&&(c=r.return(),Object(c)!==c))return}finally{if(v)throw l}}return d}}function pj(t){if(Array.isArray(t))return t}function Uh(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function Vh(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?Uh(Object(r),!0).forEach(function(i){vj(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):Uh(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function vj(t,e,r){return e=yj(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function yj(t){var e=hj(t,"string");return Ja(e)=="symbol"?e:e+""}function hj(t,e){if(Ja(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(Ja(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}var Kh=function(e,r,i,l,s){var c=i-l,d;return d="M ".concat(e,",").concat(r),d+="L ".concat(e+i,",").concat(r),d+="L ".concat(e+i-c/2,",").concat(r+s),d+="L ".concat(e+i-c/2-l,",").concat(r+s),d+="L ".concat(e,",").concat(r," Z"),d},mj={x:0,y:0,upperWidth:0,lowerWidth:0,height:0,isUpdateAnimationActive:!1,animationBegin:0,animationDuration:1500,animationEasing:"ease"},gj=function(e){var r=Vh(Vh({},mj),e),i=te.useRef(),l=te.useState(-1),s=sj(l,2),c=s[0],d=s[1];te.useEffect(function(){if(i.current&&i.current.getTotalLength)try{var _=i.current.getTotalLength();_&&d(_)}catch{}},[]);var p=r.x,v=r.y,h=r.upperWidth,m=r.lowerWidth,b=r.height,O=r.className,S=r.animationEasing,P=r.animationDuration,g=r.animationBegin,A=r.isUpdateAnimationActive;if(p!==+p||v!==+v||h!==+h||m!==+m||b!==+b||h===0&&m===0||b===0)return null;var j=ge("recharts-trapezoid",O);return A?$.createElement(lr,{canBegin:c>0,from:{upperWidth:0,lowerWidth:0,height:b,x:p,y:v},to:{upperWidth:h,lowerWidth:m,height:b,x:p,y:v},duration:P,animationEasing:S,isActive:A},function(_){var k=_.upperWidth,w=_.lowerWidth,E=_.height,C=_.x,M=_.y;return $.createElement(lr,{canBegin:c>0,from:"0px ".concat(c===-1?1:c,"px"),to:"".concat(c,"px 0px"),attributeName:"strokeDasharray",begin:g,duration:P,easing:S},$.createElement("path",Ql({},fe(r,!0),{className:j,d:Kh(C,M,k,w,E),ref:i})))}):$.createElement("g",null,$.createElement("path",Ql({},fe(r,!0),{className:j,d:Kh(p,v,h,m,b)})))},bj=["option","shapeType","propTransformer","activeClassName","isActive"];function eo(t){"@babel/helpers - typeof";return eo=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},eo(t)}function wj(t,e){if(t==null)return{};var r=xj(t,e),i,l;if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(t);for(l=0;l<s.length;l++)i=s[l],!(e.indexOf(i)>=0)&&Object.prototype.propertyIsEnumerable.call(t,i)&&(r[i]=t[i])}return r}function xj(t,e){if(t==null)return{};var r={};for(var i in t)if(Object.prototype.hasOwnProperty.call(t,i)){if(e.indexOf(i)>=0)continue;r[i]=t[i]}return r}function Hh(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function Zl(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?Hh(Object(r),!0).forEach(function(i){Oj(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):Hh(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function Oj(t,e,r){return e=Sj(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function Sj(t){var e=Pj(t,"string");return eo(e)=="symbol"?e:e+""}function Pj(t,e){if(eo(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(eo(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}function Aj(t,e){return Zl(Zl({},e),t)}function Ej(t,e){return t==="symbols"}function Gh(t){var e=t.shapeType,r=t.elementProps;switch(e){case"rectangle":return $.createElement(Zf,r);case"trapezoid":return $.createElement(gj,r);case"sector":return $.createElement(hg,r);case"symbols":if(Ej(e))return $.createElement(Xf,r);break;default:return null}}function jj(t){return te.isValidElement(t)?t.props:t}function _j(t){var e=t.option,r=t.shapeType,i=t.propTransformer,l=i===void 0?Aj:i,s=t.activeClassName,c=s===void 0?"recharts-active-shape":s,d=t.isActive,p=wj(t,bj),v;if(te.isValidElement(e))v=te.cloneElement(e,Zl(Zl({},p),jj(e)));else if(he(e))v=e(p);else if(qw(e)&&!Qw(e)){var h=l(e,p);v=$.createElement(Gh,{shapeType:r,elementProps:h})}else{var m=p;v=$.createElement(Gh,{shapeType:r,elementProps:m})}return d?$.createElement(Se,{className:c},v):v}function Ou(t,e){return e!=null&&"trapezoids"in t.props}function Su(t,e){return e!=null&&"sectors"in t.props}function to(t,e){return e!=null&&"points"in t.props}function kj(t,e){var r,i,l=t.x===(e==null||(r=e.labelViewBox)===null||r===void 0?void 0:r.x)||t.x===e.x,s=t.y===(e==null||(i=e.labelViewBox)===null||i===void 0?void 0:i.y)||t.y===e.y;return l&&s}function Tj(t,e){var r=t.endAngle===e.endAngle,i=t.startAngle===e.startAngle;return r&&i}function Cj(t,e){var r=t.x===e.x,i=t.y===e.y,l=t.z===e.z;return r&&i&&l}function $j(t,e){var r;return Ou(t,e)?r=kj:Su(t,e)?r=Tj:to(t,e)&&(r=Cj),r}function Ij(t,e){var r;return Ou(t,e)?r="trapezoids":Su(t,e)?r="sectors":to(t,e)&&(r="points"),r}function Mj(t,e){if(Ou(t,e)){var r;return(r=e.tooltipPayload)===null||r===void 0||(r=r[0])===null||r===void 0||(r=r.payload)===null||r===void 0?void 0:r.payload}if(Su(t,e)){var i;return(i=e.tooltipPayload)===null||i===void 0||(i=i[0])===null||i===void 0||(i=i.payload)===null||i===void 0?void 0:i.payload}return to(t,e)?e.payload:{}}function Nj(t){var e=t.activeTooltipItem,r=t.graphicalItem,i=t.itemData,l=Ij(r,e),s=Mj(r,e),c=i.filter(function(p,v){var h=li(s,p),m=r.props[l].filter(function(S){var P=$j(r,e);return P(S,e)}),b=r.props[l].indexOf(m[m.length-1]),O=v===b;return h&&O}),d=i.indexOf(c[c.length-1]);return d}var Dj=["key"];function gi(t){"@babel/helpers - typeof";return gi=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},gi(t)}function Lj(t,e){if(t==null)return{};var r=Rj(t,e),i,l;if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(t);for(l=0;l<s.length;l++)i=s[l],!(e.indexOf(i)>=0)&&Object.prototype.propertyIsEnumerable.call(t,i)&&(r[i]=t[i])}return r}function Rj(t,e){if(t==null)return{};var r={};for(var i in t)if(Object.prototype.hasOwnProperty.call(t,i)){if(e.indexOf(i)>=0)continue;r[i]=t[i]}return r}function Jl(){return Jl=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},Jl.apply(this,arguments)}function Xh(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function ct(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?Xh(Object(r),!0).forEach(function(i){br(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):Xh(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function Bj(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")}function Yh(t,e){for(var r=0;r<e.length;r++){var i=e[r];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(t,Tg(i.key),i)}}function zj(t,e,r){return e&&Yh(t.prototype,e),r&&Yh(t,r),Object.defineProperty(t,"prototype",{writable:!1}),t}function Fj(t,e,r){return e=eu(e),Wj(t,kg()?Reflect.construct(e,r||[],eu(t).constructor):e.apply(t,r))}function Wj(t,e){if(e&&(gi(e)==="object"||typeof e=="function"))return e;if(e!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return Uj(t)}function Uj(t){if(t===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return t}function kg(){try{var t=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(kg=function(){return!!t})()}function eu(t){return eu=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(r){return r.__proto__||Object.getPrototypeOf(r)},eu(t)}function Vj(t,e){if(typeof e!="function"&&e!==null)throw new TypeError("Super expression must either be null or a function");t.prototype=Object.create(e&&e.prototype,{constructor:{value:t,writable:!0,configurable:!0}}),Object.defineProperty(t,"prototype",{writable:!1}),e&&Pf(t,e)}function Pf(t,e){return Pf=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(i,l){return i.__proto__=l,i},Pf(t,e)}function br(t,e,r){return e=Tg(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function Tg(t){var e=Kj(t,"string");return gi(e)=="symbol"?e:e+""}function Kj(t,e){if(gi(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(gi(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(t)}var Pu=(function(t){function e(){var r;Bj(this,e);for(var i=arguments.length,l=new Array(i),s=0;s<i;s++)l[s]=arguments[s];return r=Fj(this,e,[].concat(l)),br(r,"state",{isAnimationFinished:!1}),br(r,"handleAnimationEnd",function(){var c=r.props.onAnimationEnd;r.setState({isAnimationFinished:!0}),he(c)&&c()}),br(r,"handleAnimationStart",function(){var c=r.props.onAnimationStart;r.setState({isAnimationFinished:!1}),he(c)&&c()}),br(r,"handleMouseEnter",function(c){var d=r.props.onMouseEnter;d&&d(r.props,c)}),br(r,"handleMouseLeave",function(c){var d=r.props.onMouseLeave;d&&d(r.props,c)}),r}return Vj(e,t),zj(e,[{key:"renderDots",value:function(i){var l=this.props,s=l.dot,c=l.dataKey,d=fe(this.props,!1),p=fe(s,!0),v=i.map(function(h,m){var b=ct(ct(ct({key:"dot-".concat(m),r:3},d),p),{},{dataKey:c,cx:h.x,cy:h.y,index:m,payload:h});return e.renderDotItem(s,b)});return $.createElement(Se,{className:"recharts-radar-dots"},v)}},{key:"renderPolygonStatically",value:function(i){var l=this.props,s=l.shape,c=l.dot,d=l.isRange,p=l.baseLinePoints,v=l.connectNulls,h;return $.isValidElement(s)?h=$.cloneElement(s,ct(ct({},this.props),{},{points:i})):he(s)?h=s(ct(ct({},this.props),{},{points:i})):h=$.createElement(Pg,Jl({},fe(this.props,!0),{onMouseEnter:this.handleMouseEnter,onMouseLeave:this.handleMouseLeave,points:i,baseLinePoints:d?p:null,connectNulls:v})),$.createElement(Se,{className:"recharts-radar-polygon"},h,c?this.renderDots(i):null)}},{key:"renderPolygonWithAnimation",value:function(){var i=this,l=this.props,s=l.points,c=l.isAnimationActive,d=l.animationBegin,p=l.animationDuration,v=l.animationEasing,h=l.animationId,m=this.state.prevPoints;return $.createElement(lr,{begin:d,duration:p,isActive:c,easing:v,from:{t:0},to:{t:1},key:"radar-".concat(h),onAnimationEnd:this.handleAnimationEnd,onAnimationStart:this.handleAnimationStart},function(b){var O=b.t,S=m&&m.length/s.length,P=s.map(function(g,A){var j=m&&m[Math.floor(A*S)];if(j){var _=ot(j.x,g.x),k=ot(j.y,g.y);return ct(ct({},g),{},{x:_(O),y:k(O)})}var w=ot(g.cx,g.x),E=ot(g.cy,g.y);return ct(ct({},g),{},{x:w(O),y:E(O)})});return i.renderPolygonStatically(P)})}},{key:"renderPolygon",value:function(){var i=this.props,l=i.points,s=i.isAnimationActive,c=i.isRange,d=this.state.prevPoints;return s&&l&&l.length&&!c&&(!d||!li(d,l))?this.renderPolygonWithAnimation():this.renderPolygonStatically(l)}},{key:"render",value:function(){var i=this.props,l=i.hide,s=i.className,c=i.points,d=i.isAnimationActive;if(l||!c||!c.length)return null;var p=this.state.isAnimationFinished,v=ge("recharts-radar",s);return $.createElement(Se,{className:v},this.renderPolygon(),(!d||p)&&Sr.renderCallByParent(this.props,c))}}],[{key:"getDerivedStateFromProps",value:function(i,l){return i.animationId!==l.prevAnimationId?{prevAnimationId:i.animationId,curPoints:i.points,prevPoints:l.curPoints}:i.points!==l.curPoints?{curPoints:i.points}:null}},{key:"renderDotItem",value:function(i,l){var s;if($.isValidElement(i))s=$.cloneElement(i,l);else if(he(i))s=i(l);else{var c=l.key,d=Lj(l,Dj);s=$.createElement(so,Jl({},d,{key:c,className:ge("recharts-radar-dot",typeof i!="boolean"?i.className:"")}))}return s}}])})(te.PureComponent);br(Pu,"displayName","Radar");br(Pu,"defaultProps",{angleAxisId:0,radiusAxisId:0,hide:!1,activeDot:!0,dot:!1,legendType:"rect",isAnimationActive:!Sn.isSsr,animationBegin:0,animationDuration:1500,animationEasing:"ease"});br(Pu,"getComposedData",function(t){var e=t.radiusAxis,r=t.angleAxis,i=t.displayedData,l=t.dataKey,s=t.bandSize,c=r.cx,d=r.cy,p=!1,v=[],h=r.type!=="number"?s??0:0;i.forEach(function(b,O){var S=ft(b,r.dataKey,O),P=ft(b,l),g=r.scale(S)+h,A=Array.isArray(P)?Mm(P):P,j=be(A)?void 0:e.scale(A);Array.isArray(P)&&P.length>=2&&(p=!0),v.push(ct(ct({},je(c,d,j,g)),{},{name:S,value:P,cx:c,cy:d,radius:j,angle:g,payload:b}))});var m=[];return p&&v.forEach(function(b){if(Array.isArray(b.value)){var O=Zw(b.value),S=be(O)?void 0:e.scale(O);m.push(ct(ct({},b),{},{radius:S},je(c,d,S,b.angle)))}else m.push(b)}),{points:v,isRange:p,baseLinePoints:m}});function ro(t){"@babel/helpers - typeof";return ro=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},ro(t)}function qh(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function Qh(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?qh(Object(r),!0).forEach(function(i){Cg(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):qh(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function Cg(t,e,r){return e=Hj(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function Hj(t){var e=Gj(t,"string");return ro(e)=="symbol"?e:e+""}function Gj(t,e){if(ro(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(ro(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}var Xj=["Webkit","Moz","O","ms"],Yj=function(e,r){var i=e.replace(/(\w)/,function(s){return s.toUpperCase()}),l=Xj.reduce(function(s,c){return Qh(Qh({},s),{},Cg({},c+i,r))},{});return l[e]=r,l};function bi(t){"@babel/helpers - typeof";return bi=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},bi(t)}function tu(){return tu=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},tu.apply(this,arguments)}function Zh(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function Mc(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?Zh(Object(r),!0).forEach(function(i){St(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):Zh(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function qj(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")}function Jh(t,e){for(var r=0;r<e.length;r++){var i=e[r];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(t,Ig(i.key),i)}}function Qj(t,e,r){return e&&Jh(t.prototype,e),r&&Jh(t,r),Object.defineProperty(t,"prototype",{writable:!1}),t}function Zj(t,e,r){return e=ru(e),Jj(t,$g()?Reflect.construct(e,r||[],ru(t).constructor):e.apply(t,r))}function Jj(t,e){if(e&&(bi(e)==="object"||typeof e=="function"))return e;if(e!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return e_(t)}function e_(t){if(t===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return t}function $g(){try{var t=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return($g=function(){return!!t})()}function ru(t){return ru=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(r){return r.__proto__||Object.getPrototypeOf(r)},ru(t)}function t_(t,e){if(typeof e!="function"&&e!==null)throw new TypeError("Super expression must either be null or a function");t.prototype=Object.create(e&&e.prototype,{constructor:{value:t,writable:!0,configurable:!0}}),Object.defineProperty(t,"prototype",{writable:!1}),e&&Af(t,e)}function Af(t,e){return Af=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(i,l){return i.__proto__=l,i},Af(t,e)}function St(t,e,r){return e=Ig(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function Ig(t){var e=r_(t,"string");return bi(e)=="symbol"?e:e+""}function r_(t,e){if(bi(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(bi(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(t)}var n_=function(e){var r=e.data,i=e.startIndex,l=e.endIndex,s=e.x,c=e.width,d=e.travellerWidth;if(!r||!r.length)return{};var p=r.length,v=Tl().domain(Cl(0,p)).range([s,s+c-d]),h=v.domain().map(function(m){return v(m)});return{isTextActive:!1,isSlideMoving:!1,isTravellerMoving:!1,isTravellerFocused:!1,startX:v(i),endX:v(l),scale:v,scaleValues:h}},em=function(e){return e.changedTouches&&!!e.changedTouches.length},wi=(function(t){function e(r){var i;return qj(this,e),i=Zj(this,e,[r]),St(i,"handleDrag",function(l){i.leaveTimer&&(clearTimeout(i.leaveTimer),i.leaveTimer=null),i.state.isTravellerMoving?i.handleTravellerMove(l):i.state.isSlideMoving&&i.handleSlideDrag(l)}),St(i,"handleTouchMove",function(l){l.changedTouches!=null&&l.changedTouches.length>0&&i.handleDrag(l.changedTouches[0])}),St(i,"handleDragEnd",function(){i.setState({isTravellerMoving:!1,isSlideMoving:!1},function(){var l=i.props,s=l.endIndex,c=l.onDragEnd,d=l.startIndex;c==null||c({endIndex:s,startIndex:d})}),i.detachDragEndListener()}),St(i,"handleLeaveWrapper",function(){(i.state.isTravellerMoving||i.state.isSlideMoving)&&(i.leaveTimer=window.setTimeout(i.handleDragEnd,i.props.leaveTimeOut))}),St(i,"handleEnterSlideOrTraveller",function(){i.setState({isTextActive:!0})}),St(i,"handleLeaveSlideOrTraveller",function(){i.setState({isTextActive:!1})}),St(i,"handleSlideDragStart",function(l){var s=em(l)?l.changedTouches[0]:l;i.setState({isTravellerMoving:!1,isSlideMoving:!0,slideMoveStartX:s.pageX}),i.attachDragEndListener()}),i.travellerDragStartHandlers={startX:i.handleTravellerDragStart.bind(i,"startX"),endX:i.handleTravellerDragStart.bind(i,"endX")},i.state={},i}return t_(e,t),Qj(e,[{key:"componentWillUnmount",value:function(){this.leaveTimer&&(clearTimeout(this.leaveTimer),this.leaveTimer=null),this.detachDragEndListener()}},{key:"getIndex",value:function(i){var l=i.startX,s=i.endX,c=this.state.scaleValues,d=this.props,p=d.gap,v=d.data,h=v.length-1,m=Math.min(l,s),b=Math.max(l,s),O=e.getIndexInRange(c,m),S=e.getIndexInRange(c,b);return{startIndex:O-O%p,endIndex:S===h?h:S-S%p}}},{key:"getTextOfTick",value:function(i){var l=this.props,s=l.data,c=l.tickFormatter,d=l.dataKey,p=ft(s[i],d,i);return he(c)?c(p,i):p}},{key:"attachDragEndListener",value:function(){window.addEventListener("mouseup",this.handleDragEnd,!0),window.addEventListener("touchend",this.handleDragEnd,!0),window.addEventListener("mousemove",this.handleDrag,!0)}},{key:"detachDragEndListener",value:function(){window.removeEventListener("mouseup",this.handleDragEnd,!0),window.removeEventListener("touchend",this.handleDragEnd,!0),window.removeEventListener("mousemove",this.handleDrag,!0)}},{key:"handleSlideDrag",value:function(i){var l=this.state,s=l.slideMoveStartX,c=l.startX,d=l.endX,p=this.props,v=p.x,h=p.width,m=p.travellerWidth,b=p.startIndex,O=p.endIndex,S=p.onChange,P=i.pageX-s;P>0?P=Math.min(P,v+h-m-d,v+h-m-c):P<0&&(P=Math.max(P,v-c,v-d));var g=this.getIndex({startX:c+P,endX:d+P});(g.startIndex!==b||g.endIndex!==O)&&S&&S(g),this.setState({startX:c+P,endX:d+P,slideMoveStartX:i.pageX})}},{key:"handleTravellerDragStart",value:function(i,l){var s=em(l)?l.changedTouches[0]:l;this.setState({isSlideMoving:!1,isTravellerMoving:!0,movingTravellerId:i,brushMoveStartX:s.pageX}),this.attachDragEndListener()}},{key:"handleTravellerMove",value:function(i){var l=this.state,s=l.brushMoveStartX,c=l.movingTravellerId,d=l.endX,p=l.startX,v=this.state[c],h=this.props,m=h.x,b=h.width,O=h.travellerWidth,S=h.onChange,P=h.gap,g=h.data,A={startX:this.state.startX,endX:this.state.endX},j=i.pageX-s;j>0?j=Math.min(j,m+b-O-v):j<0&&(j=Math.max(j,m-v)),A[c]=v+j;var _=this.getIndex(A),k=_.startIndex,w=_.endIndex,E=function(){var M=g.length-1;return c==="startX"&&(d>p?k%P===0:w%P===0)||d<p&&w===M||c==="endX"&&(d>p?w%P===0:k%P===0)||d>p&&w===M};this.setState(St(St({},c,v+j),"brushMoveStartX",i.pageX),function(){S&&E()&&S(_)})}},{key:"handleTravellerMoveKeyboard",value:function(i,l){var s=this,c=this.state,d=c.scaleValues,p=c.startX,v=c.endX,h=this.state[l],m=d.indexOf(h);if(m!==-1){var b=m+i;if(!(b===-1||b>=d.length)){var O=d[b];l==="startX"&&O>=v||l==="endX"&&O<=p||this.setState(St({},l,O),function(){s.props.onChange(s.getIndex({startX:s.state.startX,endX:s.state.endX}))})}}}},{key:"renderBackground",value:function(){var i=this.props,l=i.x,s=i.y,c=i.width,d=i.height,p=i.fill,v=i.stroke;return $.createElement("rect",{stroke:v,fill:p,x:l,y:s,width:c,height:d})}},{key:"renderPanorama",value:function(){var i=this.props,l=i.x,s=i.y,c=i.width,d=i.height,p=i.data,v=i.children,h=i.padding,m=te.Children.only(v);return m?$.cloneElement(m,{x:l,y:s,width:c,height:d,margin:h,compact:!0,data:p}):null}},{key:"renderTravellerLayer",value:function(i,l){var s,c,d=this,p=this.props,v=p.y,h=p.travellerWidth,m=p.height,b=p.traveller,O=p.ariaLabel,S=p.data,P=p.startIndex,g=p.endIndex,A=Math.max(i,this.props.x),j=Mc(Mc({},fe(this.props,!1)),{},{x:A,y:v,width:h,height:m}),_=O||"Min value: ".concat((s=S[P])===null||s===void 0?void 0:s.name,", Max value: ").concat((c=S[g])===null||c===void 0?void 0:c.name);return $.createElement(Se,{tabIndex:0,role:"slider","aria-label":_,"aria-valuenow":i,className:"recharts-brush-traveller",onMouseEnter:this.handleEnterSlideOrTraveller,onMouseLeave:this.handleLeaveSlideOrTraveller,onMouseDown:this.travellerDragStartHandlers[l],onTouchStart:this.travellerDragStartHandlers[l],onKeyDown:function(w){["ArrowLeft","ArrowRight"].includes(w.key)&&(w.preventDefault(),w.stopPropagation(),d.handleTravellerMoveKeyboard(w.key==="ArrowRight"?1:-1,l))},onFocus:function(){d.setState({isTravellerFocused:!0})},onBlur:function(){d.setState({isTravellerFocused:!1})},style:{cursor:"col-resize"}},e.renderTraveller(b,j))}},{key:"renderSlide",value:function(i,l){var s=this.props,c=s.y,d=s.height,p=s.stroke,v=s.travellerWidth,h=Math.min(i,l)+v,m=Math.max(Math.abs(l-i)-v,0);return $.createElement("rect",{className:"recharts-brush-slide",onMouseEnter:this.handleEnterSlideOrTraveller,onMouseLeave:this.handleLeaveSlideOrTraveller,onMouseDown:this.handleSlideDragStart,onTouchStart:this.handleSlideDragStart,style:{cursor:"move"},stroke:"none",fill:p,fillOpacity:.2,x:h,y:c,width:m,height:d})}},{key:"renderText",value:function(){var i=this.props,l=i.startIndex,s=i.endIndex,c=i.y,d=i.height,p=i.travellerWidth,v=i.stroke,h=this.state,m=h.startX,b=h.endX,O=5,S={pointerEvents:"none",fill:v};return $.createElement(Se,{className:"recharts-brush-texts"},$.createElement(pi,tu({textAnchor:"end",verticalAnchor:"middle",x:Math.min(m,b)-O,y:c+d/2},S),this.getTextOfTick(l)),$.createElement(pi,tu({textAnchor:"start",verticalAnchor:"middle",x:Math.max(m,b)+p+O,y:c+d/2},S),this.getTextOfTick(s)))}},{key:"render",value:function(){var i=this.props,l=i.data,s=i.className,c=i.children,d=i.x,p=i.y,v=i.width,h=i.height,m=i.alwaysShowText,b=this.state,O=b.startX,S=b.endX,P=b.isTextActive,g=b.isSlideMoving,A=b.isTravellerMoving,j=b.isTravellerFocused;if(!l||!l.length||!ie(d)||!ie(p)||!ie(v)||!ie(h)||v<=0||h<=0)return null;var _=ge("recharts-brush",s),k=$.Children.count(c)===1,w=Yj("userSelect","none");return $.createElement(Se,{className:_,onMouseLeave:this.handleLeaveWrapper,onTouchMove:this.handleTouchMove,style:w},this.renderBackground(),k&&this.renderPanorama(),this.renderSlide(O,S),this.renderTravellerLayer(O,"startX"),this.renderTravellerLayer(S,"endX"),(P||g||A||j||m)&&this.renderText())}}],[{key:"renderDefaultTraveller",value:function(i){var l=i.x,s=i.y,c=i.width,d=i.height,p=i.stroke,v=Math.floor(s+d/2)-1;return $.createElement($.Fragment,null,$.createElement("rect",{x:l,y:s,width:c,height:d,fill:p,stroke:"none"}),$.createElement("line",{x1:l+1,y1:v,x2:l+c-1,y2:v,fill:"none",stroke:"#fff"}),$.createElement("line",{x1:l+1,y1:v+2,x2:l+c-1,y2:v+2,fill:"none",stroke:"#fff"}))}},{key:"renderTraveller",value:function(i,l){var s;return $.isValidElement(i)?s=$.cloneElement(i,l):he(i)?s=i(l):s=e.renderDefaultTraveller(l),s}},{key:"getDerivedStateFromProps",value:function(i,l){var s=i.data,c=i.width,d=i.x,p=i.travellerWidth,v=i.updateId,h=i.startIndex,m=i.endIndex;if(s!==l.prevData||v!==l.prevUpdateId)return Mc({prevData:s,prevTravellerWidth:p,prevUpdateId:v,prevX:d,prevWidth:c},s&&s.length?n_({data:s,width:c,x:d,travellerWidth:p,startIndex:h,endIndex:m}):{scale:null,scaleValues:null});if(l.scale&&(c!==l.prevWidth||d!==l.prevX||p!==l.prevTravellerWidth)){l.scale.range([d,d+c-p]);var b=l.scale.domain().map(function(O){return l.scale(O)});return{prevData:s,prevTravellerWidth:p,prevUpdateId:v,prevX:d,prevWidth:c,startX:l.scale(i.startIndex),endX:l.scale(i.endIndex),scaleValues:b}}return null}},{key:"getIndexInRange",value:function(i,l){for(var s=i.length,c=0,d=s-1;d-c>1;){var p=Math.floor((c+d)/2);i[p]>l?d=p:c=p}return l>=i[d]?d:c}}])})(te.PureComponent);St(wi,"displayName","Brush");St(wi,"defaultProps",{height:40,travellerWidth:5,gap:1,fill:"#fff",stroke:"#666",padding:{top:1,right:1,bottom:1,left:1},leaveTimeOut:1e3,alwaysShowText:!1});var or=function(e,r){var i=e.alwaysShow,l=e.ifOverflow;return i&&(l="extendDomain"),l===r},i_=["x","y"];function no(t){"@babel/helpers - typeof";return no=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},no(t)}function Ef(){return Ef=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},Ef.apply(this,arguments)}function tm(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function Pa(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?tm(Object(r),!0).forEach(function(i){a_(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):tm(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function a_(t,e,r){return e=o_(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function o_(t){var e=l_(t,"string");return no(e)=="symbol"?e:e+""}function l_(t,e){if(no(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(no(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}function u_(t,e){if(t==null)return{};var r=s_(t,e),i,l;if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(t);for(l=0;l<s.length;l++)i=s[l],!(e.indexOf(i)>=0)&&Object.prototype.propertyIsEnumerable.call(t,i)&&(r[i]=t[i])}return r}function s_(t,e){if(t==null)return{};var r={};for(var i in t)if(Object.prototype.hasOwnProperty.call(t,i)){if(e.indexOf(i)>=0)continue;r[i]=t[i]}return r}function c_(t,e){var r=t.x,i=t.y,l=u_(t,i_),s="".concat(r),c=parseInt(s,10),d="".concat(i),p=parseInt(d,10),v="".concat(e.height||l.height),h=parseInt(v,10),m="".concat(e.width||l.width),b=parseInt(m,10);return Pa(Pa(Pa(Pa(Pa({},e),l),c?{x:c}:{}),p?{y:p}:{}),{},{height:h,width:b,name:e.name,radius:e.radius})}function rm(t){return $.createElement(_j,Ef({shapeType:"rectangle",propTransformer:c_,activeClassName:"recharts-active-bar"},t))}var f_=function(e){var r=arguments.length>1&&arguments[1]!==void 0?arguments[1]:0;return function(i,l){if(typeof e=="number")return e;var s=ie(i)||cx(i);return s?e(i,l):(s||xn(),r)}},d_=["value","background"],Mg;function xi(t){"@babel/helpers - typeof";return xi=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},xi(t)}function p_(t,e){if(t==null)return{};var r=v_(t,e),i,l;if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(t);for(l=0;l<s.length;l++)i=s[l],!(e.indexOf(i)>=0)&&Object.prototype.propertyIsEnumerable.call(t,i)&&(r[i]=t[i])}return r}function v_(t,e){if(t==null)return{};var r={};for(var i in t)if(Object.prototype.hasOwnProperty.call(t,i)){if(e.indexOf(i)>=0)continue;r[i]=t[i]}return r}function nu(){return nu=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},nu.apply(this,arguments)}function nm(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function We(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?nm(Object(r),!0).forEach(function(i){Qr(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):nm(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function y_(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")}function im(t,e){for(var r=0;r<e.length;r++){var i=e[r];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(t,Dg(i.key),i)}}function h_(t,e,r){return e&&im(t.prototype,e),r&&im(t,r),Object.defineProperty(t,"prototype",{writable:!1}),t}function m_(t,e,r){return e=iu(e),g_(t,Ng()?Reflect.construct(e,r||[],iu(t).constructor):e.apply(t,r))}function g_(t,e){if(e&&(xi(e)==="object"||typeof e=="function"))return e;if(e!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return b_(t)}function b_(t){if(t===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return t}function Ng(){try{var t=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(Ng=function(){return!!t})()}function iu(t){return iu=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(r){return r.__proto__||Object.getPrototypeOf(r)},iu(t)}function w_(t,e){if(typeof e!="function"&&e!==null)throw new TypeError("Super expression must either be null or a function");t.prototype=Object.create(e&&e.prototype,{constructor:{value:t,writable:!0,configurable:!0}}),Object.defineProperty(t,"prototype",{writable:!1}),e&&jf(t,e)}function jf(t,e){return jf=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(i,l){return i.__proto__=l,i},jf(t,e)}function Qr(t,e,r){return e=Dg(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function Dg(t){var e=x_(t,"string");return xi(e)=="symbol"?e:e+""}function x_(t,e){if(xi(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(xi(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(t)}var co=(function(t){function e(){var r;y_(this,e);for(var i=arguments.length,l=new Array(i),s=0;s<i;s++)l[s]=arguments[s];return r=m_(this,e,[].concat(l)),Qr(r,"state",{isAnimationFinished:!1}),Qr(r,"id",uo("recharts-bar-")),Qr(r,"handleAnimationEnd",function(){var c=r.props.onAnimationEnd;r.setState({isAnimationFinished:!0}),c&&c()}),Qr(r,"handleAnimationStart",function(){var c=r.props.onAnimationStart;r.setState({isAnimationFinished:!1}),c&&c()}),r}return w_(e,t),h_(e,[{key:"renderRectanglesStatically",value:function(i){var l=this,s=this.props,c=s.shape,d=s.dataKey,p=s.activeIndex,v=s.activeBar,h=fe(this.props,!1);return i&&i.map(function(m,b){var O=b===p,S=O?v:c,P=We(We(We({},h),m),{},{isActive:O,option:S,index:b,dataKey:d,onAnimationStart:l.handleAnimationStart,onAnimationEnd:l.handleAnimationEnd});return $.createElement(Se,nu({className:"recharts-bar-rectangle"},ui(l.props,m,b),{key:"rectangle-".concat(m==null?void 0:m.x,"-").concat(m==null?void 0:m.y,"-").concat(m==null?void 0:m.value,"-").concat(b)}),$.createElement(rm,P))})}},{key:"renderRectanglesWithAnimation",value:function(){var i=this,l=this.props,s=l.data,c=l.layout,d=l.isAnimationActive,p=l.animationBegin,v=l.animationDuration,h=l.animationEasing,m=l.animationId,b=this.state.prevData;return $.createElement(lr,{begin:p,duration:v,isActive:d,easing:h,from:{t:0},to:{t:1},key:"bar-".concat(m),onAnimationEnd:this.handleAnimationEnd,onAnimationStart:this.handleAnimationStart},function(O){var S=O.t,P=s.map(function(g,A){var j=b&&b[A];if(j){var _=ot(j.x,g.x),k=ot(j.y,g.y),w=ot(j.width,g.width),E=ot(j.height,g.height);return We(We({},g),{},{x:_(S),y:k(S),width:w(S),height:E(S)})}if(c==="horizontal"){var C=ot(0,g.height),M=C(S);return We(We({},g),{},{y:g.y+g.height-M,height:M})}var D=ot(0,g.width),U=D(S);return We(We({},g),{},{width:U})});return $.createElement(Se,null,i.renderRectanglesStatically(P))})}},{key:"renderRectangles",value:function(){var i=this.props,l=i.data,s=i.isAnimationActive,c=this.state.prevData;return s&&l&&l.length&&(!c||!li(c,l))?this.renderRectanglesWithAnimation():this.renderRectanglesStatically(l)}},{key:"renderBackground",value:function(){var i=this,l=this.props,s=l.data,c=l.dataKey,d=l.activeIndex,p=fe(this.props.background,!1);return s.map(function(v,h){v.value;var m=v.background,b=p_(v,d_);if(!m)return null;var O=We(We(We(We(We({},b),{},{fill:"#eee"},m),p),ui(i.props,v,h)),{},{onAnimationStart:i.handleAnimationStart,onAnimationEnd:i.handleAnimationEnd,dataKey:c,index:h,className:"recharts-bar-background-rectangle"});return $.createElement(rm,nu({key:"background-bar-".concat(h),option:i.props.background,isActive:h===d},O))})}},{key:"renderErrorBar",value:function(i,l){if(this.props.isAnimationActive&&!this.state.isAnimationFinished)return null;var s=this.props,c=s.data,d=s.xAxis,p=s.yAxis,v=s.layout,h=s.children,m=Xt(h,mu);if(!m)return null;var b=v==="vertical"?c[0].height/2:c[0].width/2,O=function(g,A){var j=Array.isArray(g.value)?g.value[1]:g.value;return{x:g.x,y:g.y,value:j,errorVal:ft(g,A)}},S={clipPath:i?"url(#clipPath-".concat(l,")"):null};return $.createElement(Se,S,m.map(function(P){return $.cloneElement(P,{key:"error-bar-".concat(l,"-").concat(P.props.dataKey),data:c,xAxis:d,yAxis:p,layout:v,offset:b,dataPointFormatter:O})}))}},{key:"render",value:function(){var i=this.props,l=i.hide,s=i.data,c=i.className,d=i.xAxis,p=i.yAxis,v=i.left,h=i.top,m=i.width,b=i.height,O=i.isAnimationActive,S=i.background,P=i.id;if(l||!s||!s.length)return null;var g=this.state.isAnimationFinished,A=ge("recharts-bar",c),j=d&&d.allowDataOverflow,_=p&&p.allowDataOverflow,k=j||_,w=be(P)?this.id:P;return $.createElement(Se,{className:A},j||_?$.createElement("defs",null,$.createElement("clipPath",{id:"clipPath-".concat(w)},$.createElement("rect",{x:j?v:v-m/2,y:_?h:h-b/2,width:j?m:m*2,height:_?b:b*2}))):null,$.createElement(Se,{className:"recharts-bar-rectangles",clipPath:k?"url(#clipPath-".concat(w,")"):null},S?this.renderBackground():null,this.renderRectangles()),this.renderErrorBar(k,w),(!O||g)&&Sr.renderCallByParent(this.props,s))}}],[{key:"getDerivedStateFromProps",value:function(i,l){return i.animationId!==l.prevAnimationId?{prevAnimationId:i.animationId,curData:i.data,prevData:l.curData}:i.data!==l.curData?{curData:i.data}:null}}])})(te.PureComponent);Mg=co;Qr(co,"displayName","Bar");Qr(co,"defaultProps",{xAxisId:0,yAxisId:0,legendType:"rect",minPointSize:0,hide:!1,data:[],layout:"vertical",activeBar:!1,isAnimationActive:!Sn.isSsr,animationBegin:0,animationDuration:400,animationEasing:"ease"});Qr(co,"getComposedData",function(t){var e=t.props,r=t.item,i=t.barPosition,l=t.bandSize,s=t.xAxis,c=t.yAxis,d=t.xAxisTicks,p=t.yAxisTicks,v=t.stackedData,h=t.dataStartIndex,m=t.displayedData,b=t.offset,O=uP(i,r);if(!O)return null;var S=e.layout,P=r.type.defaultProps,g=P!==void 0?We(We({},P),r.props):r.props,A=g.dataKey,j=g.children,_=g.minPointSize,k=S==="horizontal"?c:s,w=v?k.scale.domain():null,E=yP({numericAxis:k}),C=Xt(j,Xm),M=m.map(function(D,U){var F,H,q,Y,V,G;v?F=sP(v[h+U],w):(F=ft(D,A),Array.isArray(F)||(F=[E,F]));var ee=f_(_,Mg.defaultProps.minPointSize)(F[1],U);if(S==="horizontal"){var z,X=[c.scale(F[0]),c.scale(F[1])],K=X[0],ae=X[1];H=Xy({axis:s,ticks:d,bandSize:l,offset:O.offset,entry:D,index:U}),q=(z=ae??K)!==null&&z!==void 0?z:void 0,Y=O.size;var re=K-ae;if(V=Number.isNaN(re)?0:re,G={x:H,y:c.y,width:Y,height:c.height},Math.abs(ee)>0&&Math.abs(V)<Math.abs(ee)){var de=Ht(V||ee)*(Math.abs(ee)-Math.abs(V));q-=de,V+=de}}else{var we=[s.scale(F[0]),s.scale(F[1])],Ce=we[0],Ie=we[1];if(H=Ce,q=Xy({axis:c,ticks:p,bandSize:l,offset:O.offset,entry:D,index:U}),Y=Ie-Ce,V=O.size,G={x:s.x,y:q,width:s.width,height:V},Math.abs(ee)>0&&Math.abs(Y)<Math.abs(ee)){var At=Ht(Y||ee)*(Math.abs(ee)-Math.abs(Y));Y+=At}}return We(We(We({},D),{},{x:H,y:q,width:Y,height:V,value:v?F:F[1],payload:D,background:G},C&&C[U]&&C[U].props),{},{tooltipPayload:[fg(r,D)],tooltipPosition:{x:H+Y/2,y:q+V/2}})});return We({data:M,layout:S},b)});function io(t){"@babel/helpers - typeof";return io=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},io(t)}function O_(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")}function am(t,e){for(var r=0;r<e.length;r++){var i=e[r];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(t,Lg(i.key),i)}}function S_(t,e,r){return e&&am(t.prototype,e),r&&am(t,r),Object.defineProperty(t,"prototype",{writable:!1}),t}function om(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function Kt(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?om(Object(r),!0).forEach(function(i){Au(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):om(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function Au(t,e,r){return e=Lg(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function Lg(t){var e=P_(t,"string");return io(e)=="symbol"?e:e+""}function P_(t,e){if(io(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(io(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}var A_=function(e,r,i,l,s){var c=e.width,d=e.height,p=e.layout,v=e.children,h=Object.keys(r),m={left:i.left,leftMirror:i.left,right:c-i.right,rightMirror:c-i.right,top:i.top,topMirror:i.top,bottom:d-i.bottom,bottomMirror:d-i.bottom},b=!!Pt(v,co);return h.reduce(function(O,S){var P=r[S],g=P.orientation,A=P.domain,j=P.padding,_=j===void 0?{}:j,k=P.mirror,w=P.reversed,E="".concat(g).concat(k?"Mirror":""),C,M,D,U,F;if(P.type==="number"&&(P.padding==="gap"||P.padding==="no-gap")){var H=A[1]-A[0],q=1/0,Y=P.categoricalDomain.sort(px);if(Y.forEach(function(we,Ce){Ce>0&&(q=Math.min((we||0)-(Y[Ce-1]||0),q))}),Number.isFinite(q)){var V=q/H,G=P.layout==="vertical"?i.height:i.width;if(P.padding==="gap"&&(C=V*G/2),P.padding==="no-gap"){var ee=Gt(e.barCategoryGap,V*G),z=V*G/2;C=z-ee-(z-ee)/G*ee}}}l==="xAxis"?M=[i.left+(_.left||0)+(C||0),i.left+i.width-(_.right||0)-(C||0)]:l==="yAxis"?M=p==="horizontal"?[i.top+i.height-(_.bottom||0),i.top+(_.top||0)]:[i.top+(_.top||0)+(C||0),i.top+i.height-(_.bottom||0)-(C||0)]:M=P.range,w&&(M=[M[1],M[0]]);var X=lg(P,s,b),K=X.scale,ae=X.realScaleType;K.domain(A).range(M),ug(K);var re=sg(K,Kt(Kt({},P),{},{realScaleType:ae}));l==="xAxis"?(F=g==="top"&&!k||g==="bottom"&&k,D=i.left,U=m[E]-F*P.height):l==="yAxis"&&(F=g==="left"&&!k||g==="right"&&k,D=m[E]-F*P.width,U=i.top);var de=Kt(Kt(Kt({},P),re),{},{realScaleType:ae,x:D,y:U,scale:K,width:l==="xAxis"?i.width:P.width,height:l==="yAxis"?i.height:P.height});return de.bandSize=Fl(de,re),!P.hide&&l==="xAxis"?m[E]+=(F?-1:1)*de.height:P.hide||(m[E]+=(F?-1:1)*de.width),Kt(Kt({},O),{},Au({},S,de))},{})},Rg=function(e,r){var i=e.x,l=e.y,s=r.x,c=r.y;return{x:Math.min(i,s),y:Math.min(l,c),width:Math.abs(s-i),height:Math.abs(c-l)}},E_=function(e){var r=e.x1,i=e.y1,l=e.x2,s=e.y2;return Rg({x:r,y:i},{x:l,y:s})},Bg=(function(){function t(e){O_(this,t),this.scale=e}return S_(t,[{key:"domain",get:function(){return this.scale.domain}},{key:"range",get:function(){return this.scale.range}},{key:"rangeMin",get:function(){return this.range()[0]}},{key:"rangeMax",get:function(){return this.range()[1]}},{key:"bandwidth",get:function(){return this.scale.bandwidth}},{key:"apply",value:function(r){var i=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{},l=i.bandAware,s=i.position;if(r!==void 0){if(s)switch(s){case"start":return this.scale(r);case"middle":{var c=this.bandwidth?this.bandwidth()/2:0;return this.scale(r)+c}case"end":{var d=this.bandwidth?this.bandwidth():0;return this.scale(r)+d}default:return this.scale(r)}if(l){var p=this.bandwidth?this.bandwidth()/2:0;return this.scale(r)+p}return this.scale(r)}}},{key:"isInRange",value:function(r){var i=this.range(),l=i[0],s=i[i.length-1];return l<=s?r>=l&&r<=s:r>=s&&r<=l}}],[{key:"create",value:function(r){return new t(r)}}])})();Au(Bg,"EPS",1e-4);var Jf=function(e){var r=Object.keys(e).reduce(function(i,l){return Kt(Kt({},i),{},Au({},l,Bg.create(e[l])))},{});return Kt(Kt({},r),{},{apply:function(l){var s=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{},c=s.bandAware,d=s.position;return Jw(l,function(p,v){return r[v].apply(p,{bandAware:c,position:d})})},isInRange:function(l){return Dm(l,function(s,c){return r[c].isInRange(s)})}})};function j_(t){return(t%180+180)%180}var __=function(e){var r=e.width,i=e.height,l=arguments.length>1&&arguments[1]!==void 0?arguments[1]:0,s=j_(l),c=s*Math.PI/180,d=Math.atan(i/r),p=c>d&&c<Math.PI-d?i/Math.sin(c):r/Math.cos(c);return Math.abs(p)},k_=ex(function(t){return{x:t.left,y:t.top,width:t.width,height:t.height}},function(t){return["l",t.left,"t",t.top,"w",t.width,"h",t.height].join("")}),ed=te.createContext(void 0),td=te.createContext(void 0),zg=te.createContext(void 0),Fg=te.createContext({}),Wg=te.createContext(void 0),Ug=te.createContext(0),Vg=te.createContext(0),lm=function(e){var r=e.state,i=r.xAxisMap,l=r.yAxisMap,s=r.offset,c=e.clipPathId,d=e.children,p=e.width,v=e.height,h=k_(s);return $.createElement(ed.Provider,{value:i},$.createElement(td.Provider,{value:l},$.createElement(Fg.Provider,{value:s},$.createElement(zg.Provider,{value:h},$.createElement(Wg.Provider,{value:c},$.createElement(Ug.Provider,{value:v},$.createElement(Vg.Provider,{value:p},d)))))))},T_=function(){return te.useContext(Wg)},Kg=function(e){var r=te.useContext(ed);r==null&&xn();var i=r[e];return i==null&&xn(),i},C_=function(){var e=te.useContext(ed);return Yr(e)},$_=function(){var e=te.useContext(td),r=tx(e,function(i){return Dm(i.domain,Number.isFinite)});return r||Yr(e)},Hg=function(e){var r=te.useContext(td);r==null&&xn();var i=r[e];return i==null&&xn(),i},I_=function(){var e=te.useContext(zg);return e},M_=function(){return te.useContext(Fg)},rd=function(){return te.useContext(Vg)},nd=function(){return te.useContext(Ug)};function Oi(t){"@babel/helpers - typeof";return Oi=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},Oi(t)}function N_(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")}function D_(t,e){for(var r=0;r<e.length;r++){var i=e[r];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(t,Xg(i.key),i)}}function L_(t,e,r){return e&&D_(t.prototype,e),Object.defineProperty(t,"prototype",{writable:!1}),t}function R_(t,e,r){return e=au(e),B_(t,Gg()?Reflect.construct(e,r||[],au(t).constructor):e.apply(t,r))}function B_(t,e){if(e&&(Oi(e)==="object"||typeof e=="function"))return e;if(e!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return z_(t)}function z_(t){if(t===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return t}function Gg(){try{var t=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(Gg=function(){return!!t})()}function au(t){return au=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(r){return r.__proto__||Object.getPrototypeOf(r)},au(t)}function F_(t,e){if(typeof e!="function"&&e!==null)throw new TypeError("Super expression must either be null or a function");t.prototype=Object.create(e&&e.prototype,{constructor:{value:t,writable:!0,configurable:!0}}),Object.defineProperty(t,"prototype",{writable:!1}),e&&_f(t,e)}function _f(t,e){return _f=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(i,l){return i.__proto__=l,i},_f(t,e)}function um(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function sm(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?um(Object(r),!0).forEach(function(i){id(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):um(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function id(t,e,r){return e=Xg(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function Xg(t){var e=W_(t,"string");return Oi(e)=="symbol"?e:e+""}function W_(t,e){if(Oi(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(Oi(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(t)}function U_(t,e){return G_(t)||H_(t,e)||K_(t,e)||V_()}function V_(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function K_(t,e){if(t){if(typeof t=="string")return cm(t,e);var r=Object.prototype.toString.call(t).slice(8,-1);if(r==="Object"&&t.constructor&&(r=t.constructor.name),r==="Map"||r==="Set")return Array.from(t);if(r==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(r))return cm(t,e)}}function cm(t,e){(e==null||e>t.length)&&(e=t.length);for(var r=0,i=new Array(e);r<e;r++)i[r]=t[r];return i}function H_(t,e){var r=t==null?null:typeof Symbol<"u"&&t[Symbol.iterator]||t["@@iterator"];if(r!=null){var i,l,s,c,d=[],p=!0,v=!1;try{if(s=(r=r.call(t)).next,e!==0)for(;!(p=(i=s.call(r)).done)&&(d.push(i.value),d.length!==e);p=!0);}catch(h){v=!0,l=h}finally{try{if(!p&&r.return!=null&&(c=r.return(),Object(c)!==c))return}finally{if(v)throw l}}return d}}function G_(t){if(Array.isArray(t))return t}function kf(){return kf=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},kf.apply(this,arguments)}var X_=function(e,r){var i;return $.isValidElement(e)?i=$.cloneElement(e,r):he(e)?i=e(r):i=$.createElement("line",kf({},r,{className:"recharts-reference-line-line"})),i},Y_=function(e,r,i,l,s,c,d,p,v){var h=s.x,m=s.y,b=s.width,O=s.height;if(i){var S=v.y,P=e.y.apply(S,{position:c});if(or(v,"discard")&&!e.y.isInRange(P))return null;var g=[{x:h+b,y:P},{x:h,y:P}];return p==="left"?g.reverse():g}if(r){var A=v.x,j=e.x.apply(A,{position:c});if(or(v,"discard")&&!e.x.isInRange(j))return null;var _=[{x:j,y:m+O},{x:j,y:m}];return d==="top"?_.reverse():_}if(l){var k=v.segment,w=k.map(function(E){return e.apply(E,{position:c})});return or(v,"discard")&&rx(w,function(E){return!e.isInRange(E)})?null:w}return null};function q_(t){var e=t.x,r=t.y,i=t.segment,l=t.xAxisId,s=t.yAxisId,c=t.shape,d=t.className,p=t.alwaysShow,v=T_(),h=Kg(l),m=Hg(s),b=I_();if(!v||!b)return null;Or(p===void 0,'The alwaysShow prop is deprecated. Please use ifOverflow="extendDomain" instead.');var O=Jf({x:h.scale,y:m.scale}),S=Ge(e),P=Ge(r),g=i&&i.length===2,A=Y_(O,S,P,g,b,t.position,h.orientation,m.orientation,t);if(!A)return null;var j=U_(A,2),_=j[0],k=_.x,w=_.y,E=j[1],C=E.x,M=E.y,D=or(t,"hidden")?"url(#".concat(v,")"):void 0,U=sm(sm({clipPath:D},fe(t,!0)),{},{x1:k,y1:w,x2:C,y2:M});return $.createElement(Se,{className:ge("recharts-reference-line",d)},X_(c,U),et.renderCallByParent(t,E_({x1:k,y1:w,x2:C,y2:M})))}var ad=(function(t){function e(){return N_(this,e),R_(this,e,arguments)}return F_(e,t),L_(e,[{key:"render",value:function(){return $.createElement(q_,this.props)}}])})($.Component);id(ad,"displayName","ReferenceLine");id(ad,"defaultProps",{isFront:!1,ifOverflow:"discard",xAxisId:0,yAxisId:0,fill:"none",stroke:"#ccc",fillOpacity:1,strokeWidth:1,position:"middle"});function Tf(){return Tf=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},Tf.apply(this,arguments)}function Si(t){"@babel/helpers - typeof";return Si=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},Si(t)}function fm(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function dm(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?fm(Object(r),!0).forEach(function(i){Eu(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):fm(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function Q_(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")}function Z_(t,e){for(var r=0;r<e.length;r++){var i=e[r];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(t,qg(i.key),i)}}function J_(t,e,r){return e&&Z_(t.prototype,e),Object.defineProperty(t,"prototype",{writable:!1}),t}function ek(t,e,r){return e=ou(e),tk(t,Yg()?Reflect.construct(e,r||[],ou(t).constructor):e.apply(t,r))}function tk(t,e){if(e&&(Si(e)==="object"||typeof e=="function"))return e;if(e!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return rk(t)}function rk(t){if(t===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return t}function Yg(){try{var t=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(Yg=function(){return!!t})()}function ou(t){return ou=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(r){return r.__proto__||Object.getPrototypeOf(r)},ou(t)}function nk(t,e){if(typeof e!="function"&&e!==null)throw new TypeError("Super expression must either be null or a function");t.prototype=Object.create(e&&e.prototype,{constructor:{value:t,writable:!0,configurable:!0}}),Object.defineProperty(t,"prototype",{writable:!1}),e&&Cf(t,e)}function Cf(t,e){return Cf=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(i,l){return i.__proto__=l,i},Cf(t,e)}function Eu(t,e,r){return e=qg(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function qg(t){var e=ik(t,"string");return Si(e)=="symbol"?e:e+""}function ik(t,e){if(Si(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(Si(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(t)}var ak=function(e){var r=e.x,i=e.y,l=e.xAxis,s=e.yAxis,c=Jf({x:l.scale,y:s.scale}),d=c.apply({x:r,y:i},{bandAware:!0});return or(e,"discard")&&!c.isInRange(d)?null:d},ju=(function(t){function e(){return Q_(this,e),ek(this,e,arguments)}return nk(e,t),J_(e,[{key:"render",value:function(){var i=this.props,l=i.x,s=i.y,c=i.r,d=i.alwaysShow,p=i.clipPathId,v=Ge(l),h=Ge(s);if(Or(d===void 0,'The alwaysShow prop is deprecated. Please use ifOverflow="extendDomain" instead.'),!v||!h)return null;var m=ak(this.props);if(!m)return null;var b=m.x,O=m.y,S=this.props,P=S.shape,g=S.className,A=or(this.props,"hidden")?"url(#".concat(p,")"):void 0,j=dm(dm({clipPath:A},fe(this.props,!0)),{},{cx:b,cy:O});return $.createElement(Se,{className:ge("recharts-reference-dot",g)},e.renderDot(P,j),et.renderCallByParent(this.props,{x:b-c,y:O-c,width:2*c,height:2*c}))}}])})($.Component);Eu(ju,"displayName","ReferenceDot");Eu(ju,"defaultProps",{isFront:!1,ifOverflow:"discard",xAxisId:0,yAxisId:0,r:10,fill:"#fff",stroke:"#ccc",fillOpacity:1,strokeWidth:1});Eu(ju,"renderDot",function(t,e){var r;return $.isValidElement(t)?r=$.cloneElement(t,e):he(t)?r=t(e):r=$.createElement(so,Tf({},e,{cx:e.cx,cy:e.cy,className:"recharts-reference-dot-dot"})),r});function $f(){return $f=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},$f.apply(this,arguments)}function Pi(t){"@babel/helpers - typeof";return Pi=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},Pi(t)}function pm(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function vm(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?pm(Object(r),!0).forEach(function(i){_u(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):pm(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function ok(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")}function lk(t,e){for(var r=0;r<e.length;r++){var i=e[r];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(t,Zg(i.key),i)}}function uk(t,e,r){return e&&lk(t.prototype,e),Object.defineProperty(t,"prototype",{writable:!1}),t}function sk(t,e,r){return e=lu(e),ck(t,Qg()?Reflect.construct(e,r||[],lu(t).constructor):e.apply(t,r))}function ck(t,e){if(e&&(Pi(e)==="object"||typeof e=="function"))return e;if(e!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return fk(t)}function fk(t){if(t===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return t}function Qg(){try{var t=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(Qg=function(){return!!t})()}function lu(t){return lu=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(r){return r.__proto__||Object.getPrototypeOf(r)},lu(t)}function dk(t,e){if(typeof e!="function"&&e!==null)throw new TypeError("Super expression must either be null or a function");t.prototype=Object.create(e&&e.prototype,{constructor:{value:t,writable:!0,configurable:!0}}),Object.defineProperty(t,"prototype",{writable:!1}),e&&If(t,e)}function If(t,e){return If=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(i,l){return i.__proto__=l,i},If(t,e)}function _u(t,e,r){return e=Zg(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function Zg(t){var e=pk(t,"string");return Pi(e)=="symbol"?e:e+""}function pk(t,e){if(Pi(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(Pi(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(t)}var vk=function(e,r,i,l,s){var c=s.x1,d=s.x2,p=s.y1,v=s.y2,h=s.xAxis,m=s.yAxis;if(!h||!m)return null;var b=Jf({x:h.scale,y:m.scale}),O={x:e?b.x.apply(c,{position:"start"}):b.x.rangeMin,y:i?b.y.apply(p,{position:"start"}):b.y.rangeMin},S={x:r?b.x.apply(d,{position:"end"}):b.x.rangeMax,y:l?b.y.apply(v,{position:"end"}):b.y.rangeMax};return or(s,"discard")&&(!b.isInRange(O)||!b.isInRange(S))?null:Rg(O,S)},ku=(function(t){function e(){return ok(this,e),sk(this,e,arguments)}return dk(e,t),uk(e,[{key:"render",value:function(){var i=this.props,l=i.x1,s=i.x2,c=i.y1,d=i.y2,p=i.className,v=i.alwaysShow,h=i.clipPathId;Or(v===void 0,'The alwaysShow prop is deprecated. Please use ifOverflow="extendDomain" instead.');var m=Ge(l),b=Ge(s),O=Ge(c),S=Ge(d),P=this.props.shape;if(!m&&!b&&!O&&!S&&!P)return null;var g=vk(m,b,O,S,this.props);if(!g&&!P)return null;var A=or(this.props,"hidden")?"url(#".concat(h,")"):void 0;return $.createElement(Se,{className:ge("recharts-reference-area",p)},e.renderRect(P,vm(vm({clipPath:A},fe(this.props,!0)),g)),et.renderCallByParent(this.props,g))}}])})($.Component);_u(ku,"displayName","ReferenceArea");_u(ku,"defaultProps",{isFront:!1,ifOverflow:"discard",xAxisId:0,yAxisId:0,r:10,fill:"#ccc",fillOpacity:.5,stroke:"none",strokeWidth:1});_u(ku,"renderRect",function(t,e){var r;return $.isValidElement(t)?r=$.cloneElement(t,e):he(t)?r=t(e):r=$.createElement(Zf,$f({},e,{className:"recharts-reference-area-rect"})),r});function Jg(t,e,r){if(e<1)return[];if(e===1&&r===void 0)return t;for(var i=[],l=0;l<t.length;l+=e)i.push(t[l]);return i}function yk(t,e,r){var i={width:t.width+e.width,height:t.height+e.height};return __(i,r)}function hk(t,e,r){var i=r==="width",l=t.x,s=t.y,c=t.width,d=t.height;return e===1?{start:i?l:s,end:i?l+c:s+d}:{start:i?l+c:s+d,end:i?l:s}}function uu(t,e,r,i,l){if(t*e<t*i||t*e>t*l)return!1;var s=r();return t*(e-t*s/2-i)>=0&&t*(e+t*s/2-l)<=0}function mk(t,e){return Jg(t,e+1)}function gk(t,e,r,i,l){for(var s=(i||[]).slice(),c=e.start,d=e.end,p=0,v=1,h=c,m=function(){var S=i==null?void 0:i[p];if(S===void 0)return{v:Jg(i,v)};var P=p,g,A=function(){return g===void 0&&(g=r(S,P)),g},j=S.coordinate,_=p===0||uu(t,j,A,h,d);_||(p=0,h=c,v+=1),_&&(h=j+t*(A()/2+l),p+=v)},b;v<=s.length;)if(b=m(),b)return b.v;return[]}function ao(t){"@babel/helpers - typeof";return ao=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},ao(t)}function ym(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function at(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?ym(Object(r),!0).forEach(function(i){bk(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):ym(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function bk(t,e,r){return e=wk(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function wk(t){var e=xk(t,"string");return ao(e)=="symbol"?e:e+""}function xk(t,e){if(ao(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(ao(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}function Ok(t,e,r,i,l){for(var s=(i||[]).slice(),c=s.length,d=e.start,p=e.end,v=function(b){var O=s[b],S,P=function(){return S===void 0&&(S=r(O,b)),S};if(b===c-1){var g=t*(O.coordinate+t*P()/2-p);s[b]=O=at(at({},O),{},{tickCoord:g>0?O.coordinate-g*t:O.coordinate})}else s[b]=O=at(at({},O),{},{tickCoord:O.coordinate});var A=uu(t,O.tickCoord,P,d,p);A&&(p=O.tickCoord-t*(P()/2+l),s[b]=at(at({},O),{},{isShow:!0}))},h=c-1;h>=0;h--)v(h);return s}function Sk(t,e,r,i,l,s){var c=(i||[]).slice(),d=c.length,p=e.start,v=e.end;if(s){var h=i[d-1],m=r(h,d-1),b=t*(h.coordinate+t*m/2-v);c[d-1]=h=at(at({},h),{},{tickCoord:b>0?h.coordinate-b*t:h.coordinate});var O=uu(t,h.tickCoord,function(){return m},p,v);O&&(v=h.tickCoord-t*(m/2+l),c[d-1]=at(at({},h),{},{isShow:!0}))}for(var S=s?d-1:d,P=function(j){var _=c[j],k,w=function(){return k===void 0&&(k=r(_,j)),k};if(j===0){var E=t*(_.coordinate-t*w()/2-p);c[j]=_=at(at({},_),{},{tickCoord:E<0?_.coordinate-E*t:_.coordinate})}else c[j]=_=at(at({},_),{},{tickCoord:_.coordinate});var C=uu(t,_.tickCoord,w,p,v);C&&(p=_.tickCoord+t*(w()/2+l),c[j]=at(at({},_),{},{isShow:!0}))},g=0;g<S;g++)P(g);return c}function od(t,e,r){var i=t.tick,l=t.ticks,s=t.viewBox,c=t.minTickGap,d=t.orientation,p=t.interval,v=t.tickFormatter,h=t.unit,m=t.angle;if(!l||!l.length||!i)return[];if(ie(p)||Sn.isSsr)return mk(l,typeof p=="number"&&ie(p)?p:0);var b=[],O=d==="top"||d==="bottom"?"width":"height",S=h&&O==="width"?Ea(h,{fontSize:e,letterSpacing:r}):{width:0,height:0},P=function(_,k){var w=he(v)?v(_.value,k):_.value;return O==="width"?yk(Ea(w,{fontSize:e,letterSpacing:r}),S,m):Ea(w,{fontSize:e,letterSpacing:r})[O]},g=l.length>=2?Ht(l[1].coordinate-l[0].coordinate):1,A=hk(s,g,O);return p==="equidistantPreserveStart"?gk(g,A,P,l,c):(p==="preserveStart"||p==="preserveStartEnd"?b=Sk(g,A,P,l,c,p==="preserveStartEnd"):b=Ok(g,A,P,l,c),b.filter(function(j){return j.isShow}))}var Pk=["viewBox"],Ak=["viewBox"],Ek=["ticks"];function Ai(t){"@babel/helpers - typeof";return Ai=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},Ai(t)}function ri(){return ri=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},ri.apply(this,arguments)}function hm(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function Ke(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?hm(Object(r),!0).forEach(function(i){ld(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):hm(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function Nc(t,e){if(t==null)return{};var r=jk(t,e),i,l;if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(t);for(l=0;l<s.length;l++)i=s[l],!(e.indexOf(i)>=0)&&Object.prototype.propertyIsEnumerable.call(t,i)&&(r[i]=t[i])}return r}function jk(t,e){if(t==null)return{};var r={};for(var i in t)if(Object.prototype.hasOwnProperty.call(t,i)){if(e.indexOf(i)>=0)continue;r[i]=t[i]}return r}function _k(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")}function mm(t,e){for(var r=0;r<e.length;r++){var i=e[r];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(t,tb(i.key),i)}}function kk(t,e,r){return e&&mm(t.prototype,e),r&&mm(t,r),Object.defineProperty(t,"prototype",{writable:!1}),t}function Tk(t,e,r){return e=su(e),Ck(t,eb()?Reflect.construct(e,r||[],su(t).constructor):e.apply(t,r))}function Ck(t,e){if(e&&(Ai(e)==="object"||typeof e=="function"))return e;if(e!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return $k(t)}function $k(t){if(t===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return t}function eb(){try{var t=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(eb=function(){return!!t})()}function su(t){return su=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(r){return r.__proto__||Object.getPrototypeOf(r)},su(t)}function Ik(t,e){if(typeof e!="function"&&e!==null)throw new TypeError("Super expression must either be null or a function");t.prototype=Object.create(e&&e.prototype,{constructor:{value:t,writable:!0,configurable:!0}}),Object.defineProperty(t,"prototype",{writable:!1}),e&&Mf(t,e)}function Mf(t,e){return Mf=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(i,l){return i.__proto__=l,i},Mf(t,e)}function ld(t,e,r){return e=tb(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function tb(t){var e=Mk(t,"string");return Ai(e)=="symbol"?e:e+""}function Mk(t,e){if(Ai(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(Ai(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(t)}var Ii=(function(t){function e(r){var i;return _k(this,e),i=Tk(this,e,[r]),i.state={fontSize:"",letterSpacing:""},i}return Ik(e,t),kk(e,[{key:"shouldComponentUpdate",value:function(i,l){var s=i.viewBox,c=Nc(i,Pk),d=this.props,p=d.viewBox,v=Nc(d,Ak);return!ii(s,p)||!ii(c,v)||!ii(l,this.state)}},{key:"componentDidMount",value:function(){var i=this.layerReference;if(i){var l=i.getElementsByClassName("recharts-cartesian-axis-tick-value")[0];l&&this.setState({fontSize:window.getComputedStyle(l).fontSize,letterSpacing:window.getComputedStyle(l).letterSpacing})}}},{key:"getTickLineCoord",value:function(i){var l=this.props,s=l.x,c=l.y,d=l.width,p=l.height,v=l.orientation,h=l.tickSize,m=l.mirror,b=l.tickMargin,O,S,P,g,A,j,_=m?-1:1,k=i.tickSize||h,w=ie(i.tickCoord)?i.tickCoord:i.coordinate;switch(v){case"top":O=S=i.coordinate,g=c+ +!m*p,P=g-_*k,j=P-_*b,A=w;break;case"left":P=g=i.coordinate,S=s+ +!m*d,O=S-_*k,A=O-_*b,j=w;break;case"right":P=g=i.coordinate,S=s+ +m*d,O=S+_*k,A=O+_*b,j=w;break;default:O=S=i.coordinate,g=c+ +m*p,P=g+_*k,j=P+_*b,A=w;break}return{line:{x1:O,y1:P,x2:S,y2:g},tick:{x:A,y:j}}}},{key:"getTickTextAnchor",value:function(){var i=this.props,l=i.orientation,s=i.mirror,c;switch(l){case"left":c=s?"start":"end";break;case"right":c=s?"end":"start";break;default:c="middle";break}return c}},{key:"getTickVerticalAnchor",value:function(){var i=this.props,l=i.orientation,s=i.mirror,c="end";switch(l){case"left":case"right":c="middle";break;case"top":c=s?"start":"end";break;default:c=s?"end":"start";break}return c}},{key:"renderAxisLine",value:function(){var i=this.props,l=i.x,s=i.y,c=i.width,d=i.height,p=i.orientation,v=i.mirror,h=i.axisLine,m=Ke(Ke(Ke({},fe(this.props,!1)),fe(h,!1)),{},{fill:"none"});if(p==="top"||p==="bottom"){var b=+(p==="top"&&!v||p==="bottom"&&v);m=Ke(Ke({},m),{},{x1:l,y1:s+b*d,x2:l+c,y2:s+b*d})}else{var O=+(p==="left"&&!v||p==="right"&&v);m=Ke(Ke({},m),{},{x1:l+O*c,y1:s,x2:l+O*c,y2:s+d})}return $.createElement("line",ri({},m,{className:ge("recharts-cartesian-axis-line",Mt(h,"className"))}))}},{key:"renderTicks",value:function(i,l,s){var c=this,d=this.props,p=d.tickLine,v=d.stroke,h=d.tick,m=d.tickFormatter,b=d.unit,O=od(Ke(Ke({},this.props),{},{ticks:i}),l,s),S=this.getTickTextAnchor(),P=this.getTickVerticalAnchor(),g=fe(this.props,!1),A=fe(h,!1),j=Ke(Ke({},g),{},{fill:"none"},fe(p,!1)),_=O.map(function(k,w){var E=c.getTickLineCoord(k),C=E.line,M=E.tick,D=Ke(Ke(Ke(Ke({textAnchor:S,verticalAnchor:P},g),{},{stroke:"none",fill:v},A),M),{},{index:w,payload:k,visibleTicksCount:O.length,tickFormatter:m});return $.createElement(Se,ri({className:"recharts-cartesian-axis-tick",key:"tick-".concat(k.value,"-").concat(k.coordinate,"-").concat(k.tickCoord)},ui(c.props,k,w)),p&&$.createElement("line",ri({},j,C,{className:ge("recharts-cartesian-axis-tick-line",Mt(p,"className"))})),h&&e.renderTickItem(h,D,"".concat(he(m)?m(k.value,w):k.value).concat(b||"")))});return $.createElement("g",{className:"recharts-cartesian-axis-ticks"},_)}},{key:"render",value:function(){var i=this,l=this.props,s=l.axisLine,c=l.width,d=l.height,p=l.ticksGenerator,v=l.className,h=l.hide;if(h)return null;var m=this.props,b=m.ticks,O=Nc(m,Ek),S=b;return he(p)&&(S=b&&b.length>0?p(this.props):p(O)),c<=0||d<=0||!S||!S.length?null:$.createElement(Se,{className:ge("recharts-cartesian-axis",v),ref:function(g){i.layerReference=g}},s&&this.renderAxisLine(),this.renderTicks(S,this.state.fontSize,this.state.letterSpacing),et.renderCallByParent(this.props))}}],[{key:"renderTickItem",value:function(i,l,s){var c,d=ge(l.className,"recharts-cartesian-axis-tick-value");return $.isValidElement(i)?c=$.cloneElement(i,Ke(Ke({},l),{},{className:d})):he(i)?c=i(Ke(Ke({},l),{},{className:d})):c=$.createElement(pi,ri({},l,{className:"recharts-cartesian-axis-tick-value"}),s),c}}])})(te.Component);ld(Ii,"displayName","CartesianAxis");ld(Ii,"defaultProps",{x:0,y:0,width:0,height:0,viewBox:{x:0,y:0,width:0,height:0},orientation:"bottom",ticks:[],stroke:"#666",tickLine:!0,axisLine:!0,tick:!0,mirror:!1,minTickGap:5,tickSize:6,tickMargin:2,interval:"preserveEnd"});var Nk=["x1","y1","x2","y2","key"],Dk=["offset"];function On(t){"@babel/helpers - typeof";return On=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},On(t)}function gm(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function lt(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?gm(Object(r),!0).forEach(function(i){Lk(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):gm(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function Lk(t,e,r){return e=Rk(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function Rk(t){var e=Bk(t,"string");return On(e)=="symbol"?e:e+""}function Bk(t,e){if(On(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(On(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}function gn(){return gn=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},gn.apply(this,arguments)}function bm(t,e){if(t==null)return{};var r=zk(t,e),i,l;if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(t);for(l=0;l<s.length;l++)i=s[l],!(e.indexOf(i)>=0)&&Object.prototype.propertyIsEnumerable.call(t,i)&&(r[i]=t[i])}return r}function zk(t,e){if(t==null)return{};var r={};for(var i in t)if(Object.prototype.hasOwnProperty.call(t,i)){if(e.indexOf(i)>=0)continue;r[i]=t[i]}return r}var Fk=function(e){var r=e.fill;if(!r||r==="none")return null;var i=e.fillOpacity,l=e.x,s=e.y,c=e.width,d=e.height,p=e.ry;return $.createElement("rect",{x:l,y:s,ry:p,width:c,height:d,stroke:"none",fill:r,fillOpacity:i,className:"recharts-cartesian-grid-bg"})};function rb(t,e){var r;if($.isValidElement(t))r=$.cloneElement(t,e);else if(he(t))r=t(e);else{var i=e.x1,l=e.y1,s=e.x2,c=e.y2,d=e.key,p=bm(e,Nk),v=fe(p,!1);v.offset;var h=bm(v,Dk);r=$.createElement("line",gn({},h,{x1:i,y1:l,x2:s,y2:c,fill:"none",key:d}))}return r}function Wk(t){var e=t.x,r=t.width,i=t.horizontal,l=i===void 0?!0:i,s=t.horizontalPoints;if(!l||!s||!s.length)return null;var c=s.map(function(d,p){var v=lt(lt({},t),{},{x1:e,y1:d,x2:e+r,y2:d,key:"line-".concat(p),index:p});return rb(l,v)});return $.createElement("g",{className:"recharts-cartesian-grid-horizontal"},c)}function Uk(t){var e=t.y,r=t.height,i=t.vertical,l=i===void 0?!0:i,s=t.verticalPoints;if(!l||!s||!s.length)return null;var c=s.map(function(d,p){var v=lt(lt({},t),{},{x1:d,y1:e,x2:d,y2:e+r,key:"line-".concat(p),index:p});return rb(l,v)});return $.createElement("g",{className:"recharts-cartesian-grid-vertical"},c)}function Vk(t){var e=t.horizontalFill,r=t.fillOpacity,i=t.x,l=t.y,s=t.width,c=t.height,d=t.horizontalPoints,p=t.horizontal,v=p===void 0?!0:p;if(!v||!e||!e.length)return null;var h=d.map(function(b){return Math.round(b+l-l)}).sort(function(b,O){return b-O});l!==h[0]&&h.unshift(0);var m=h.map(function(b,O){var S=!h[O+1],P=S?l+c-b:h[O+1]-b;if(P<=0)return null;var g=O%e.length;return $.createElement("rect",{key:"react-".concat(O),y:b,x:i,height:P,width:s,stroke:"none",fill:e[g],fillOpacity:r,className:"recharts-cartesian-grid-bg"})});return $.createElement("g",{className:"recharts-cartesian-gridstripes-horizontal"},m)}function Kk(t){var e=t.vertical,r=e===void 0?!0:e,i=t.verticalFill,l=t.fillOpacity,s=t.x,c=t.y,d=t.width,p=t.height,v=t.verticalPoints;if(!r||!i||!i.length)return null;var h=v.map(function(b){return Math.round(b+s-s)}).sort(function(b,O){return b-O});s!==h[0]&&h.unshift(0);var m=h.map(function(b,O){var S=!h[O+1],P=S?s+d-b:h[O+1]-b;if(P<=0)return null;var g=O%i.length;return $.createElement("rect",{key:"react-".concat(O),x:b,y:c,width:P,height:p,stroke:"none",fill:i[g],fillOpacity:l,className:"recharts-cartesian-grid-bg"})});return $.createElement("g",{className:"recharts-cartesian-gridstripes-vertical"},m)}var Hk=function(e,r){var i=e.xAxis,l=e.width,s=e.height,c=e.offset;return og(od(lt(lt(lt({},Ii.defaultProps),i),{},{ticks:wr(i,!0),viewBox:{x:0,y:0,width:l,height:s}})),c.left,c.left+c.width,r)},Gk=function(e,r){var i=e.yAxis,l=e.width,s=e.height,c=e.offset;return og(od(lt(lt(lt({},Ii.defaultProps),i),{},{ticks:wr(i,!0),viewBox:{x:0,y:0,width:l,height:s}})),c.top,c.top+c.height,r)},Jn={horizontal:!0,vertical:!0,stroke:"#ccc",fill:"none",verticalFill:[],horizontalFill:[]};function Xk(t){var e,r,i,l,s,c,d=rd(),p=nd(),v=M_(),h=lt(lt({},t),{},{stroke:(e=t.stroke)!==null&&e!==void 0?e:Jn.stroke,fill:(r=t.fill)!==null&&r!==void 0?r:Jn.fill,horizontal:(i=t.horizontal)!==null&&i!==void 0?i:Jn.horizontal,horizontalFill:(l=t.horizontalFill)!==null&&l!==void 0?l:Jn.horizontalFill,vertical:(s=t.vertical)!==null&&s!==void 0?s:Jn.vertical,verticalFill:(c=t.verticalFill)!==null&&c!==void 0?c:Jn.verticalFill,x:ie(t.x)?t.x:v.left,y:ie(t.y)?t.y:v.top,width:ie(t.width)?t.width:v.width,height:ie(t.height)?t.height:v.height}),m=h.x,b=h.y,O=h.width,S=h.height,P=h.syncWithTicks,g=h.horizontalValues,A=h.verticalValues,j=C_(),_=$_();if(!ie(O)||O<=0||!ie(S)||S<=0||!ie(m)||m!==+m||!ie(b)||b!==+b)return null;var k=h.verticalCoordinatesGenerator||Hk,w=h.horizontalCoordinatesGenerator||Gk,E=h.horizontalPoints,C=h.verticalPoints;if((!E||!E.length)&&he(w)){var M=g&&g.length,D=w({yAxis:_?lt(lt({},_),{},{ticks:M?g:_.ticks}):void 0,width:d,height:p,offset:v},M?!0:P);Or(Array.isArray(D),"horizontalCoordinatesGenerator should return Array but instead it returned [".concat(On(D),"]")),Array.isArray(D)&&(E=D)}if((!C||!C.length)&&he(k)){var U=A&&A.length,F=k({xAxis:j?lt(lt({},j),{},{ticks:U?A:j.ticks}):void 0,width:d,height:p,offset:v},U?!0:P);Or(Array.isArray(F),"verticalCoordinatesGenerator should return Array but instead it returned [".concat(On(F),"]")),Array.isArray(F)&&(C=F)}return $.createElement("g",{className:"recharts-cartesian-grid"},$.createElement(Fk,{fill:h.fill,fillOpacity:h.fillOpacity,x:h.x,y:h.y,width:h.width,height:h.height,ry:h.ry}),$.createElement(Wk,gn({},h,{offset:v,horizontalPoints:E,xAxis:j,yAxis:_})),$.createElement(Uk,gn({},h,{offset:v,verticalPoints:C,xAxis:j,yAxis:_})),$.createElement(Vk,gn({},h,{horizontalPoints:E})),$.createElement(Kk,gn({},h,{verticalPoints:C})))}Xk.displayName="CartesianGrid";var Yk=["layout","type","stroke","connectNulls","isRange","ref"],qk=["key"],nb;function Ei(t){"@babel/helpers - typeof";return Ei=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},Ei(t)}function ib(t,e){if(t==null)return{};var r=Qk(t,e),i,l;if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(t);for(l=0;l<s.length;l++)i=s[l],!(e.indexOf(i)>=0)&&Object.prototype.propertyIsEnumerable.call(t,i)&&(r[i]=t[i])}return r}function Qk(t,e){if(t==null)return{};var r={};for(var i in t)if(Object.prototype.hasOwnProperty.call(t,i)){if(e.indexOf(i)>=0)continue;r[i]=t[i]}return r}function bn(){return bn=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},bn.apply(this,arguments)}function wm(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function Xr(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?wm(Object(r),!0).forEach(function(i){ar(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):wm(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function Zk(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")}function xm(t,e){for(var r=0;r<e.length;r++){var i=e[r];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(t,ob(i.key),i)}}function Jk(t,e,r){return e&&xm(t.prototype,e),r&&xm(t,r),Object.defineProperty(t,"prototype",{writable:!1}),t}function eT(t,e,r){return e=cu(e),tT(t,ab()?Reflect.construct(e,r||[],cu(t).constructor):e.apply(t,r))}function tT(t,e){if(e&&(Ei(e)==="object"||typeof e=="function"))return e;if(e!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return rT(t)}function rT(t){if(t===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return t}function ab(){try{var t=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(ab=function(){return!!t})()}function cu(t){return cu=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(r){return r.__proto__||Object.getPrototypeOf(r)},cu(t)}function nT(t,e){if(typeof e!="function"&&e!==null)throw new TypeError("Super expression must either be null or a function");t.prototype=Object.create(e&&e.prototype,{constructor:{value:t,writable:!0,configurable:!0}}),Object.defineProperty(t,"prototype",{writable:!1}),e&&Nf(t,e)}function Nf(t,e){return Nf=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(i,l){return i.__proto__=l,i},Nf(t,e)}function ar(t,e,r){return e=ob(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function ob(t){var e=iT(t,"string");return Ei(e)=="symbol"?e:e+""}function iT(t,e){if(Ei(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(Ei(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(t)}var An=(function(t){function e(){var r;Zk(this,e);for(var i=arguments.length,l=new Array(i),s=0;s<i;s++)l[s]=arguments[s];return r=eT(this,e,[].concat(l)),ar(r,"state",{isAnimationFinished:!0}),ar(r,"id",uo("recharts-area-")),ar(r,"handleAnimationEnd",function(){var c=r.props.onAnimationEnd;r.setState({isAnimationFinished:!0}),he(c)&&c()}),ar(r,"handleAnimationStart",function(){var c=r.props.onAnimationStart;r.setState({isAnimationFinished:!1}),he(c)&&c()}),r}return nT(e,t),Jk(e,[{key:"renderDots",value:function(i,l,s){var c=this.props.isAnimationActive,d=this.state.isAnimationFinished;if(c&&!d)return null;var p=this.props,v=p.dot,h=p.points,m=p.dataKey,b=fe(this.props,!1),O=fe(v,!0),S=h.map(function(g,A){var j=Xr(Xr(Xr({key:"dot-".concat(A),r:3},b),O),{},{index:A,cx:g.x,cy:g.y,dataKey:m,value:g.value,payload:g.payload,points:h});return e.renderDotItem(v,j)}),P={clipPath:i?"url(#clipPath-".concat(l?"":"dots-").concat(s,")"):null};return $.createElement(Se,bn({className:"recharts-area-dots"},P),S)}},{key:"renderHorizontalRect",value:function(i){var l=this.props,s=l.baseLine,c=l.points,d=l.strokeWidth,p=c[0].x,v=c[c.length-1].x,h=i*Math.abs(p-v),m=qr(c.map(function(b){return b.y||0}));return ie(s)&&typeof s=="number"?m=Math.max(s,m):s&&Array.isArray(s)&&s.length&&(m=Math.max(qr(s.map(function(b){return b.y||0})),m)),ie(m)?$.createElement("rect",{x:p<v?p:p-h,y:0,width:h,height:Math.floor(m+(d?parseInt("".concat(d),10):1))}):null}},{key:"renderVerticalRect",value:function(i){var l=this.props,s=l.baseLine,c=l.points,d=l.strokeWidth,p=c[0].y,v=c[c.length-1].y,h=i*Math.abs(p-v),m=qr(c.map(function(b){return b.x||0}));return ie(s)&&typeof s=="number"?m=Math.max(s,m):s&&Array.isArray(s)&&s.length&&(m=Math.max(qr(s.map(function(b){return b.x||0})),m)),ie(m)?$.createElement("rect",{x:0,y:p<v?p:p-h,width:m+(d?parseInt("".concat(d),10):1),height:Math.floor(h)}):null}},{key:"renderClipRect",value:function(i){var l=this.props.layout;return l==="vertical"?this.renderVerticalRect(i):this.renderHorizontalRect(i)}},{key:"renderAreaStatically",value:function(i,l,s,c){var d=this.props,p=d.layout,v=d.type,h=d.stroke,m=d.connectNulls,b=d.isRange;d.ref;var O=ib(d,Yk);return $.createElement(Se,{clipPath:s?"url(#clipPath-".concat(c,")"):null},$.createElement(_a,bn({},fe(O,!0),{points:i,connectNulls:m,type:v,baseLine:l,layout:p,stroke:"none",className:"recharts-area-area"})),h!=="none"&&$.createElement(_a,bn({},fe(this.props,!1),{className:"recharts-area-curve",layout:p,type:v,connectNulls:m,fill:"none",points:i})),h!=="none"&&b&&$.createElement(_a,bn({},fe(this.props,!1),{className:"recharts-area-curve",layout:p,type:v,connectNulls:m,fill:"none",points:l})))}},{key:"renderAreaWithAnimation",value:function(i,l){var s=this,c=this.props,d=c.points,p=c.baseLine,v=c.isAnimationActive,h=c.animationBegin,m=c.animationDuration,b=c.animationEasing,O=c.animationId,S=this.state,P=S.prevPoints,g=S.prevBaseLine;return $.createElement(lr,{begin:h,duration:m,isActive:v,easing:b,from:{t:0},to:{t:1},key:"area-".concat(O),onAnimationEnd:this.handleAnimationEnd,onAnimationStart:this.handleAnimationStart},function(A){var j=A.t;if(P){var _=P.length/d.length,k=d.map(function(M,D){var U=Math.floor(D*_);if(P[U]){var F=P[U],H=ot(F.x,M.x),q=ot(F.y,M.y);return Xr(Xr({},M),{},{x:H(j),y:q(j)})}return M}),w;if(ie(p)&&typeof p=="number"){var E=ot(g,p);w=E(j)}else if(be(p)||Ci(p)){var C=ot(g,0);w=C(j)}else w=p.map(function(M,D){var U=Math.floor(D*_);if(g[U]){var F=g[U],H=ot(F.x,M.x),q=ot(F.y,M.y);return Xr(Xr({},M),{},{x:H(j),y:q(j)})}return M});return s.renderAreaStatically(k,w,i,l)}return $.createElement(Se,null,$.createElement("defs",null,$.createElement("clipPath",{id:"animationClipPath-".concat(l)},s.renderClipRect(j))),$.createElement(Se,{clipPath:"url(#animationClipPath-".concat(l,")")},s.renderAreaStatically(d,p,i,l)))})}},{key:"renderArea",value:function(i,l){var s=this.props,c=s.points,d=s.baseLine,p=s.isAnimationActive,v=this.state,h=v.prevPoints,m=v.prevBaseLine,b=v.totalLength;return p&&c&&c.length&&(!h&&b>0||!li(h,c)||!li(m,d))?this.renderAreaWithAnimation(i,l):this.renderAreaStatically(c,d,i,l)}},{key:"render",value:function(){var i,l=this.props,s=l.hide,c=l.dot,d=l.points,p=l.className,v=l.top,h=l.left,m=l.xAxis,b=l.yAxis,O=l.width,S=l.height,P=l.isAnimationActive,g=l.id;if(s||!d||!d.length)return null;var A=this.state.isAnimationFinished,j=d.length===1,_=ge("recharts-area",p),k=m&&m.allowDataOverflow,w=b&&b.allowDataOverflow,E=k||w,C=be(g)?this.id:g,M=(i=fe(c,!1))!==null&&i!==void 0?i:{r:3,strokeWidth:2},D=M.r,U=D===void 0?3:D,F=M.strokeWidth,H=F===void 0?2:F,q=Ox(c)?c:{},Y=q.clipDot,V=Y===void 0?!0:Y,G=U*2+H;return $.createElement(Se,{className:_},k||w?$.createElement("defs",null,$.createElement("clipPath",{id:"clipPath-".concat(C)},$.createElement("rect",{x:k?h:h-O/2,y:w?v:v-S/2,width:k?O:O*2,height:w?S:S*2})),!V&&$.createElement("clipPath",{id:"clipPath-dots-".concat(C)},$.createElement("rect",{x:h-G/2,y:v-G/2,width:O+G,height:S+G}))):null,j?null:this.renderArea(E,C),(c||j)&&this.renderDots(E,V,C),(!P||A)&&Sr.renderCallByParent(this.props,d))}}],[{key:"getDerivedStateFromProps",value:function(i,l){return i.animationId!==l.prevAnimationId?{prevAnimationId:i.animationId,curPoints:i.points,curBaseLine:i.baseLine,prevPoints:l.curPoints,prevBaseLine:l.curBaseLine}:i.points!==l.curPoints||i.baseLine!==l.curBaseLine?{curPoints:i.points,curBaseLine:i.baseLine}:null}}])})(te.PureComponent);nb=An;ar(An,"displayName","Area");ar(An,"defaultProps",{stroke:"#3182bd",fill:"#3182bd",fillOpacity:.6,xAxisId:0,yAxisId:0,legendType:"line",connectNulls:!1,points:[],dot:!1,activeDot:!0,hide:!1,isAnimationActive:!Sn.isSsr,animationBegin:0,animationDuration:1500,animationEasing:"ease"});ar(An,"getBaseValue",function(t,e,r,i){var l=t.layout,s=t.baseValue,c=e.props.baseValue,d=c??s;if(ie(d)&&typeof d=="number")return d;var p=l==="horizontal"?i:r,v=p.scale.domain();if(p.type==="number"){var h=Math.max(v[0],v[1]),m=Math.min(v[0],v[1]);return d==="dataMin"?m:d==="dataMax"||h<0?h:Math.max(Math.min(v[0],v[1]),0)}return d==="dataMin"?v[0]:d==="dataMax"?v[1]:v[0]});ar(An,"getComposedData",function(t){var e=t.props,r=t.item,i=t.xAxis,l=t.yAxis,s=t.xAxisTicks,c=t.yAxisTicks,d=t.bandSize,p=t.dataKey,v=t.stackedData,h=t.dataStartIndex,m=t.displayedData,b=t.offset,O=e.layout,S=v&&v.length,P=nb.getBaseValue(e,r,i,l),g=O==="horizontal",A=!1,j=m.map(function(k,w){var E;S?E=v[h+w]:(E=ft(k,p),Array.isArray(E)?A=!0:E=[P,E]);var C=E[1]==null||S&&ft(k,p)==null;return g?{x:Gy({axis:i,ticks:s,bandSize:d,entry:k,index:w}),y:C?null:l.scale(E[1]),value:E,payload:k}:{x:C?null:i.scale(E[1]),y:Gy({axis:l,ticks:c,bandSize:d,entry:k,index:w}),value:E,payload:k}}),_;return S||A?_=j.map(function(k){var w=Array.isArray(k.value)?k.value[0]:null;return g?{x:k.x,y:w!=null&&k.y!=null?l.scale(w):null}:{x:w!=null?i.scale(w):null,y:k.y}}):_=g?l.scale(P):i.scale(P),Xr({points:j,baseLine:_,layout:O,isRange:A},b)});ar(An,"renderDotItem",function(t,e){var r;if($.isValidElement(t))r=$.cloneElement(t,e);else if(he(t))r=t(e);else{var i=ge("recharts-area-dot",typeof t!="boolean"?t.className:""),l=e.key,s=ib(e,qk);r=$.createElement(so,bn({},s,{key:l,className:i}))}return r});function ji(t){"@babel/helpers - typeof";return ji=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},ji(t)}function aT(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")}function oT(t,e){for(var r=0;r<e.length;r++){var i=e[r];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(t,sb(i.key),i)}}function lT(t,e,r){return e&&oT(t.prototype,e),Object.defineProperty(t,"prototype",{writable:!1}),t}function uT(t,e,r){return e=fu(e),sT(t,lb()?Reflect.construct(e,r||[],fu(t).constructor):e.apply(t,r))}function sT(t,e){if(e&&(ji(e)==="object"||typeof e=="function"))return e;if(e!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return cT(t)}function cT(t){if(t===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return t}function lb(){try{var t=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(lb=function(){return!!t})()}function fu(t){return fu=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(r){return r.__proto__||Object.getPrototypeOf(r)},fu(t)}function fT(t,e){if(typeof e!="function"&&e!==null)throw new TypeError("Super expression must either be null or a function");t.prototype=Object.create(e&&e.prototype,{constructor:{value:t,writable:!0,configurable:!0}}),Object.defineProperty(t,"prototype",{writable:!1}),e&&Df(t,e)}function Df(t,e){return Df=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(i,l){return i.__proto__=l,i},Df(t,e)}function ub(t,e,r){return e=sb(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function sb(t){var e=dT(t,"string");return ji(e)=="symbol"?e:e+""}function dT(t,e){if(ji(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(ji(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(t)}function Lf(){return Lf=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},Lf.apply(this,arguments)}function pT(t){var e=t.xAxisId,r=rd(),i=nd(),l=Kg(e);return l==null?null:te.createElement(Ii,Lf({},l,{className:ge("recharts-".concat(l.axisType," ").concat(l.axisType),l.className),viewBox:{x:0,y:0,width:r,height:i},ticksGenerator:function(c){return wr(c,!0)}}))}var ud=(function(t){function e(){return aT(this,e),uT(this,e,arguments)}return fT(e,t),lT(e,[{key:"render",value:function(){return te.createElement(pT,this.props)}}])})(te.Component);ub(ud,"displayName","XAxis");ub(ud,"defaultProps",{allowDecimals:!0,hide:!1,orientation:"bottom",width:0,height:30,mirror:!1,xAxisId:0,tickCount:5,type:"category",padding:{left:0,right:0},allowDataOverflow:!1,scale:"auto",reversed:!1,allowDuplicatedCategory:!0});function _i(t){"@babel/helpers - typeof";return _i=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},_i(t)}function vT(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")}function yT(t,e){for(var r=0;r<e.length;r++){var i=e[r];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(t,db(i.key),i)}}function hT(t,e,r){return e&&yT(t.prototype,e),Object.defineProperty(t,"prototype",{writable:!1}),t}function mT(t,e,r){return e=du(e),gT(t,cb()?Reflect.construct(e,r||[],du(t).constructor):e.apply(t,r))}function gT(t,e){if(e&&(_i(e)==="object"||typeof e=="function"))return e;if(e!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return bT(t)}function bT(t){if(t===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return t}function cb(){try{var t=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(cb=function(){return!!t})()}function du(t){return du=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(r){return r.__proto__||Object.getPrototypeOf(r)},du(t)}function wT(t,e){if(typeof e!="function"&&e!==null)throw new TypeError("Super expression must either be null or a function");t.prototype=Object.create(e&&e.prototype,{constructor:{value:t,writable:!0,configurable:!0}}),Object.defineProperty(t,"prototype",{writable:!1}),e&&Rf(t,e)}function Rf(t,e){return Rf=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(i,l){return i.__proto__=l,i},Rf(t,e)}function fb(t,e,r){return e=db(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function db(t){var e=xT(t,"string");return _i(e)=="symbol"?e:e+""}function xT(t,e){if(_i(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(_i(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(t)}function Bf(){return Bf=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},Bf.apply(this,arguments)}var OT=function(e){var r=e.yAxisId,i=rd(),l=nd(),s=Hg(r);return s==null?null:te.createElement(Ii,Bf({},s,{className:ge("recharts-".concat(s.axisType," ").concat(s.axisType),s.className),viewBox:{x:0,y:0,width:i,height:l},ticksGenerator:function(d){return wr(d,!0)}}))},sd=(function(t){function e(){return vT(this,e),mT(this,e,arguments)}return wT(e,t),hT(e,[{key:"render",value:function(){return te.createElement(OT,this.props)}}])})(te.Component);fb(sd,"displayName","YAxis");fb(sd,"defaultProps",{allowDuplicatedCategory:!0,allowDecimals:!0,hide:!1,orientation:"left",width:60,height:0,mirror:!1,yAxisId:0,tickCount:5,type:"number",padding:{top:0,bottom:0},allowDataOverflow:!1,scale:"auto",reversed:!1});function Om(t){return ET(t)||AT(t)||PT(t)||ST()}function ST(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function PT(t,e){if(t){if(typeof t=="string")return zf(t,e);var r=Object.prototype.toString.call(t).slice(8,-1);if(r==="Object"&&t.constructor&&(r=t.constructor.name),r==="Map"||r==="Set")return Array.from(t);if(r==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(r))return zf(t,e)}}function AT(t){if(typeof Symbol<"u"&&t[Symbol.iterator]!=null||t["@@iterator"]!=null)return Array.from(t)}function ET(t){if(Array.isArray(t))return zf(t)}function zf(t,e){(e==null||e>t.length)&&(e=t.length);for(var r=0,i=new Array(e);r<e;r++)i[r]=t[r];return i}var Ff=function(e,r,i,l,s){var c=Xt(e,ad),d=Xt(e,ju),p=[].concat(Om(c),Om(d)),v=Xt(e,ku),h="".concat(l,"Id"),m=l[0],b=r;if(p.length&&(b=p.reduce(function(P,g){if(g.props[h]===i&&or(g.props,"extendDomain")&&ie(g.props[m])){var A=g.props[m];return[Math.min(P[0],A),Math.max(P[1],A)]}return P},b)),v.length){var O="".concat(m,"1"),S="".concat(m,"2");b=v.reduce(function(P,g){if(g.props[h]===i&&or(g.props,"extendDomain")&&ie(g.props[O])&&ie(g.props[S])){var A=g.props[O],j=g.props[S];return[Math.min(P[0],A,j),Math.max(P[1],A,j)]}return P},b)}return s&&s.length&&(b=s.reduce(function(P,g){return ie(g)?[Math.min(P[0],g),Math.max(P[1],g)]:P},b)),b},Dc={exports:{}},Sm;function jT(){return Sm||(Sm=1,(function(t){var e=Object.prototype.hasOwnProperty,r="~";function i(){}Object.create&&(i.prototype=Object.create(null),new i().__proto__||(r=!1));function l(p,v,h){this.fn=p,this.context=v,this.once=h||!1}function s(p,v,h,m,b){if(typeof h!="function")throw new TypeError("The listener must be a function");var O=new l(h,m||p,b),S=r?r+v:v;return p._events[S]?p._events[S].fn?p._events[S]=[p._events[S],O]:p._events[S].push(O):(p._events[S]=O,p._eventsCount++),p}function c(p,v){--p._eventsCount===0?p._events=new i:delete p._events[v]}function d(){this._events=new i,this._eventsCount=0}d.prototype.eventNames=function(){var v=[],h,m;if(this._eventsCount===0)return v;for(m in h=this._events)e.call(h,m)&&v.push(r?m.slice(1):m);return Object.getOwnPropertySymbols?v.concat(Object.getOwnPropertySymbols(h)):v},d.prototype.listeners=function(v){var h=r?r+v:v,m=this._events[h];if(!m)return[];if(m.fn)return[m.fn];for(var b=0,O=m.length,S=new Array(O);b<O;b++)S[b]=m[b].fn;return S},d.prototype.listenerCount=function(v){var h=r?r+v:v,m=this._events[h];return m?m.fn?1:m.length:0},d.prototype.emit=function(v,h,m,b,O,S){var P=r?r+v:v;if(!this._events[P])return!1;var g=this._events[P],A=arguments.length,j,_;if(g.fn){switch(g.once&&this.removeListener(v,g.fn,void 0,!0),A){case 1:return g.fn.call(g.context),!0;case 2:return g.fn.call(g.context,h),!0;case 3:return g.fn.call(g.context,h,m),!0;case 4:return g.fn.call(g.context,h,m,b),!0;case 5:return g.fn.call(g.context,h,m,b,O),!0;case 6:return g.fn.call(g.context,h,m,b,O,S),!0}for(_=1,j=new Array(A-1);_<A;_++)j[_-1]=arguments[_];g.fn.apply(g.context,j)}else{var k=g.length,w;for(_=0;_<k;_++)switch(g[_].once&&this.removeListener(v,g[_].fn,void 0,!0),A){case 1:g[_].fn.call(g[_].context);break;case 2:g[_].fn.call(g[_].context,h);break;case 3:g[_].fn.call(g[_].context,h,m);break;case 4:g[_].fn.call(g[_].context,h,m,b);break;default:if(!j)for(w=1,j=new Array(A-1);w<A;w++)j[w-1]=arguments[w];g[_].fn.apply(g[_].context,j)}}return!0},d.prototype.on=function(v,h,m){return s(this,v,h,m,!1)},d.prototype.once=function(v,h,m){return s(this,v,h,m,!0)},d.prototype.removeListener=function(v,h,m,b){var O=r?r+v:v;if(!this._events[O])return this;if(!h)return c(this,O),this;var S=this._events[O];if(S.fn)S.fn===h&&(!b||S.once)&&(!m||S.context===m)&&c(this,O);else{for(var P=0,g=[],A=S.length;P<A;P++)(S[P].fn!==h||b&&!S[P].once||m&&S[P].context!==m)&&g.push(S[P]);g.length?this._events[O]=g.length===1?g[0]:g:c(this,O)}return this},d.prototype.removeAllListeners=function(v){var h;return v?(h=r?r+v:v,this._events[h]&&c(this,h)):(this._events=new i,this._eventsCount=0),this},d.prototype.off=d.prototype.removeListener,d.prototype.addListener=d.prototype.on,d.prefixed=r,d.EventEmitter=d,t.exports=d})(Dc)),Dc.exports}var _T=jT();const kT=Cm(_T);var Lc=new kT,Rc="recharts.syncMouseEvents";function oo(t){"@babel/helpers - typeof";return oo=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},oo(t)}function TT(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")}function CT(t,e){for(var r=0;r<e.length;r++){var i=e[r];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(t,pb(i.key),i)}}function $T(t,e,r){return e&&CT(t.prototype,e),Object.defineProperty(t,"prototype",{writable:!1}),t}function Bc(t,e,r){return e=pb(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function pb(t){var e=IT(t,"string");return oo(e)=="symbol"?e:e+""}function IT(t,e){if(oo(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(oo(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(t)}var MT=(function(){function t(){TT(this,t),Bc(this,"activeIndex",0),Bc(this,"coordinateList",[]),Bc(this,"layout","horizontal")}return $T(t,[{key:"setDetails",value:function(r){var i,l=r.coordinateList,s=l===void 0?null:l,c=r.container,d=c===void 0?null:c,p=r.layout,v=p===void 0?null:p,h=r.offset,m=h===void 0?null:h,b=r.mouseHandlerCallback,O=b===void 0?null:b;this.coordinateList=(i=s??this.coordinateList)!==null&&i!==void 0?i:[],this.container=d??this.container,this.layout=v??this.layout,this.offset=m??this.offset,this.mouseHandlerCallback=O??this.mouseHandlerCallback,this.activeIndex=Math.min(Math.max(this.activeIndex,0),this.coordinateList.length-1)}},{key:"focus",value:function(){this.spoofMouse()}},{key:"keyboardEvent",value:function(r){if(this.coordinateList.length!==0)switch(r.key){case"ArrowRight":{if(this.layout!=="horizontal")return;this.activeIndex=Math.min(this.activeIndex+1,this.coordinateList.length-1),this.spoofMouse();break}case"ArrowLeft":{if(this.layout!=="horizontal")return;this.activeIndex=Math.max(this.activeIndex-1,0),this.spoofMouse();break}}}},{key:"setIndex",value:function(r){this.activeIndex=r}},{key:"spoofMouse",value:function(){var r,i;if(this.layout==="horizontal"&&this.coordinateList.length!==0){var l=this.container.getBoundingClientRect(),s=l.x,c=l.y,d=l.height,p=this.coordinateList[this.activeIndex].coordinate,v=((r=window)===null||r===void 0?void 0:r.scrollX)||0,h=((i=window)===null||i===void 0?void 0:i.scrollY)||0,m=s+p+v,b=c+this.offset.top+d/2+h;this.mouseHandlerCallback({pageX:m,pageY:b})}}}])})();function NT(t,e,r){if(r==="number"&&e===!0&&Array.isArray(t)){var i=t==null?void 0:t[0],l=t==null?void 0:t[1];if(i&&l&&ie(i)&&ie(l))return!0}return!1}function DT(t,e,r,i){var l=i/2;return{stroke:"none",fill:"#ccc",x:t==="horizontal"?e.x-l:r.left+.5,y:t==="horizontal"?r.top+.5:e.y-l,width:t==="horizontal"?i:r.width-1,height:t==="horizontal"?r.height-1:i}}function vb(t){var e=t.cx,r=t.cy,i=t.radius,l=t.startAngle,s=t.endAngle,c=je(e,r,i,l),d=je(e,r,i,s);return{points:[c,d],cx:e,cy:r,radius:i,startAngle:l,endAngle:s}}function LT(t,e,r){var i,l,s,c;if(t==="horizontal")i=e.x,s=i,l=r.top,c=r.top+r.height;else if(t==="vertical")l=e.y,c=l,i=r.left,s=r.left+r.width;else if(e.cx!=null&&e.cy!=null)if(t==="centric"){var d=e.cx,p=e.cy,v=e.innerRadius,h=e.outerRadius,m=e.angle,b=je(d,p,v,m),O=je(d,p,h,m);i=b.x,l=b.y,s=O.x,c=O.y}else return vb(e);return[{x:i,y:l},{x:s,y:c}]}function lo(t){"@babel/helpers - typeof";return lo=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},lo(t)}function Pm(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function _l(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?Pm(Object(r),!0).forEach(function(i){RT(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):Pm(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function RT(t,e,r){return e=BT(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function BT(t){var e=zT(t,"string");return lo(e)=="symbol"?e:e+""}function zT(t,e){if(lo(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(lo(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}function FT(t){var e,r,i=t.element,l=t.tooltipEventType,s=t.isActive,c=t.activeCoordinate,d=t.activePayload,p=t.offset,v=t.activeTooltipIndex,h=t.tooltipAxisBandSize,m=t.layout,b=t.chartName,O=(e=i.props.cursor)!==null&&e!==void 0?e:(r=i.type.defaultProps)===null||r===void 0?void 0:r.cursor;if(!i||!O||!s||!c||b!=="ScatterChart"&&l!=="axis")return null;var S,P=_a;if(b==="ScatterChart")S=c,P=$E;else if(b==="BarChart")S=DT(m,c,p,h),P=Zf;else if(m==="radial"){var g=vb(c),A=g.cx,j=g.cy,_=g.radius,k=g.startAngle,w=g.endAngle;S={cx:A,cy:j,startAngle:k,endAngle:w,innerRadius:_,outerRadius:_},P=hg}else S={points:LT(m,c,p)},P=_a;var E=_l(_l(_l(_l({stroke:"#ccc",pointerEvents:"none"},p),S),fe(O,!1)),{},{payload:d,payloadIndex:v,className:ge("recharts-tooltip-cursor",O.className)});return te.isValidElement(O)?te.cloneElement(O,E):te.createElement(P,E)}var WT=["item"],UT=["children","className","width","height","style","compact","title","desc"];function ki(t){"@babel/helpers - typeof";return ki=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(e){return typeof e}:function(e){return e&&typeof Symbol=="function"&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},ki(t)}function ni(){return ni=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(t[i]=r[i])}return t},ni.apply(this,arguments)}function Am(t,e){return HT(t)||KT(t,e)||hb(t,e)||VT()}function VT(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function KT(t,e){var r=t==null?null:typeof Symbol<"u"&&t[Symbol.iterator]||t["@@iterator"];if(r!=null){var i,l,s,c,d=[],p=!0,v=!1;try{if(s=(r=r.call(t)).next,e!==0)for(;!(p=(i=s.call(r)).done)&&(d.push(i.value),d.length!==e);p=!0);}catch(h){v=!0,l=h}finally{try{if(!p&&r.return!=null&&(c=r.return(),Object(c)!==c))return}finally{if(v)throw l}}return d}}function HT(t){if(Array.isArray(t))return t}function Em(t,e){if(t==null)return{};var r=GT(t,e),i,l;if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(t);for(l=0;l<s.length;l++)i=s[l],!(e.indexOf(i)>=0)&&Object.prototype.propertyIsEnumerable.call(t,i)&&(r[i]=t[i])}return r}function GT(t,e){if(t==null)return{};var r={};for(var i in t)if(Object.prototype.hasOwnProperty.call(t,i)){if(e.indexOf(i)>=0)continue;r[i]=t[i]}return r}function XT(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")}function YT(t,e){for(var r=0;r<e.length;r++){var i=e[r];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(t,mb(i.key),i)}}function qT(t,e,r){return e&&YT(t.prototype,e),Object.defineProperty(t,"prototype",{writable:!1}),t}function QT(t,e,r){return e=pu(e),ZT(t,yb()?Reflect.construct(e,r||[],pu(t).constructor):e.apply(t,r))}function ZT(t,e){if(e&&(ki(e)==="object"||typeof e=="function"))return e;if(e!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return JT(t)}function JT(t){if(t===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return t}function yb(){try{var t=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(yb=function(){return!!t})()}function pu(t){return pu=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(r){return r.__proto__||Object.getPrototypeOf(r)},pu(t)}function eC(t,e){if(typeof e!="function"&&e!==null)throw new TypeError("Super expression must either be null or a function");t.prototype=Object.create(e&&e.prototype,{constructor:{value:t,writable:!0,configurable:!0}}),Object.defineProperty(t,"prototype",{writable:!1}),e&&Wf(t,e)}function Wf(t,e){return Wf=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(i,l){return i.__proto__=l,i},Wf(t,e)}function Ti(t){return nC(t)||rC(t)||hb(t)||tC()}function tC(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function hb(t,e){if(t){if(typeof t=="string")return Uf(t,e);var r=Object.prototype.toString.call(t).slice(8,-1);if(r==="Object"&&t.constructor&&(r=t.constructor.name),r==="Map"||r==="Set")return Array.from(t);if(r==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(r))return Uf(t,e)}}function rC(t){if(typeof Symbol<"u"&&t[Symbol.iterator]!=null||t["@@iterator"]!=null)return Array.from(t)}function nC(t){if(Array.isArray(t))return Uf(t)}function Uf(t,e){(e==null||e>t.length)&&(e=t.length);for(var r=0,i=new Array(e);r<e;r++)i[r]=t[r];return i}function jm(t,e){var r=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(l){return Object.getOwnPropertyDescriptor(t,l).enumerable})),r.push.apply(r,i)}return r}function W(t){for(var e=1;e<arguments.length;e++){var r=arguments[e]!=null?arguments[e]:{};e%2?jm(Object(r),!0).forEach(function(i){ye(t,i,r[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(r)):jm(Object(r)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(r,i))})}return t}function ye(t,e,r){return e=mb(e),e in t?Object.defineProperty(t,e,{value:r,enumerable:!0,configurable:!0,writable:!0}):t[e]=r,t}function mb(t){var e=iC(t,"string");return ki(e)=="symbol"?e:e+""}function iC(t,e){if(ki(t)!="object"||!t)return t;var r=t[Symbol.toPrimitive];if(r!==void 0){var i=r.call(t,e);if(ki(i)!="object")return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return(e==="string"?String:Number)(t)}var aC={xAxis:["bottom","top"],yAxis:["left","right"]},oC={width:"100%",height:"100%"},gb={x:0,y:0};function kl(t){return t}var lC=function(e,r){return r==="horizontal"?e.x:r==="vertical"?e.y:r==="centric"?e.angle:e.radius},uC=function(e,r,i,l){var s=r.find(function(h){return h&&h.index===i});if(s){if(e==="horizontal")return{x:s.coordinate,y:l.y};if(e==="vertical")return{x:l.x,y:s.coordinate};if(e==="centric"){var c=s.coordinate,d=l.radius;return W(W(W({},l),je(l.cx,l.cy,d,c)),{},{angle:c,radius:d})}var p=s.coordinate,v=l.angle;return W(W(W({},l),je(l.cx,l.cy,p,v)),{},{angle:v,radius:p})}return gb},Tu=function(e,r){var i=r.graphicalItems,l=r.dataStartIndex,s=r.dataEndIndex,c=(i??[]).reduce(function(d,p){var v=p.props.data;return v&&v.length?[].concat(Ti(d),Ti(v)):d},[]);return c.length>0?c:e&&e.length&&ie(l)&&ie(s)?e.slice(l,s+1):[]};function bb(t){return t==="number"?[0,"auto"]:void 0}var Vf=function(e,r,i,l){var s=e.graphicalItems,c=e.tooltipAxis,d=Tu(r,e);return i<0||!s||!s.length||i>=d.length?null:s.reduce(function(p,v){var h,m=(h=v.props.data)!==null&&h!==void 0?h:r;m&&e.dataStartIndex+e.dataEndIndex!==0&&e.dataEndIndex-e.dataStartIndex>=i&&(m=m.slice(e.dataStartIndex,e.dataEndIndex+1));var b;if(c.dataKey&&!c.allowDuplicatedCategory){var O=m===void 0?d:m;b=$l(O,c.dataKey,l)}else b=m&&m[i]||d[i];return b?[].concat(Ti(p),[fg(v,b)]):p},[])},_m=function(e,r,i,l){var s=l||{x:e.chartX,y:e.chartY},c=lC(s,i),d=e.orderedTooltipTicks,p=e.tooltipAxis,v=e.tooltipTicks,h=rP(c,d,v,p);if(h>=0&&v){var m=v[h]&&v[h].value,b=Vf(e,r,h,m),O=uC(i,d,h,s);return{activeTooltipIndex:h,activeLabel:m,activePayload:b,activeCoordinate:O}}return null},sC=function(e,r){var i=r.axes,l=r.graphicalItems,s=r.axisType,c=r.axisIdKey,d=r.stackGroups,p=r.dataStartIndex,v=r.dataEndIndex,h=e.layout,m=e.children,b=e.stackOffset,O=ag(h,s);return i.reduce(function(S,P){var g,A=P.type.defaultProps!==void 0?W(W({},P.type.defaultProps),P.props):P.props,j=A.type,_=A.dataKey,k=A.allowDataOverflow,w=A.allowDuplicatedCategory,E=A.scale,C=A.ticks,M=A.includeHidden,D=A[c];if(S[D])return S;var U=Tu(e.data,{graphicalItems:l.filter(function(re){var de,we=c in re.props?re.props[c]:(de=re.type.defaultProps)===null||de===void 0?void 0:de[c];return we===D}),dataStartIndex:p,dataEndIndex:v}),F=U.length,H,q,Y;NT(A.domain,k,j)&&(H=af(A.domain,null,k),O&&(j==="number"||E!=="auto")&&(Y=ja(U,_,"category")));var V=bb(j);if(!H||H.length===0){var G,ee=(G=A.domain)!==null&&G!==void 0?G:V;if(_){if(H=ja(U,_,j),j==="category"&&O){var z=dx(H);w&&z?(q=H,H=Cl(0,F)):w||(H=Qy(ee,H,P).reduce(function(re,de){return re.indexOf(de)>=0?re:[].concat(Ti(re),[de])},[]))}else if(j==="category")w?H=H.filter(function(re){return re!==""&&!be(re)}):H=Qy(ee,H,P).reduce(function(re,de){return re.indexOf(de)>=0||de===""||be(de)?re:[].concat(Ti(re),[de])},[]);else if(j==="number"){var X=lP(U,l.filter(function(re){var de,we,Ce=c in re.props?re.props[c]:(de=re.type.defaultProps)===null||de===void 0?void 0:de[c],Ie="hide"in re.props?re.props.hide:(we=re.type.defaultProps)===null||we===void 0?void 0:we.hide;return Ce===D&&(M||!Ie)}),_,s,h);X&&(H=X)}O&&(j==="number"||E!=="auto")&&(Y=ja(U,_,"category"))}else O?H=Cl(0,F):d&&d[D]&&d[D].hasStack&&j==="number"?H=b==="expand"?[0,1]:cg(d[D].stackGroups,p,v):H=ig(U,l.filter(function(re){var de=c in re.props?re.props[c]:re.type.defaultProps[c],we="hide"in re.props?re.props.hide:re.type.defaultProps.hide;return de===D&&(M||!we)}),j,h,!0);if(j==="number")H=Ff(m,H,D,s,C),ee&&(H=af(ee,H,k));else if(j==="category"&&ee){var K=ee,ae=H.every(function(re){return K.indexOf(re)>=0});ae&&(H=K)}}return W(W({},S),{},ye({},D,W(W({},A),{},{axisType:s,domain:H,categoricalDomain:Y,duplicateDomain:q,originalDomain:(g=A.domain)!==null&&g!==void 0?g:V,isCategorical:O,layout:h})))},{})},cC=function(e,r){var i=r.graphicalItems,l=r.Axis,s=r.axisType,c=r.axisIdKey,d=r.stackGroups,p=r.dataStartIndex,v=r.dataEndIndex,h=e.layout,m=e.children,b=Tu(e.data,{graphicalItems:i,dataStartIndex:p,dataEndIndex:v}),O=b.length,S=ag(h,s),P=-1;return i.reduce(function(g,A){var j=A.type.defaultProps!==void 0?W(W({},A.type.defaultProps),A.props):A.props,_=j[c],k=bb("number");if(!g[_]){P++;var w;return S?w=Cl(0,O):d&&d[_]&&d[_].hasStack?(w=cg(d[_].stackGroups,p,v),w=Ff(m,w,_,s)):(w=af(k,ig(b,i.filter(function(E){var C,M,D=c in E.props?E.props[c]:(C=E.type.defaultProps)===null||C===void 0?void 0:C[c],U="hide"in E.props?E.props.hide:(M=E.type.defaultProps)===null||M===void 0?void 0:M.hide;return D===_&&!U}),"number",h),l.defaultProps.allowDataOverflow),w=Ff(m,w,_,s)),W(W({},g),{},ye({},_,W(W({axisType:s},l.defaultProps),{},{hide:!0,orientation:Mt(aC,"".concat(s,".").concat(P%2),null),domain:w,originalDomain:k,isCategorical:S,layout:h})))}return g},{})},fC=function(e,r){var i=r.axisType,l=i===void 0?"xAxis":i,s=r.AxisComp,c=r.graphicalItems,d=r.stackGroups,p=r.dataStartIndex,v=r.dataEndIndex,h=e.children,m="".concat(l,"Id"),b=Xt(h,s),O={};return b&&b.length?O=sC(e,{axes:b,graphicalItems:c,axisType:l,axisIdKey:m,stackGroups:d,dataStartIndex:p,dataEndIndex:v}):c&&c.length&&(O=cC(e,{Axis:s,graphicalItems:c,axisType:l,axisIdKey:m,stackGroups:d,dataStartIndex:p,dataEndIndex:v})),O},dC=function(e){var r=Yr(e),i=wr(r,!1,!0);return{tooltipTicks:i,orderedTooltipTicks:Kf(i,function(l){return l.coordinate}),tooltipAxis:r,tooltipAxisBandSize:Fl(r,i)}},km=function(e){var r=e.children,i=e.defaultShowTooltip,l=Pt(r,wi),s=0,c=0;return e.data&&e.data.length!==0&&(c=e.data.length-1),l&&l.props&&(l.props.startIndex>=0&&(s=l.props.startIndex),l.props.endIndex>=0&&(c=l.props.endIndex)),{chartX:0,chartY:0,dataStartIndex:s,dataEndIndex:c,activeTooltipIndex:-1,isTooltipActive:!!i}},pC=function(e){return!e||!e.length?!1:e.some(function(r){var i=xr(r&&r.type);return i&&i.indexOf("Bar")>=0})},Tm=function(e){return e==="horizontal"?{numericAxisName:"yAxis",cateAxisName:"xAxis"}:e==="vertical"?{numericAxisName:"xAxis",cateAxisName:"yAxis"}:e==="centric"?{numericAxisName:"radiusAxis",cateAxisName:"angleAxis"}:{numericAxisName:"angleAxis",cateAxisName:"radiusAxis"}},vC=function(e,r){var i=e.props,l=e.graphicalItems,s=e.xAxisMap,c=s===void 0?{}:s,d=e.yAxisMap,p=d===void 0?{}:d,v=i.width,h=i.height,m=i.children,b=i.margin||{},O=Pt(m,wi),S=Pt(m,ai),P=Object.keys(p).reduce(function(w,E){var C=p[E],M=C.orientation;return!C.mirror&&!C.hide?W(W({},w),{},ye({},M,w[M]+C.width)):w},{left:b.left||0,right:b.right||0}),g=Object.keys(c).reduce(function(w,E){var C=c[E],M=C.orientation;return!C.mirror&&!C.hide?W(W({},w),{},ye({},M,Mt(w,"".concat(M))+C.height)):w},{top:b.top||0,bottom:b.bottom||0}),A=W(W({},g),P),j=A.bottom;O&&(A.bottom+=O.props.height||wi.defaultProps.height),S&&r&&(A=aP(A,l,i,r));var _=v-A.left-A.right,k=h-A.top-A.bottom;return W(W({brushBottom:j},A),{},{width:Math.max(_,0),height:Math.max(k,0)})},yC=function(e,r){if(r==="xAxis")return e[r].width;if(r==="yAxis")return e[r].height},wb=function(e){var r=e.chartName,i=e.GraphicalChild,l=e.defaultTooltipEventType,s=l===void 0?"axis":l,c=e.validateTooltipEventTypes,d=c===void 0?["axis"]:c,p=e.axisComponents,v=e.legendContent,h=e.formatAxisMap,m=e.defaultProps,b=function(A,j){var _=j.graphicalItems,k=j.stackGroups,w=j.offset,E=j.updateId,C=j.dataStartIndex,M=j.dataEndIndex,D=A.barSize,U=A.layout,F=A.barGap,H=A.barCategoryGap,q=A.maxBarSize,Y=Tm(U),V=Y.numericAxisName,G=Y.cateAxisName,ee=pC(_),z=[];return _.forEach(function(X,K){var ae=Tu(A.data,{graphicalItems:[X],dataStartIndex:C,dataEndIndex:M}),re=X.type.defaultProps!==void 0?W(W({},X.type.defaultProps),X.props):X.props,de=re.dataKey,we=re.maxBarSize,Ce=re["".concat(V,"Id")],Ie=re["".concat(G,"Id")],At={},Re=p.reduce(function(dt,Dt){var jn=j["".concat(Dt.axisType,"Map")],Mi=re["".concat(Dt.axisType,"Id")];jn&&jn[Mi]||Dt.axisType==="zAxis"||xn();var Ni=jn[Mi];return W(W({},dt),{},ye(ye({},Dt.axisType,Ni),"".concat(Dt.axisType,"Ticks"),wr(Ni)))},At),Yt=Re[G],En=Re["".concat(G,"Ticks")],Nt=k&&k[Ce]&&k[Ce].hasStack&&hP(X,k[Ce].stackGroups),Ar=xr(X.type).indexOf("Bar")>=0,Et=Fl(Yt,En),qt=[],Er=ee&&nP({barSize:D,stackGroups:k,totalSize:yC(Re,G)});if(Ar){var jr,Qt,_r=be(we)?q:we,Zt=(jr=(Qt=Fl(Yt,En,!0))!==null&&Qt!==void 0?Qt:_r)!==null&&jr!==void 0?jr:0;qt=iP({barGap:F,barCategoryGap:H,bandSize:Zt!==Et?Zt:Et,sizeList:Er[Ie],maxBarSize:_r}),Zt!==Et&&(qt=qt.map(function(dt){return W(W({},dt),{},{position:W(W({},dt.position),{},{offset:dt.position.offset-Zt/2})})}))}var Jt=X&&X.type&&X.type.getComposedData;Jt&&z.push({props:W(W({},Jt(W(W({},Re),{},{displayedData:ae,props:A,dataKey:de,item:X,bandSize:Et,barPosition:qt,offset:w,stackedData:Nt,layout:U,dataStartIndex:C,dataEndIndex:M}))),{},ye(ye(ye({key:X.key||"item-".concat(K)},V,Re[V]),G,Re[G]),"animationId",E)),childIndex:Ax(X,A.children),item:X})}),z},O=function(A,j){var _=A.props,k=A.dataStartIndex,w=A.dataEndIndex,E=A.updateId;if(!fy({props:_}))return null;var C=_.children,M=_.layout,D=_.stackOffset,U=_.data,F=_.reverseStackOrder,H=Tm(M),q=H.numericAxisName,Y=H.cateAxisName,V=Xt(C,i),G=vP(U,V,"".concat(q,"Id"),"".concat(Y,"Id"),D,F),ee=p.reduce(function(re,de){var we="".concat(de.axisType,"Map");return W(W({},re),{},ye({},we,fC(_,W(W({},de),{},{graphicalItems:V,stackGroups:de.axisType===q&&G,dataStartIndex:k,dataEndIndex:w}))))},{}),z=vC(W(W({},ee),{},{props:_,graphicalItems:V}),j==null?void 0:j.legendBBox);Object.keys(ee).forEach(function(re){ee[re]=h(_,ee[re],z,re.replace("Map",""),r)});var X=ee["".concat(Y,"Map")],K=dC(X),ae=b(_,W(W({},ee),{},{dataStartIndex:k,dataEndIndex:w,updateId:E,graphicalItems:V,stackGroups:G,offset:z}));return W(W({formattedGraphicalItems:ae,graphicalItems:V,offset:z,stackGroups:G},K),ee)},S=(function(g){function A(j){var _,k,w;return XT(this,A),w=QT(this,A,[j]),ye(w,"eventEmitterSymbol",Symbol("rechartsEventEmitter")),ye(w,"accessibilityManager",new MT),ye(w,"handleLegendBBoxUpdate",function(E){if(E){var C=w.state,M=C.dataStartIndex,D=C.dataEndIndex,U=C.updateId;w.setState(W({legendBBox:E},O({props:w.props,dataStartIndex:M,dataEndIndex:D,updateId:U},W(W({},w.state),{},{legendBBox:E}))))}}),ye(w,"handleReceiveSyncEvent",function(E,C,M){if(w.props.syncId===E){if(M===w.eventEmitterSymbol&&typeof w.props.syncMethod!="function")return;w.applySyncEvent(C)}}),ye(w,"handleBrushChange",function(E){var C=E.startIndex,M=E.endIndex;if(C!==w.state.dataStartIndex||M!==w.state.dataEndIndex){var D=w.state.updateId;w.setState(function(){return W({dataStartIndex:C,dataEndIndex:M},O({props:w.props,dataStartIndex:C,dataEndIndex:M,updateId:D},w.state))}),w.triggerSyncEvent({dataStartIndex:C,dataEndIndex:M})}}),ye(w,"handleMouseEnter",function(E){var C=w.getMouseInfo(E);if(C){var M=W(W({},C),{},{isTooltipActive:!0});w.setState(M),w.triggerSyncEvent(M);var D=w.props.onMouseEnter;he(D)&&D(M,E)}}),ye(w,"triggeredAfterMouseMove",function(E){var C=w.getMouseInfo(E),M=C?W(W({},C),{},{isTooltipActive:!0}):{isTooltipActive:!1};w.setState(M),w.triggerSyncEvent(M);var D=w.props.onMouseMove;he(D)&&D(M,E)}),ye(w,"handleItemMouseEnter",function(E){w.setState(function(){return{isTooltipActive:!0,activeItem:E,activePayload:E.tooltipPayload,activeCoordinate:E.tooltipPosition||{x:E.cx,y:E.cy}}})}),ye(w,"handleItemMouseLeave",function(){w.setState(function(){return{isTooltipActive:!1}})}),ye(w,"handleMouseMove",function(E){E.persist(),w.throttleTriggeredAfterMouseMove(E)}),ye(w,"handleMouseLeave",function(E){w.throttleTriggeredAfterMouseMove.cancel();var C={isTooltipActive:!1};w.setState(C),w.triggerSyncEvent(C);var M=w.props.onMouseLeave;he(M)&&M(C,E)}),ye(w,"handleOuterEvent",function(E){var C=Px(E),M=Mt(w.props,"".concat(C));if(C&&he(M)){var D,U;/.*touch.*/i.test(C)?U=w.getMouseInfo(E.changedTouches[0]):U=w.getMouseInfo(E),M((D=U)!==null&&D!==void 0?D:{},E)}}),ye(w,"handleClick",function(E){var C=w.getMouseInfo(E);if(C){var M=W(W({},C),{},{isTooltipActive:!0});w.setState(M),w.triggerSyncEvent(M);var D=w.props.onClick;he(D)&&D(M,E)}}),ye(w,"handleMouseDown",function(E){var C=w.props.onMouseDown;if(he(C)){var M=w.getMouseInfo(E);C(M,E)}}),ye(w,"handleMouseUp",function(E){var C=w.props.onMouseUp;if(he(C)){var M=w.getMouseInfo(E);C(M,E)}}),ye(w,"handleTouchMove",function(E){E.changedTouches!=null&&E.changedTouches.length>0&&w.throttleTriggeredAfterMouseMove(E.changedTouches[0])}),ye(w,"handleTouchStart",function(E){E.changedTouches!=null&&E.changedTouches.length>0&&w.handleMouseDown(E.changedTouches[0])}),ye(w,"handleTouchEnd",function(E){E.changedTouches!=null&&E.changedTouches.length>0&&w.handleMouseUp(E.changedTouches[0])}),ye(w,"handleDoubleClick",function(E){var C=w.props.onDoubleClick;if(he(C)){var M=w.getMouseInfo(E);C(M,E)}}),ye(w,"handleContextMenu",function(E){var C=w.props.onContextMenu;if(he(C)){var M=w.getMouseInfo(E);C(M,E)}}),ye(w,"triggerSyncEvent",function(E){w.props.syncId!==void 0&&Lc.emit(Rc,w.props.syncId,E,w.eventEmitterSymbol)}),ye(w,"applySyncEvent",function(E){var C=w.props,M=C.layout,D=C.syncMethod,U=w.state.updateId,F=E.dataStartIndex,H=E.dataEndIndex;if(E.dataStartIndex!==void 0||E.dataEndIndex!==void 0)w.setState(W({dataStartIndex:F,dataEndIndex:H},O({props:w.props,dataStartIndex:F,dataEndIndex:H,updateId:U},w.state)));else if(E.activeTooltipIndex!==void 0){var q=E.chartX,Y=E.chartY,V=E.activeTooltipIndex,G=w.state,ee=G.offset,z=G.tooltipTicks;if(!ee)return;if(typeof D=="function")V=D(z,E);else if(D==="value"){V=-1;for(var X=0;X<z.length;X++)if(z[X].value===E.activeLabel){V=X;break}}var K=W(W({},ee),{},{x:ee.left,y:ee.top}),ae=Math.min(q,K.x+K.width),re=Math.min(Y,K.y+K.height),de=z[V]&&z[V].value,we=Vf(w.state,w.props.data,V),Ce=z[V]?{x:M==="horizontal"?z[V].coordinate:ae,y:M==="horizontal"?re:z[V].coordinate}:gb;w.setState(W(W({},E),{},{activeLabel:de,activeCoordinate:Ce,activePayload:we,activeTooltipIndex:V}))}else w.setState(E)}),ye(w,"renderCursor",function(E){var C,M=w.state,D=M.isTooltipActive,U=M.activeCoordinate,F=M.activePayload,H=M.offset,q=M.activeTooltipIndex,Y=M.tooltipAxisBandSize,V=w.getTooltipEventType(),G=(C=E.props.active)!==null&&C!==void 0?C:D,ee=w.props.layout,z=E.key||"_recharts-cursor";return $.createElement(FT,{key:z,activeCoordinate:U,activePayload:F,activeTooltipIndex:q,chartName:r,element:E,isActive:G,layout:ee,offset:H,tooltipAxisBandSize:Y,tooltipEventType:V})}),ye(w,"renderPolarAxis",function(E,C,M){var D=Mt(E,"type.axisType"),U=Mt(w.state,"".concat(D,"Map")),F=E.type.defaultProps,H=F!==void 0?W(W({},F),E.props):E.props,q=U&&U[H["".concat(D,"Id")]];return te.cloneElement(E,W(W({},q),{},{className:ge(D,q.className),key:E.key||"".concat(C,"-").concat(M),ticks:wr(q,!0)}))}),ye(w,"renderPolarGrid",function(E){var C=E.props,M=C.radialLines,D=C.polarAngles,U=C.polarRadius,F=w.state,H=F.radiusAxisMap,q=F.angleAxisMap,Y=Yr(H),V=Yr(q),G=V.cx,ee=V.cy,z=V.innerRadius,X=V.outerRadius;return te.cloneElement(E,{polarAngles:Array.isArray(D)?D:wr(V,!0).map(function(K){return K.coordinate}),polarRadius:Array.isArray(U)?U:wr(Y,!0).map(function(K){return K.coordinate}),cx:G,cy:ee,innerRadius:z,outerRadius:X,key:E.key||"polar-grid",radialLines:M})}),ye(w,"renderLegend",function(){var E=w.state.formattedGraphicalItems,C=w.props,M=C.children,D=C.width,U=C.height,F=w.props.margin||{},H=D-(F.left||0)-(F.right||0),q=rg({children:M,formattedGraphicalItems:E,legendWidth:H,legendContent:v});if(!q)return null;var Y=q.item,V=Em(q,WT);return te.cloneElement(Y,W(W({},V),{},{chartWidth:D,chartHeight:U,margin:F,onBBoxUpdate:w.handleLegendBBoxUpdate}))}),ye(w,"renderTooltip",function(){var E,C=w.props,M=C.children,D=C.accessibilityLayer,U=Pt(M,mr);if(!U)return null;var F=w.state,H=F.isTooltipActive,q=F.activeCoordinate,Y=F.activePayload,V=F.activeLabel,G=F.offset,ee=(E=U.props.active)!==null&&E!==void 0?E:H;return te.cloneElement(U,{viewBox:W(W({},G),{},{x:G.left,y:G.top}),active:ee,label:V,payload:ee?Y:[],coordinate:q,accessibilityLayer:D})}),ye(w,"renderBrush",function(E){var C=w.props,M=C.margin,D=C.data,U=w.state,F=U.offset,H=U.dataStartIndex,q=U.dataEndIndex,Y=U.updateId;return te.cloneElement(E,{key:E.key||"_recharts-brush",onChange:Pl(w.handleBrushChange,E.props.onChange),data:D,x:ie(E.props.x)?E.props.x:F.left,y:ie(E.props.y)?E.props.y:F.top+F.height+F.brushBottom-(M.bottom||0),width:ie(E.props.width)?E.props.width:F.width,startIndex:H,endIndex:q,updateId:"brush-".concat(Y)})}),ye(w,"renderReferenceElement",function(E,C,M){if(!E)return null;var D=w,U=D.clipPathId,F=w.state,H=F.xAxisMap,q=F.yAxisMap,Y=F.offset,V=E.type.defaultProps||{},G=E.props,ee=G.xAxisId,z=ee===void 0?V.xAxisId:ee,X=G.yAxisId,K=X===void 0?V.yAxisId:X;return te.cloneElement(E,{key:E.key||"".concat(C,"-").concat(M),xAxis:H[z],yAxis:q[K],viewBox:{x:Y.left,y:Y.top,width:Y.width,height:Y.height},clipPathId:U})}),ye(w,"renderActivePoints",function(E){var C=E.item,M=E.activePoint,D=E.basePoint,U=E.childIndex,F=E.isRange,H=[],q=C.props.key,Y=C.item.type.defaultProps!==void 0?W(W({},C.item.type.defaultProps),C.item.props):C.item.props,V=Y.activeDot,G=Y.dataKey,ee=W(W({index:U,dataKey:G,cx:M.x,cy:M.y,r:4,fill:Qf(C.item),strokeWidth:2,stroke:"#fff",payload:M.payload,value:M.value},fe(V,!1)),Il(V));return H.push(A.renderActiveDot(V,ee,"".concat(q,"-activePoint-").concat(U))),D?H.push(A.renderActiveDot(V,W(W({},ee),{},{cx:D.x,cy:D.y}),"".concat(q,"-basePoint-").concat(U))):F&&H.push(null),H}),ye(w,"renderGraphicChild",function(E,C,M){var D=w.filterFormatItem(E,C,M);if(!D)return null;var U=w.getTooltipEventType(),F=w.state,H=F.isTooltipActive,q=F.tooltipAxis,Y=F.activeTooltipIndex,V=F.activeLabel,G=w.props.children,ee=Pt(G,mr),z=D.props,X=z.points,K=z.isRange,ae=z.baseLine,re=D.item.type.defaultProps!==void 0?W(W({},D.item.type.defaultProps),D.item.props):D.item.props,de=re.activeDot,we=re.hide,Ce=re.activeBar,Ie=re.activeShape,At=!!(!we&&H&&ee&&(de||Ce||Ie)),Re={};U!=="axis"&&ee&&ee.props.trigger==="click"?Re={onClick:Pl(w.handleItemMouseEnter,E.props.onClick)}:U!=="axis"&&(Re={onMouseLeave:Pl(w.handleItemMouseLeave,E.props.onMouseLeave),onMouseEnter:Pl(w.handleItemMouseEnter,E.props.onMouseEnter)});var Yt=te.cloneElement(E,W(W({},D.props),Re));function En(Dt){return typeof q.dataKey=="function"?q.dataKey(Dt.payload):null}if(At)if(Y>=0){var Nt,Ar;if(q.dataKey&&!q.allowDuplicatedCategory){var Et=typeof q.dataKey=="function"?En:"payload.".concat(q.dataKey.toString());Nt=$l(X,Et,V),Ar=K&&ae&&$l(ae,Et,V)}else Nt=X==null?void 0:X[Y],Ar=K&&ae&&ae[Y];if(Ie||Ce){var qt=E.props.activeIndex!==void 0?E.props.activeIndex:Y;return[te.cloneElement(E,W(W(W({},D.props),Re),{},{activeIndex:qt})),null,null]}if(!be(Nt))return[Yt].concat(Ti(w.renderActivePoints({item:D,activePoint:Nt,basePoint:Ar,childIndex:Y,isRange:K})))}else{var Er,jr=(Er=w.getItemByXY(w.state.activeCoordinate))!==null&&Er!==void 0?Er:{graphicalItem:Yt},Qt=jr.graphicalItem,_r=Qt.item,Zt=_r===void 0?E:_r,Jt=Qt.childIndex,dt=W(W(W({},D.props),Re),{},{activeIndex:Jt});return[te.cloneElement(Zt,dt),null,null]}return K?[Yt,null,null]:[Yt,null]}),ye(w,"renderCustomized",function(E,C,M){return te.cloneElement(E,W(W({key:"recharts-customized-".concat(M)},w.props),w.state))}),ye(w,"renderMap",{CartesianGrid:{handler:kl,once:!0},ReferenceArea:{handler:w.renderReferenceElement},ReferenceLine:{handler:kl},ReferenceDot:{handler:w.renderReferenceElement},XAxis:{handler:kl},YAxis:{handler:kl},Brush:{handler:w.renderBrush,once:!0},Bar:{handler:w.renderGraphicChild},Line:{handler:w.renderGraphicChild},Area:{handler:w.renderGraphicChild},Radar:{handler:w.renderGraphicChild},RadialBar:{handler:w.renderGraphicChild},Scatter:{handler:w.renderGraphicChild},Pie:{handler:w.renderGraphicChild},Funnel:{handler:w.renderGraphicChild},Tooltip:{handler:w.renderCursor,once:!0},PolarGrid:{handler:w.renderPolarGrid,once:!0},PolarAngleAxis:{handler:w.renderPolarAxis},PolarRadiusAxis:{handler:w.renderPolarAxis},Customized:{handler:w.renderCustomized}}),w.clipPathId="".concat((_=j.id)!==null&&_!==void 0?_:uo("recharts"),"-clip"),w.throttleTriggeredAfterMouseMove=Im(w.triggeredAfterMouseMove,(k=j.throttleDelay)!==null&&k!==void 0?k:1e3/60),w.state={},w}return eC(A,g),qT(A,[{key:"componentDidMount",value:function(){var _,k;this.addListener(),this.accessibilityManager.setDetails({container:this.container,offset:{left:(_=this.props.margin.left)!==null&&_!==void 0?_:0,top:(k=this.props.margin.top)!==null&&k!==void 0?k:0},coordinateList:this.state.tooltipTicks,mouseHandlerCallback:this.triggeredAfterMouseMove,layout:this.props.layout}),this.displayDefaultTooltip()}},{key:"displayDefaultTooltip",value:function(){var _=this.props,k=_.children,w=_.data,E=_.height,C=_.layout,M=Pt(k,mr);if(M){var D=M.props.defaultIndex;if(!(typeof D!="number"||D<0||D>this.state.tooltipTicks.length-1)){var U=this.state.tooltipTicks[D]&&this.state.tooltipTicks[D].value,F=Vf(this.state,w,D,U),H=this.state.tooltipTicks[D].coordinate,q=(this.state.offset.top+E)/2,Y=C==="horizontal",V=Y?{x:H,y:q}:{y:H,x:q},G=this.state.formattedGraphicalItems.find(function(z){var X=z.item;return X.type.name==="Scatter"});G&&(V=W(W({},V),G.props.points[D].tooltipPosition),F=G.props.points[D].tooltipPayload);var ee={activeTooltipIndex:D,isTooltipActive:!0,activeLabel:U,activePayload:F,activeCoordinate:V};this.setState(ee),this.renderCursor(M),this.accessibilityManager.setIndex(D)}}}},{key:"getSnapshotBeforeUpdate",value:function(_,k){if(!this.props.accessibilityLayer)return null;if(this.state.tooltipTicks!==k.tooltipTicks&&this.accessibilityManager.setDetails({coordinateList:this.state.tooltipTicks}),this.props.layout!==_.layout&&this.accessibilityManager.setDetails({layout:this.props.layout}),this.props.margin!==_.margin){var w,E;this.accessibilityManager.setDetails({offset:{left:(w=this.props.margin.left)!==null&&w!==void 0?w:0,top:(E=this.props.margin.top)!==null&&E!==void 0?E:0}})}return null}},{key:"componentDidUpdate",value:function(_){Wc([Pt(_.children,mr)],[Pt(this.props.children,mr)])||this.displayDefaultTooltip()}},{key:"componentWillUnmount",value:function(){this.removeListener(),this.throttleTriggeredAfterMouseMove.cancel()}},{key:"getTooltipEventType",value:function(){var _=Pt(this.props.children,mr);if(_&&typeof _.props.shared=="boolean"){var k=_.props.shared?"axis":"item";return d.indexOf(k)>=0?k:s}return s}},{key:"getMouseInfo",value:function(_){if(!this.container)return null;var k=this.container,w=k.getBoundingClientRect(),E=aS(w),C={chartX:Math.round(_.pageX-E.left),chartY:Math.round(_.pageY-E.top)},M=w.width/k.offsetWidth||1,D=this.inRange(C.chartX,C.chartY,M);if(!D)return null;var U=this.state,F=U.xAxisMap,H=U.yAxisMap,q=this.getTooltipEventType(),Y=_m(this.state,this.props.data,this.props.layout,D);if(q!=="axis"&&F&&H){var V=Yr(F).scale,G=Yr(H).scale,ee=V&&V.invert?V.invert(C.chartX):null,z=G&&G.invert?G.invert(C.chartY):null;return W(W({},C),{},{xValue:ee,yValue:z},Y)}return Y?W(W({},C),Y):null}},{key:"inRange",value:function(_,k){var w=arguments.length>2&&arguments[2]!==void 0?arguments[2]:1,E=this.props.layout,C=_/w,M=k/w;if(E==="horizontal"||E==="vertical"){var D=this.state.offset,U=C>=D.left&&C<=D.left+D.width&&M>=D.top&&M<=D.top+D.height;return U?{x:C,y:M}:null}var F=this.state,H=F.angleAxisMap,q=F.radiusAxisMap;if(H&&q){var Y=Yr(H);return eh({x:C,y:M},Y)}return null}},{key:"parseEventsOfWrapper",value:function(){var _=this.props.children,k=this.getTooltipEventType(),w=Pt(_,mr),E={};w&&k==="axis"&&(w.props.trigger==="click"?E={onClick:this.handleClick}:E={onMouseEnter:this.handleMouseEnter,onDoubleClick:this.handleDoubleClick,onMouseMove:this.handleMouseMove,onMouseLeave:this.handleMouseLeave,onTouchMove:this.handleTouchMove,onTouchStart:this.handleTouchStart,onTouchEnd:this.handleTouchEnd,onContextMenu:this.handleContextMenu});var C=Il(this.props,this.handleOuterEvent);return W(W({},C),E)}},{key:"addListener",value:function(){Lc.on(Rc,this.handleReceiveSyncEvent)}},{key:"removeListener",value:function(){Lc.removeListener(Rc,this.handleReceiveSyncEvent)}},{key:"filterFormatItem",value:function(_,k,w){for(var E=this.state.formattedGraphicalItems,C=0,M=E.length;C<M;C++){var D=E[C];if(D.item===_||D.props.key===_.key||k===xr(D.item.type)&&w===D.childIndex)return D}return null}},{key:"renderClipPath",value:function(){var _=this.clipPathId,k=this.state.offset,w=k.left,E=k.top,C=k.height,M=k.width;return $.createElement("defs",null,$.createElement("clipPath",{id:_},$.createElement("rect",{x:w,y:E,height:C,width:M})))}},{key:"getXScales",value:function(){var _=this.state.xAxisMap;return _?Object.entries(_).reduce(function(k,w){var E=Am(w,2),C=E[0],M=E[1];return W(W({},k),{},ye({},C,M.scale))},{}):null}},{key:"getYScales",value:function(){var _=this.state.yAxisMap;return _?Object.entries(_).reduce(function(k,w){var E=Am(w,2),C=E[0],M=E[1];return W(W({},k),{},ye({},C,M.scale))},{}):null}},{key:"getXScaleByAxisId",value:function(_){var k;return(k=this.state.xAxisMap)===null||k===void 0||(k=k[_])===null||k===void 0?void 0:k.scale}},{key:"getYScaleByAxisId",value:function(_){var k;return(k=this.state.yAxisMap)===null||k===void 0||(k=k[_])===null||k===void 0?void 0:k.scale}},{key:"getItemByXY",value:function(_){var k=this.state,w=k.formattedGraphicalItems,E=k.activeItem;if(w&&w.length)for(var C=0,M=w.length;C<M;C++){var D=w[C],U=D.props,F=D.item,H=F.type.defaultProps!==void 0?W(W({},F.type.defaultProps),F.props):F.props,q=xr(F.type);if(q==="Bar"){var Y=(U.data||[]).find(function(z){return pE(_,z)});if(Y)return{graphicalItem:D,payload:Y}}else if(q==="RadialBar"){var V=(U.data||[]).find(function(z){return eh(_,z)});if(V)return{graphicalItem:D,payload:V}}else if(Ou(D,E)||Su(D,E)||to(D,E)){var G=Nj({graphicalItem:D,activeTooltipItem:E,itemData:H.data}),ee=H.activeIndex===void 0?G:H.activeIndex;return{graphicalItem:W(W({},D),{},{childIndex:ee}),payload:to(D,E)?H.data[G]:D.props.data[G]}}}return null}},{key:"render",value:function(){var _=this;if(!fy(this))return null;var k=this.props,w=k.children,E=k.className,C=k.width,M=k.height,D=k.style,U=k.compact,F=k.title,H=k.desc,q=Em(k,UT),Y=fe(q,!1);if(U)return $.createElement(lm,{state:this.state,width:this.props.width,height:this.props.height,clipPathId:this.clipPathId},$.createElement(Vc,ni({},Y,{width:C,height:M,title:F,desc:H}),this.renderClipPath(),py(w,this.renderMap)));if(this.props.accessibilityLayer){var V,G;Y.tabIndex=(V=this.props.tabIndex)!==null&&V!==void 0?V:0,Y.role=(G=this.props.role)!==null&&G!==void 0?G:"application",Y.onKeyDown=function(z){_.accessibilityManager.keyboardEvent(z)},Y.onFocus=function(){_.accessibilityManager.focus()}}var ee=this.parseEventsOfWrapper();return $.createElement(lm,{state:this.state,width:this.props.width,height:this.props.height,clipPathId:this.clipPathId},$.createElement("div",ni({className:ge("recharts-wrapper",E),style:W({position:"relative",cursor:"default",width:C,height:M},D)},ee,{ref:function(X){_.container=X}}),$.createElement(Vc,ni({},Y,{width:C,height:M,title:F,desc:H,style:oC}),this.renderClipPath(),py(w,this.renderMap)),this.renderLegend(),this.renderTooltip()))}}])})(te.Component);ye(S,"displayName",r),ye(S,"defaultProps",W({layout:"horizontal",stackOffset:"none",barCategoryGap:"10%",barGap:4,margin:{top:5,right:5,bottom:5,left:5},reverseStackOrder:!1,syncMethod:"index"},m)),ye(S,"getDerivedStateFromProps",function(g,A){var j=g.dataKey,_=g.data,k=g.children,w=g.width,E=g.height,C=g.layout,M=g.stackOffset,D=g.margin,U=A.dataStartIndex,F=A.dataEndIndex;if(A.updateId===void 0){var H=km(g);return W(W(W({},H),{},{updateId:0},O(W(W({props:g},H),{},{updateId:0}),A)),{},{prevDataKey:j,prevData:_,prevWidth:w,prevHeight:E,prevLayout:C,prevStackOffset:M,prevMargin:D,prevChildren:k})}if(j!==A.prevDataKey||_!==A.prevData||w!==A.prevWidth||E!==A.prevHeight||C!==A.prevLayout||M!==A.prevStackOffset||!ii(D,A.prevMargin)){var q=km(g),Y={chartX:A.chartX,chartY:A.chartY,isTooltipActive:A.isTooltipActive},V=W(W({},_m(A,_,C)),{},{updateId:A.updateId+1}),G=W(W(W({},q),Y),V);return W(W(W({},G),O(W({props:g},G),A)),{},{prevDataKey:j,prevData:_,prevWidth:w,prevHeight:E,prevLayout:C,prevStackOffset:M,prevMargin:D,prevChildren:k})}if(!Wc(k,A.prevChildren)){var ee,z,X,K,ae=Pt(k,wi),re=ae&&(ee=(z=ae.props)===null||z===void 0?void 0:z.startIndex)!==null&&ee!==void 0?ee:U,de=ae&&(X=(K=ae.props)===null||K===void 0?void 0:K.endIndex)!==null&&X!==void 0?X:F,we=re!==U||de!==F,Ce=!be(_),Ie=Ce&&!we?A.updateId:A.updateId+1;return W(W({updateId:Ie},O(W(W({props:g},A),{},{updateId:Ie,dataStartIndex:re,dataEndIndex:de}),A)),{},{prevChildren:k,dataStartIndex:re,dataEndIndex:de})}return null}),ye(S,"renderActiveDot",function(g,A,j){var _;return te.isValidElement(g)?_=te.cloneElement(g,A):he(g)?_=g(A):_=$.createElement(so,A),$.createElement(Se,{className:"recharts-active-dot",key:j},_)});var P=te.forwardRef(function(A,j){return $.createElement(S,ni({},A,{ref:j}))});return P.displayName=S.displayName,P},bC=wb({chartName:"RadarChart",GraphicalChild:Pu,axisComponents:[{axisType:"angleAxis",AxisComp:xu},{axisType:"radiusAxis",AxisComp:bu}],formatAxisMap:jP,defaultProps:{layout:"centric",startAngle:90,endAngle:-270,cx:"50%",cy:"50%",innerRadius:0,outerRadius:"80%"}}),wC=wb({chartName:"AreaChart",GraphicalChild:An,axisComponents:[{axisType:"xAxis",AxisComp:ud},{axisType:"yAxis",AxisComp:sd}],formatAxisMap:A_});export{wC as A,Xk as C,VE as P,gC as R,mr as T,ud as X,sd as Y,bC as a,xu as b,bu as c,Pu as d,An as e,ox as r};
